{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_assembly_gtpv1;

interface

uses base_structures,
     class_parser_gtpv1,
     class_assembly_baseclass;


type tGTPv1Assembler = class (tFrameAssembler)

                                constructor create;

                                private
                                       priv_Version      : byte;
                                       priv_ProtocolType : tGTPProtocolType;
                                       priv_flags        : TGTPFlagSet;
                                       priv_MessageType  : byte;
                                       priv_TotalLength  : word;
                                       priv_TEID         : longword;
                                       priv_SequenceNo   : word;
                                       priv_NPDUNumber   : byte;
                                       priv_NextExtHdr   : byte;

                                       function VersionProtocolAndFlagsToByte : byte;

                                       procedure setSequenceNo (const Number : word);
                                       procedure setNPDUNumber (const Number : byte);
                                       procedure setNextExtHdr (const Number : byte);

                                protected
                                         // called from base class from GetAssembly

                                         procedure DetermineAssemblySize; override;
                                         procedure Assemble; override;

                                public
                                      property Version : byte read priv_Version write priv_version;
                                      property ProtocolType : tGTPProtocolType read priv_ProtocolType write priv_ProtocolType;
                                      property Flags : TGTPFlagSet read priv_flags write priv_flags;
                                      property MessageType : byte read priv_MessageType write priv_MessageType;
                                      property TotalLength : word read priv_TotalLength write priv_TotalLength;
                                      property TEID : longword read priv_TEID write priv_TEID;
                                      property SequenceNo : word read priv_SequenceNo write setSequenceNo;
                                      property NPDUNumber : byte read priv_NPDUNumber write setNPDUNumber;
                                      property NextExtHdr : byte read priv_NextExtHdr write SetNextExtHdr;

                                      procedure CopyFromParser (Parser : tGTPv1Parser);
                           end;


implementation

uses toolkit;

type tGTPOptionValues = record
                              SequenceNo  : word;
                              NPDUNumber  : byte;
                              NextExtHdr  : byte;
                        end;
     pGTPOptionValues = ^tGTPOptionValues;

(* ------------------------------------------- *)
{ tGTPv1Assembler }
(* ------------------------------------------- *)

constructor tGTPv1Assembler.create;

begin
     inherited create;

     priv_Version := 0;
     priv_ProtocolType := gtpptGTP;
     priv_flags := [];
     priv_MessageType := 0;
     priv_TotalLength := 0;
     priv_TEID := 0;
     priv_SequenceNo := 0;
     priv_NPDUNumber := 0;
     priv_NextExtHdr := 0;
end;

(* ------------------------------------------- *)

procedure tGTPv1Assembler.DetermineAssemblySize;

begin
     prot_length := sizeOf (tGTPv1Struct);

     if priv_flags <> [] then
        prot_length := prot_length + 4;
end;

(* ------------------------------------------- *)

function tGTPv1Assembler.VersionProtocolAndFlagsToByte : byte;

var value : byte;
    flag  : tGTPFlags;

begin
     value := (priv_version shl 5);

     if priv_ProtocolType = gtpptGTP then
        value := value or $10;

     for flag in priv_flags do
         begin
              case flag of
               gtpReserved         : value := value or $8;
               gtpExtensionPresent : value := value or $4;
               gtpSequencePresent  : value := value or $2;
               gtpNPDUPresent      : value := value or $1;
              end;
         end;

     result := value;
end;

(* ------------------------------------------- *)

procedure tGTPv1Assembler.setSequenceNo (const Number : word);

begin
     priv_SequenceNo := Number;

     if Number = 0 then
        priv_flags := priv_flags - [gtpSequencePresent]
     else
         priv_flags := priv_flags + [gtpSequencePresent];
end;

(* ------------------------------------------- *)

procedure tGTPv1Assembler.setNPDUNumber (const Number: byte);

begin
     priv_NPDUNumber := Number;

     if Number = 0 then
        priv_flags := priv_flags - [gtpNPDUPresent]
     else
         priv_flags := priv_flags + [gtpNPDUPresent];
end;

(* ------------------------------------------- *)

procedure tGTPv1Assembler.setNextExtHdr (const Number: byte);

begin
     priv_NextExtHdr := Number;

     if Number = 0 then
        priv_flags := priv_flags - [gtpExtensionPresent]
     else
         priv_flags := priv_flags + [gtpExtensionPresent];
end;

(* ------------------------------------------- *)

procedure tGTPv1Assembler.Assemble;

var pData : pGTPv1Struct;
    help  : pGTPOptionValues;

begin
     pData := pGTPv1Struct (prot_data);

     pData^.Flags := VersionProtocolAndFlagsToByte;
     pData^.MessageType := priv_MessageType;
     pData^.Length := swap16 (priv_TotalLength);
     pData^.TEID := swap32 (priv_TEID);

     // Write options, if any

     if priv_flags <> [] then
        begin
             help := pGTPOptionValues (prot_data + sizeOf (tGTPv1Struct));

             help^.SequenceNo := priv_SequenceNo;
             help^.NPDUNumber := priv_NPDUNumber;
             help^.NextExtHdr := priv_NextExtHdr;

             assert (sizeOf (tGTPv1Struct) + sizeOf (tGTPOptionValues) = prot_length, 'Memory offset check mismatch (tGTPv1Assembler.Assemble)');
        end;
end;

(* ------------------------------------------- *)

procedure tGTPv1Assembler.CopyFromParser (Parser: tGTPv1Parser);

begin
     priv_Version  := Parser.Version;
     priv_ProtocolType := Parser.ProtocolType;
     priv_flags := Parser.Flags;
     priv_MessageType := Parser.MessageType;
     priv_TotalLength := Parser.TotalLength;
     priv_TEID := Parser.TEID;
     priv_SequenceNo  := Parser.SequenceNo;
     priv_NPDUNumber  := Parser.NPDUNumber;
     priv_NextExtHdr  := Parser.NextExtHdr;
end;

(* ------------------------------------------- *)

end.
