{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_assembly_hsrp;

interface

uses base_structures,
     definition_ip_addresses,
     class_parser_hsrp,
     class_assembly_baseclass;

type tHSRPAssembler = class (tFrameAssembler)

                                constructor create;

                                private
                                       priv_version    : byte;
                                       priv_opcode     : tHSRPOpCode;
                                       priv_state      : tHSRPState;
                                       priv_hellotime  : byte;
                                       priv_holdtime   : byte;
                                       priv_priority   : byte;
                                       priv_group      : byte;
                                       priv_authdata   : tHSRPAuthData;
                                       priv_virtualip  : tIPv4Address;

                                       // Advertisement stuff

                                       priv_AdvTLVType           : word;
                                       priv_AdvTLVLength         : word;
                                       priv_AdvActiveGroupCount  : word;
                                       priv_AdvPassiveGroupCount : word;

                                protected
                                         // called from base class from GetAssembly

                                         procedure DetermineAssemblySize; override;
                                         procedure Assemble; override;

                                public
                                      property Version : byte read priv_version write priv_version;
                                      property OpCode : tHSRPOpCode read priv_opcode write priv_opcode;
                                      property State : tHSRPState read priv_state write priv_state;
                                      property HelloTime : byte read priv_hellotime write priv_hellotime;
                                      property HoldTime : byte read priv_holdtime write priv_holdtime;
                                      property Priority : byte read priv_priority write priv_priority;
                                      property AuthData : tHSRPAuthData read priv_authdata write priv_authdata;
                                      property VirtualIP : tIPv4Address read priv_virtualip write priv_virtualip;

                                      property AdvTLVType : word read priv_AdvTLVType;
                                      property AdvTLVLength : word read priv_AdvTLVLength;
                                      property AdvActiveGroupCount : word read priv_AdvActiveGroupCount;
                                      property AdvPassiveGroupCount : word read priv_AdvPassiveGroupCount;

                                      procedure CopyFromParser (Parser : tHSRPParser);
                           end;


implementation

uses toolkit;

(* ------------------------------------------- *)
{ tHSRPAssembler }
(* ------------------------------------------- *)

constructor tHSRPAssembler.create;

begin
     priv_version := 0;
     priv_opcode := hoUnknown;
     priv_state := hsUnknown;
     priv_hellotime := 0;
     priv_holdtime := 0;
     priv_priority := 0;
     priv_group := 0;
     priv_authdata := DefaultAuthData;
     priv_virtualip.initialize;
end;

(* ------------------------------------------- *)

procedure tHSRPAssembler.DetermineAssemblySize;

begin
     if priv_opcode = hoAdvertise then
        prot_length := sizeOf (tHSRPAdvStruct)
     else
         prot_length := sizeOf (tHSRPStruct);
end;

(* ------------------------------------------- *)

procedure tHSRPAssembler.Assemble;

var pData     : pHSRPStruct;
    pAdvData  : pHSRPAdvStruct;
    help      : word;
    help2     : word;
    pLongHelp : pLongword;
    LongHelp  : Longword;
    offset    : integer;

begin
     if not (priv_opcode = hoAdvertise) then
        begin
             pData := pHSRPStruct (prot_data);

             pData^.Version := priv_version;

             case priv_opcode of
              hoHello     : pData^.OpCode := 0;
              hoCoup      : pData^.OpCode := 1;
              hoResign    : pData^.OpCode := 2;
              hoAdvertise : pData^.OpCode := 3;
              hoUnknown   : pData^.OpCode := 0;
             end;

             case priv_state of
              hsInitial : pData^.State := 0;
              hsLearn   : pData^.State := 1;
              hsListen  : pData^.State := 2;
              hsSpeak   : pData^.State := 4;
              hsStandby : pData^.State := 8;
              hsActive  : pData^.State := 16;
              hsUnknown : pData^.State := 0;
             end;

             pData^.HelloTime := priv_hellotime;
             pData^.HoldTime := priv_holdtime;
             pData^.Priority := priv_priority;
             pData^.Group := priv_group;
             pData^.Reserved := 0;
             pData^.AuthData := priv_authdata;
             pData^.VirtualIP := swap32 (priv_virtualip.ValueAsLong);
        end
     else
         begin
              pAdvData := pHSRPAdvStruct (prot_data);

              pAdvData^.Version := priv_version;
              pAdvData^.OpCode := 3;

              pAdvData^.TLVType := swap16 (priv_AdvTLVType);
              pAdvData^.TLVLength := swap16 (priv_AdvTLVLength);

              case priv_state of
               hsInitial : pAdvData^.State := 0;
               hsLearn   : pAdvData^.State := 1;
               hsListen  : pAdvData^.State := 2;
               hsSpeak   : pAdvData^.State := 4;
               hsStandby : pAdvData^.State := 8;
               hsActive  : pAdvData^.State := 16;
               hsUnknown : pAdvData^.State := 0;
              end;

              pAdvData^.Reserved1 := 0;
              pAdvData^.ActiveGroupCount := swap16 (priv_AdvActiveGroupCount);
              pAdvData^.PassiveGroupCount := swap16 (priv_AdvPassiveGroupCount);
              pAdvData^.Reserved2 := 0;
         end;
end;

(* ------------------------------------------- *)

procedure tHSRPAssembler.CopyFromParser (Parser: tHSRPParser);

begin
     priv_version := Parser.Version;
     priv_opcode := Parser.OpCode;
     priv_state := Parser.State;
     priv_hellotime := Parser.HelloTime;
     priv_holdtime := Parser.HoldTime;
     priv_priority := Parser.Priority;
     priv_group := Parser.Group;
     priv_authdata := Parser.AuthData;
     priv_virtualip := Parser.VirtualIP;

     priv_AdvTLVType := Parser.AdvTLVType;
     priv_AdvTLVLength := Parser.AdvTLVLength;
     priv_AdvActiveGroupCount := Parser.AdvActiveGroupCount;
     priv_AdvPassiveGroupCount := Parser.AdvPassiveGroupCount;
end;

(* ------------------------------------------- *)

end.
