{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_assembly_netflow;

interface

uses base_structures,
     class_parser_netflow,
     class_assembly_baseclass;


type tNetFlowAssembler = class (tFrameAssembler)

                                constructor create;

                                private
                                       priv_NetFlowVersion       : word;
                                       priv_NetFlowV5Header      : tNetFlowV5Header;
                                       priv_NetFlowV5Records     : tNetFlowV5Array;
                                       priv_NetFlowV5RecordCount : integer;

                                       procedure SetNetFlowVersion (Const Version : word);

                                protected
                                         // called from base class from GetAssembly

                                         procedure DetermineAssemblySize; override;
                                         procedure Assemble; override;

                                public
                                      property Version : word read priv_NetFlowVersion write SetNetFlowVersion;
                                      property NetFlowV5Header : tNetFlowV5Header read priv_NetFlowV5Header;
                                      property NetFlowV5Records : tNetFlowV5Array read priv_NetFlowV5Records;

                                      procedure CopyFromParser (Parser : tNetFlowParser);
                           end;


implementation

uses toolkit,
     definition_tcp_structures;

(* ------------------------------------------- *)
{ tNetFlowAssembler }
(* ------------------------------------------- *)

constructor tNetFlowAssembler.create;

begin
     inherited;

     priv_NetFlowVersion := 0;
     priv_NetFlowV5Header.initalize;
     setLength (priv_NetFlowV5Records, 0);
     priv_NetFlowV5RecordCount := 0;
end;

(* ------------------------------------------- *)

procedure tNetFlowAssembler.DetermineAssemblySize;

begin
     case Version of
      5 : begin
               prot_length := sizeOf (tNetFlowV5HdrStruct) + priv_NetFlowV5RecordCount * sizeOf (tNetFlowV5RecordStruct);
          end;
     end;
end;

(* ------------------------------------------- *)

procedure tNetFlowAssembler.SetNetFlowVersion (const Version: word);

begin
     assert (version = 5, 'cannot handle netflow versions other than 5, sorry (tNetFlowAssembler.SetNetFlowVersion)');

     priv_NetFlowVersion := Version;
end;

(* ------------------------------------------- *)

procedure tNetFlowAssembler.Assemble;

var pV5Hdr    : pNetFlowV5HdrStruct;
    pV5Rec    : pNetFlowV5RecordStruct;
    entry     : tNetFlowV5Record;
    help      : word;
    help2     : word;
    pLongHelp : pLongword;
    LongHelp  : Longword;
    offset    : integer;
    i         : integer;

begin
     offset := 0;

     pV5Hdr := pNetFlowV5HdrStruct (prot_data);

     // Header

     pV5Hdr^.version := swap16 (priv_NetFlowV5Header.Version);
     pV5Hdr^.count := swap16 (priv_NetFlowV5Header.RecordCount);
     pV5Hdr^.sys_uptime := swap32 (priv_NetFlowV5Header.SystemUptime);
     pV5Hdr^.unix_secs := swap32 (priv_NetFlowV5Header.UnixSeconds);
     pV5Hdr^.unix_nsecs := swap32 (priv_NetFlowV5Header.UnixNanoSeconds);
     pV5Hdr^.flow_sequence := swap32 (priv_NetFlowV5Header.FlowSequence);
     pV5Hdr^.engine_type := priv_NetFlowV5Header.EngineType;
     pV5Hdr^.engine_id := priv_NetFlowV5Header.EngineID;

     case priv_NetFlowV5Header.SamplingMode of
      nfsmNone           : help := 0;
      nfsmPacketInterval : help := $4000;
      nfsmRandom         : help := $8000;
     end;

     help := help or priv_NetFlowV5Header.SamplingInterval;

     pV5Hdr^.sampling_interval := swap16 (help);

     // Records

     offset := sizeOf (tNetFlowV5HdrStruct);

     for i := 1 to priv_NetFlowV5RecordCount do
         begin
              entry := priv_NetFlowV5Records[i-1];

              pV5Rec := pNetFlowV5RecordStruct (prot_data + offset);

              pV5Rec^.sourceaddr := swap32 (entry.SourceAddress.ValueAsLong);
              pV5Rec^.destaddr := swap32 (entry.DestinationAddress.ValueAsLong);
              pV5Rec^.nexthop := swap32 (entry.NextHop.ValueAsLong);
              pV5Rec^.input := swap16 (entry.SNMPInputInterfaceId);
              pV5Rec^.output := swap16 (entry.SNMPOutputInterfaceId);
              pV5Rec^.dpkts := swap32 (entry.FlowPacketCount);
              pV5Rec^.doctets := swap32 (entry.FlowPacketOctets);
              pV5Rec^.first := swap32 (entry.FlowFirstSysUptime);
              pV5Rec^.last := swap32 (entry.FlowLastSysUptime);
              pV5Rec^.srcport := swap16 (entry.SourcePort);
              pV5Rec^.dstport := swap16 (entry.DestinationPort);
              pV5Rec^.pad1 := 0;
              pV5Rec^.tcp_flags := getWordFromTCPFlagSet (entry.TCPFlags);
              pV5Rec^.prot := entry.IPProtocolType;
              pV5Rec^.tos := entry.IPToS;
              pV5Rec^.src_as := swap16 (entry.SourceAS);
              pV5Rec^.dst_as := swap16 (entry.DestinationAS);
              pV5Rec^.src_mask := entry.SourceMask;
              pV5Rec^.dst_mask := entry.DestinationMask;
              pV5Rec^.pad2 := 0;

              inc (offset, sizeOf (tNetFlowV5RecordStruct));
         end;

     assert (offset = prot_length, 'Memory offset check mismatch (tNetFlowAssembler.Assemble)');
end;

(* ------------------------------------------- *)

procedure tNetFlowAssembler.CopyFromParser (Parser: tNetFlowParser);

var i : integer;

begin
     SetNetFlowVersion (Parser.NetFlowVersion);

     case priv_NetFlowVersion of
       5 : begin
                priv_NetFlowV5Header := Parser.NetFlowV5Header;

                priv_NetFlowV5RecordCount := length (Parser.NetFlowV5Records);
                setLength (priv_NetFlowV5Records, priv_NetFlowV5RecordCount);

                for i := 1 to priv_NetFlowV5RecordCount do
                    priv_NetFlowV5Records[i-1] := Parser.NetFlowV5Records[i-1];

           end;
     end;
end;

(* ------------------------------------------- *)

end.
