{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_parse_csvfilterfile;

interface

uses base_structures,
     definition_ip_addresses,
     definition_filter_structures;

type tCSVFilterParseMode = (cpmNone, cpmSingleIP, cpmIPPair, cpmSingleSocket, cpmSocketPair, cpm5Tuple, cpmAuto);

     tCSVFilterFile = class (tObject)

                            class function getFilterParseMode (const Filename : string; const CSVSeparator : string = ',') : tCSVFilterParseMode;

                            constructor create (const Filename : string; const ParseMode : tCSVFilterParseMode; const CSVSeparator : string = ',');

                            private
                                   priv_filename      : string;
                                   priv_parsemode     : tCSVFilterParseMode;
                                   priv_separator     : string;
                                   priv_filterlist    : tFilterSet;

                                   function getSingleIPFilter (const sLine : string) : tFrameFilter;
                                   function getIPPairFilter (const sLine : string) : tFrameFilter;
                                   function getSingleSocketFilter (const sLine : string) : tFrameFilter;
                                   function getSocketPairFilter (const sLine : string) : tFrameFilter;
                                   function getTupleFilter (const sLine : string) : tFrameFilter;

                            public
                                  property FilterList : tFilterSet read priv_filterlist;
                                  property CSVSeparator : string read priv_separator write priv_separator;

                                  procedure clear;
                                  procedure execute;

                                  destructor Destroy; override;
                      end;


implementation

uses SysUtils,
     Toolkit;

(* ----------------------------------- *)
{ tCSVFilterFile }
(* ----------------------------------- *)

class function tCSVFilterFile.getFilterParseMode (const Filename: string; const CSVSeparator : string = ',') : tCSVFilterParseMode;

var fHandle : textfile;
    sLine   : string;
    count   : Integer;
    mode    : tCSVFilterParseMode;
    help1   : string;
    help2   : string;
    help3   : integer;
    help4   : integer;
    dummy   : integer;

begin
     mode := cpmNone;

     // read the first line

     assignfile (fHandle, Filename);
     reset (fHandle);
     readln (fHandle, sLine);
     sLine := trim (sLine);
     closefile (fHandle);

     // determine if we can use it

     count := numEntries (sLine, CSVSeparator);

     case count of
      1 : begin
               // Single IP?

               if isValidIPv4String (sLine) or isValidIPv6String (sLine) then
                  mode := cpmSingleIP;
          end;
      2 : begin
               // IP Pair or Single Socket

               help1 := getEntry (sLine, CSVSeparator, 1);
               help2 := getEntry (sLine, CSVSeparator, 2);

               if (isValidIPv4String (help1) and isValidIPv4String (help2)) or (isValidIPv6String (help1) and isValidIPv6String (help2)) then
                  mode := cpmIPPair
               else
                   begin
                        val (help2, help3, dummy);

                        if (isValidIPv4String (help1) or isValidIPv6String (help1)) and ((help3 > 0) and (help3 < 65536)) then
                           mode := cpmSingleSocket;
                   end;
          end;
      4,
      5 : begin
               // Socket Pairs or a tuple

               help1 := getEntry (sLine, CSVSeparator, 1);
               help2 := getEntry (sLine, CSVSeparator, 3);

               if (isValidIPv4String (help1) and isValidIPv4String (help2)) or (isValidIPv6String (help1) and isValidIPv6String (help2)) then
                  begin
                       help1 := getEntry (sLine, CSVSeparator, 2);
                       help2 := getEntry (sLine, CSVSeparator, 4);

                       val (help1, help3, dummy);
                       val (help2, help4, dummy);

                       if ((help3 > 0) and (help3 < 65536)) and ((help4 > 0) and (help4 < 65536)) then
                          begin
                               // it may still be a 5-tuple

                               if count = 5 then
                                  begin
                                       help1 := getEntry (sLine, CSVSeparator, 5);

                                       if StrToLayer4Protocol (help1) <> l4Unknown then
                                          mode := cpm5Tuple;
                                  end
                               else
                                   mode := cpmSocketPair;
                          end;
                  end;
          end;
      else
          mode := cpmNone;
     end;

     result := mode;
end;

(* ----------------------------------- *)

constructor tCSVFilterFile.create (const Filename: string; const ParseMode : tCSVFilterParseMode; const CSVSeparator : string = ',');

begin
     inherited create;

     priv_filename := Filename;
     priv_separator := CSVSeparator;

     // For Drag&Drop the type is set to auto, otherwise the file dialog has already determined the type

     if ParseMode = cpmAuto then
        priv_parsemode := getFilterParseMode (priv_filename)
     else
         priv_parsemode := ParseMode;

     clear;

//   setLength (priv_filterlist, 0);
     priv_filterlist.create;
end;

(* ----------------------------------- *)

destructor tCSVFilterFile.Destroy;

begin
     clear;
     priv_filterlist.Free;

     inherited;
end;

(* ----------------------------------- *)

procedure tCSVFilterFile.clear;

begin
//   setLength (priv_filterlist, 0);
     priv_filterlist.Clear;
end;

(* ----------------------------------- *)

function tCSVFilterFile.getSingleIPFilter (const sLine: string): tFrameFilter;

var IPv4Address : tIPv4Address;
    IPv6Address : tIPv6Address;
    filter      : tFrameFilter;

begin
     filter.initialize;
     IPv4Address.initialize;
     IPv6Address.initialize;

     if isValidIPv4String (sLine) then
        begin
             IPv4Address.setValueFromString (sLine);

             filter.IPv4Filter.initialize (IPv4Address);
        end
     else
         if isValidIPv6String (sLine) then
            begin
                IPv6Address.setValueFromString (sLine);

                filter.IPv6Filter.initialize (IPv6Address);
            end;

     result := filter;
end;

(* ----------------------------------- *)

function tCSVFilterFile.getIPPairFilter (const sLine: string): tFrameFilter;

var IPv4AddressPair : tIPv4AddressPair;
    IPv6AddressPair : tIPv6AddressPair;
    filter          : tFrameFilter;
    help            : string;

begin
     filter.initialize;
     IPv4AddressPair.initialize;
     IPv6AddressPair.initialize;

     if numEntries (sLine, priv_separator) = 2 then
        begin
             help := getEntry (sLine, priv_separator, 1);

             if isValidIPv4String (help) then
                begin
                     // Seems like we create an IPv4 pair

                     IPv4AddressPair.Source.setValueFromString (help);

                     help := getEntry (sLine, priv_separator, 2);

                     if isValidIPv4String (help) then
                        begin
                             IPv4AddressPair.Destination.setValueFromString (help);

                             filter.IPv4Filter.initialize (IPv4AddressPair);
                        end;
                end
             else
                 if isValidIPv6String (help) then
                    begin
                         // Seems like we create an IPv4 pair

                         IPv6AddressPair.Source.setValueFromString (help);

                         help := getEntry (sLine, priv_separator, 2);

                         if isValidIPv6String (help) then
                            begin
                                 IPv6AddressPair.Destination.setValueFromString (help);

                                 filter.IPv6Filter.initialize (IPv6AddressPair);
                            end;
                    end;
        end;

     result := filter;
end;

(* ----------------------------------- *)

function tCSVFilterFile.getSingleSocketFilter (const sLine: string): tFrameFilter;

var IPv4Socket : tIPv4Socket;
    IPv6Socket : tIPv6Socket;
    help        : string;
    dummy       : integer;
    filter      : tFrameFilter;

begin
     filter.initialize;
     IPv4Socket.initialize;
     IPv6Socket.initialize;

     if numEntries (sLine, priv_separator) = 2 then
        begin
             help := getEntry (sLine, priv_separator, 1);

             if isValidIPv4String (help) then
                begin
                     // Seems like we create an IPv4 Socket

                     IPv4Socket.IPAddress.setValueFromString (help);

                     help := getEntry (sLine, priv_separator, 2);
                     val (help, IPv4Socket.Port, dummy);

                     if IPv4Socket.Port > 0 then
                        begin
                             filter.IPv4Filter.initialize (IPv4Socket.IPAddress);
                             filter.PortFilter.initialize (IPv4Socket.Port);
                        end;
                end
             else
                 if isValidIPv6String (help) then
                    begin
                         // Seems like we create an IPv4 Socket

                         IPv6Socket.IPAddress.setValueFromString (help);

                         help := getEntry (sLine, priv_separator, 2);
                         val (help, IPv6Socket.Port, dummy);

                         if IPv6Socket.Port > 0 then
                            begin
                                 filter.IPv6Filter.initialize (IPv6Socket.IPAddress);
                                 filter.PortFilter.initialize (IPv6Socket.Port);
                            end;
                    end;
        end;

     result := filter;
end;

(* ----------------------------------- *)

function tCSVFilterFile.getSocketPairFilter (const sLine: string): tFrameFilter;

var IPv4SocketPair : tIPv4SocketPair;
    IPv6SocketPair : tIPv6Socketpair;
    help           : string;
    dummy          : integer;
    filter         : tFrameFilter;

begin
     filter.initialize;
     IPv4SocketPair.initialize;
     IPv6SocketPair.initialize;

     if numEntries (sLine, priv_separator) = 4 then
        begin
             help := getEntry (sLine, priv_separator, 1);

             if isValidIPv4String (help) then
                begin
                     // Seems like we create an IPv4 Socket

                     IPv4SocketPair.Source.IPAddress.setValueFromString (help);

                     help := getEntry (sLine, priv_separator, 2);
                     val (help, IPv4SocketPair.Source.Port, dummy);

                     help := getEntry (sLine, priv_separator, 3);

                     if isValidIPv4String (help) then
                        begin
                             IPv4SocketPair.Destination.IPAddress.setValueFromString (help);

                             help := getEntry (sLine, priv_separator, 4);
                             val (help, IPv4SocketPair.Destination.Port, dummy);
                        end;

                     // if both ports are greater than zero, this seems to be a valid socket pair

                     if (IPv4SocketPair.Source.Port > 0) and (IPv4SocketPair.Destination.Port > 0) then
                        begin
                             filter.IPv4Filter.initialize (IPv4SocketPair);
                             filter.PortFilter.initialize (IPv4SocketPair, l4Any);
                        end;
                end
             else
                 if isValidIPv6String (help) then
                    begin
                         // Seems like we create an IPv6 Socket

                         IPv6SocketPair.Source.IPAddress.setValueFromString (help);

                         help := getEntry (sLine, priv_separator, 2);
                         val (help, IPv6SocketPair.Source.Port, dummy);

                         help := getEntry (sLine, priv_separator, 3);

                         if isValidIPv6String (help) then
                            begin
                                 IPv6SocketPair.Destination.IPAddress.setValueFromString (help);

                                 help := getEntry (sLine, priv_separator, 4);
                                 val (help, IPv6SocketPair.Destination.Port, dummy);
                            end;

                         // if both ports are greater than zero, this seems to be a valid socket pair

                         if (IPv6SocketPair.Source.Port > 0) and (IPv6SocketPair.Destination.Port > 0) then
                            begin
                                 filter.IPv6Filter.initialize (IPv6SocketPair);
                                 filter.PortFilter.initialize (IPv6SocketPair, l4Any);
                            end;
                    end;
        end;

     result := filter;
end;

(* ----------------------------------- *)

function tCSVFilterFile.getTupleFilter (const sLine: string): tFrameFilter;

var IPv4SocketPair : tIPv4SocketPair;
    IPv6SocketPair : tIPv6Socketpair;
    L4Prot         : tLayer4Protocol;
    help           : string;
    dummy          : integer;
    filter         : tFrameFilter;

begin
     filter.initialize;
     IPv4SocketPair.initialize;
     IPv6SocketPair.initialize;

     if numEntries (sLine, priv_separator) = 5 then
        begin
             help := getEntry (sLine, priv_separator, 1);

             if isValidIPv4String (help) then
                begin
                     // Seems like we create an IPv4 Socket

                     IPv4SocketPair.Source.IPAddress.setValueFromString (help);

                     help := getEntry (sLine, priv_separator, 2);
                     val (help, IPv4SocketPair.Source.Port, dummy);

                     help := getEntry (sLine, priv_separator, 3);

                     if isValidIPv4String (help) then
                        begin
                             IPv4SocketPair.Destination.IPAddress.setValueFromString (help);

                             help := getEntry (sLine, priv_separator, 4);
                             val (help, IPv4SocketPair.Destination.Port, dummy);
                        end;

                     help := lowercase (getEntry (sLine, priv_separator, 5));
                     L4Prot := StrToLayer4Protocol (help);

                     // if both ports are greater than zero and we have a valid layer 4 protocol, this seems to be a valid tuple

                     if (IPv4SocketPair.Source.Port > 0) and (IPv4SocketPair.Destination.Port > 0) and (L4Prot <> l4unknown)  then
                        begin
                             filter.IPv4Filter.initialize (IPv4SocketPair);
                             filter.PortFilter.initialize (IPv4SocketPair, L4Prot);
                        end;
                end
             else
                 if isValidIPv6String (help) then
                    begin
                         // Seems like we create an IPv6 Socket

                         IPv6SocketPair.Source.IPAddress.setValueFromString (help);

                         help := getEntry (sLine, priv_separator, 2);
                         val (help, IPv6SocketPair.Source.Port, dummy);

                         help := getEntry (sLine, priv_separator, 3);

                         if isValidIPv6String (help) then
                            begin
                                 IPv6SocketPair.Destination.IPAddress.setValueFromString (help);

                                 help := getEntry (sLine, priv_separator, 4);
                                 val (help, IPv6SocketPair.Destination.Port, dummy);
                            end;

                         help := lowercase (getEntry (sLine, priv_separator, 5));
                         L4Prot := StrToLayer4Protocol (help);

                         // if both ports are greater than zero and we have a valid layer 4 protocol, this seems to be a valid tuple

                         if (IPv6SocketPair.Source.Port > 0) and (IPv6SocketPair.Destination.Port > 0) and (L4Prot <> l4unknown)  then
                            begin
                                 filter.IPv6Filter.initialize (IPv6SocketPair);
                                 filter.PortFilter.initialize (IPv6SocketPair, L4Prot);
                            end;
                    end;
        end;

     result := filter;
end;

(* ----------------------------------- *)

procedure tCSVFilterFile.execute;

var fHandle : textfile;
    sLine   : string;
    filter  : tFrameFilter;
    i       : Integer;

begin
     assignfile (fHandle, priv_filename);
     reset (fHandle);

     i := 0;

     while not eof (fHandle) do
           begin
                inc (i);
                readln (fHandle, sLine);
                sLine := trim (sLine);

                if pos ('#', sLine) <> 1 then
                   begin
                        case priv_parsemode of
                         cpmSingleIP     : filter := getSingleIPFilter (sLine);
                         cpmIPPair       : filter := getIPPairFilter (sLine);
                         cpmSingleSocket : filter := getSingleSocketFilter (sLine);
                         cpmSocketPair   : filter := getSocketPairFilter (sLine);
                         cpm5Tuple       : filter := getTupleFilter (sLine);
                        end;
                   end;

                if not filter.isUnspecified then
//                 mergeFilterToSet (Filter, priv_filterlist);
                   priv_filterlist.AddFilter (filter);
           end;

     closefile (fHandle);
end;

(* ----------------------------------- *)

end.
