{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_rawframe;

interface

uses SysUtils,
     Classes,
     IdHashMessageDigest,
     IdHash,
     base_structures;

const NoInitRawBufferDefault = true;

type tRawFrame    = class (tObject)

                          constructor create (const TimeDetails : tTimingDetails; const WireDetails : tWireDetails; const FilePosition : tFileFrameLocation; const FrameBuffer : tRawBuffer; const DoNotInitRawBuffer : boolean = NoInitRawBufferDefault); overload;
                          constructor create (const CopyFromFrame : tRawFrame; const ForceWireSize : longword = 0; const ForceCaptureSize : longword = 0); overload;
//                        procedure free;

                          private
                                 iOriginalTracefileIndex  : integer;         // Index of the Tracefile (the actual list is kept in a higher structure, cannot change)
                                 iFileFrameIndex          : uint64;          // Index of the frame in the file
                                 iFileOffset              : int64;           // Offset of the Filehandle where the frame starts

                                 priv_donotinitrawbuffer  : boolean;         // If true, the tRawBuffer of the "Buffer" property is not zeroed
                                 priv_InterfaceLinkID     : tGUID;

                                 priv_filterlabel         : string;          // optional: if frame was matched by a filter, this can carry the label of that filter

                                 procedure initialize;
                                 function TicksToTime     : tDeltaTickTime;   // gives the *absolute* time but the *delta* days to the trace file start
                                 function getFrameBuffer  : tRawBuffer;       // Returns the actual frame data as fixed size array
                                 function getCRC32        : longword;         // Calculates the CRC32 checksum if not already contained in lwCRC32
                                 function getMD5          : tMD5HashString;   // Calculates the MD5 checksum if not already contained in prot_MD5Hash;
                                 function getFilePosition : tFileFrameLocation;    // Returns the file index, frame index and offset as a combined set

                          strict protected
                                 iTracefileIndex  : integer;                  // Index of the Tracefile (the actual list is kept in a higher structure). Can be changed when renumbering files
                                 lwLinkType       : longword;                 // Should usually be Ethernet
                                 wCaptureSize     : longword;                 // Size of the frame
                                 wWireSize        : longword;                 // Size on the wire - can be larger than iFrameSize if it was sliced
                                 i64Ticks         : int64;                    // Microseconds since midnight
                                 bBufferFlags     : byte;                     // Buffer Flags (whatever they do)
                                 FrameTimestamp   : tFrameTimeStamp;          // Absolute Time downto nanoseconds
                                 InterfaceIndex   : longword;                 // Interface index of the capture card the frame was recorded on
                                 priv_FrameFlags  : tFrameFlags;              // Direction of packet, Error Indicator etc., see PCAPng Specs
                                 iDropped         : uInt64;                   // Specifies the number of packets lost between this packet and the preceding one.
                                 FrameType        : tTracefileFrameType;      // Specifies the origin file type (enc, cap, pcap, pcap-ng etc.)

                                 pFrameData       : pointer;                  // Dynamic Storage for the actual frame data

                                 lwCRC32          : longword;                 // CRC32 calculated for the frame data, internal use only
                                 prot_MD5Hash     : tMD5HashString;           // MD5 calculated for the frame data, internal use only
                                 priv_commentlist : tStringlist;              // List of comments for the frame, usually none or one.

                                /// <summary>
                                ///   Copies an existing frame into a new frame
                                ///   which must have been created before the
                                ///   copy is invoked. If the destination frame
                                ///   already has frame bytes allocated they
                                ///   will be freed before adding the new bytes
                                /// </summary>
                                /// <param name="DestinationFrame">
                                ///   existing RawFrame object
                                /// </param>
                                /// <param name="ForceWireSize">
                                ///   if zero the original wiresize is
                                ///   transferred, otherwise the specified
                                ///   value
                                /// </param>
                                procedure copyTo (DestinationFrame : tRawFrame; const ForceWireSize : longword = 0; const ForceCaptureSize : longword = 0);

                          public
                                property Buffer : tRawBuffer read getFrameBuffer;

                                property Ticks : int64 read i64Ticks;
                                property TracefileID : integer read iTracefileIndex;
                                property FileFrameID : uint64 read iFileFrameIndex;
                                property FilePosition : tFileFrameLocation read getFilePosition;
                                property FrameSize : longword read wCaptureSize;
                                property WireSize : longword read wWireSize write wWiresize;                    // Writeable to allow fixing bad wiresize values in edit tasks
                                property FileOffset : int64 read iFileOffset;
                                property TickTime : tDeltaTickTime read TicksToTime;
                                property Timestamp : tFrameTimeStamp read FrameTimeStamp;
                                property LinkType : longword read lwLinkType;
                                property InterfaceID : longword read InterfaceIndex write InterfaceIndex;       // Only exception with r/w property, because a frame may need to have the interface index adjusted when writing it back out
                                property FrameFlags : tFrameFlags read priv_frameflags;
                                property CRC32 : longword read getCRC32;
                                property MD5Hash : tMD5HashString read getMD5;
                                property FrameFileType : tTracefileFrameType read FrameType write FrameType;
                                property DropCount : uint64 read iDropped;
                                property Data : pointer read pFrameData;
                                property Commentlist : tStringlist read priv_commentlist write priv_commentlist;
                                property InterfaceLinkID : tGUID read priv_InterfaceLinkID write priv_InterfaceLinkID;
                                property FilterLabel : string read priv_filterlabel write priv_filterlabel;

                                function FrameDataIsEqualTo (CompareFrame : tRawFrame) : boolean;
                                function isEqualTo (CompareFrame : tRawFrame) : boolean;
                                /// <summary>
                                ///   Returns part of the frame, starting from
                                ///   the offset specified and containing the
                                ///   amount of bytes specified in length, or
                                ///   less if not enough bytes present. This
                                ///   creates a new framechunk that needs to be
                                ///   discarded by the calling process later. <br />
                                /// </summary>
                                /// <param name="FromOffset">
                                ///   Offset from which the data copy starts
                                /// </param>
                                /// <param name="Length">
                                ///   amount of bytes to return, or all
                                ///   available bytes starting from the offset
                                ///   if set to zero
                                /// </param>
                                function FrameDataAsChunk (const FromOffset: integer = 0; const Length : integer = 0) : tFrameChunk;

                                procedure DiscardFrameData;
                                destructor Destroy; override;

                                // ***** WARNING: If changing something in the RawFrame Declaration you need to modify the copyto routine as well!!! *****
                    end;


implementation

uses SiAuto,
     DateUtils,
     IdGlobal,
     Math,
     Toolkit,
     base_checksums;


(* ------------------------------------------- *)
{ tRawFrame }
(* ------------------------------------------- *)

constructor tRawFrame.create (const TimeDetails : tTimingDetails; const WireDetails : tWireDetails; const FilePosition : tFileFrameLocation; const FrameBuffer : tRawBuffer; const DoNotInitRawBuffer : boolean = NoInitRawBufferDefault);

var DeltaTime : tDeltaTickTime;
    HelpTime  : tDateTime;

begin
     inherited create;

     initialize;

     priv_donotinitrawbuffer := DoNotInitRawBuffer;
     priv_interfacelinkid := WireDetails.LinkID;                  // GUID that is unique per Interface object

     try
        // Assign Details relevant Tracefile details

        iOriginalTracefileIndex := FilePosition.FileIndex;
        iTracefileIndex := FilePosition.FileIndex;
        iFileFrameIndex := FilePosition.FileFrameIndex;
        iFileOffset := FilePosition.FileOffset;
     except
           On E : Exception do
              SiMain.LogError ('Error when assigning Tracefiledetails for Frame #' + IntToStr (Fileposition.FileIndex) + '/' + IntToStr (Fileposition.FileFrameIndex) + ': ' + E.Message + ' (tRawFrame.create)');
     end;

     try
        // Assign Capture details

        i64Ticks := TimeDetails.TraceTicks;
        lwLinkType := WireDetails.LinkType;
        wCaptureSize := min (maxFrameSize, WireDetails.CaptureSize);
        wWireSize := WireDetails.WireSize;
        priv_FrameFlags := WireDetails.FrameFlags;
        iDropped := WireDetails.Dropped;
        FrameType := WireDetails.FrameType;
        InterfaceIndex := WireDetails.InterfaceID;

        // Now lets calculate the absolute Frame Date/Time, first by setting it from a "normal" tDatetime because it's easy to handle

        if i64Ticks >= 0 then
           DeltaTime := TicksToTime
        else
            begin
                 DeltaTime.initialize;

                 SiMain.LogWarning('Encountered negative i64Ticks value: ' + IntToStr (i64Ticks));
            end;

        HelpTime := (incDay (TimeDetails.AbsoluteTime, DeltaTime.deltaDay));
        HelpTime := recodeTime (HelpTime, DeltaTime.iHour, DeltaTime.iMin, DeltaTime.iSec, 0);

        FrameTimestamp.datetime := HelpTime;

        // Now adjust the nanoseconds (it has to include microseconds, because Helptime is only accurate to milliseconds)

        FrameTimestamp.nanoseconds := 1000 * DeltaTime.iMicSec + TimeDetails.NanoTicks;
     except
           On E : Exception do
              SiMain.LogError ('Error when assigning Capture and Timing Details for Frame #' + IntToStr (Fileposition.FileIndex) + '/' + IntToStr (Fileposition.FileFrameIndex) + ': ' + E.Message + ' (tRawFrame.create)');
     end;


     try
        // New dynamic storage - takes only as many bytes as the frame is in size

        getMem (pFrameData, wCaptureSize);

        // To copy the data to a dynamic memory area we first need to move it to a static
        // array - in this case this has been done before calling this constructor to
        // get the tRawbuffer for the parameter list.
        // If we do not do this first and copy from pointer to pointer we will slam
        // into a wall as soon as we try to dispose the memory, because the move procedure
        // seems to align pointer locations instead of actually copying the content when called
        // with two pointer types. Took me 2 hours to figure this problem out. Sigh.

        move (FrameBuffer, pFrameData^, min (maxFrameSize, wCaptureSize));
     except
           On E : Exception do
              SiMain.LogError ('Error when reserving and copying bytes for Frame #' + IntToStr (Fileposition.FileIndex) + '/' + IntToStr (Fileposition.FileFrameIndex) + ': ' + E.Message + ' (tRawFrame.create)');
     end;
end;

(* ------------------------------------------- *)

constructor tRawFrame.create (const CopyFromFrame : tRawFrame; const ForceWireSize : longword = 0; const ForceCaptureSize : longword = 0);

begin
     inherited create;

     CopyFromFrame.copyTo (self, ForceWireSize, ForceCaptureSize);
end;

(* ------------------------------------------- *)

destructor tRawFrame.Destroy;

begin
     DiscardFrameData;

     FreeAndNil (priv_commentlist);

     inherited;
end;

(* ------------------------------------------- *)

procedure tRawFrame.initialize;

begin
     FrameType := tftUnknown;
     priv_commentlist := tStringlist.Create;

     pFrameData := nil;
     lwCRC32 := 0;
     prot_MD5Hash := '';
     priv_filterlabel := '';
end;

(* ------------------------------------------- *)

procedure tRawFrame.DiscardFrameData;

begin
     try
        freeMem (pFrameData, wCaptureSize);
        pFrameData := nil;
     except
           SiMain.LogFatal ('Freeing frame data failed (tRawFrame.DiscardFrameData)');
     end;
end;

(* ------------------------------------------- *)
{
procedure tRawFrame.free;

begin
     inherited;
end;
}
(* ------------------------------------------- *)

function tRawFrame.getCRC32: longword;

var crc : longword;

begin
     if lwCRC32 <> 0 then
        crc := lwCRC32
     else
         begin
              crc := getCRC32FromPointer (pFrameData, wCaptureSize - 4);

              crc := swap32 (crc);
         end;

     result := crc;
end;

(* ------------------------------------------- *)

function tRawFrame.getMD5: tMD5HashString;

var hashengine : tIdHashMessageDigest5;
    bBuf       : TIdBytes;
    p          : PByte;
    i          : Integer;

begin
     if prot_MD5Hash = '' then
        begin
             hashengine := tIdHashMessageDigest5.Create;
             setlength (bBuf, wCaptureSize);

             p := PByte(pFrameData);

             for i := 0 to wCaptureSize - 1 do
                 begin
                      bBuf[i] := p^;
                      inc (p);
                 end;

             prot_MD5Hash := hashengine.HashBytesAsHex (bBuf);

             hashengine.Free;
        end;

     result := prot_MD5Hash;
end;

(* ------------------------------------------- *)

function tRawFrame.getFilePosition: tFileFrameLocation;

begin
     result.FileIndex := iTracefileIndex;
     result.FileFrameIndex := iFileFrameIndex;
     result.FileOffset := iFileOffset;
end;

(* ------------------------------------------- *)

function tRawFrame.getFrameBuffer: tRawBuffer;

var temp : tRawBuffer;
    i    : integer;
    boOK : boolean;

begin
     // Clear the array first - this may be disabled IF and only IF there is 100% sure that
     // random content beyond the size of the actual frame is NEVER accessed!

     if not priv_donotinitrawbuffer then
        begin
{
             for i := 1 to maxFrameSize do
                 temp[i] := 0;
}
             InitializeRawBuffer (temp);
        end;

     if pFrameData <> nil then
        begin
             // transfer dynamic stored frame buffer to fixed size array

             try
                Move(pFrameData^, temp,  wCaptureSize);
             except
                   SiMain.LogFatal('Error while moving ' + IntToStr (wCaptureSize) + ' bytes from pFrameData to tRawBuffer in tRawFrame.getFrameBuffer');
             end;
        end
     else
         begin
              // TODO: exception
         end;

     result := temp;
end;

(* ------------------------------------------- *)

function tRawFrame.TicksToTime: tDeltaTickTime;

var temp        : Int64;

begin
     result.iMicSec  := i64Ticks mod 1000000;
     temp            := i64Ticks div 1000000;
     result.iSec     := temp mod 60;
     temp            := temp div 60;
     result.iMin     := temp mod 60;
     temp            := temp div 60;
     result.iHour    := temp mod 24;
     result.deltaDay := temp div 24;
end;

(* ------------------------------------------- *)

// This routine is used to copy a tRawFrame to another tRawFrame, because if you just
// assign them with ":=" Delphi simply aligns object pointers
// Destinationframe must have been created before copying to it.

procedure tRawFrame.copyTo (DestinationFrame: tRawFrame; const ForceWireSize : longword = 0; const ForceCaptureSize : longword = 0);

var i : integer;
    a,b : pByte;

begin
     DestinationFrame.iOriginalTracefileIndex := self.iOriginalTracefileIndex;
     DestinationFrame.iTracefileIndex := self.iTracefileIndex;
     DestinationFrame.iFileFrameIndex := self.iFileFrameIndex;
     DestinationFrame.iFileOffset := self.iFileOffset;

     DestinationFrame.lwLinkType := self.lwLinkType;

     if ForceCaptureSize = 0 then
        DestinationFrame.wCaptureSize := self.wCaptureSize
     else
         DestinationFrame.wCaptureSize := min (self.wCaptureSize, ForceCaptureSize);

     if ForceWireSize = 0 then
        DestinationFrame.wWireSize := self.wWireSize
     else
         DestinationFrame.wWireSize := ForceWireSize;

     DestinationFrame.i64Ticks := self.i64Ticks;

     DestinationFrame.priv_FrameFlags := self.priv_FrameFlags;
     DestinationFrame.bBufferFlags := self.bBufferFlags;

     DestinationFrame.FrameTimestamp := self.FrameTimestamp;

     DestinationFrame.InterfaceIndex := self.InterfaceIndex;
     DestinationFrame.InterfaceLinkID := self.InterfaceLinkID;

     DestinationFrame.iDropped := self.iDropped;
     DestinationFrame.FrameType := self.FrameType;

     DestinationFrame.Commentlist := tStringlist.Create;

     for i := 1 to self.priv_commentlist.Count do
         DestinationFrame.Commentlist.Add (self.priv_commentlist[i-1]);

     a := self.pFrameData;

     if (DestinationFrame.pFrameData <> nil) then
        freeAndNil (DestinationFrame.pFrameData);

     getMem (DestinationFrame.pFrameData, self.wCaptureSize);
     b := DestinationFrame.pFrameData;

     // TODO: this copy routine is stupid because its sloooooww.

     for i := 1 to DestinationFrame.wCaptureSize do
         begin
              b^ := a^;

              inc (a);
              inc (b);
         end;
end;

(* ------------------------------------------- *)

// This routine returns true if the frame data are identical

function tRawFrame.FrameDataIsEqualTo (CompareFrame: tRawFrame): boolean;

var i : integer;
    a,b : pByte;

begin
     result := (self.wCaptureSize = CompareFrame.wCaptureSize);

     if result then
        begin
             a := self.pFrameData;
             b := CompareFrame.pFrameData;

             i := 0;

             while result and (i < self.wCaptureSize) do
                   begin
                        inc (i);

                        if not (a^ = b^) then
                           result := false;

                        // Move Pointer if bytes are left

                        if i < self.FrameSize then
                           begin
                                inc (a);
                                inc (b);
                           end;
                   end;
        end;
end;

(* ------------------------------------------- *)

// This routine compares complete RawFrames including all meta data

function tRawFrame.isEqualTo (CompareFrame: tRawFrame) : boolean;

begin
     result := (CompareFrame.InterfaceLinkID = self.InterfaceLinkID) and
               (CompareFrame.iTracefileIndex = self.iTracefileIndex) and
               (CompareFrame.iFileFrameIndex = self.iFileFrameIndex) and
               (CompareFrame.iFileOffset = self.iFileOffset) and
               (CompareFrame.wCaptureSize = self.wCaptureSize) and
               (CompareFrame.wWireSize = self.wWireSize) and
               (CompareFrame.i64Ticks = self.i64Ticks) and
               (CompareFrame.priv_FrameFlags.Value = self.priv_FrameFlags.Value) and
               (CompareFrame.bBufferFlags = self.bBufferFlags) and
               (CompareFrame.FrameTimestamp = self.FrameTimestamp) and
               (CompareFrame.InterfaceIndex = self.InterfaceIndex) and
               (CompareFrame.iDropped = self.iDropped) and
               (CompareFrame.FrameType = self.FrameType) and
               self.FrameDataIsEqualTo(CompareFrame);
end;

(* ------------------------------------------- *)

function tRawFrame.FrameDataAsChunk (const FromOffset: integer = 0; const Length : integer = 0) : tFrameChunk;

var Chunk : tFrameChunk;
    pHelp : pByte;

begin
     if Length = 0 then
        Chunk.Len := wCaptureSize - FromOffset
     else
         Chunk.Len := min (Length, wCaptureSize - FromOffset);

     if Chunk.Len <= 0 then
        begin
             Chunk.Len := 0;
             Chunk.Data := nil
        end
     else
         begin
              pHelp := pByte (pFrameData) + FromOffset;

              // Reserve memory for chunk and copy

              getmem (Chunk.Data, Chunk.Len);
              move (pHelp^, Chunk.Data^, Chunk.Len);
         end;

     result := Chunk;
end;

(* ------------------------------------------- *)

end.
