{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_timezonedetails;

interface

uses Windows,
     Contnrs;

type tTimeZoneDetails  = class (tObject)

                               constructor create (Year : word);
                               procedure free;

                               private
                                      priv_TZInfo        : TIME_ZONE_INFORMATION;

                                      priv_year          : word;
                                      priv_daylightstart : tDateTime;
                                      priv_standardstart : tDateTime;
                                      priv_systemisUTC   : boolean;
                                      priv_noDST         : boolean;

                                      function GetBias: integer;
                                      function GetDaylightBias: integer;

                                      function GetTzStandardDate(const year : word; const TzInfo: TTimeZoneInformation): TDateTime;
                                      function GetTzDaylightDate(const Year : word; TzInfo: TTimeZoneInformation): TDateTime;
                                      function GetStandardBias: integer;

                               public
                                     property Year : word read priv_year;
                                     property DaylightStart : tDateTime read priv_daylightstart;
                                     property StandardStart : tDateTime read priv_standardstart;
                                     property Bias : integer read GetBias;
                                     property StandardBias : integer read GetStandardBias;
                                     property DaylightBias : integer read GetDaylightBias;
                                     property SystemIsOnUTC : boolean read priv_SystemIsUTC;
                                     property SystemHasNoDST : boolean read priv_noDST;
                         end;

     tTimeZoneHandler  = class (tObject)

                               constructor create;
                               procedure free;

                               private
                                      priv_detaillist : tObjectlist;
                                      currententry    : tTimeZoneDetails;
                               public
                                     function GetBiasForDate (DateToTest : tDateTime) : integer;
                         end;

implementation

Uses SysUtils,
     DateUtils;

(* ----------------------------------- *)
{ tTimeZoneDetails }
(* ----------------------------------- *)

constructor tTimeZoneDetails.create(Year: word);

begin
     inherited create;

     priv_year := Year;
     GetTimeZoneInformation (priv_TZInfo);

     priv_systemisUTC := (priv_TZInfo.Bias = 0) and (priv_TZInfo.StandardBias = 0) and (priv_TZInfo.DaylightBias = 0);
     priv_noDST := (priv_TZInfo.StandardDate.wMonth = 0);

     if not priv_systemIsUTC then
        begin
             if priv_noDST then
                begin
                     priv_standardstart := encodeDateTime (1970,1,1,0,0,0,0);
                     priv_daylightstart := encodeDateTime (1970,1,1,0,0,0,0);
                end
             else
                 begin
                      priv_standardstart := GetTzStandardDate (priv_year, priv_TZInfo);
                      priv_daylightstart := GetTzDaylightDate (priv_year, priv_TZInfo);
                 end;
        end;
end;

(* ----------------------------------- *)

procedure tTimeZoneDetails.free;

begin
     inherited free;
end;

(* -------------------------------- *)

// Get start of standard time in specified time-zone

function tTimeZoneDetails.GetTzStandardDate(const year: word; const TzInfo: TTimeZoneInformation): TDateTime;

var  LocalTime  : TSystemTime;

begin
     LocalTime := TzInfo.StandardDate;

     if LocalTime.wYear > 0 then
        Result := SystemTimeToDateTime(LocalTime)
     else
         begin
              LocalTime.wYear := year;

              if LocalTime.wDay < 5 then
                 begin
                      LocalTime.wDay := (LocalTime.wDay - 1) * 7 + 1;
                      Result := SystemTimeToDateTime(LocalTime);
                 end
              else
                  begin
                       LocalTime.wMonth := (LocalTime.wMonth mod 12) + 1;
                       LocalTime.wDay := 1;
                       Result := SystemTimeToDateTime(LocalTime) - 7;
                  end;

              Result := Result + (7 - Abs(Trunc(Result) - 1) mod 7 + LocalTime.wDayOfWeek) mod 7;
         end;
end;

(* -------------------------------- *)

// Get start of daylight time in specified time-zone

function tTimeZoneDetails.GetTzDaylightDate(const Year: word; TzInfo: TTimeZoneInformation): TDateTime;

var LocalTime: TSystemTime;

begin
     if TzInfo.DaylightDate.wMonth = 0 then
        raise Exception.Create('No start of daylight time defined');

     LocalTime := TzInfo.DaylightDate;

     if LocalTime.wYear > 0 then
        Result := SystemTimeToDateTime(LocalTime)
     else
         begin
              LocalTime.wYear := Year;

              if LocalTime.wDay < 5 then
                 begin
                      LocalTime.wDay := (LocalTime.wDay - 1) * 7 + 1;
                      Result := SystemTimeToDateTime(LocalTime);
                 end
              else
                  begin
                       LocalTime.wMonth := (LocalTime.wMonth mod 12) + 1;
                       LocalTime.wDay := 1;
                       Result := SystemTimeToDateTime(LocalTime) - 7;
                  end;

              Result := Result + (7 - Abs(Trunc(Result) - 1) mod 7 + LocalTime.wDayOfWeek) mod 7;
         end;
end;


(* ----------------------------------- *)

function tTimeZoneDetails.GetBias: integer;

begin
     result := priv_TZInfo.Bias;
end;

(* ----------------------------------- *)

function tTimeZoneDetails.GetStandardBias: integer;

begin
     result := priv_TZInfo.StandardBias;
end;

(* ----------------------------------- *)

function tTimeZoneDetails.GetDaylightBias: integer;

begin
     result := priv_TZInfo.DaylightBias;
end;

(* ----------------------------------- *)
{ tTimeZoneHandler }
(* ----------------------------------- *)

constructor tTimeZoneHandler.create;

begin
     inherited create;

     priv_detaillist := tObjectList.Create (true);
     currententry := nil;
end;

(* ----------------------------------- *)

procedure tTimeZoneHandler.free;

begin
     priv_detaillist.Free;

     inherited free;
end;

(* ----------------------------------- *)

function tTimeZoneHandler.GetBiasForDate(DateToTest: tDateTime): integer;

var seek         : tTimeZoneDetails;
    i            : integer;

begin
     if (currententry = nil) or (currententry.Year <> yearOf (DateToTest)) then
        begin
             for i := 1 to priv_detaillist.Count do
                 begin
                      seek := priv_detaillist[i-1] as tTimeZoneDetails;

                      if seek.Year = yearOf (DateToTest) then
                         currententry := seek;
                 end;

             if currententry = nil then
                begin
                     currententry := tTimeZoneDetails.create(yearOf (DateToTest));
                     priv_detaillist.Add(currententry);
                end;
        end;

     if not (currententry.SystemIsOnUTC or currententry.SystemHasNoDST) then
        begin
             if (DateToTest > currententry.StandardStart) or (DateToTest < currententry.DaylightStart) then
                result := currententry.Bias
             else
                 result := currententry.Bias + currententry.DaylightBias;
        end
     else
         result := 0;
end;

(* ----------------------------------- *)

end.
