{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit definition_5tuple;

interface

uses base_structures,
     definition_ip_addresses;

type t5Tuple   = record
                       IPVersion      : byte;
                       Layer4Protocol : tLayer4Protocol;
                       IPv4SocketPair : tIPv4SocketPair;
                       IPv6SocketPair : tIPv6SocketPair;

                       class operator equal (a, b : t5Tuple) : boolean;
                       class operator notequal (a, b : t5Tuple) : boolean;
                       class operator greaterthan (a, b : t5Tuple) : boolean;
                       class operator lessthan (a, b : t5Tuple) : boolean;
                       class operator greaterthanorequal (a, b : t5Tuple) : boolean;
                       class operator lessthanorequal (a, b : t5Tuple) : boolean;

                       procedure initialize; overload;
                       procedure initialize (const SocketPair : tIPv4SocketPair; const Layer4 : byte); overload;
                       procedure initialize (const SocketPair : tIPv6SocketPair; const Layer4 : byte); overload;
                       procedure initialize (const SocketPair : tIPv4SocketPair; const Layer4 : tLayer4Protocol); overload;
                       procedure initialize (const SocketPair : tIPv6SocketPair; const Layer4 : tLayer4Protocol); overload;
                       /// <summary>
                       ///   The key consist of the layer 4 protocol plus the
                       ///   dictionary key of the IPv4 or IPv6 socket pair.
                       ///   The socket pair dictionary key is already
                       ///   normalized.
                       /// </summary>
                       function DictionaryKey : string;
                       function isUndefined : boolean;
                 end;
     t5TupleSet = array of t5Tuple;

     tTCPBaseTuple = record
                           Anchor        : tConversationAnchor;
                           Tuple         : t5Tuple;

                           procedure initialize; overload;
                           procedure initialize (ConAnchor : tConversationAnchor; Value : t5Tuple); overload;
                           procedure initialize (Value : t5Tuple); overload;

                           function DictionaryKey : string;
                     end;

implementation

(* -------------------------------------- *)
{ t5Tuple }
(* -------------------------------------- *)

procedure t5Tuple.initialize;

begin
     IPVersion := 0;
     Layer4Protocol := l4Any;
     IPv4SocketPair.initialize;
     IPv6SocketPair.initialize;
end;

(* -------------------------------------- *)

procedure t5Tuple.initialize (const SocketPair: tIPv4SocketPair; const Layer4: tLayer4Protocol);

begin
     IPVersion := 4;
     Layer4Protocol := Layer4;
     IPv4SocketPair := SocketPair;
     IPv6SocketPair.initialize;
end;

(* -------------------------------------- *)

procedure t5Tuple.initialize (const SocketPair: tIPv6SocketPair; const Layer4: tLayer4Protocol);

begin
     IPVersion := 6;
     Layer4Protocol := Layer4;
     IPv4SocketPair.initialize;
     IPv6SocketPair := SocketPair;
end;

(* -------------------------------------- *)

procedure t5Tuple.initialize (const SocketPair: tIPv4SocketPair; const Layer4: byte);

begin
     initialize (SocketPair, ByteToLayer4Protocol (Layer4));
end;

(* -------------------------------------- *)

procedure t5Tuple.initialize (const SocketPair: tIPv6SocketPair; const Layer4: byte);

begin
     initialize (SocketPair, ByteToLayer4Protocol (Layer4));
end;

(* -------------------------------------- *)

function t5Tuple.DictionaryKey: string;

begin
     case IPVersion of
      0 : result := 'undefined';
      4 : result := Layer4ProtocolToStr (Layer4Protocol) + IPv4SocketPair.DictionaryKey;
      6 : result := Layer4ProtocolToStr (Layer4Protocol) + IPv6SocketPair.DictionaryKey;
     end;
end;

(* -------------------------------------- *)

function t5Tuple.isUndefined: boolean;

begin
     result := (IPVersion = 0);
end;

(* -------------------------------------- *)

class operator t5Tuple.equal (a, b: t5Tuple): boolean;

begin
     result := (a.IPVersion = b.IPVersion) and
               (a.Layer4Protocol = b.Layer4Protocol) and
               (a.IPv4SocketPair = b.IPv4SocketPair) and
               (a.IPv6SocketPair = b.IPv6SocketPair);
end;

(* -------------------------------------- *)

class operator t5Tuple.notequal (a, b: t5Tuple): boolean;

begin
     result := not (a=b);
end;

(* -------------------------------------- *)

class operator t5Tuple.greaterthan (a, b: t5Tuple): boolean;

begin
     if (a.IPVersion = b.IPVersion) then
        begin
             if (a.Layer4Protocol = b.Layer4Protocol) then
                begin
                     case a.IPVersion of
                      0: result := false;
                      4 : begin
                               if b.IPVersion = 4 then
                                  result := (a.IPv4SocketPair > b.IPv4SocketPair)
                               else
                                   result := False;
                          end;
                      6 : begin
                               if b.IPVersion = 4 then
                                  result := true
                               else
                                   result := (a.IPv6SocketPair > b.IPv6SocketPair);
                          end;
                     end;
                end
             else
                 result := (a.Layer4Protocol > b.Layer4Protocol)
        end
     else
         result := (a.IPVersion > b.IPVersion);
end;

(* -------------------------------------- *)

class operator t5Tuple.greaterthanorequal (a, b: t5Tuple): boolean;

begin
     result := (a = b) or (a > b);
end;

(* -------------------------------------- *)

class operator t5Tuple.lessthan (a, b: t5Tuple): boolean;

begin
     result := not (a >= b);
end;

(* -------------------------------------- *)

class operator t5Tuple.lessthanorequal (a, b: t5Tuple): boolean;

begin
     result := not (a > b);
end;

(* ------------------------------------------- *)
{ tTCPBaseTuple }
(* ------------------------------------------- *)

procedure tTCPBaseTuple.initialize;

begin
     Anchor.initialize;
     Tuple.initialize;
end;

(* ------------------------------------------- *)

procedure tTCPBaseTuple.initialize (Value: t5Tuple);

begin
     Anchor.initialize;
     Tuple := Value;
end;

(* ------------------------------------------- *)

procedure tTCPBaseTuple.initialize (ConAnchor : tConversationAnchor; Value : t5Tuple);

begin
     Anchor := ConAnchor;
     Tuple := Value;
end;

(* ------------------------------------------- *)

function tTCPBaseTuple.DictionaryKey: string;

begin
     result := Anchor.DictionaryKey + '-' + Tuple.DictionaryKey;
end;

(* ------------------------------------------- *)
end.
