{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit definition_ip_addresses;

interface

uses Types;
//   base_structures;           // should not be necessary to avoid big dependencies in other use cases

type tIPFlags = (ipfReserved, ipfDontFragment, ipfMoreFragments, ipfLastFragment);
     tIPFlagSet = set of tIPFlags;

     // MAC Address

     pMACBytes   = ^tMACBytes;
     tMACBytes   = array[1..6] of byte;

     pMACAddress = ^tMACAddress;
     tMACAddress = record
                         public
                               RangeBytes  : tMACBytes;

                               class operator equal (a, b : tMACAddress) : boolean;
                               class operator notequal (a, b : tMACAddress) : boolean;
                               class operator greaterthan (a, b : tMACAddress) : boolean;
                               class operator lessthan (a, b : tMACAddress) : boolean;
                               class operator greaterthanorequal (a, b : tMACAddress) : boolean;
                               class operator lessthanorequal (a, b : tMACAddress) : boolean;

                               procedure initialize;
                               function asString (const Separator : string = ':'; const ForceLowercase : boolean = false) : string;
                               {$REGION 'Documentation'}
                               /// <summary>
                               ///   Sets the MAC address from string which
                               ///   includes a separator, e.g. "00:01:02:03:04:
                               ///   05"
                               /// </summary>
                               /// <param name="Address">
                               ///   MAC address with separators between octets
                               /// </param>
                               /// <param name="separator">
                               ///   character that is used as separator, e.g. "
                               ///   :" or "-"
                               /// </param>
                               {$ENDREGION}
                               procedure setfromString (const Address : string; const separator : string); overload;
                               {$REGION 'Documentation'}
                               /// <summary>
                               ///   Sets a MAC address from string that has no
                               ///   separators between the octets, e.g. "
                               ///   000102030405"
                               /// </summary>
                               /// <param name="Address">
                               ///   MAC adress string, without separator
                               /// </param>
                               {$ENDREGION}
                               procedure setfromString (const Address : string); overload;
                               function DictionaryKey : string;
                               function isBroadcast : boolean;
                               function isMulticast : boolean;
                   end;

     tMACAddressPair = record
                             Source      : tMACAddress;
                             Destination : tMACAddress;
                             VLANID      : word;

                             procedure initialize; overload;
                             procedure initialize (value : string); overload;
                             procedure normalize;
                             function asString (const Separator : string = ':'; const ForceLowercase : boolean = false) : string;
                             function DictionaryKey : string;

                             class operator equal (a, b : tMACAddressPair) : boolean;
                             class operator notequal (a, b : tMACAddressPair) : boolean;
                             class operator greaterthan (a, b : tMACAddressPair) : boolean;
                             class operator lessthan (a, b : tMACAddressPair) : boolean;
                             class operator greaterthanorequal (a, b : tMACAddressPair) : boolean;
                             class operator lessthanorequal (a, b : tMACAddressPair) : boolean;
                       end;
     pMACAddressPair = ^tMACAddressPair;

     // Conversation Anchor

     tVLANID = integer;

     tConversationAnchor = record
                                 MACPairN          : tMACAddressPair;
                                 VLANId            : tVLANID;
//                               TTLAnchorPair     : tTTLAnchorPair;             // not really usefull
                                 InterfaceID       : longword;

                                 class operator equal (a, b : tConversationAnchor) : boolean;
                                 class operator notequal (a, b : tConversationAnchor) : boolean;

                                 procedure initialize;
                                 /// <summary>
                                 ///   The dictionary key consists of the VLAN
                                 ///   ID and the TTL
                                 /// </summary>
                                 function DictionaryKey : string;
                           end;

     // IPv4 Address

     tIPRandomScope = (iprsPrivate, iprsAll);

     tIPv4ByteArray = array [1..4] of byte;
     tIPv4Address = record
                          ValueAsLong     : longword;
                          MaskBitCount    : byte;

                          class operator equal (a, b : tIPv4Address) : boolean;
                          class operator notequal (a, b : tIPv4Address) : boolean;
                          class operator greaterthan (a, b : tIPv4Address) : boolean;
                          class operator lessthan (a, b : tIPv4Address) : boolean;
                          class operator greaterthanorequal (a, b : tIPv4Address) : boolean;
                          class operator lessthanorequal (a, b : tIPv4Address) : boolean;

                          {$REGION 'Documentation'}
                          ///	<summary>
                          ///	  Initializes the IP address and mask bits
                          ///	</summary>
                          ///	<param name="SetRandomIP">
                          ///	  if true, a random IP address will be generated.
                          ///	  No sanity checks are in place, so there may be
                          ///	  loopback or broadcast addresses returned. If
                          ///	  false, a zero address will be generated.
                          ///	</param>
                          {$ENDREGION}
                          procedure initialize (const SetRandomIP : boolean = false; const Scope : tIPRandomScope = iprsAll);
                          procedure setValueFromString (const IPAddress: string; const MaskBits : byte = 32);
                          procedure setValueFromCIDRString (const IPAddress: string);
                          procedure setValueFromByteArray (const IPAddress : tIPv4ByteArray);
                          function ValueAsString : string;
                          function ValueAsCIDRString : string;
                          function ValueAsByteArray : tIPv4ByteArray;
                          function isZeroAddress : boolean;
                          function IsLegalSource : boolean;
                          function IsLegalDestination : boolean;
                          function IsMulticast : boolean;
                          function IsBroadcast : boolean;
                          function IsPrivate : boolean;
                          function IsAPIPA : boolean;
                          function isLoopback : boolean;
                          function isDocumentationAddress : boolean;
                          function isCommon : boolean;
                          function MaskAsString: string;
                          procedure setMaskFromString(const Value: string);
                          function MaskAsLong: longword;
                          procedure setMaskFromLong(const Value: longword);
                          function LowestNetIP : tIPv4Address;
                          function HighestNetIP : tIPv4Address;
                          function NetWildCard : tIPv4Address;
                          function BelongsToSubnet (Subnet : tIPv4Address) : boolean;
                          function isUndefined : boolean;
                          function DictionaryKey : string;
                    end;

     tIPv4AddressPair = record
                              Source      : tIPv4Address;
                              Destination : tIPv4Address;

                              class operator equal (a, b : tIPv4AddressPair) : boolean;
                              class operator notequal (a, b : tIPv4AddressPair) : boolean;
                              class operator greaterthan (a, b : tIPv4AddressPair) : boolean;
                              class operator lessthan (a, b : tIPv4AddressPair) : boolean;
                              class operator greaterthanorequal (a, b : tIPv4AddressPair) : boolean;
                              class operator lessthanorequal (a, b : tIPv4AddressPair) : boolean;

                              procedure initialize; overload;
                              procedure initialize (value : string; const SetCIDR : boolean); overload;
                              procedure normalize;
                              procedure swap;
                              function ValueAsString (const NoSpaces : boolean) : string;
                              function isUndefined : boolean;
                              function DictionaryKey : string;
                        end;
     pIPv4AddressPair = ^tIPv4AddressPair;

     tIPv4Socket = record
                         VLANID    : word;
                         IPAddress : tIPv4Address;
                         Port      : word;

                         class operator equal (a, b : tIPv4Socket) : boolean;
                         class operator notequal (a, b : tIPv4Socket) : boolean;
                         class operator greaterthan (a, b : tIPv4Socket) : boolean;
                         class operator lessthan (a, b : tIPv4Socket) : boolean;
                         class operator greaterthanorequal (a, b : tIPv4Socket) : boolean;
                         class operator lessthanorequal (a, b : tIPv4Socket) : boolean;

                         procedure initialize; overload;
                         procedure initialize (const Value : string); overload;
                         function isUndefined : boolean;
                         function ValueAsString : string;
                   end;

     tIPv4SocketPair = record
                             Source : tIPv4Socket;
                             Destination : tIPv4Socket;

                             class operator equal (a, b : tIPv4SocketPair) : boolean;
                             class operator notequal (a, b : tIPv4SocketPair) : boolean;
                             class operator greaterthan (a, b : tIPv4SocketPair) : boolean;
                             class operator lessthan (a, b : tIPv4SocketPair) : boolean;
                             class operator greaterthanorequal (a, b : tIPv4SocketPair) : boolean;
                             class operator lessthanorequal (a, b : tIPv4SocketPair) : boolean;

                             procedure initialize;
                             procedure normalize;
                             procedure swap;
                             function ValueAsString (const NoSpaces : boolean) : string;
                             function IPAddressPair : tIPv4AddressPair;
                             function isUndefined : boolean;
                             function DictionaryKey : string;
                       end;
     pIPv4SocketPair = ^tIPv4SocketPair;

     tIPv4AddrArray = array of tIPv4Address;

     tIPv4SimpleAddressList = record
                                    list : tIPv4AddrArray;

                                    function count : integer;
                                    procedure Add (NewEntry : tIPv4Address);
                                    function GetEntry (index : integer) : tIPv4Address;
                                    procedure copyFrom (SourceList : tIPv4AddrArray);
                              end;

     // IPv6 Address

     IPv6MulticastScope = (msReserved, msInterfaceLocal, msLinkLocal, msAdminLocal, msSiteLocal, msNotAssigned, msOrganizationLocal, msGlobal);

     pIPv6AddressWordStruct = ^tIPv6AddressWordStruct;
     tIPv6AddressWordStruct = array [0..7] of word;

     pIPv6AddressByteStruct = ^tIPv6AddressByteStruct;
     tIPv6AddressByteStruct = array [0..15] of byte;


     tIPv6Address = record
                          ValueAsWordStruct : tIPv6AddressWordStruct;
                          MaskBitCount      : byte;

                          class operator equal (a, b : tIPv6Address) : boolean;
                          class operator notequal (a, b : tIPv6Address) : boolean;
                          class operator greaterthan (a, b : tIPv6Address) : boolean;
                          class operator lessthan (a, b : tIPv6Address) : boolean;
                          class operator greaterthanorequal (a, b : tIPv6Address) : boolean;
                          class operator lessthanorequal (a, b : tIPv6Address) : boolean;

                          procedure initialize (const SetRandomIP : boolean = false; const Scope : tIPRandomScope = iprsAll); overload;
                          procedure initialize (const Prefix : tIPv6Address; const SetRandomIP : boolean = false); overload;
                          procedure setValueFromString(const IPAddress : string; const PrefixLength : byte = 64);
                          procedure setValueFromStringWithMask (const IPAddress : string);
                          {$REGION 'Documentation'}
                          /// <summary>
                          ///   Creates an IPv6 address in EUI64 form, based on
                          ///   a MAC address
                          /// </summary>
                          /// <param name="Prefix">
                          ///   IPv6 network prefix
                          /// </param>
                          /// <param name="MACAddress">
                          ///   MAC address to generate the EUI64 address from
                          /// </param>
                          {$ENDREGION}
                          procedure setValueFromMAC (const Prefix : tIPv6Address; const MACAddress : tMACAddress);
                          function ValueAsString : string;
                          function ValueAsSlashedNotationString : string;
                          procedure SetValueFromByteStruct(const Value: tIPv6AddressByteStruct);
                          {$REGION 'Documentation'}
                          /// <summary>
                          ///   Copies the prefix bits from the specified
                          ///   address. The remaining bits will not be touched.
                          /// </summary>
                          /// <param name="Address">
                          ///   IPv6 address to copy the prefix from. This can
                          ///   also be just a prefix if all host bits are zero.
                          /// </param>
                          /// <param name="PrefixLength">
                          ///   Optional prefix length to copy. If this is zero
                          ///   the MaskBitCount from the specified address
                          ///   will be used.
                          /// </param>
                          {$ENDREGION}
                          procedure CopyPrefixFrom (Address : tIPv6Address; const PrefixLength : byte = 0);
                          function ValueAsByteStruct : tIPv6AddressByteStruct;
                          function getExpandedValue : string;
                          function getIPv4NotationValue :  string;
                          {$REGION 'Documentation'}
                          /// <summary>
                          ///   Returns the solicited node multicast address
                          ///   that is derived from the given IP address
                          /// </summary>
                          {$ENDREGION}
                          function getSolicitedNodeAddress : tIPv6Address;
                          {$REGION 'Documentation'}
                          /// <summary>
                          ///   Calculates the MAC address that was used to
                          ///   generate the EUI64 form of the IP address
                          /// </summary>
                          {$ENDREGION}
                          function getEUI64BaseMAC : tMACAddress;
                          function getMCastBaseMAC : tMACAddress;
                          function IsEUI64Address: boolean;
                          function IsLinkLocalAddress : boolean;
                          function IsLegalSource : boolean;
                          function IsLegalDestination : boolean;
                          function IsMulticast : boolean;
                          function isSolicitedNodeMCast : boolean;
                          function isCommon : boolean;
                          function IsTemporaryMCast: boolean;
                          function getMulticastScope: IPv6MulticastScope;
                          function IsLocalHost: boolean;
                          function IsUnspecifiedAddress: boolean;
                          function IsDocumentationAddress : boolean;

                          {$REGION 'Documentation'}
                          ///	<summary>
                          ///	  This function returns the number of zero bits
                          ///	  at the end of an IP address. This is used to
                          ///	  match the bit mask count to the actual empty
                          ///	  bits to see if the "address" is a prefix
                          ///	</summary>
                          {$ENDREGION}
                          function getEmptyHostBitCount : integer;
                          function IsPrefixInformation : boolean;
                          function BelongsToSubnet (Subnet : tIPv6Address) : boolean;
                          function isUndefined : boolean;
                          function DictionaryKey : string;
                    end;

     tIPv6AddressPair = record
                              Source      : tIPv6Address;
                              Destination : tIPv6Address;

                              class operator equal (a, b : tIPv6AddressPair) : boolean;
                              class operator notequal (a, b : tIPv6AddressPair) : boolean;
                              class operator greaterthan (a, b : tIPv6AddressPair) : boolean;
                              class operator lessthan (a, b : tIPv6AddressPair) : boolean;
                              class operator greaterthanorequal (a, b : tIPv6AddressPair) : boolean;
                              class operator lessthanorequal (a, b : tIPv6AddressPair) : boolean;

                              procedure initialize; overload;
                              procedure initialize (value : string; const SetCIDR : boolean); overload;
                              procedure normalize;
                              procedure swap;
                              function ValueAsString (const NoSpaces : boolean) : string;
                              function getExpandedValue (const NoSpaces: boolean) : string;
                              function isUndefined : boolean;
                              function DictionaryKey : string;
                        end;
     pIPv6AddressPair = ^tIPv6AddressPair;

     tIPv6Socket = record
                         VLANID    : word;
                         IPAddress : tIPv6Address;
                         Port      : word;

                         class operator equal (a, b : tIPv6Socket) : boolean;
                         class operator notequal (a, b : tIPv6Socket) : boolean;
                         class operator greaterthan (a, b : tIPv6Socket) : boolean;
                         class operator lessthan (a, b : tIPv6Socket) : boolean;
                         class operator greaterthanorequal (a, b : tIPv6Socket) : boolean;
                         class operator lessthanorequal (a, b : tIPv6Socket) : boolean;

                         procedure initialize; overload;
                         procedure initialize (const Value : string); overload;
                         function ValueAsString : string;
                         function getExpandedValue (const NoSpaces: boolean) : string;
                         function isUndefined : boolean;
                   end;

     tIPv6SocketPair = record
                             Source : tIPv6Socket;
                             Destination : tIPv6Socket;

                             class operator equal (a, b : tIPv6SocketPair) : boolean;
                             class operator notequal (a, b : tIPv6SocketPair) : boolean;
                             class operator greaterthan (a, b : tIPv6SocketPair) : boolean;
                             class operator lessthan (a, b : tIPv6SocketPair) : boolean;
                             class operator greaterthanorequal (a, b : tIPv6SocketPair) : boolean;
                             class operator lessthanorequal (a, b : tIPv6SocketPair) : boolean;

                             procedure initialize;
                             procedure normalize;
                             procedure swap;
                             function ValueAsString (const NoSpaces : boolean) : string;
                             function getExpandedValue (const NoSpaces: boolean) : string;
                             function IPAddressPair : tIPv6AddressPair;
                             function isUndefined : boolean;
                             function DictionaryKey : string;
                       end;

     tIPAddress = record
                        IPVersion     : byte;
                        IPv4Address   : tIPv4Address;
                        IPv6Address   : tIPv6Address;
{$ifdef debug}
                        IPString      : string;
{$endif}

                        class operator equal (a, b : tIPAddress) : boolean;
                        class operator notequal (a, b : tIPAddress) : boolean;
                        class operator greaterthan (a, b : tIPAddress) : boolean;
                        class operator lessthan (a, b : tIPAddress) : boolean;
                        class operator greaterthanorequal (a, b : tIPAddress) : boolean;
                        class operator lessthanorequal (a, b : tIPAddress) : boolean;

                        procedure initialize; overload;
                        procedure initialize (const IPAddress : tIPv4Address); overload;
                        procedure initialize (const IPAddress : tIPv6Address); overload;
                        procedure initialize (const IPv4AsLong : longword); overload;
                        procedure initialize (const Value: tIPv6AddressByteStruct); overload;
                        procedure setValueFromString (const IPAddress: string; const MaskBits : byte = 0);
                        function isZeroAddress : boolean;
                        function ValueAsString : string;
                        function isUndefined : boolean;
                        function DictionaryKey : string;
                  end;
     pIPAddress = ^tIPAddress;

     tIPAddressPair = record
                            Source      : tIPAddress;
                            Destination : tIPAddress;

                            class operator equal (a, b : tIPAddressPair) : boolean;
                            class operator notequal (a, b : tIPAddressPair) : boolean;
                            class operator greaterthan (a, b : tIPAddressPair) : boolean;
                            class operator lessthan (a, b : tIPAddressPair) : boolean;
                            class operator greaterthanorequal (a, b : tIPAddressPair) : boolean;
                            class operator lessthanorequal (a, b : tIPAddressPair) : boolean;

                            procedure initialize;
                            procedure normalize;
                            procedure swap;
                            function ValueAsString (const NoSpaces : boolean) : string;
                            function isUndefined : boolean;
                            function DictionaryKey : string;
                      end;
     pIPAddressPair = ^tIPAddressPair;

     tIPSocket = record
                       VLANID      : word;
                       IPAddress   : tIPAddress;
                       Port        : word;

                       class operator equal (a, b : tIPSocket) : boolean;
                       class operator notequal (a, b : tIPSocket) : boolean;
                       class operator greaterthan (a, b : tIPSocket) : boolean;
                       class operator lessthan (a, b : tIPSocket) : boolean;
                       class operator greaterthanorequal (a, b : tIPSocket) : boolean;
                       class operator lessthanorequal (a, b : tIPSocket) : boolean;

                       procedure initialize; overload;
                       procedure initialize (const NewAddress : tIPAddress; const NewPort : word; const NewVLANID : word = 0); overload;
                       procedure initialize (const IPv4Address : tIPv4Address; const NewPort : word; const NewVLANID : word = 0); overload;
                       procedure initialize (const IPv6Address : tIPv6Address; const NewPort : word; const NewVLANID : word = 0); overload;
                       procedure initialize (const IPv4Socket : tIPv4Socket); overload;
                       procedure initialize (const IPv6Socket : tIPv6Socket); overload;
                       procedure initialize (const Value : string); overload;
                       function ValueAsString : string;
                       function getIPv4Socket : tIPv4Socket;
                       function getIPv6Socket : tIPv6Socket;
                       function DictionaryKey : string;
                       function isUndefined : boolean;
                 end;
     pIPSocket = ^tIPSocket;

     tIPSocketPair = record
                           Source : tIPSocket;
                           Destination : tIPSocket;

                           class operator equal (a, b : tIPSocketPair) : boolean;
                           class operator notequal (a, b : tIPSocketPair) : boolean;
                           class operator greaterthan (a, b : tIPSocketPair) : boolean;
                           class operator lessthan (a, b : tIPSocketPair) : boolean;
                           class operator greaterthanorequal (a, b : tIPSocketPair) : boolean;
                           class operator lessthanorequal (a, b : tIPSocketPair) : boolean;

                           procedure initialize; overload;
                           procedure initialize (const IPv4SocketPair : tIPv4SocketPair); overload;
                           procedure initialize (const IPv6SocketPair : tIPv6SocketPair); overload;
                           procedure initialize (const value : string); overload;
                           procedure normalize;
                           procedure swap;
                           function ValueAsString (const NoSpaces : boolean) : string;
                           function IPAddressPair : tIPAddressPair;
                           function isUndefined : boolean;
                           function DictionaryKey : string;
                           function toCSV (const Separator : string = ',') : string;
                           /// <summary>
                           ///   Returns a Wireshark filter for the socketpair
                           /// </summary>
                           /// <param name="DisplayFilter">
                           ///   if true, the string returned is in display
                           ///   filter format, otherwise BPF
                           /// </param>
                           function toFilter (const DisplayFilter : boolean = true) : string;
                     end;
     pIPSocketPair = ^tIPSocketpair;

     tIPv6AddrArray = array of tIPv6Address;

     tIPv6SimpleAddressList = record
                                    list : tIPv6AddrArray;

                                    function count : integer;
                                    procedure Add (NewEntry : tIPv6Address);
                                    function GetEntry (index : integer) : tIPv6Address;
                                    procedure copyFrom (SourceList : tIPv6AddrArray);
                              end;

     tByteArray = array of byte;

     tWordPair   = record
                         Value1      : word;
                         Value2      : word;

                         class operator equal (a, b : tWordPair) : boolean;
                         class operator notequal (a, b : tWordPair) : boolean;
                         class operator greaterthan (a, b : tWordPair) : boolean;
                         class operator lessthan (a, b : tWordPair) : boolean;
                         class operator greaterthanorequal (a, b : tWordPair) : boolean;
                         class operator lessthanorequal (a, b : tWordPair) : boolean;

                         procedure initialize; overload;
                         procedure initialize (value : string); overload;
                   end;

     tVLAN       = word;
     tVLANPair   = tWordPair;
     pVLANPair   = ^tVLANPair;
     tPortPair   = tWordPair;
     pPortPair   = ^tPortPair;
{
     tSocketPortPair = record
                             IPVersion          : byte;
                             OriginalIPv4Socket : tIPv4Socket;
                             OriginalIPv6Socket : tIPv6Socket;
                             NewPort            : word;

                             procedure initialize;
                       end;
     pSocketPortPair = ^tSocketPortPair;
}
     tFQDNLabel = string[63];

     tFQDN = record
                   FullFQDN          : string;
                   FullFQDNBackwards : string;
                   ElementCount      : integer;
                   Elements          : array of tFQDNLabel;

                   procedure setFQDN (NewValue : string);
                   function getPartialFQDN (PartCount : integer) : string;
                   function getAsFrameBufferArray : tByteArray;
                   procedure initialize;

                   class operator equal (a, b : tFQDN) : boolean;
                   class operator notequal (a, b : tFQDN) : boolean;
                   class operator greaterthan (a, b : tFQDN) : boolean;
                   class operator lessthan (a, b : tFQDN) : boolean;
                   class operator greaterthanorequal (a, b : tFQDN) : boolean;
                   class operator lessthanorequal (a, b : tFQDN) : boolean;
             end;

     tFQDNPair = record
                       Original    : tFQDN;
                       Replacement : tFQDN;

                       procedure initialize;
                       procedure normalize;
                 end;

     tNRAnswerType = (nraUnknown, nraIPAddress, nraFQDN, nraText);

     tNameResolutionAnswer = record
                                   AnswerType   : tNRAnswerType;
                                   IPAddress    : tIPAddress;
                                   FQDN         : tFQDN;
                                   TextValue    : string;

                                   procedure initialize; overload;
                                   procedure initialize (const IPv4Address : tIPv4Address); overload;
                                   procedure initialize (const IPv6Address : tIPv6Address); overload;
                                   procedure initialize (const NewIPAddress : tIPAddress); overload;
                                   procedure initialize (const NewFQDN : tFQDN); overload;
                                   procedure initialize (const NewTextValue : string); overload;

                                   function DictionaryKey : string;
                                   function asString : string;
                             end;

//     tDummyClass = class (tObject)

//                   end;


const
      VALUE_ZERO_MAC       : tMACaddress = (rangebytes : ($00, $00, $00, $00, $00, $00));
      VALUE_MCAST_MAC      : tMACaddress = (rangebytes : ($01, $00, $5E, $00, $00, $00));
      VALUE_HSRP_DEST      : tMACaddress = (rangebytes : ($01, $00, $5E, $00, $00, $02));
      VALUE_OSPF_DESTdot5  : tMACaddress = (rangebytes : ($01, $00, $5E, $00, $00, $05));
      VALUE_OSPF_DESTdot6  : tMACaddress = (rangebytes : ($01, $00, $5E, $00, $00, $06));
      VALUE_VRRP_DEST      : tMACAddress = (rangebytes : ($01, $00, $5E, $00, $00, $12));
      VALUE_EIGRP_DEST     : tMACAddress = (rangebytes : ($01, $00, $5E, $00, $00, $0A));
      VALUE_BPDU_DEST      : tMACaddress = (rangebytes : ($01, $80, $C2, $00, $00, $00));
      VALUE_PVST_DEST      : tMACaddress = (rangebytes : ($01, $00, $0C, $CC, $CC, $CD));
      VALUE_BROADCAST_DEST : tMACaddress = (rangebytes : ($FF, $FF, $FF, $FF, $FF, $FF));
      VALUE_VRRP_MACGRP    : tMACaddress = (rangebytes : ($00, $00, $5E, $00, $01, $00));

      IPv4_Broadcast = $ffffffff;                      // 255.255.255.255
      IPv4_APIPANet = $a9fe0000;                       // 169.254.x.y
      IPv4_NODE_UNSPECIFIED                      : tIPv4Address = (ValueAsLong : 0; MaskBitCount : 32);

      IPv6_NODE_UNSPECIFIED                      : tIPv6AddressWordStruct = (0,0,0,0,0,0,0,0);             // ::
      IPv6_NODE_UNSPECIFIED_BYTES                : tIPv6AddressByteStruct = (0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0);             // ::
      IPv6_NODE_LOCALHOST                        : tIPv6AddressWordStruct = (0,0,0,0,0,0,0,1);             // ::1

      IPv6_NODE_LOCAL_ALL_NODES_MULTICAST        : tIPv6AddressWordStruct = ($ff01,0,0,0,0,0,0,1);         // ff01::1
      IPv6_NODE_LOCAL_ALL_ROUTERS_MULTICAST      : tIPv6AddressWordStruct = ($ff01,0,0,0,0,0,0,2);         // ff01::2
      IPv6_NODE_LOCAL_MDNSV6                     : tIPv6AddressWordStruct = ($ff01,0,0,0,0,0,0,$fb);       // ff01::fb

      IPv6_LINK_LOCAL_ALL_NODES_MULTICAST        : tIPv6AddressWordStruct = ($ff02,0,0,0,0,0,0,1);         // ff02::1
      IPv6_LINK_LOCAL_ALL_ROUTERS_MULTICAST      : tIPv6AddressWordStruct = ($ff02,0,0,0,0,0,0,2);         // ff02::2
      IPv6_LINK_LOCAL_DVMRP_ROUTERS              : tIPv6AddressWordStruct = ($ff02,0,0,0,0,0,0,4);         // ff02::4
      IPv6_LINK_LOCAL_OSPFIGP                    : tIPv6AddressWordStruct = ($ff02,0,0,0,0,0,0,5);         // ff02::5
      IPv6_LINK_LOCAL_OSPFIGP_DESIGNATED_ROUTERS : tIPv6AddressWordStruct = ($ff02,0,0,0,0,0,0,6);         // ff02::6
      IPv6_LINK_LOCAL_ST_ROUTERS                 : tIPv6AddressWordStruct = ($ff02,0,0,0,0,0,0,7);         // ff02::7
      IPv6_LINK_LOCAL_ST_HOSTS                   : tIPv6AddressWordStruct = ($ff02,0,0,0,0,0,0,8);         // ff02::8
      IPv6_LINK_LOCAL_RIP_ROUTERS                : tIPv6AddressWordStruct = ($ff02,0,0,0,0,0,0,9);         // ff02::9
      IPv6_LINK_LOCAL_EIGRP_ROUTERS              : tIPv6AddressWordStruct = ($ff02,0,0,0,0,0,0,$a);        // ff02::a
      IPv6_LINK_LOCAL_MOBILE_AGENTS              : tIPv6AddressWordStruct = ($ff02,0,0,0,0,0,0,$b);        // ff02::b
      IPv6_LINK_LOCAL_SSDP                       : tIPv6AddressWordStruct = ($ff02,0,0,0,0,0,0,$c);        // ff02::c
      IPv6_LINK_LOCAL_ALL_PIM_ROUTERS            : tIPv6AddressWordStruct = ($ff02,0,0,0,0,0,0,$d);        // ff02::d
      IPv6_LINK_LOCAL_RSVP_ENCAPSULATION         : tIPv6AddressWordStruct = ($ff02,0,0,0,0,0,0,$e);        // ff02::e
      IPv6_LINK_LOCAL_UPNP                       : tIPv6AddressWordStruct = ($ff02,0,0,0,0,0,0,$f);        // ff02::f
      IPv6_LINK_LOCAL_VRRP                       : tIPv6AddressWordStruct = ($ff02,0,0,0,0,0,0,$12);       // ff02::12
      IPv6_LINK_LOCAL_ALL_MLDV2_CAPABLE_ROUTERS  : tIPv6AddressWordStruct = ($ff02,0,0,0,0,0,0,$16);       // ff02::16
      IPv6_LINK_LOCAL_ALL_RPL_NODES              : tIPv6AddressWordStruct = ($ff02,0,0,0,0,0,0,$1a);       // ff02::1a
      IPv6_LINK_LOCAL_ALL_SNOOPERS               : tIPv6AddressWordStruct = ($ff02,0,0,0,0,0,0,$6a);       // ff02::6a
      IPv6_LINK_LOCAL_PTP_PDELAY                 : tIPv6AddressWordStruct = ($ff02,0,0,0,0,0,0,$6b);       // ff02::6b
      IPv6_LINK_LOCAL_SARATOGA                   : tIPv6AddressWordStruct = ($ff02,0,0,0,0,0,0,$6c);       // ff02::6c
      IPv6_LINK_LOCAL_LL_MANET_ROUTERS           : tIPv6AddressWordStruct = ($ff02,0,0,0,0,0,0,$6d);       // ff02::6d
      IPv6_LINK_LOCAL_IGRS                       : tIPv6AddressWordStruct = ($ff02,0,0,0,0,0,0,$6e);       // ff02::6e
      IPv6_LINK_LOCAL_IADT_DISCOVERY             : tIPv6AddressWordStruct = ($ff02,0,0,0,0,0,0,$6f);       // ff02::6f
      IPv6_LINK_LOCAL_MDNSV6                     : tIPv6AddressWordStruct = ($ff02,0,0,0,0,0,0,$fb);       // ff02::fb
      IPv6_LINK_LOCAL_LINK_NAME                  : tIPv6AddressWordStruct = ($ff02,0,0,0,0,0,1,1);         // ff02::1:1
      IPv6_LINK_LOCAL_ALL_DHCP_AGENTS            : tIPv6AddressWordStruct = ($ff02,0,0,0,0,0,1,2);         // ff02::1:2
      IPv6_LINK_LOCAL_MULTICAST_NAME_RESOLUTION  : tIPv6AddressWordStruct = ($ff02,0,0,0,0,0,1,3);         // ff02::1:3
      IPv6_LINK_LOCAL_DTCP_ANNOUNCEMENT          : tIPv6AddressWordStruct = ($ff02,0,0,0,0,0,1,4);         // ff02::1:4
      IPv6_LINK_LOCAL_AFORE_VDP                  : tIPv6AddressWordStruct = ($ff02,0,0,0,0,0,1,5);         // ff02::1:5
      IPv6_LINK_LOCAL_BABEL                      : tIPv6AddressWordStruct = ($ff02,0,0,0,0,0,1,6);         // ff02::1:6
      IPv6_LINK_LOCAL_SOLICITED_NODE_ADDRESS     : tIPv6AddressWordStruct = ($ff02,0,0,0,0,1,$ff00,0);     // ff02::1:ff00:0/104
      IPv6_LINK_LOCAL_NODE_INFORMATION_QUERIES   : tIPv6AddressWordStruct = ($ff02,0,0,0,0,2,$ff00,0);     // ff02:0:0:0:0:2:ff00::/104

      IPv6_SITE_LOCAL_ALL_ROUTERS_ADDRESS        : tIPv6AddressWordStruct = ($ff05,0,0,0,0,0,0,2);         // ff05::2
      IPv6_SITE_LOCAL_MDNSV6                     : tIPv6AddressWordStruct = ($ff05,0,0,0,0,0,0,$fb);       // ff05::fb
      IPv6_SITE_LOCAL_ALL_DHCP_SERVERS           : tIPv6AddressWordStruct = ($ff05,0,0,0,0,0,1,3);         // ff05::1:3

      function CompareValue (const A, B: tMACAddress) : tValueRelationship; overload;
      function CompareValue (const A, B: tIPv4Address): TValueRelationship; overload;
      function CompareValue (const A, B: tIPv6Address): TValueRelationship; overload;
      function CompareValue (const A, B: tIPAddress): TValueRelationship; overload;
      function isValidIPv4String (IPaddress : string) : boolean;
      function isValidIPv6String (IPv6Address : string): boolean;
      function isValidIPv4Subnet (Subnet : string) : boolean;
      function isValidIPv6Subnet (Subnet : string) : boolean;
      function isValidIPv4Socket (Socket : string) : boolean;
      function isValidIPv6Socket (Socket : string; const AllowSnortNotation : boolean = false) : boolean;
      function isValidIPString (Value : string) : boolean;
      function isValidSocketString (Value : string) : boolean;
      function isValidMACString (MACAddress : string) : boolean;
      function isValidMACaddressPair (value : string; const CheckNull : boolean) : boolean;
      function isValidIPv4AddressPair (value : string; const CheckNull : boolean) : boolean;
      function isValidIPv4SubnetPair (value : string; const CheckNull : boolean) : boolean;
      function isValidIPv6AddressPair (value : string; const CheckNull : boolean) : boolean;
      function isValidIPv6SubnetPair (value : string; const CheckNull : boolean) : boolean;
      function isValidSocketPair (value : string) : boolean;

      function CompareIPv4SocketPair (a, b : tIPv4SocketPair) : integer;
      function CompareIPv6SocketPair (a, b : tIPv6SocketPair) : integer;

      function AnswerTypeToInt (value : tNRAnswerType) : integer;
      function IntegerToAnswerType (value : integer) : tNRAnswerType;

      /// <summary>
      ///   Determines if a port is likely a server port, by looking up a list of
      ///   well known server ports
      /// </summary>
      /// <param name="Port">
      ///   Port to check
      /// </param>
      function isWellKnowServerPort (Port : word) : boolean;

      function DetermineProbableClientAddress (const SocketPair : tIPSocketPair) : tIPAddress;


implementation

uses Classes,
     SysUtils,
     StrUtils,
     DateUtils,
     Math,
     toolkit;

(* -------------------------------------- *)
{ Common functions and procedures }
(* -------------------------------------- *)

function CompareValue (const A, B: tMACAddress) : tValueRelationship; overload;

begin
    if A = B then
       Result := EqualsValue
     else if A < B then
       Result := LessThanValue
     else
       Result := GreaterThanValue;
end;

(* -------------------------------------- *)

function CompareValue (const A, B: tIPv4Address): TValueRelationship; overload;

begin
     if A = B then
       Result := EqualsValue
     else if A < B then
       Result := LessThanValue
     else
       Result := GreaterThanValue;
end;

(* -------------------------------------- *)
// converts an IP address in longword format to a human readable string value

function LongToIP(IP : Longword) : string;

var
   temp : Longword;
   oneval : Byte;
   i : Integer;

begin
     result := '';
     temp := IP;

     for i := 0 to 2 do
     begin
          oneval := temp mod 256;
          result := '.' + IntToStr (oneval) + result;
          temp := temp div 256;
     end;

     result := IntToStr (temp) + result;

end;

(* -------------------------------------- *)

function IPToLong (sIPString : String) : Longword;

var subbyte : byte;
    temp    : int64;
    list    : tStrings;

begin
     result := 0;

     list := tStringList.Create;
     ExtractStrings (['.'], [], pChar (sIPString), list);

     if (sIPString <> '') then
        begin
             subbyte := StrToInt (list[0]);
             temp := subByte;
             temp := temp * 256 * 256 * 256;

             subbyte := StrToInt (list[1]);
             temp := temp + subByte * 256 * 256;

             subbyte := StrToInt (list[2]);
             temp := temp + subByte * 256;

             subbyte := StrToInt (list[3]);
             temp := temp + subByte;

             result := temp;
        end;

     list.Free;
end;

(* -------------------------------------- *)

function ConvertMaskBitsToLong (BitCount : byte) : longword;

var help  : longword;
    value : longword;
    i     : integer;

begin
     value := 0;

     if BitCount > 0 then
        for i := 1 to BitCount do
            begin
                 value := value shr 1;
                 value := value or $80000000;
            end;

    result := value;
end;

(* -------------------------------------- *)

function ConvertLongToMaskBits (Mask : longword) : byte;

var i        : integer;
    checker  : Int64;
    value    : longword;

begin
     value := 0;
     checker := 1;

     for i := 0 to 31 do
         begin
              if ((Mask and Checker) <> 0 ) then
                 inc (value);

              checker := checker*2;
         end;

     result := byte (value);
end;

(* -------------------------------------- *)

function isValidIPv4String (IPaddress : string) : boolean;

var sOctet  : string;
    iOctet  : integer;
    i       : Integer;
    list    : tStrings;
    isValid : boolean;

begin
     list := tStringlist.Create;
     ExtractStrings (['.'], [], pChar (IPaddress), list);

     // Check the string contains 4 octets

     isValid := (list.Count = 4);

     // Check if IPaddress string contains dots and numbers only

     if isValid then
        begin
             i := 0;

             repeat
                   inc (i);
                   iOctet := ord (IPAddress[i]);

                   isValid := (iOctet = 46) or ((iOctet > 47) and (iOctet < 58));

             until not isValid or (i >= length (IPAddress));
        end;

     // Check

     if isValid then
         begin
              for i := 1 to 4 do
                  begin
                       sOctet := list[i-1];

                       // Check for leading zeros in octet - not allowed.
                       // The BNF in RFC-3986 explicitly defines the dec-octet (for IPv4 addresses) not to have a leading zero

                       if (sOctet <> '0') and (pos ('0', sOctet) = 1) then
                          isValid := false;

                       // Test if octet is in the valid byte range

                       tryStrToInt (sOctet, iOctet);

                          if (iOctet < 0) or (iOctet > 255) then
                             isValid := false;
                  end;
         end;

     list.Free;

     result := isValid;
end;

(* -------------------------------------- *)

function CompareValue (const A, B: tIPv6Address): TValueRelationship;

begin
     if A = B then
       Result := EqualsValue
     else if A < B then
       Result := LessThanValue
     else
       Result := GreaterThanValue;
end;

(* -------------------------------------- *)

function CompareValue (const A, B: tIPAddress) : TValueRelationship; overload;

begin
     if A.IPVersion = 4 then
        begin
             if B.IPVersion = 4 then
                result := CompareValue (a.IPv4Address, b.IPv4Address)
             else
                 Result := LessThanValue;
        end
     else
         begin
              if B.IPVersion = 6 then
                 result := CompareValue (a.IPv6Address, b.IPv6Address)
              else
                  Result := GreaterThanValue;
         end;
end;

(* -------------------------------------- *)

// returns the hex string representing an IPv6 block of 16 bits

function getIPv6TextField (value: word): string;

begin
     // leading zeros MUST be omitted

     if value < $10 then
        result := IntToHex (value, 1)
     else
     if value < $100 then
        result := IntToHex (value, 2)
     else
     if value< $1000 then
        result := IntToHex (value, 3)
     else
         result := IntToHex (value, 4);

     // RFC 5952 Section 4.3: Characters must be lowercase

     result := lowercase (result);
end;

(* -------------------------------------- *)

function ByteStructToWordStruct (value: tIPv6AddressByteStruct): tIPv6AddressWordStruct;

var i    : integer;
    addr : tIPv6AddressWordStruct;

begin
     for i := 0 to 7 do
         addr[i] := (value [2*i] shl 8) + value [(2*i) + 1];

     result := addr;
end;

(* -------------------------------------- *)

function WordStructToByteStruct (value: tIPv6AddressWordStruct): tIPv6AddressByteStruct;

var i    : integer;
    addr : tIPv6AddressByteStruct;

begin
     for i := 0 to 7 do
         begin
              addr[2*i] := value[i] shr 8;
              addr[(2*i) + 1] := value [i] and $ff;
         end;

     result := addr;
end;

(* -------------------------------------- *)

function IPv6AddressIsEqual (value1, value2: tIPv6AddressWordStruct): boolean;

var i : integer;

begin
     result := true;

     for i := 0 to 7 do
         if value1[i] <> value2[i] then
            result := false;
end;

(* -------------------------------------- *)

function ReplaceIPV6DoubleColons (IPaddress: string): string;

var colonblockcount : integer;
    replacecount    : integer;
    i               : integer;
    sHelp           : string;
    value           : string;
    ipv4notation    : boolean;
    list            : tStrings;

begin
     if pos ('::', IPaddress) > 0 then
        begin
             if IPaddress = '::' then
                value := '0:0:0:0:0:0:0:0'
             else
                 begin
                      sHelp := '';

                      // Count single colon blocks. Replace double colons with # mark to avoid confusion when counting blocks

                      IPAddress := stringReplace (IPAddress, '::', '#', [rfReplaceAll]);
                      colonblockcount := numEntries (IPAddress, ':');

                      // Count how may blocks we need to create. With IPv4 notatition its only 6, otherwise 8

                      if pos ('.', IPaddress) > 0 then
                         begin
                              replacecount := 6;
                              ipv4notation := true;
                         end
                      else
                          begin
                               replacecount := 8;
                               ipv4notation := false;
                          end;

                      if pos ('#', IPAddress) = 1 then
                         begin
                              // IP address starts with double colons.

                              if not ipv4notation then
                                 // Easy with normal IPv6 addresses,

                                 for i := 1 to replacecount - colonblockcount do
                                     sHelp := sHelp + '0:'
                              else
                                  // a little more work with IPv4 at the end. In this case we also need to
                                  // create one extra block because the IPv4 block was found as colon block

                                  for i := 1 to (replacecount + 1) - colonblockcount do
                                      sHelp := sHelp + '0:';
                         end
                      else
                          if (pos ('#', IPaddress) = length (IPaddress)) then
                             begin
                                  // IP address ends with double colons

                                  for i := 1 to replacecount - colonblockcount do
                                      sHelp := sHelp + ':0';
                             end
                          else
                              begin
                                   // Double colons are somewhere in the middle

                                   if replacecount = 6 then
                                      begin
                                           // The colon part is terminated by an IPv4 address

                                            for i := 1 to replacecount - colonblockcount do
                                                sHelp := sHelp + ':0';
                                      end
                                   else
                                       begin
                                            // Here we need to go to coloncount + 1 because
                                            // replacing :: with # turns two blocks into one

                                            for i := 1 to replacecount - (colonblockcount + 1) do
                                                sHelp := sHelp + ':0';
                                       end;

                                   sHelp := sHelp + ':';
                              end;

                      // replace double colons with missing blocks

                      value := stringReplace (IPAddress, '#', sHelp, [rfReplaceAll]);
                 end;
        end
     else
         value := IPaddress;

     result := value;
end;

(* -------------------------------------- *)

function isValidIPv6String (IPv6Address : string): boolean;

var isValid             : boolean;
    i                   : integer;
    compressZeroEntries : integer;
    colonblockcount     : integer;
    doublecolonpos      : integer;
    firstdotpos         : integer;
    colonpos            : integer;
    lastcolonpos        : integer;
    slashcount          : integer;
    slashpos            : integer;
    testchar1           : char;
    testchar2           : char;
    block               : string;
    ipv4address         : string;
    sHelp               : string;
    iHelp               : integer;

begin
     isValid := true;

     // Check if it's only hex and colons - and dots for IPv4 represenstation stuff
     // as well as slashes for the mask bits

     for i := 1 to length (IPv6Address) do
         if not (IPv6Address[i] in  ['0'..'9', 'a'..'f', 'A'..'F', ':', '.','/']) then
            isValid := false;

     // Check min length and no single dots or colons at beginning or end

     if length (IPv6Address) > 1 then
        begin
             testchar1 := IPv6Address[1];
             testchar2 := IPv6Address[2];

             if testchar1 = '.' then
                isValid := false;

             if (testchar1 = ':') and (testchar2 <> ':') then
                isValid := false;

             testchar1 := IPv6Address[length (IPv6Address) - 1];
             testchar2 := IPv6Address[length (IPv6Address)];

             if testchar2 = '.' then
                isValid := false;

             if (testchar1 <> ':') and (testchar2 = ':') then
                isValid := false;
        end
     else
         isValid := false;

     // Check if there is zero compression, and if it is, check if there's more than one block

     compressZeroEntries := 0;
     doublecolonpos := PosEx ('::', IPv6Address, 1);

     while doublecolonpos > 0 do
           begin
                inc (compressZeroEntries);
                doublecolonpos := PosEx ('::', IPv6Address, doublecolonpos + 1);
           end;

     if compressZeroEntries > 1 then
        isValid := false;

     // Check if dots exists and if they appear after the last colon

     firstdotpos := pos ('.', IPv6Address);

     if firstdotpos > 0 then
        begin
             colonpos := PosEx (':', IPv6Address, 1);

             while colonpos > 0 do
                   begin
                        lastcolonpos := colonpos;
                        colonpos := PosEx (':', IPv6Address, colonpos + 1);
                   end;

             if firstdotpos < lastcolonpos then
                isValid := false;

             // Check IP address for being valid

             IPv4Address := rightstring (IPv6Address, lastcolonpos + 1);

             if not isValidIPv4String (IPv4Address) then
                isValid := false;
        end;

     // Check if there's a bit mask, and that it's the last entry

     slashcount := pos ('/', IPv6Address);

     if slashcount > 1 then
        isValid := false
     else
         if slashcount = 1 then
            begin
                 slashpos := pos ('/', IPv6Address);

                 if (slashpos < lastcolonpos) then
                    isValid := false;
            end;

     // Check if there are not more than 8 hex blocks
     // Replace double colons with # mark to avoid confusion when counting blocks

     IPv6Address := stringReplace (IPv6Address, '::', '#', [rfReplaceAll]);
     colonblockcount := numEntries (IPv6Address, ':');
     IPv6Address := stringReplace (IPv6Address, '#', '::', [rfReplaceAll]);

     if firstdotpos = 0 then
        begin
             // No IPv4 notation found, so we simply check the 8 blocks

             if (colonblockcount > 8) or ((colonblockcount = 8) and (compressZeroEntries > 0)) then
                isValid := false
             else
                 begin
                      if (colonblockcount < 8) and (compressZeroEntries = 0) then
                         isValid := false
                      else
                          begin
                               IPv6Address := ReplaceIPV6DoubleColons (IPv6Address);

                               // Check if the replacement of double colons has
                               // resulted in more than 8 blocks, like in 1:2:3::4:5:6:7:8:9

                               if NumEntries (IPv6Address, ':') > 8 then
                                  isValid := false;
                          end;
                 end;

             // Check block length

             for i := 1 to 8 do
                 begin
                      block := getEntry (IPv6Address, ':', i);

                      if length (block) > 4 then
                         isValid := false;
                 end;
        end
     else
         begin
              // Check for invalid double colon combinations. Extract the colon string first.

              sHelp := leftstring (IPv6Address, ':', true);
              iHelp := NumEntries (sHelp, ':');

              if pos ('::', sHelp) > 0 then
                 dec (iHelp);

              if iHelp > 6 then
                 isValid := false;

              // The last two blocks are replaced by one IPv4 notation block

              if (colonblockcount > 7) then
                 isValid := false
              else
                  begin
                       if (colonblockcount < 6) and (compressZeroEntries = 0) then
                          isValid := false
                       else
                           IPv6Address := ReplaceIPV6DoubleColons (IPv6Address);
                  end;

              // Check block length

              for i := 1 to 6 do
                  begin
                       block := getEntry (IPv6Address, ':', i);

                       if length (block) > 4 then
                          isValid := false;
                  end;
         end;

     result := isValid;
end;

(* -------------------------------------- *)

function isValidIPv4Subnet (Subnet : string) : boolean;

var net     : string;
    mask    : string;
    bits    : integer;

begin
     net := getEntry (Subnet, '/', 1);
     mask := getEntry (Subnet, '/', 2);

     TryStrToInt (mask, bits);

     result := isValidIPv4String (net) and (bits > 0) and (bits <= 32);
end;

(* -------------------------------------- *)

function isValidIPv6Subnet (Subnet : string) : boolean;

var net     : string;
    mask    : string;
    bits    : integer;

begin
     net := getEntry (Subnet, '/', 1);
     mask := getEntry (Subnet, '/', 2);

     TryStrToInt (mask, bits);

     result := isValidIPv6String (net) and (bits > 0) and (bits <= 128);
end;

(* ----------------------------------- *)

function isValidIPv4Socket (Socket : string) : boolean;

var address : string;
    portstr : string;
    port    : integer;
    index   : integer;
    i       : integer;

begin
     index := pos (':', Socket);

     if index = 0 then
        result := false
     else
         begin
              address := copy (Socket, 1, index - 1);
              portstr := copy (Socket, index + 1, length (Socket) - (index - 1));

              for i := 1 to length(portstr) do
                  begin
                       if (portstr[i] < '0') or (portstr[i] > '9') then
                          exit (false);
                  end;

              tryStrToInt (portstr, port);

              result := isValidIPv4String (address) and ((port >0) and (port < 65536));
         end;
end;

(* ----------------------------------- *)

function isValidIPv6Socket (Socket : string; const AllowSnortNotation : boolean = false) : boolean;

var address : string;
    portstr : string;
    port    : integer;
    block   : string;
    index   : integer;
    i       : integer;

begin
     index := pos (']:', Socket);

     if (index = 0) or (Socket[1] <> '[') then
        begin
             if AllowSnortNotation and (pos ('::', Socket) = 0) then
                begin
                     // This is for notations that use a ninth colon to add the port, like e.g. in a Snort alert.ids log

                     if numEntries (Socket, ':') = 9 then
                        begin
                             // Check if the last block is a valid port

                             block := getEntry (Socket, ':', 9);
                             TryStrToInt (block, port);

                             if (port > 0) and (port <= 65535) then
                                begin
                                     // Last step: check address

                                     address := getEntry (Socket, ':', 1);

                                     for i := 2 to 8 do
                                         address := address + ':' + getEntry (Socket, ':', i);

                                     result := isValidIPv6String (address);
                                end
                             else
                                 result := False;
                        end
                     else
                         result := False;
                end
             else
                 result := false
        end
     else
         begin
              address := copy (Socket, 2, index - 2);
              portstr := copy (Socket, index + 2, length (Socket) - (index - 2));

              for i := 1 to length(portstr) do
                  begin
                       if (portstr[i] < '0') or (portstr[i] > '9') then
                          exit (false);
                  end;

              tryStrToInt (portstr, port);

              result := isValidIPv6String (address) and ((port >0) and (port < 65536));
         end;
end;

(* ----------------------------------- *)

function isValidIPString (Value : string) : boolean;

begin
     result := isValidIPv4String (value) or isValidIPv6String (value);
end;

(* ----------------------------------- *)

function isValidSocketString (Value : string) : boolean;

begin
     result := isValidIPv4Socket (value) or isValidIPv6Socket (value);
end;

(* ----------------------------------- *)

function AnswerTypeToInt (value : tNRAnswerType) : integer;

begin
     case value of
      nraUnknown   : result := 0;
      nraIPAddress : result := 1;
      nraFQDN      : result := 2;
      nraText      : result := 3;
     end;
end;

(* ----------------------------------- *)

function IntegerToAnswerType (value : integer) : tNRAnswerType;

begin
     case Value of
      0 : result := nraUnknown;
      1 : result := nraIPAddress;
      2 : result := nraFQDN;
      3 : result := nraText;
     end;
end;

(* ----------------------------------- *)
{ tMACAddress }
(* ----------------------------------- *)

procedure tMACAddress.initialize;

const ZeroMac : tMACBytes = ($00, $00, $00, $00, $00, $00);

begin
     RangeBytes := ZeroMac;
end;

(* ----------------------------------- *)

function tMACAddress.isBroadcast: boolean;

begin
     result := (RangeBytes[1] = $FF) and
               (RangeBytes[2] = $FF) and
               (RangeBytes[3] = $FF) and
               (RangeBytes[4] = $FF) and
               (RangeBytes[5] = $FF) and
               (RangeBytes[6] = $FF);
end;

(* ----------------------------------- *)

function tMACAddress.isMulticast: boolean;

begin
     result := ((RangeBytes[1] = $01) and
                (RangeBytes[2] = $00) and
                (RangeBytes[3] = $5e) and
                ((RangeBytes[4] >= $00) and (RangeBytes[4] <= $7f))) or

               ((RangeBytes[1] = $33) and
                (RangeBytes[2] = $33));
end;

(* ----------------------------------- *)

function tMACAddress.asString (const Separator : string = ':'; const ForceLowercase : boolean = false) : string;

var value : string;
    i     : integer;

begin
     value := '';

     for i := 1 to 6 do
         begin
              if i < 6 then
                 value := value + IntToHex (rangebytes[i], 2) + separator
              else
                  value := value + IntToHex (rangebytes[i], 2);
         end;

     if ForceLowerCase then
        result := LowerCase (value)
     else
         result := UpperCase (value);
end;

(* ----------------------------------- *)

function tMACAddress.DictionaryKey: string;

begin
     result := asString;
end;

(* ----------------------------------- *)

// for strings like e.g. '01:02:03:04:05:06' with a separator of ':'

procedure tMACAddress.setfromString (const Address : string; const separator : string);

var i         : Integer;
    value     : string;
    temp      : integer;

begin
     if (separator = '') or (pos (separator, Address) = 0) then
        setfromString (Address)
     else
         begin
              for i := 1 to 6 do
                  begin
                       value := '$' + getEntry (Address, Separator, i);

                       tryStrToInt (value, temp);
                       rangebytes[i] := temp;
                  end;
         end;
end;

(* ----------------------------------- *)

// for strings like '010203040506'

procedure tMACAddress.setfromString (const Address: string);

var i         : Integer;
    value     : string;
    temp      : integer;

begin
     for i := 0 to 5 do
         begin
              value := '$' + copy (Address, (i*2) + 1, 2);

              tryStrToInt (value, temp);
              rangebytes[i+1] := temp;
         end;
end;

(* ----------------------------------- *)

class operator tMACAddress.equal(a, b: tMACAddress): boolean;

var i : integer;

begin
     result := true;
     i := 1;

     for i := 1 to 6 do
         if not (a.rangebytes[i] = b.rangebytes[i]) then
            result := false;
end;

(* ----------------------------------- *)

class operator tMACAddress.greaterthan(a, b: tMACAddress): boolean;

var i : integer;

begin
     result := false;
     i := 1;

     while  (i < 7) and (a.rangebytes[i] = b.rangebytes[i]) do
            begin
                 inc (i);
            end;

     if i < 7 then
        result := a.rangebytes[i] > b.rangebytes[i];
end;

(* ----------------------------------- *)

class operator tMACAddress.greaterthanorequal(a, b: tMACAddress): boolean;

var i : integer;

begin
     result := false;
     i := 1;

     while  (i < 7) and (a.rangebytes[i] = b.rangebytes[i]) do
            begin
                 inc (i);
            end;

     if i < 7 then
        result := a.rangebytes[i] >= b.rangebytes[i];
end;

(* ----------------------------------- *)

class operator tMACAddress.lessthan(a, b: tMACAddress): boolean;

var i : integer;

begin
     result := false;
     i := 1;

     while  (i < 7) and (a.rangebytes[i] = b.rangebytes[i]) do
            begin
                 inc (i);
            end;

     if i < 7 then
        result := a.rangebytes[i] < b.rangebytes[i];
end;

(* ----------------------------------- *)

class operator tMACAddress.lessthanorequal(a, b: tMACAddress): boolean;

var i : integer;

begin
     result := false;
     i := 1;

     while  (i < 7) and (a.rangebytes[i] = b.rangebytes[i]) do
            begin
                 inc (i);
            end;

     if i < 7 then
        result := a.rangebytes[i] <= b.rangebytes[i];
end;

(* ----------------------------------- *)

class operator tMACAddress.notequal(a, b: tMACAddress): boolean;

begin
     result := not (a=b);
end;

(* ----------------------------------- *)
{ tMACAddressPair }
(* ----------------------------------- *)

procedure tMACAddressPair.initialize;

begin
     Source.initialize;
     Destination.initialize;

     VLANID := 0;
end;

(* ----------------------------------- *)

procedure tMACAddressPair.initialize (value: string);

var entry : string;

begin
     entry := getEntry (value, '=', 1);
     entry := stringReplace (entry, '-', ':', [rfReplaceAll]);

     Source.setfromString (entry, ':');

     entry := getEntry (value, '=', 2);
     entry := stringReplace (entry, '-', ':', [rfReplaceAll]);

     Destination.setfromString (entry, ':');

     VLANID := 0;
end;

(* ----------------------------------- *)

procedure tMACAddressPair.normalize;

var help : tMACAddress;

begin
     if Source > Destination then
        begin
             help := Source;
             Source := Destination;
             Destination := help;
        end;
end;

(* ----------------------------------- *)

function tMACAddressPair.asString (const Separator: string = ':'; const ForceLowercase: boolean = false): string;

var value : string;

begin
     value := Source.asString (Separator) + '-' + Destination.asString (Separator);

     if ForceLowerCase then
        value := lowercase (value);

     result := lowercase (value)
end;

(* ----------------------------------- *)

function tMACAddressPair.DictionaryKey: string;

var help : tMACAddressPair;

begin
     help := self;
     help.normalize;

     result := help.asString (':', true);
end;

(* ----------------------------------- *)

class operator tMACAddressPair.equal(a, b: tMACAddressPair): boolean;

begin
     result := (a.VLANID = b.VLANID) and
               (((a.Source = b.Source) and (a.Destination = b.Destination)) or
               ((a.Source = b.Destination) and (a.Destination = b.Source)));
end;

(* ----------------------------------- *)

class operator tMACAddressPair.notequal(a, b: tMACAddressPair): boolean;

begin
     result := not (a=b);
end;

(* ----------------------------------- *)

class operator tMACAddressPair.greaterthan (a, b: tMACAddressPair): boolean;

begin
     if a.VLANID = b.VLANID then
        begin
             if a.Source = b.Source then
                result := a.Destination > b.Destination
             else
                 result := a.Source > b.Source;
        end
     else
         result := a.VLANID > b.VLANID;
end;

(* ----------------------------------- *)

class operator tMACAddressPair.greaterthanorequal (a, b: tMACAddressPair): boolean;

begin
     if a.VLANID = b.VLANID then
        begin
             if a.Source = b.Source then
                result := a.Destination >= b.Destination
             else
                 result := a.Source > b.Source;
        end
     else
         result := a.VLANID > b.VLANID;
end;

(* ----------------------------------- *)

class operator tMACAddressPair.lessthan(a, b: tMACAddressPair): boolean;

begin
     if a.VLANID = b.VLANID then
        begin
             if a.Source = b.Source then
                result := a.Destination < b.Destination
             else
                 result := a.Source < b.Source;
        end
     else
         result := a.VLANID < b.VLANID;
end;

(* ----------------------------------- *)

class operator tMACAddressPair.lessthanorequal(a, b: tMACAddressPair): boolean;

begin
     if a.VLANID = b.VLANID then
        begin
             if a.Source = b.Source then
                result := a.Destination <= b.Destination
             else
                 result := a.Source < b.Source;
        end
     else
         result := a.VLANID < b.VLANID;
end;

(* ----------------------------------- *)
{ tConversationAnchor }
(* ----------------------------------- *)

class operator tConversationAnchor.equal (a, b: tConversationAnchor): boolean;

begin
     result := (a.VLANId = b.VLANId) and
//             (a.TTLAnchorPair = b.TTLAnchorPair) and
               (a.MACPairN = b.MACPairN) and
               (a.InterfaceID = b.InterfaceID);
end;

(* ----------------------------------- *)

class operator tConversationAnchor.notequal (a, b: tConversationAnchor): boolean;

begin
     result := not (a=b);
end;


(* ----------------------------------- *)

procedure tConversationAnchor.initialize;

begin
     VLANId := 0;
     MACPairN.initialize;
//   TTLAnchorPair.initialize;
     InterfaceID := 0;
end;

(* ----------------------------------- *)

function tConversationAnchor.DictionaryKey: string;

begin
     result := 'VLAN' + IntToStr (VLANId) +
               'Interface' + IntToStr (InterfaceID) +
//             '-SourceTTL' + IntToStr (TTLAnchorPair.SourceTTL) +
//             '-DestinationTTL' +
//             IntToStr (TTLAnchorPair.DestinationTTL);
               MACPairN.DictionaryKey;
end;

(* -------------------------------------- *)
{ tIPv4Address }
(* -------------------------------------- *)

procedure tIPv4Address.initialize (const SetRandomIP : boolean = false; const Scope : tIPRandomScope = iprsAll);

var subByte  : byte;
    subRange : integer;
    boOkay   : boolean;

begin
     if SetRandomIP then
        begin
             boOkay := true;

             case Scope of
              iprsPrivate : begin
                                 subrange := random (3);

                                 case subrange of
                                  0 : begin
                                           // 10.x.y.z network

                                           ValueAsLong := $A000000;     // 10.0.0.0

                                           subByte := random (256);
                                           ValueAsLong := ValueAsLong + (subByte * 256 * 256);

                                           subByte := random (256);
                                           ValueAsLong := ValueAsLong + (subByte * 256);

                                           subByte := random (255);
                                           ValueAsLong := ValueAsLong + subByte;
                                      end;
                                  1 : begin
                                           // 172.x.y.z network

                                           ValueAsLong := $AC000000;   // 172.0.0.0

                                           subByte := 16 + random (16);
                                           ValueAsLong := ValueAsLong + (subByte * 256 * 256);

                                           subByte := random (256);
                                           ValueAsLong := ValueAsLong + (subByte * 256);

                                           subByte := random (255);
                                           ValueAsLong := ValueAsLong + subByte;
                                      end;
                                  2 : begin
                                           // 192.168.x.y network

                                           ValueAsLong := $C0A80000;  // 192.168.0.0

                                           subByte := random (256);
                                           ValueAsLong := ValueAsLong + (subByte * 256);

                                           subByte := random (255);
                                           ValueAsLong := ValueAsLong + subByte;
                                      end;
                                 end;
                            end;
              iprsAll     : begin
                                 repeat
                                       // Avoid Loopback and 0.0.0.0/8

                                       repeat
                                             subByte := random (255) + 1;
                                             ValueAsLong := subByte;
                                       until (ValueAsLong <> 127);

                                       ValueAsLong := ValueAsLong * 256 * 256 * 256;

                                       subByte := random (256);
                                       ValueAsLong := ValueAsLong + subByte * 256 * 256;

                                       subByte := random (256);
                                       ValueAsLong := ValueAsLong + subByte * 256;

                                       subByte := random (256);
                                       ValueAsLong := ValueAsLong + subByte;

                                       // Check the result

                                       boOkay := not (self.IsMulticast or self.IsBroadcast);

                                 until boOkay;
                            end;
             end;
        end
     else
         ValueAsLong := 0;

     MaskBitCount := 32;
end;

(* -------------------------------------- *)

class operator tIPv4Address.equal(a, b: tIPv4Address): boolean;

begin
     result := (a.ValueAsLong = b.ValueAsLong) and (a.MaskBitCount = b.MaskBitCount);
end;

(* -------------------------------------- *)

class operator tIPv4Address.notequal(a, b: tIPv4Address): boolean;

begin
     result := not (a=b);
end;

(* -------------------------------------- *)

class operator tIPv4Address.greaterthan (a, b: tIPv4Address): boolean;

begin
     result := (a.ValueAsLong > b.ValueAsLong) or ((a.ValueAsLong = b.ValueAsLong) and (a.MaskBitCount > b.MaskBitCount));
end;

(* -------------------------------------- *)

class operator tIPv4Address.greaterthanorequal (a,  b: tIPv4Address): boolean;

begin
     result := a.ValueAsLong >= b.ValueAsLong;
end;

(* -------------------------------------- *)

class operator tIPv4Address.lessthan (a, b: tIPv4Address): boolean;

begin
     result := (a.ValueAsLong < b.ValueAsLong) or ((a.ValueAsLong = b.ValueAsLong) and (a.MaskBitCount < b.MaskBitCount));
end;

(* -------------------------------------- *)

class operator tIPv4Address.lessthanorequal (a, b: tIPv4Address): boolean;

begin
     result := a.ValueAsLong <= b.ValueAsLong;
end;

(* -------------------------------------- *)

function tIPv4Address.IsAPIPA: boolean;

begin
      result := (ValueAsLong and IPv4_APIPANet) = IPv4_APIPANet;
end;

(* -------------------------------------- *)

function tIPv4Address.IsBroadcast: boolean;

var mask     : longword;
    hostbits : longword;
    network  : longword;

begin
     if MaskBitCount = 32 then
        result := (ValueAsLong = IPv4_Broadcast)
     else
         begin
              mask := ConvertMaskBitsToLong (MaskBitCount);
              network := ValueAsLong AND mask;
              hostbits := ValueAsLong AND NOT network;

              result := ((hostbits OR mask) = IPv4_Broadcast)
         end;
end;

(* -------------------------------------- *)

function tIPv4Address.isCommon: boolean;

begin
     result := not (isBroadcast or isMulticast or isLoopback or isZeroAddress);
end;

(* -------------------------------------- *)

function tIPv4Address.isDocumentationAddress: boolean;

begin
     // Check if IP is in the ranges defined for documentation purposes, which are
     // 192.0.2.0/24 (TEST-NET-1), 198.51.100.0/24 (TEST-NET-2) and 203.0.113.0/24
     // see RFC 5737, section 3

     result := ((ValueAsLong >= 3221225984) and (ValueAsLong <= 3221226239)) or
               ((ValueAsLong >= 3325256704) and (ValueAsLong <= 3325256959)) or
               ((ValueAsLong >= 3405803776) and (ValueAsLong <= 3405804031));
end;

(* -------------------------------------- *)

function tIPv4Address.IsLegalDestination: boolean;

var bFirstOctet : byte;

begin
     // IP must be >= 1.0.0.0 and <= 224.0.0.0 and <> 127.x.y.z

     bFirstOctet := ValueAsLong shr 24;

     result := ((bFirstOctet > 0) and (bFirstOctet <> 127) and (bFirstOctet < 240)) or (ValueAsLong = IPv4_Broadcast);
end;

(* -------------------------------------- *)

function tIPv4Address.IsLegalSource: boolean;

var bFirstOctet : byte;

begin
     // IP must be >= 1.0.0.0 and <= 224.0.0.0 and <> 127.x.y.z

     bFirstOctet := ValueAsLong shr 24;

     result := (bFirstOctet > 0) and (bFirstOctet <> 127) and (bFirstOctet < 224);
end;

(* -------------------------------------- *)

function tIPv4Address.IsMulticast: boolean;

var bFirstOctet : byte;

begin
     // first octet must be > 224 and < 240;

     bFirstOctet := ValueAsLong shr 24;

     result := ((bFirstOctet > 223) and (bFirstOctet < 240));
end;

(* -------------------------------------- *)

function tIPv4Address.IsPrivate: boolean;

var bFirstOctet : byte;
    bSecondOctet : byte;

begin
     bFirstOctet := ValueAsLong shr 24;
     bSecondOctet := ((ValueAsLong shr 16) and $ff);

     result := (bFirstOctet = 10) or
               ((bFirstOctet = 172) and (bSecondOctet > 15) and (bSecondOctet < 32)) or
               ((bFirstOctet = 192) and (bSecondOctet = 168));
end;

(* -------------------------------------- *)

function tIPv4Address.isZeroAddress: boolean;

begin
     result := (ValueAsLong = 0);
end;

(* -------------------------------------- *)

function tIPv4Address.isLoopback: boolean;

begin
     result := (ValueAsLong and $7F000000) = $7F000000
end;

(* -------------------------------------- *)

function tIPv4Address.MaskAsLong: longword;

begin
     result := ConvertMaskBitsToLong (MaskBitCount);
end;

(* -------------------------------------- *)

function tIPv4Address.MaskAsString: string;

var mask : longword;

begin
     mask := ConvertMaskBitsToLong (MaskBitCount);

     result := LongToIP (mask);
end;

(* -------------------------------------- *)

function tIPv4Address.ValueAsByteArray: tIPv4ByteArray;

var help  : longword;
    value : tIPv4ByteArray;

begin
     help := (ValueAsLong and $ff000000) shr 24;
     value[1] := help;

     help := (ValueAsLong and $ff0000) shr 16;
     value[2] := help;

     help := (ValueAsLong and $ff00) shr 8;
     value[3] := help;

     help := (ValueAsLong and $ff);
     value[4] := help;
end;

(* -------------------------------------- *)

procedure tIPv4Address.setMaskFromLong (const Value: longword);

begin
     MaskBitCount := ConvertLongToMaskBits (Value);
end;

(* -------------------------------------- *)

procedure tIPv4Address.setMaskFromString (const Value: string);

var mask : longword;

begin
     mask := IPToLong (Value);

     MaskBitCount := ConvertLongToMaskBits (mask);
end;

(* -------------------------------------- *)

procedure tIPv4Address.setValueFromString (const IPAddress: string; const MaskBits : byte = 32);

begin
     if isValidIPv4String (IPAddress) then
        begin
             ValueAsLong := IPToLong (IPAddress);
             MaskBitCount := MaskBits;
        end
     else
         begin
              valueAsLong := 0;
              MaskBitCount := 32;
         end;
end;

(* -------------------------------------- *)

procedure tIPv4Address.setValueFromCIDRString (const IPAddress: string);

var Addr   : string;
    BitStr : string;
    Bits   : integer;

begin
     if pos ('/', IPAddress) > 0 then
        begin
             Addr := GetEntry (IPAddress, '/', 1);
             BitStr := GetEntry (IPAddress, '/', 2);
             TryStrToInt (BitStr, Bits);

             setValueFromString (Addr, Bits);
        end
     else
         raise EInOutError.Create ('Error: Invalid IPv4 address CIDR format detected.');
end;

(* -------------------------------------- *)

procedure tIPv4Address.setValueFromByteArray (const IPAddress: tIPv4ByteArray);

var i : integer;

begin
     valueAsLong := IPAddress[4];

     for i := 3 downto 1 do
         begin
              ValueAsLong := ValueAsLong shl 8;
              ValueAsLong := ValueAsLong + IPAddress[i];
         end;
end;

(* -------------------------------------- *)

function tIPv4Address.ValueAsCIDRString: string;

begin
     result := ValueAsString + '/' + IntToStr (MaskBitCount);
end;

(* -------------------------------------- *)

function tIPv4Address.ValueAsString: string;

begin
     result := LongToIP (ValueAsLong);
end;

(* -------------------------------------- *)

function tIPv4Address.LowestNetIP: tIPv4Address;

begin
     result.ValueAsLong := (ValueAsLong AND MaskAsLong);
     result.MaskBitCount := 32;
end;

(* -------------------------------------- *)

function tIPv4Address.HighestNetIP: tIPv4Address;

begin
     result.ValueAsLong := ((ValueAsLong AND MaskAsLong) or (IPv4_Broadcast - MaskAsLong));
     result.MaskBitCount := 32;
end;

(* -------------------------------------- *)

function tIPv4Address.NetWildCard: tIPv4Address;

begin
     result.ValueAsLong := IPv4_Broadcast - MaskAsLong;
     result.MaskBitCount := 32;
end;

(* -------------------------------------- *)

function tIPv4Address.BelongsToSubnet (Subnet: tIPv4Address): boolean;

var help     : tIPv4Address;
    baseaddr : tIPv4Address;

begin
     help := self;

     // Take mask of the subnet and apply it to the IPaddress

     help.MaskBitCount := subnet.MaskBitCount;
     baseaddr := help.LowestNetIP;

     // See if the address matches the subnet

     result := (baseaddr.ValueAsLong = subnet.ValueAsLong);
end;

(* -------------------------------------- *)

function tIPv4Address.isUndefined: boolean;

begin
     result := (ValueAsLong = 0) and (MaskBitCount = 32);
end;

(* -------------------------------------- *)

function tIPv4Address.DictionaryKey: string;

begin
     result := ValueAsString;
end;

(* -------------------------------------- *)
{ tIPv4AddressPair }
(* -------------------------------------- *)

procedure tIPv4AddressPair.initialize;

begin
     Source.initialize;
     Destination.initialize;
end;

(* -------------------------------------- *)

procedure tIPv4AddressPair.initialize (value: string; const SetCIDR : boolean);

var entry : string;

begin
     entry := getEntry (value, '=', 1);

     if SetCIDR then
        Source.setValueFromCIDRString (entry)
     else
         Source.setValueFromString (entry);

     entry := getEntry (value, '=', 2);

     if SetCIDR then
        Destination.setValueFromCIDRString (entry)
     else
         Destination.setValueFromString (entry);
end;

(* -------------------------------------- *)

class operator tIPv4AddressPair.equal (a, b: tIPv4AddressPair): boolean;

begin
     result := ((a.Source = b.Source) and (a.Destination = b.Destination)) or
               ((a.Source = b.Destination) and (a.Destination = b.Source))
end;

(* -------------------------------------- *)

class operator tIPv4AddressPair.greaterthan (a, b: tIPv4AddressPair): boolean;

begin
     if a.Source = b.Source then
        result := a.Destination > b.Destination
     else
         result := a.Source > b.Source;
end;

(* -------------------------------------- *)

class operator tIPv4AddressPair.greaterthanorequal (a, b: tIPv4AddressPair): boolean;

begin
     if a.Source = b.Source then
        result := a.Destination > b.Destination
     else
         result := a.Source >= b.Source;
end;

(* -------------------------------------- *)

class operator tIPv4AddressPair.lessthan (a, b: tIPv4AddressPair): boolean;

begin
     if a.Source = b.Source then
        result := a.Destination < b.Destination
     else
         result := a.Source < b.Source;
end;

(* -------------------------------------- *)

class operator tIPv4AddressPair.lessthanorequal (a, b: tIPv4AddressPair): boolean;

begin
     if a.Source = b.Source then
        result := a.Destination <= b.Destination
     else
         result := a.Source <= b.Source;
end;

(* -------------------------------------- *)

class operator tIPv4AddressPair.notequal (a, b: tIPv4AddressPair): boolean;

begin
     result := not (a = b);
end;

(* -------------------------------------- *)

(* -------------------------------------- *)

procedure tIPv4AddressPair.normalize;

var help : tIPv4Address;

begin
     if Source > Destination then
        swap;
end;

(* -------------------------------------- *)

procedure tIPv4AddressPair.swap;

var help : tIPv4Address;

begin
     help := Source;
     Source := Destination;
     Destination := help;
end;

(* -------------------------------------- *)

function tIPv4AddressPair.ValueAsString (const NoSpaces : boolean) : string;

begin
     if NoSpaces then
        result := Source.ValueAsString + '-' + Destination.ValueAsString
     else
         result := Source.ValueAsString + ' - ' + Destination.ValueAsString
end;

(* -------------------------------------- *)

function tIPv4AddressPair.isUndefined: boolean;

begin
     result := source.isUndefined and destination.isUndefined;
end;

(* -------------------------------------- *)

function tIPv4AddressPair.DictionaryKey: string;

var help : tIPv4AddressPair;

begin
     help := self;
     help.normalize;

     result := help.ValueAsString(true);
end;

(* -------------------------------------- *)
{ tIPv6Address }
(* -------------------------------------- *)

procedure tIPv6Address.initialize (const SetRandomIP : boolean = false; const Scope : tIPRandomScope = iprsAll);

var i : integer;

begin
     MaskBitCount := 64;

     if not SetRandomIP then
        ValueAsWordStruct := IPv6_NODE_UNSPECIFIED
     else
         begin
              case Scope of
               iprsPrivate : begin
                                  // The range is fc00::/7, so fc00-fdff

                                  ValueAsWordStruct[0] := $fc00 + random ($200);
                             end;
               iprsAll     : begin
                                  // Make sure the first block is either private or in $2000/3

                                  if random (1) = 0 then
                                     ValueAsWordStruct[0] := $fc00 + random ($200)
                                  else
                                      ValueAsWordStruct[0] := $2000 + random ($2000);
                             end;
              end;

              for i := 1 to 7 do
                  ValueAsWordStruct[i] := random (65536);
         end;
end;

(* -------------------------------------- *)

procedure tIPv6Address.initialize (const Prefix: tIPv6Address; const SetRandomIP: boolean);

var i : integer;

begin
     MaskBitCount := 64;

     // Init host part

     initialize (SetRandomIP);

     // Transfer Prefix, overwriting the prefix part

     copyPrefixFrom (Prefix);
end;

(* -------------------------------------- *)

class operator tIPv6Address.equal(a, b: tIPv6Address): boolean;

var i : integer;

begin
     result := true;

     // Check Address

     for i := 0 to 7 do
         if a.ValueAsWordStruct[i] <> b.ValueAsWordStruct[i] then
            result := false;

     // Check Prefix

     if result then
        result := a.MaskBitCount = b.MaskBitCount;
end;

(* -------------------------------------- *)

class operator tIPv6Address.notequal(a, b: tIPv6Address): boolean;

begin
     result := not (a=b);
end;

(* -------------------------------------- *)

class operator tIPv6Address.greaterthan (a, b: tIPv6Address): boolean;

var isGreater : boolean;
    i         : integer;

begin
     if not (a=b) then
        isGreater := not (a < b)
     else
         isGreater := false;

     result := isGreater;
end;

(* -------------------------------------- *)

class operator tIPv6Address.greaterthanorequal (a, b: tIPv6Address): boolean;

begin
     result := (a=b) or (a > b);
end;

(* -------------------------------------- *)

class operator tIPv6Address.lessthan (a, b: tIPv6Address): boolean;

var isLesser : boolean;
    i        : integer;

begin
     if not (a=b) then
        begin
             i := 0;

             while (i < 7) and (a.ValueAsWordStruct[i] = b.ValueAsWordStruct[i]) do
                   inc (i);

             isLesser := a.ValueAsWordStruct[i] < b.ValueAsWordStruct[i];
        end
     else
         isLesser := a.MaskBitCount > b.MaskBitCount;                     // larger mask = smaller net

     result := isLesser;
end;

(* -------------------------------------- *)

class operator tIPv6Address.lessthanorequal (a, b: tIPv6Address): boolean;

begin
     result := not (a > b);
end;

(* -------------------------------------- *)

function tIPv6Address.getExpandedValue: string;

var value : string;
    part  : string;
    i     : integer;

begin
     value := '';

     for i := 0 to 7 do
         begin
              part := getIPv6TextField (ValueAsWordStruct[i]);

              if (i < 7) then
                 value := value + part + ':'
              else
                  value := value + part;
         end;

     result := value;
end;

(* -------------------------------------- *)

function tIPv6Address.getIPv4NotationValue: string;

var value       : string;
    ipv4address : tIPv4Address;
    ipv4value   : longword;
    i           : integer;

begin
     value := '';

     // Create IPv6 part

     for i := 0 to 5 do
         value := value + getIPv6TextField (ValueAsWordStruct[i]) + ':';

     ipv4value := ValueAsWordStruct[6] * 65535 + ValueAsWordStruct[7];

     ipv4address.ValueAsLong := ipv4value;
     value := value + ipv4address.ValueAsString;

     result := value;
end;

(* -------------------------------------- *)

function tIPv6Address.getMulticastScope: IPv6MulticastScope;

var scope : longword;

begin
     // set all bits of the first word to zero except the last 4 (containing the scope)

     scope := (ValueAsWordStruct[1] and 15);

     case scope of
      0, 3, 15  : result := msReserved;
      1         : result := msInterfaceLocal;
      2         : result := msLinkLocal;
      4         : result := msAdminLocal;
      5         : result := msSiteLocal;
      6,7,9..13 : result := msNotAssigned;
      8         : result := msOrganizationLocal;
      14        : result := msGlobal;
     end;
end;

(* -------------------------------------- *)

function tIPv6Address.ValueAsByteStruct: tIPv6AddressByteStruct;

begin
     result := WordStructToByteStruct (ValueAsWordStruct);
end;

(* -------------------------------------- *)

// According to rfc5952

function tIPv6Address.ValueAsString: string;

var value             : string;
    field              : string;
    BlockActive       : boolean;
    zeroblockstart    : integer;
    zeroblocklen      : integer;
    longestblocklen   : integer;
    longestblockstart : integer;
    i                 : integer;
    count             : integer;

begin
     longestblocklen := 0;
     longestblockstart := 0;
     zeroblocklen := 0;

     // Find longest zero block range

     for i := 0 to 7 do
         begin
              if ValueAsWordStruct[i] = 0 then
                 begin
                      if not BlockActive then
                         begin
                              BlockActive := true;

                              // New Zero found, maybe a block

                              zeroblockstart := i;
                              zeroblocklen := 1;
                         end
                      else
                          begin
                               inc (zeroblocklen);
                          end;
                 end
              else
                  begin
                       BlockActive := false;

                       // RFC 5952 Section 4.2.2: The symbol "::" MUST NOT be used to shorten just one 16-bit 0 field.
                       // So lets ignore zero block length of 1.

                       if (zeroblocklen > 1) and (zeroblocklen > longestblocklen) then
                          begin
                               longestblocklen := zeroblocklen;
                               longestblockstart := zeroblockstart;
                          end;

                       zeroblockstart := 0;
                       zeroblocklen := 0;
                  end;
         end;

     if (longestblocklen = 0) and (zeroblocklen > 1) then
        begin
             longestblocklen := zeroblocklen;
             longestblockstart := zeroblockstart;
        end;

     if zeroblocklen = 8 then
        value := '::'
     else
         // Found a valid block range, lets shorten it

         if longestblocklen > 1 then
            begin
                 i := -1;

                 while i < 7 do
                       begin
                            inc (i);

                            if i = longestblockstart then
                               begin
                                    // Insert double colons and skip the according blocks

                                    field := '::';
                                    inc (i, longestblocklen - 1);
                               end
                            else
                                begin
                                     field := getIPv6TextField (ValueAsWordStruct[i]);
                                end;

                            // Add field to value

                            if field = '::' then
                               begin
                                    // Only add two colons if the value is empty
                                    // otherwise the "first" of the double colons is already
                                    // in there from the last value

                                    if value = '' then
                                       value := '::'
                                    else
                                        value := value + ':';
                               end
                            else
                                begin
                                     if (i < 7) then
                                        value := value + field + ':'
                                     else
                                         value := value + field;
                                end;
                       end;
            end
         else
             value := getExpandedValue;

     result := value;
end;

(* -------------------------------------- *)

function tIPv6Address.ValueAsSlashedNotationString: string;

begin
     result := ValueAsString + '/' + IntToStr (MaskBitCount);
end;

(* -------------------------------------- *)

function tIPv6Address.getSolicitedNodeAddress: tIPv6Address;

begin
     result.ValueAsWordStruct[0] := $ff02;
     result.ValueAsWordStruct[1] := 0;
     result.ValueAsWordStruct[2] := 0;
     result.ValueAsWordStruct[3] := 0;
     result.ValueAsWordStruct[4] := 0;
     result.ValueAsWordStruct[5] := 1;
     result.ValueAsWordStruct[6] := $ff00 or (ValueAsWordStruct[6] and $ff);
     result.ValueAsWordStruct[7] := ValueAsWordStruct[7];
     result.MaskBitCount := 64;
end;

(* -------------------------------------- *)

procedure tIPv6Address.CopyPrefixFrom (Address: tIPv6Address; const PrefixLength: byte);

var len   : integer;
    i     : integer;
    count : integer;
    bits  : integer;
    mask  : word;

begin
     MaskBitCount := PrefixLength;

     if PrefixLength = 0 then
        len := 64
     else
         len := PrefixLength;

     if len > 128 then
        raise Exception.Create ('Error: Prefix length exceeds 128 bits (tIPv6Address.CopyPrefixFrom)');

     // Do the simple cases quickly

     case len of
      16 : ValueAsWordStruct[0] := Address.ValueAsWordStruct[0];
      32 : begin
                ValueAsWordStruct[0] := Address.ValueAsWordStruct[0];
                ValueAsWordStruct[1] := Address.ValueAsWordStruct[1];
           end;
      48 : begin
                ValueAsWordStruct[0] := Address.ValueAsWordStruct[0];
                ValueAsWordStruct[1] := Address.ValueAsWordStruct[1];
                ValueAsWordStruct[2] := Address.ValueAsWordStruct[2];
           end;
      56 : begin
                ValueAsWordStruct[0] := Address.ValueAsWordStruct[0];
                ValueAsWordStruct[1] := Address.ValueAsWordStruct[1];
                ValueAsWordStruct[2] := Address.ValueAsWordStruct[2];

                // Wipe the first octet, then copy

                ValueAsWordStruct[3] := ValueAsWordStruct[3] and $00FF;
                ValueAsWordStruct[3] := ValueAsWordStruct[3] or (Address.ValueAsWordStruct[2] and $FF00);
           end;
      64 : begin
                ValueAsWordStruct[0] := Address.ValueAsWordStruct[0];
                ValueAsWordStruct[1] := Address.ValueAsWordStruct[1];
                ValueAsWordStruct[2] := Address.ValueAsWordStruct[2];
                ValueAsWordStruct[3] := Address.ValueAsWordStruct[3];
           end;
     else
         begin
              // darn... the more complicated stuff

              // copy the easy blocks first

              count := len div 16;

              for i := 1 to count do
                  ValueAsWordStruct[i-1] := Address.ValueAsWordStruct[i-1];

              // now do the rest, which must be less than 16 bits

              bits := len - (count * 16);

              case bits of
                0 : mask := $0000;
                1 : mask := $8000;
                2 : mask := $C000;
                3 : mask := $E000;
                4 : mask := $F000;
                5 : mask := $F800;
                6 : mask := $FC00;
                7 : mask := $FE00;
                8 : mask := $FF00;
                9 : mask := $FF80;
               10 : mask := $FFC0;
               11 : mask := $FFE0;
               12 : mask := $FFF0;
               13 : mask := $FFF8;
               14 : mask := $FFFC;
               15 : mask := $FFFE;
              end;

              // Wipe, then copy

              ValueAsWordStruct[count] := ValueAsWordStruct[count] and ($FFFF - mask);
              ValueAsWordStruct[count] := ValueAsWordStruct[count] or (Address.ValueAsWordStruct[count] and mask);
         end;
     end;
end;

(* -------------------------------------- *)

function tIPv6Address.getEUI64baseMAC : tMACAddress;

var value : tMACAddress;
    help  : tIPv6AddressByteStruct;

begin
     help := ValueAsByteStruct;

     value.initialize;

     // Flip locally administered bit

     if (help[8] and $2) = $2 then
        value.RangeBytes[1] := help[8] and $FD    // $FD = 11111101
     else
         value.RangeBytes[1] := help[8] or $2;

     value.RangeBytes[2] := help[9];
     value.RangeBytes[3] := help[10];

     // Skipping over the "$FF $FE" sequence here

     value.RangeBytes[4] := help[13];
     value.RangeBytes[5] := help[14];
     value.RangeBytes[6] := help[15];

     result := value;
end;

(* -------------------------------------- *)

function tIPv6Address.getMCastBaseMAC: tMACAddress;

var value : tMACAddress;
    help  : tIPv6AddressByteStruct;

begin
     help := ValueAsByteStruct;

     value.initialize;

     value.RangeBytes[1] := $33;
     value.RangeBytes[2] := $33;
     value.RangeBytes[3] := help[12];
     value.RangeBytes[4] := help[13];
     value.RangeBytes[5] := help[14];
     value.RangeBytes[6] := help[15];

     result := value;
end;

(* -------------------------------------- *)

function tIPv6Address.IsEUI64Address: boolean;

var help : tIPv6AddressByteStruct;

begin
     help := ValueAsByteStruct;

     result := (help[11] = $ff) and (help[12] = $fe);
end;

(* -------------------------------------- *)

function tIPv6Address.IsLinkLocalAddress: boolean;

begin
     result := (ValueAsWordStruct[0] = $fe80) and
               (ValueAsWordStruct[1] = 0) and
               (ValueAsWordStruct[2] = 0) and
               (ValueAsWordStruct[3] = 0);
end;

(* -------------------------------------- *)

function tIPv6Address.IsLegalDestination: boolean;

begin
     result := true;

     if isLocalhost then
        result := false;
end;

(* -------------------------------------- *)

function tIPv6Address.IsLegalSource: boolean;

begin
     result := true;

     if IsLocalHost or IsMulticast then
        result := false;
end;

(* -------------------------------------- *)

function tIPv6Address.IsLocalHost: boolean;

begin
      result := IPv6AddressIsEqual (ValueAsWordStruct, IPv6_NODE_LOCALHOST);
end;

(* -------------------------------------- *)

function tIPv6Address.IsMulticast: boolean;

begin
     result := (ValueAsWordStruct[0] shr 8) = $ff;
end;

(* -------------------------------------- *)

function tIPv6Address.isSolicitedNodeMCast: boolean;

begin
     result := (ValueAsWordStruct[0] = $ff02) and
               (ValueAsWordStruct[1] = 0) and
               (ValueAsWordStruct[2] = 0) and
               (ValueAsWordStruct[3] = 0) and
               (ValueAsWordStruct[4] = 0) and
               (ValueAsWordStruct[5] = 1) and
               ((ValueAsWordStruct[6] shr 8) = $ff);
end;

(* -------------------------------------- *)

function tIPv6Address.IsTemporaryMCast: boolean;

begin
     // Temporary Multicast addresses have a flag set to one at the first field at position 0000 0000 000x 0000 (decimal 16)

     result := (ValueAsWordStruct[0] and 16) = 16;
end;

(* -------------------------------------- *)

function tIPv6Address.IsUnspecifiedAddress: boolean;

begin
     result := IPv6AddressIsEqual (ValueAsWordStruct, IPv6_NODE_UNSPECIFIED);
end;

(* -------------------------------------- *)

function tIPv6Address.IsDocumentationAddress: boolean;

begin
     result := (ValueAsWordStruct[0] = $2001) and (ValueAsWordStruct[1] = $0db8);
end;

(* -------------------------------------- *)

function tIPv6Address.isCommon: boolean;

begin
     result := not (IsMulticast or isUnspecifiedAddress)
end;

(* -------------------------------------- *)

procedure tIPv6Address.SetValueFromByteStruct (const Value: tIPv6AddressByteStruct);

begin
     ValueAsWordStruct := ByteStructToWordStruct (Value);
end;

(* -------------------------------------- *)

procedure tIPv6Address.setValueFromString (const IPAddress : string; const PrefixLength : byte = 64);

var value  : string;
    help   : integer;
    octet1 : integer;
    octet2 : integer;
    i      : integer;
    block  : string;

begin
     if isValidIPv6String (IPAddress) then
        begin
             MaskBitCount := PrefixLength;

             value := ReplaceIPV6DoubleColons (IPAddress);

             if pos ('.', IPAddress) = 0 then
                begin
                     // Assign Ipv6 part

                     for i := 1 to 8 do
                         begin
                              block := '$' + GetEntry (value, ':', i);

                              tryStrToInt (block, help);
                              ValueAsWordStruct[i-1] := help;
                         end;
                end
             else
                 begin
                      // Assign Ipv6 part

                      for i := 1 to 6 do
                          begin
                               block := '$' + GetEntry (value, ':', i);

                               tryStrToInt (block, help);
                               ValueAsWordStruct[i-1] := help;
                          end;

                      // Assign IPv4 part

                      value := getEntry (value, ':', 7);

                      tryStrToInt (GetEntry (value, '.', 1), octet1);
                      tryStrToInt (GetEntry (value, '.', 2), octet2);

                      ValueAsWordStruct[6] := (octet1 * 256 + octet2);

                      tryStrToInt (GetEntry (value, '.', 3), octet1);
                      tryStrToInt (GetEntry (value, '.', 4), octet2);

                      ValueAsWordStruct[7] := (octet1 * 256 + octet2);
                 end;
        end
     else
         raise EInOutError.Create ('Error: Invalid IPv6 address format detected.');
end;

(* -------------------------------------- *)

procedure tIPv6Address.setValueFromStringWithMask(const IPAddress: string);

var Addr   : string;
    BitStr : string;
    Bits   : integer;

begin
     if pos ('/', IPAddress) > 0 then
        begin
             Addr := GetEntry (IPAddress, '/', 1);
             BitStr := GetEntry (IPAddress, '/', 2);
             TryStrToInt (BitStr, Bits);

             setValueFromString (Addr, Bits);
        end
     else
         raise EInOutError.Create ('Error: Invalid IPv6 address slashed notation format detected.');
end;

(* -------------------------------------- *)

procedure tIPv6Address.setValueFromMAC (const Prefix: tIPv6Address; const MACAddress: tMACAddress);

var help : byte;

begin
     ValueAsWordStruct := IPv6_NODE_UNSPECIFIED;
     MaskBitCount := 64;

     // Flip the locally administered bit

     if (MacAddress.RangeBytes[1] and $2) = $2 then
        help := MACAddress.RangeBytes[1] and $FD    // $FD = 11111101
     else
         help := MACAddress.RangeBytes[1] or $2;

     ValueAsWordStruct[4] := help shl 8;
     ValueAsWordStruct[4] := ValueAsWordStruct[4] or MACAddress.RangeBytes[2];

     ValueAsWordStruct[5] := MACAddress.RangeBytes[3] shl 8;
     ValueAsWordStruct[5] := ValueAsWordStruct[5] or $ff;

     ValueAsWordStruct[6] := $fe00;
     ValueAsWordStruct[6] := ValueAsWordStruct[6] or MACAddress.RangeBytes[4];

     ValueAsWordStruct[7] := MACAddress.RangeBytes[5] shl 8;
     ValueAsWordStruct[7] := ValueAsWordStruct[7] or MACAddress.RangeBytes[6];

     CopyPrefixFrom (Prefix);
end;

(* -------------------------------------- *)

function tIPv6Address.getEmptyHostBitCount: integer;

var count  : integer;
    i,j    : integer;
    block  : word;
    boDone : boolean;

begin
     count := 0;
     i := 7;

     // Go through all block from end to front
     // count consequtive bits as long as possible
     // continue to the next block when the current
     // is $0000

     repeat
           block := ValueAsWordStruct[i];
           boDone := true;

           case block of
            $0000 : begin
                         inc (count, 16);
                         boDone := false;
                    end;
            $8000 : inc (count, 15);
            $C000 : inc (count, 14);
            $E000 : inc (count, 13);
            $F000 : inc (count, 12);
            $F800 : inc (count, 11);
            $FC00 : inc (count, 10);
            $FE00 : inc (count, 9);
            $FF00 : inc (count, 8);
            $FF80 : inc (count, 7);
            $FFC0 : inc (count, 6);
            $FFE0 : inc (count, 5);
            $FFF0 : inc (count, 4);
            $FFF8 : inc (count, 3);
            $FFFC : inc (count, 2);
            $FFFE : inc (count, 1);
           end;

           if i = 0 then
              boDone := true;

           dec (i);

     until boDone;

     result := count;
end;

(* -------------------------------------- *)

function tIPv6Address.IsPrefixInformation: boolean;

begin
     result := (MaskBitCount <= 64) and (MaskBitCount = getEmptyHostBitCount);
end;

(* -------------------------------------- *)

function tIPv6Address.BelongsToSubnet (Subnet: tIPv6Address): boolean;

var help : tIPv6Address;

begin
     // Create a subnet only version of self with the same prefix length as the subnet

     help.initialize;
     help.CopyPrefixFrom (self, Subnet.MaskBitCount);

     result := (help = Subnet);
end;

(* -------------------------------------- *)

function tIPv6Address.isUndefined: boolean;

begin
     result := self.IsUnspecifiedAddress;
end;

(* -------------------------------------- *)

function tIPv6Address.DictionaryKey: string;

begin
     result := getExpandedValue;
end;

(* -------------------------------------- *)
{ tIPv6AddressPair }
(* -------------------------------------- *)

class operator tIPv6AddressPair.equal (a, b: tIPv6AddressPair): boolean;

begin
     result := ((a.Source = b.Source) and (a.Destination = b.Destination)) or
               ((a.Source = b.Destination) and (a.Destination = b.Source))
end;

(* -------------------------------------- *)

class operator tIPv6AddressPair.greaterthan (a, b: tIPv6AddressPair): boolean;

begin
     if a.Source = b.Source then
        result := a.Destination > b.Destination
     else
         result := a.Source > b.Source;

end;

(* -------------------------------------- *)

class operator tIPv6AddressPair.greaterthanorequal (a, b: tIPv6AddressPair): boolean;

begin
     if a.Source = b.Source then
        result := a.Destination > b.Destination
     else
         result := a.Source >= b.Source;
end;

(* -------------------------------------- *)

class operator tIPv6AddressPair.lessthan (a, b: tIPv6AddressPair): boolean;

begin
     if a.Source = b.Source then
        result := a.Destination < b.Destination
     else
         result := a.Source < b.Source;
end;

(* -------------------------------------- *)

class operator tIPv6AddressPair.lessthanorequal (a, b: tIPv6AddressPair): boolean;

begin
     if a.Source = b.Source then
        result := a.Destination <= b.Destination
     else
         result := a.Source <= b.Source;
end;

(* -------------------------------------- *)

procedure tIPv6AddressPair.swap;

var help : tIPv6Address;

begin
     help := Source;
     Source := Destination;
     Destination := help;

end;

(* -------------------------------------- *)

procedure tIPv6AddressPair.normalize;

begin
     if Source > Destination then
        swap;
end;

(* -------------------------------------- *)

class operator tIPv6AddressPair.notequal (a, b: tIPv6AddressPair): boolean;

begin
     result := not (a = b);
end;

(* -------------------------------------- *)

procedure tIPv6AddressPair.initialize (value: string; const SetCIDR: boolean);

var entry : string;

begin
     entry := getEntry (value, '=', 1);

     if SetCIDR then
        Source.setValueFromStringWithMask (entry)
     else
         Source.setValueFromString (entry);

     entry := getEntry (value, '=', 2);

     if SetCIDR then
        Destination.setValueFromStringWithMask (entry)
     else
         Destination.setValueFromString (entry);
end;

(* -------------------------------------- *)

procedure tIPv6AddressPair.initialize;

begin
     Source.initialize;
     Destination.initialize;
end;

(* -------------------------------------- *)

function tIPv6AddressPair.ValueAsString (const NoSpaces: boolean) : string;

begin
     if NoSpaces then
        result := Source.ValueAsString + '-' + Destination.ValueAsString
     else
         result := Source.ValueAsString + ' - ' + Destination.ValueAsString;
end;

(* -------------------------------------- *)

function tIPv6AddressPair.getExpandedValue (const NoSpaces: boolean) : string;

begin
     if NoSpaces then
        result := Source.getExpandedValue + '-' + Destination.getExpandedValue
     else
         result := Source.getExpandedValue + ' - ' + Destination.getExpandedValue;
end;

(* -------------------------------------- *)

function tIPv6AddressPair.isUndefined: boolean;

begin
     result := source.isUndefined and destination.isUndefined;
end;

(* -------------------------------------- *)

function tIPv6AddressPair.DictionaryKey: string;

var help : tIPv6AddressPair;

begin
     help := self;
     help.normalize;

     result := help.getExpandedValue (true);
end;

(* -------------------------------------- *)
{ tIPv4SimpleAddressList }
(* -------------------------------------- *)

procedure tIPv4SimpleAddressList.Add(NewEntry: tIPv4Address);

var count : integer;

begin
     count := length (list);
     setLength (list, count + 1);
     list[count] := NewEntry;
end;

(* -------------------------------------- *)

procedure tIPv4SimpleAddressList.copyFrom (SourceList: tIPv4AddrArray);

begin
     setLength (list, 0);

     list := copy (SourceList, 0, high (SourceList));
end;

(* -------------------------------------- *)

function tIPv4SimpleAddressList.count: integer;

begin
     result := length (list);
end;

(* -------------------------------------- *)

function tIPv4SimpleAddressList.GetEntry (index: integer): tIPv4Address;

begin
     if (index > 0) and (index <= length (list)) then
        result := list[index-1];
end;

(* -------------------------------------- *)
{ tIPv6SimpleAddressList }
(* -------------------------------------- *)

procedure tIPv6SimpleAddressList.Add(NewEntry: tIPv6Address);

var count : integer;

begin
     count := length (list);
     setLength (list, count + 1);
     list[count] := NewEntry;
end;

(* -------------------------------------- *)

procedure tIPv6SimpleAddressList.copyFrom(SourceList: tIPv6AddrArray);

begin
     setLength (list, 0);

     list := copy (SourceList, 0, high (SourceList));
end;

(* -------------------------------------- *)

function tIPv6SimpleAddressList.count: integer;

begin
     result := length (list);
end;

(* -------------------------------------- *)

function tIPv6SimpleAddressList.GetEntry(index: integer): tIPv6Address;

begin
     if (index > 0) and (index <= length (list)) then
        result := list[index-1];
end;

(* -------------------------------------- *)
{ tIPv4Socket }
(* -------------------------------------- *)

procedure tIPv4Socket.initialize;

begin
     VLANID := 0;
     IPAddress.initialize;
     Port := 0;
end;

(* -------------------------------------- *)

procedure tIPv4Socket.initialize (const Value: string);

var help  : string;
    index : integer;
    num   : integer;

begin
     VLANID := 0;
     index := pos (':', Value);

     if index = 0 then
        begin
             IPAddress.setValueFromString (Value);
             Port := 0;
        end
     else
         begin
              help := copy (Value, 1, index - 1);
              IPAddress.setValueFromString (help);

              help := copy (Value, index + 1, (length (Value) - (Index - 1)));
              TryStrToInt (help, num);
              port := num;
         end;
end;

(* -------------------------------------- *)

class operator tIPv4Socket.equal(a, b: tIPv4Socket): boolean;

begin
     result := (a.VLANID = b.VLANID) and (a.IPAddress = b.IPAddress) and (a.Port = b.Port);
end;

(* -------------------------------------- *)

class operator tIPv4Socket.notequal(a, b: tIPv4Socket): boolean;

begin
     result := not (a=b);
end;

(* -------------------------------------- *)

class operator tIPv4Socket.greaterthan(a, b: tIPv4Socket): boolean;

begin
     if a.VLANID = b.VLANID then
        begin
             if a.IPAddress = b.IPAddress then
                result := a.Port > b.Port
             else
                 result := (a.IPAddress > b.IPAddress);
        end
     else
         result := (a.VLANID > b.VLANID);
end;

(* -------------------------------------- *)

class operator tIPv4Socket.greaterthanorequal(a, b: tIPv4Socket): boolean;

begin
     if a.VLANID = b.VLANID then
        begin
             if a.IPAddress = b.IPAddress then
                result := a.Port >= b.Port
             else
                 result := (a.IPAddress > b.IPAddress);
        end
     else
         result := (a.VLANID >= b.VLANID);
end;

(* -------------------------------------- *)

class operator tIPv4Socket.lessthan(a, b: tIPv4Socket): boolean;

begin
     if a.VLANID = b.VLANID then
        begin
             if a.IPAddress = b.IPAddress then
                result := a.Port < b.Port
             else
                 result := (a.IPAddress < b.IPAddress);
        end
     else
         result := a.VLANID < b.VLANID;
end;

(* -------------------------------------- *)

class operator tIPv4Socket.lessthanorequal(a, b: tIPv4Socket): boolean;

begin
     if a.VLANID = b.VLANID then
        begin
             if a.IPAddress = b.IPAddress then
                result := a.Port <= b.Port
             else
                 result := (a.IPAddress < b.IPAddress);
        end
     else
         result := a.VLANID <= b.VLANID;
end;

(* -------------------------------------- *)

function tIPv4Socket.ValueAsString: string;

begin
     result := IPAddress.ValueAsString + ':' + IntToStr (Port);
end;

(* -------------------------------------- *)

function tIPv4Socket.isUndefined: boolean;

begin
     result := IPAddress.isUndefined and (Port = 0)
end;

(* -------------------------------------- *)
{ tIPv4SocketPair }
(* -------------------------------------- *)

procedure tIPv4SocketPair.initialize;

begin
     Source.initialize;
     Destination.initialize;
end;

(* -------------------------------------- *)

procedure tIPv4SocketPair.swap;

var help : tIPv4Socket;

begin
     help := Source;
     Source := Destination;
     Destination := help;
end;

(* -------------------------------------- *)

procedure tIPv4SocketPair.normalize;

begin
     // Create a socket pair with the lower IP as Source. This does not make any sense when talking about who is client or server,
     // but we do it anyway to get the details into a comparable order when we lookup existing flows.

     if Destination < Source then
        swap;
end;

(* -------------------------------------- *)

class operator tIPv4SocketPair.equal (a, b: tIPv4SocketPair): boolean;

begin
     result := ((a.Source = b.Source) and (a.Destination = b.Destination)) or
               ((a.Source = b.Destination) and (a.Destination = b.Source));
end;

(* -------------------------------------- *)

class operator tIPv4SocketPair.notequal (a, b: tIPv4SocketPair): boolean;

begin
     result := not (a=b);
end;

(* -------------------------------------- *)

class operator tIPv4SocketPair.greaterthan (a, b: tIPv4SocketPair): boolean;

begin
     if a.Destination = b.Destination then
        result := a.Source > b.Source
     else
         result := a.Destination > b.Destination;
end;

(* -------------------------------------- *)

class operator tIPv4SocketPair.greaterthanorequal (a,  b: tIPv4SocketPair): boolean;

begin
     if a.Destination = b.Destination then
        result := a.Source >= b.Source
     else
         result := a.Destination > b.Destination;
end;

(* -------------------------------------- *)

class operator tIPv4SocketPair.lessthan (a, b: tIPv4SocketPair): boolean;

begin
     if a.Destination = b.Destination then
        result := a.Source < b.Source
     else
         result := a.Destination < b.Destination;
end;

(* -------------------------------------- *)

class operator tIPv4SocketPair.lessthanorequal (a, b: tIPv4SocketPair): boolean;

begin
     if a.Destination = b.Destination then
        result := a.Source <= b.Source
     else
         result := a.Destination <= b.Destination;
end;

(* -------------------------------------- *)

function CompareIPv4SocketPair (a, b : tIPv4SocketPair) : integer;

begin
     if a > b then
        result := 1
     else
         if a < b then
            result := -1
         else
             result := 0;
end;

(* -------------------------------------- *)

function tIPv4SocketPair.ValueAsString (const NoSpaces : boolean) : string;

begin
     if NoSpaces then
        result := Source.ValueAsString + '-' + Destination.ValueAsString
     else
         result := Source.ValueAsString + ' - ' + Destination.ValueAsString;
end;

(* -------------------------------------- *)

function tIPv4SocketPair.IPAddressPair: tIPv4AddressPair;

begin
     result.initialize;

     result.Source := Source.IPAddress;
     result.Destination := Destination.IPAddress;
end;

(* -------------------------------------- *)

function tIPv4SocketPair.isUndefined: boolean;

begin
     result := source.isUndefined and destination.isUndefined;
end;

(* -------------------------------------- *)

function tIPv4SocketPair.DictionaryKey: string;

var help : tIPv4SocketPair;

begin
     help := self;
     help.normalize;

     result := help.ValueAsString (true);
end;

(* -------------------------------------- *)
{ tIPv6Socket }
(* -------------------------------------- *)

procedure tIPv6Socket.initialize;

begin
     VLANID := 0;
     IPAddress.initialize;
     Port := 0;
end;

(* -------------------------------------- *)

procedure tIPv6Socket.initialize (const Value: string);

var help  : string;
    index : integer;
    num   : integer;
    i     : integer;

begin
     VLANID := 0;

     index := pos (']:', Value);

     if index > 0 then
        begin
             help := copy (Value, 2, index - 2);

             IPAddress.setValueFromString (help);

             help := copy (Value, index + 2, (length (Value) - (Index - 2)));
             TryStrToInt (help, num);
             port := num;
        end
     else
         if numEntries (Value, ':') = 9 then
            begin
                 // Snort alert.ids style notation

                 help := getEntry (Value, ':', 9);
                 TryStrToInt (help, num);
                 port := num;

                 help := getEntry (Value, ':', 1);

                 for i := 2 to 8 do
                     help := help + ':' + getEntry (Value, ':', i);

                 IPAddress.setValueFromString(help);
            end
         else
             if isValidIPv6String (value) then
                IPAddress.setValueFromString(value)
             else
                 raise Exception.Create('Cannot initialize IPv6 Socket (tIPv6Socket.initialize)');
end;

(* -------------------------------------- *)

class operator tIPv6Socket.equal(a, b: tIPv6Socket): boolean;

begin
     result := (a.VLANID = b.VLANID) and (a.IPAddress = b.IPAddress) and (a.Port = b.Port);
end;

(* -------------------------------------- *)

class operator tIPv6Socket.notequal(a, b: tIPv6Socket): boolean;

begin
     result := not (a=b);
end;

(* -------------------------------------- *)

class operator tIPv6Socket.greaterthan(a, b: tIPv6Socket): boolean;

begin
     if a.VLANID = b.VLANID then
        begin
             if a.IPAddress = b.IPAddress then
                result := a.Port > b.Port
             else
                 result := (a.IPAddress > b.IPAddress);
        end
     else
         result := a.VLANID > b.VLANID;
end;

(* -------------------------------------- *)

class operator tIPv6Socket.greaterthanorequal(a, b: tIPv6Socket): boolean;

begin
     if a.VLANID = b.VLANID then
        begin
             if a.IPAddress = b.IPAddress then
                result := a.Port > b.Port
             else
                 result := (a.IPAddress >= b.IPAddress);
        end
     else
         result := a.VLANID >= b.VLANID;
end;

(* -------------------------------------- *)

class operator tIPv6Socket.lessthan(a, b: tIPv6Socket): boolean;

begin
     if a.VLANID = b.VLANID then
        begin
             if a.IPAddress = b.IPAddress then
                result := a.Port < b.Port
             else
                 result := (a.IPAddress < b.IPAddress);
        end
     else
         result := a.VLANID < b.VLANID;
end;

(* -------------------------------------- *)

class operator tIPv6Socket.lessthanorequal(a, b: tIPv6Socket): boolean;

begin
     if a.VLANID = b.VLANID then
        begin
             if a.IPAddress = b.IPAddress then
                result := a.Port <= b.Port
             else
                 result := (a.IPAddress <= b.IPAddress);
        end
     else
         result := a.VLANID <= b.VLANID;
end;

(* -------------------------------------- *)

function tIPv6Socket.ValueAsString: string;

begin
     result := '[' + IPAddress.ValueAsString + ']:' + IntToStr (Port);
end;

(* -------------------------------------- *)

function tIPv6Socket.getExpandedValue (const NoSpaces: boolean) : string;

begin
     result := '[' + IPAddress.getExpandedValue + ']:' + IntToStr (Port);
end;

(* -------------------------------------- *)

function tIPv6Socket.isUndefined: boolean;

begin
     result := IPAddress.isUndefined and (port = 0);
end;

(* -------------------------------------- *)
{ tIPv6SocketPair }
(* -------------------------------------- *)

procedure tIPv6SocketPair.initialize;

begin
     Destination.initialize;
     Source.initialize;
end;

(* -------------------------------------- *)

procedure tIPv6SocketPair.swap;

var help : tIPv6Socket;

begin
     help := Source;
     Source := Destination;
     Destination := help;
end;

(* -------------------------------------- *)

procedure tIPv6SocketPair.normalize;

begin
     // Create a socket pair with the lower IP as Client. This does not make any sense when talking about who is client or server,
     // but we do it anyway to get the details into a comparable order when we lookup existing flows.

     if Destination < Source then
        swap;
end;

(* -------------------------------------- *)

class operator tIPv6SocketPair.equal(a, b: tIPv6SocketPair): boolean;

begin
     result := ((a.Source = b.Source) and (a.Destination = b.Destination)) or
               ((a.Source = b.Destination) and (a.Destination = b.Source));
end;

(* -------------------------------------- *)

class operator tIPv6SocketPair.notequal(a, b: tIPv6SocketPair): boolean;

begin
     result := not (a=b);
end;

(* -------------------------------------- *)

class operator tIPv6SocketPair.greaterthan(a, b: tIPv6SocketPair): boolean;

begin
     if a.Destination = b.Destination then
        result := a.Source > b.Source
     else
         result := a.Destination > b.Destination;
end;

(* -------------------------------------- *)

class operator tIPv6SocketPair.greaterthanorequal (a, b: tIPv6SocketPair): boolean;

begin
     if a.Destination = b.Destination then
        result := a.Source >= b.Source
     else
         result := a.Destination >= b.Destination;
end;

(* -------------------------------------- *)

class operator tIPv6SocketPair.lessthan(a, b: tIPv6SocketPair): boolean;

begin
     if a.Destination = b.Destination then
        result := a.Source < b.Source
     else
         result := a.Destination < b.Destination;
end;

(* -------------------------------------- *)

class operator tIPv6SocketPair.lessthanorequal(a, b: tIPv6SocketPair): boolean;

begin
     if a.Destination = b.Destination then
        result := a.Source <= b.Source
     else
         result := a.Destination <= b.Destination;
end;

(* -------------------------------------- *)

function tIPv6SocketPair.ValueAsString (const NoSpaces : boolean) : string;

var help1 : string;
    help2 : string;

begin
     help1 := Source.ValueAsString;
     help2 := Destination.ValueAsString;

     if NoSpaces then
        result := help1 + '-' + help2
     else
         result := help1 + ' - ' + help2;
end;

(* -------------------------------------- *)

function tIPv6SocketPair.getExpandedValue (const NoSpaces: boolean) : string;

var help1 : string;
    help2 : string;

begin
     help1 := Source.getExpandedValue (true);
     help2 := Destination.getExpandedValue (true);

     if NoSpaces then
        result := help1 + '-' + help2
     else
         result := help1 + ' - ' + help2;
end;

(* -------------------------------------- *)

function tIPv6SocketPair.IPAddressPair: tIPv6AddressPair;

begin
     result.initialize;

     result.Source := Source.IPAddress;
     result.Destination := Destination.IPAddress;
end;

(* -------------------------------------- *)

function tIPv6SocketPair.isUndefined: boolean;

begin
     result := source.isUndefined and destination.isUndefined;
end;

(* -------------------------------------- *)

function tIPv6SocketPair.DictionaryKey: string;

var help  : tIPv6SocketPair;
    value : string;

begin
     help := self;
     help.normalize;

     value := help.getExpandedValue (true);

     // Remove the colons before returning the address

     value := stringReplace (value, ':', '', [rfReplaceAll]);
     result := value;
end;

(* -------------------------------------- *)

function CompareIPv6SocketPair (a, b : tIPv6SocketPair) : integer;

begin
     if a > b then
        result := 1
     else
         if a < b then
            result := -1
         else
             result := 0;
end;

(* -------------------------------------- *)
{ tIPv4Conversation }
(* -------------------------------------- *)
{
procedure tIPv4Conversation.initialize;

begin
     DBIndex := 0;
     FileIndex := 0;
     IPv4SocketPair.initialize;
     Address1DBIndex := 0;
     Address2DBIndex := 0;
     IPVersion := 0;
     Protocol := 0;
     FirstFoundInFileID := 0;
     LastFoundInFileID := 0;
     FirstFoundInFrameID := 0;
     LastFoundInFrameID := 0;
     VLANID := 0;
     FrameCount := 0;

     StartTimeStamp := 0;
     EndTimeStamp := 0;
end;
}
(* -------------------------------------- *)
{ tIPv6Conversation }
(* -------------------------------------- *)
{
procedure tIPv6Conversation.initialize;

begin
     DBIndex := 0;
     FileIndex := 0;
     IPv6SocketPair.initialize;
     Address1DBIndex := 0;
     Address2DBIndex := 0;
     IPVersion := 0;
     Protocol := 0;
     FirstFoundInFileID := 0;
     LastFoundInFileID := 0;
     FirstFoundInFrameID := 0;
     LastFoundInFrameID := 0;
     VLANID := 0;
     FrameCount := 0;

     StartTimeStamp := 0;
     EndTimeStamp := 0;
end;
}
(* ----------------------------------- *)
{ tVLANPair }
(* ----------------------------------- *)

procedure tWordPair.initialize;

begin
     Value1 := 0;
     Value2 := 0;
end;

(* -------------------------------------- *)

procedure tWordPair.initialize (value : string);

var help  : string;
    dummy : integer;

begin
     help := getEntry (value, '=', 1);
     val (help, value1, dummy);

     help := getEntry (value, '=', 2);
     val (help, value2, dummy);
end;

(* ----------------------------------- *)

class operator tWordPair.equal(a, b: tWordPair): boolean;

begin
     result := ((a.Value1 = b.Value1) and (a.Value2=b.Value2)) or
               ((a.Value1 = b.Value2) and (a.Value2=b.Value1)) ;
end;

(* ----------------------------------- *)

class operator tWordPair.greaterthan(a, b: tWordPair): boolean;

begin
     if a.Value1 = b.Value1 then
        result := a.Value2 > b.Value2
     else
         result := a.Value1 > b.Value1;
end;

(* ----------------------------------- *)

class operator tWordPair.greaterthanorequal(a, b: tWordPair): boolean;

begin
     if a.Value1 = b.Value1 then
        result := a.Value2 >= b.Value2
     else
         result := a.Value1 >= b.Value1;
end;

(* ----------------------------------- *)

class operator tWordPair.lessthan(a, b: tWordPair): boolean;

begin
     if a.Value1 = b.Value1 then
        result := a.Value2 < b.Value2
     else
         result := a.Value1 < b.Value1;
end;

(* ----------------------------------- *)

class operator tWordPair.lessthanorequal(a, b: tWordPair): boolean;

begin
     if a.Value1 = b.Value1 then
        result := a.Value2 <= b.Value2
     else
         result := a.Value1 <= b.Value1;
end;

(* ----------------------------------- *)

class operator tWordPair.notequal(a, b: tWordPair): boolean;

begin
     result := not (a=b);
end;

(* -------------------------------------- *)
{ tSocketPortPair   deprecated; used by anonymization with socket and port
(* -------------------------------------- *)

procedure tSocketPortPair.initialize;

begin
     IPVersion := 0;
     OriginalIPv4Socket.initialize;
     OriginalIPv6Socket.initialize;
     NewPort := 0;
end;
}
(* -------------------------------------- *)
{ tIPAddress }
(* -------------------------------------- *)

procedure tIPAddress.initialize;

begin
     IPVersion := 0;
     IPv4Address.initialize;
     IPv6Address.initialize;
end;

(* -------------------------------------- *)

procedure tIPAddress.initialize (const IPAddress: tIPv4Address);

begin
     IPVersion := 4;
     IPv4Address := IPAddress;
     IPv6Address.initialize;

{$ifdef debug}
     IPString := IPv4Address.ValueAsString;
{$endif}
end;

(* -------------------------------------- *)

procedure tIPAddress.initialize (const IPAddress: tIPv6Address);

begin
     IPVersion := 6;
     IPv4Address.initialize;
     IPv6Address := IPAddress;

{$ifdef debug}
     IPString := IPv6Address.ValueAsString;
{$endif}
end;

(* -------------------------------------- *)

procedure tIPAddress.initialize (const IPv4AsLong: longword);

begin
     IPVersion := 4;
     IPv4Address.ValueAsLong := IPv4AsLong;
     IPv6Address.initialize;

{$ifdef debug}
     IPString := IPv4Address.ValueAsString;
{$endif}
end;

(* -------------------------------------- *)

procedure tIPAddress.initialize(const Value: tIPv6AddressByteStruct);

begin
     IPVersion := 6;
     IPV4Address.initialize;
     IPv6Address.SetValueFromByteStruct (value);

{$ifdef debug}
     IPString := IPv6Address.ValueAsString;
{$endif}
end;

(* -------------------------------------- *)

class operator tIPAddress.equal (a, b: tIPAddress): boolean;

begin
     if a.IPVersion = b.IPVersion then
        begin
             if a.IPVersion = 4 then
                result := a.IPv4Address = b.IPv4Address
             else
                 result := a.IPv6Address = b.IPv6Address;
        end
     else
         result := False;
end;

(* -------------------------------------- *)

class operator tIPAddress.notequal (a, b: tIPAddress): boolean;

begin
     result := not (a=b);
end;

(* -------------------------------------- *)

class operator tIPAddress.greaterthan (a, b: tIPAddress): boolean;

begin
     if a.IPVersion = b.IPVersion then
        begin
             if a.IPVersion = 4 then
                result := a.IPv4Address > b.IPv4Address
             else
                 result := a.IPv6Address > b.IPv6Address;
        end
     else
         result := (a.IPVersion > b.IPVersion);
end;

(* -------------------------------------- *)

class operator tIPAddress.greaterthanorequal (a, b: tIPAddress): boolean;

begin
     result := (a > b) or (a=b);
end;

(* -------------------------------------- *)

class operator tIPAddress.lessthan (a, b: tIPAddress): boolean;

begin
     result := not (a >= b);
end;

(* -------------------------------------- *)

class operator tIPAddress.lessthanorequal (a, b: tIPAddress): boolean;

begin
     result := not (a > b);
end;

(* -------------------------------------- *)

procedure tIPAddress.setValueFromString (const IPAddress: string; const MaskBits: byte);

begin
     if isValidIPv4String (IPAddress) then
        begin
             IPVersion := 4;
             IPv6Address.initialize;

             if MaskBits = 0 then
                IPv4Address.setValueFromString (IPAddress)
             else
                 IPv4Address.setValueFromString (IPAddress, MaskBits);
        end
     else
         if isValidIPv6String (IPAddress) then
            begin
                 IPVersion := 6;
                 IPv4Address.initialize;

                 if MaskBits = 0 then
                    IPv6Address.setValueFromString (IPAddress)
                 else
                     IPv6Address.setValueFromString (IPAddress, MaskBits);
            end
         else
             initialize;
end;

(* -------------------------------------- *)

function tIPAddress.isZeroAddress: boolean;

begin
     if IPVersion = 6 then
        result := IPv6Address.IsUnspecifiedAddress
     else
         result := IPv4Address.isZeroAddress;
end;

(* -------------------------------------- *)

function tIPAddress.ValueAsString: string;

begin
     if IPVersion = 6 then
        result := IPv6Address.ValueAsString
     else
         result := IPv4Address.ValueAsString;
end;

(* -------------------------------------- *)

function tIPAddress.DictionaryKey: string;

begin
     if IPVersion = 6 then
        result := IPv6Address.DictionaryKey
     else
         result := IPv4Address.DictionaryKey
end;

(* -------------------------------------- *)

function tIPAddress.isUndefined: boolean;

begin
     result := isZeroAddress;
end;

(* -------------------------------------- *)
{ tIPAddressPair }
(* -------------------------------------- *)

procedure tIPAddressPair.initialize;

begin
     Source.initialize;
     Destination.initialize;
end;

(* -------------------------------------- *)

class operator tIPAddressPair.equal (a, b: tIPAddressPair): boolean;

begin
     result := ((a.Source = b.Source) and (a.Destination = b.Destination)) or
               ((a.Source = b.Destination) and (a.Destination = b.Source))
end;

(* -------------------------------------- *)

class operator tIPAddressPair.notequal (a, b: tIPAddressPair): boolean;

begin
     result := not (a = b);
end;

(* -------------------------------------- *)

class operator tIPAddressPair.greaterthan (a, b: tIPAddressPair): boolean;

begin
     if a.Source = b.Source then
        result := a.Destination > b.Destination
     else
         result := a.Source > b.Source;
end;

(* -------------------------------------- *)

class operator tIPAddressPair.greaterthanorequal (a, b: tIPAddressPair): boolean;

begin
     if a.Source = b.Source then
        result := a.Destination > b.Destination
     else
         result := a.Source >= b.Source;
end;

(* -------------------------------------- *)

class operator tIPAddressPair.lessthan (a, b: tIPAddressPair): boolean;

begin
     if a.Source = b.Source then
        result := a.Destination < b.Destination
     else
         result := a.Source < b.Source;
end;

(* -------------------------------------- *)

class operator tIPAddressPair.lessthanorequal (a, b: tIPAddressPair): boolean;

begin
     if a.Source = b.Source then
        result := a.Destination <= b.Destination
     else
         result := a.Source <= b.Source;
end;

(* -------------------------------------- *)

procedure tIPAddressPair.swap;

var help : tIPAddress;

begin
     help := Source;
     Source := Destination;
     Destination := help;
end;

(* -------------------------------------- *)

procedure tIPAddressPair.normalize;

begin
     if Source > Destination then
        swap;
end;

(* -------------------------------------- *)

function tIPAddressPair.ValueAsString (const NoSpaces: boolean): string;

begin
     if NoSpaces then
        result := Source.ValueAsString + '-' + Destination.ValueAsString
     else
         result := Source.ValueAsString + ' - ' + Destination.ValueAsString
end;

(* -------------------------------------- *)

function tIPAddressPair.isUndefined: boolean;

begin
     result := Source.isUndefined and Destination.isUndefined;
end;

(* -------------------------------------- *)

function tIPAddressPair.DictionaryKey: string;

var help : tIPAddressPair;

begin
     help := self;
     help.normalize;

     result := help.ValueAsString (true);
end;

(* -------------------------------------- *)
{ tIPSocket }
(* -------------------------------------- *)

procedure tIPSocket.initialize;

begin
     VLANID := 0;
     IPAddress.initialize;
     Port := 0;
end;

(* -------------------------------------- *)

procedure tIPSocket.initialize (const NewAddress : tIPAddress; const NewPort : word; const NewVLANID : word = 0);

begin
     VLANID := NewVLANID;
     IPAddress := NewAddress;
     Port := NewPort;
end;

(* -------------------------------------- *)

procedure tIPSocket.initialize (const IPv4Address: tIPv4Address; const NewPort, NewVLANID: word);

begin
     VLANID := NewVLANID;
     IPAddress.initialize (IPv4Address);
     Port := NewPort;
end;

(* -------------------------------------- *)

procedure tIPSocket.initialize (const IPv6Address: tIPv6Address; const NewPort, NewVLANID: word);

begin
     VLANID := NewVLANID;
     IPAddress.initialize (IPv6Address);
     Port := NewPort;
end;

(* -------------------------------------- *)

procedure tIPSocket.initialize (const IPv4Socket: tIPv4Socket);

begin
     VLANID := IPv4Socket.VLANID;
     IPAddress.initialize (IPv4Socket.IPAddress);
     Port := IPv4Socket.Port;
end;

(* -------------------------------------- *)

procedure tIPSocket.initialize (const IPv6Socket: tIPv6Socket);

begin
     VLANID := IPv6Socket.VLANID;
     IPAddress.initialize (IPv6Socket.IPAddress);
     Port := IPv6Socket.Port;
end;

(* -------------------------------------- *)

procedure tIPSocket.initialize (const Value: string);

var IPv4Socket : tIPv4Socket;
    IPv6Socket : tIPv6Socket;

begin
     if isValidIPv4Socket (Value) then
        begin
             IPv4Socket.initialize (Value);

             initialize (IPv4Socket);
        end
     else
         if isValidIPv6Socket (value) then
            begin
                 IPv6Socket.initialize (Value);

                 initialize (IPv6Socket);
            end
         else
             initialize;
end;

(* -------------------------------------- *)

class operator tIPSocket.equal (a, b: tIPSocket): boolean;

begin
     result := (a.VLANID = b.VLANID) and (a.IPAddress = b.IPAddress) and (a.Port = b.Port);
end;

(* -------------------------------------- *)

class operator tIPSocket.greaterthan (a, b: tIPSocket): boolean;

begin
     if a.VLANID = b.VLANID then
        begin
             if a.IPAddress = b.IPAddress then
                result := a.Port > b.Port
             else
                 result := (a.IPAddress > b.IPAddress);
        end
     else
         result := (a.VLANID > b.VLANID);
end;

(* -------------------------------------- *)

class operator tIPSocket.greaterthanorequal (a, b: tIPSocket): boolean;

begin
     result := (a = b) or (a > b);
end;

(* -------------------------------------- *)

class operator tIPSocket.lessthan (a, b: tIPSocket): boolean;

begin
     result := not (a >= b);
end;

(* -------------------------------------- *)

class operator tIPSocket.lessthanorequal (a, b: tIPSocket): boolean;

begin
     result := not (a > b);
end;

(* -------------------------------------- *)

class operator tIPSocket.notequal (a, b: tIPSocket): boolean;

begin
     result := not (a=b);
end;

(* -------------------------------------- *)

function tIPSocket.getIPv4Socket: tIPv4Socket;

begin
     result.IPAddress := IPAddress.IPv4Address;
     result.Port := Port;
end;

(* -------------------------------------- *)

function tIPSocket.getIPv6Socket: tIPv6Socket;

begin
     result.IPAddress := IPAddress.IPv6Address;
     result.Port := Port;
end;

(* -------------------------------------- *)

function tIPSocket.DictionaryKey: string;

begin
     result := ValueAsString;
end;

(* -------------------------------------- *)

function tIPSocket.isUndefined: boolean;

begin
     result := IPAddress.isUndefined and (Port = 0);
end;

(* -------------------------------------- *)

function tIPSocket.ValueAsString: string;

begin
     result := IPAddress.ValueAsString + ':' + IntToStr (Port);
end;

(* -------------------------------------- *)
{ tIPSocketPair }
(* -------------------------------------- *)

procedure tIPSocketPair.initialize;

begin
     Source.initialize;
     Destination.initialize;
end;

(* -------------------------------------- *)

procedure tIPSocketPair.initialize (const IPv4SocketPair: tIPv4SocketPair);

begin
     Source.initialize (IPv4SocketPair.Source);
     Destination.initialize (IPv4SocketPair.Destination);
end;

(* -------------------------------------- *)

procedure tIPSocketPair.initialize (const IPv6SocketPair: tIPv6SocketPair);

begin
     Source.initialize (IPv6SocketPair.Source);
     Destination.initialize (IPv6SocketPair.Destination);
end;

(* -------------------------------------- *)

procedure tIPSocketPair.initialize (const value: string);

var help : string;

begin
     assert (isValidSocketPair (value), 'Not a valid socket pair: "' + value + '" (tIPSocketPair.initialize)');

     help := getEntry (value, '=', 1);
     Source.initialize (help);

     help := getEntry (value, '=', 2);
     Destination.initialize (help);
end;

(* -------------------------------------- *)

class operator tIPSocketPair.equal (a, b: tIPSocketPair): boolean;

begin
     result := ((a.Source = b.Source) and (a.Destination = b.Destination)) or
               ((a.Source = b.Destination) and (a.Destination = b.Source));
end;

(* -------------------------------------- *)

class operator tIPSocketPair.notequal (a, b: tIPSocketPair): boolean;

begin
     result := not (a = b);
end;

(* -------------------------------------- *)

class operator tIPSocketPair.greaterthan (a, b: tIPSocketPair): boolean;

begin
     if a.Destination = b.Destination then
        result := a.Source > b.Source
     else
         result := a.Destination > b.Destination;
end;

(* -------------------------------------- *)

class operator tIPSocketPair.greaterthanorequal (a, b: tIPSocketPair): boolean;

begin
     if a.Destination = b.Destination then
        result := a.Source >= b.Source
     else
         result := a.Destination > b.Destination;
end;

(* -------------------------------------- *)

class operator tIPSocketPair.lessthan (a, b: tIPSocketPair): boolean;

begin
     if a.Destination = b.Destination then
        result := a.Source < b.Source
     else
         result := a.Destination < b.Destination;
end;

(* -------------------------------------- *)

class operator tIPSocketPair.lessthanorequal (a, b: tIPSocketPair): boolean;

begin
     if a.Destination = b.Destination then
        result := a.Source <= b.Source
     else
         result := a.Destination <= b.Destination;
end;

(* -------------------------------------- *)

function tIPSocketPair.IPAddressPair: tIPAddressPair;

begin
     result.initialize;

     result.Source := Source.IPAddress;
     result.Destination := Destination.IPAddress;
end;

(* -------------------------------------- *)

procedure tIPSocketPair.swap;

var help : tIPSocket;

begin
     help := Source;
     Source := Destination;
     Destination := help;
end;

(* -------------------------------------- *)

procedure tIPSocketPair.normalize;

begin
     // Create a socket pair with the lower IP as Source. This does not make any sense when talking about who is client or server,
     // but we do it anyway to get the details into a comparable order when we lookup existing flows.

     if Destination < Source then
        swap;
end;

(* -------------------------------------- *)

function tIPSocketPair.ValueAsString (const NoSpaces: boolean): string;

var IPv4SocketPair : tIPv4SocketPair;
    IPv6SocketPair : tIPv6SocketPair;

begin
     assert (Source.IPAddress.IPVersion <> 0, 'Invalid socket IP version 0 (tIPSocketPair.ValueAsString)');

     if Source.IPAddress.IPVersion = 4 then
        begin
             IPv4SocketPair.Source.VLANID := Source.VLANID;
             IPv4SocketPair.Source.IPAddress := Source.IPAddress.IPv4Address;
             IPv4SocketPair.Source.Port := Source.Port;
             IPv4SocketPair.Destination.VLANID := Destination.VLANID;
             IPv4SocketPair.Destination.IPAddress := Destination.IPAddress.IPv4Address;
             IPv4SocketPair.Destination.Port := Destination.Port;

             result := IPv4SocketPair.DictionaryKey;
        end
     else
         begin
              IPv6SocketPair.Source.VLANID := Source.VLANID;
              IPv6SocketPair.Source.IPAddress := Source.IPAddress.IPv6Address;
              IPv6SocketPair.Source.Port := Source.Port;
              IPv6SocketPair.Destination.VLANID := Destination.VLANID;
              IPv6SocketPair.Destination.IPAddress := Destination.IPAddress.IPv6Address;
              IPv6SocketPair.Destination.Port := Destination.Port;

              result := IPv6SocketPair.DictionaryKey;
         end;
end;

(* -------------------------------------- *)

function tIPSocketPair.isUndefined: boolean;

begin
     result := Source.isUndefined and Destination.isUndefined;
end;

(* -------------------------------------- *)

function tIPSocketPair.toCSV (const Separator : string = ',') : string;

begin
     result := Source.IPAddress.ValueAsString + Separator + Source.Port.ToString + Separator +
               Destination.IPAddress.ValueAsString + Separator + Destination.Port.ToString;
end;

(* -------------------------------------- *)

function tIPSocketPair.toFilter (const DisplayFilter: boolean): string;

begin
     if DisplayFilter then
        result := 'ip.addr==' + Source.IPAddress.ValueAsString + ' and ' +
                  'ip.addr==' + Destination.IPAddress.ValueAsString + ' and ' +
                  'tcp.port==' + Source.Port.ToString + ' and ' +
                  'tcp.port==' + Destination.Port.ToString
     else
         result := 'host ' + Source.IPAddress.ValueAsString + ' and ' +  Destination.IPAddress.ValueAsString + ' and ' +
                   'tcp port ' + Source.Port.ToString + ' and ' + Destination.Port.ToString;
end;

(* -------------------------------------- *)

function tIPSocketPair.DictionaryKey: string;

var help : tIPSocketPair;

begin
     help := self;
     help.normalize;

     result := help.ValueAsString (true);
end;

(* -------------------------------------- *)
{ tFQDN }
(* -------------------------------------- *)

class operator tFQDN.equal(a, b: tFQDN): boolean;

begin
     result := a.FullFQDNBackwards = b.FullFQDNBackwards;
end;

(* -------------------------------------- *)

class operator tFQDN.greaterthan(a, b: tFQDN): boolean;

begin
     result := a.FullFQDNBackwards > b.FullFQDNBackwards;
end;

(* -------------------------------------- *)

class operator tFQDN.greaterthanorequal(a, b: tFQDN): boolean;

begin
     result := a.FullFQDNBackwards >= b.FullFQDNBackwards;
end;

(* -------------------------------------- *)

class operator tFQDN.lessthan(a, b: tFQDN): boolean;

begin
     result := a.FullFQDNBackwards < b.FullFQDNBackwards;
end;

(* -------------------------------------- *)

class operator tFQDN.lessthanorequal(a, b: tFQDN): boolean;

begin
     result := a.FullFQDNBackwards <= b.FullFQDNBackwards;
end;

(* -------------------------------------- *)

class operator tFQDN.notequal(a, b: tFQDN): boolean;

begin
     result := a.FullFQDNBackwards <> b.FullFQDNBackwards;
end;

(* -------------------------------------- *)

procedure tFQDN.initialize;

var i : integer;

begin
     FullFQDN := '';
     FullFQDNBackwards := '';
     ElementCount := 0;

     setlength (elements, 0);
end;

(* -------------------------------------- *)

procedure tFQDN.setFQDN (NewValue: string);

var i : integer;

begin
     FullFQDN := NewValue;
     FullFQDNBackwards := '';

     ElementCount := NumEntries (FullFQDN, '.');

     setLength (elements, elementcount);

     // Create reverse FQDN

     for i := 1 to ElementCount do
         begin
              Elements[i-1] := GetEntry (FullFQDN, '.', i);

              if i=1 then
                 FullFQDNBackwards := Elements[i-1]
              else
                  FullFQDNBackwards := Elements[i-1] + '.' + FullFQDNBackwards;
         end;
end;

(* -------------------------------------- *)

function tFQDN.getPartialFQDN (PartCount: integer): string;

var value : string;
    limit : integer;
    parts : integer;
    i     : integer;

begin
     value := '';
     parts := NumEntries (FullFQDN, '.');
     limit := min (PartCount, parts);

     // we want the FQDN from back to front

     for i := 1 to limit do
         begin
              if i = 1 then
                 value := GetEntry (FullFQDN, '.', parts)
              else
                  value := GetEntry (FullFQDN, '.', parts - (i-1)) + '.' + value;
         end;

     result := value;
end;

(* -------------------------------------- *)

function tFQDN.getAsFrameBufferArray: tByteArray;

var i,j   : integer;
    len   : integer;
    part  : string;
    count : integer;

begin
     setLength (result, length (FullFQDN)  + 2); // initial len and terminating 0
     count := 0;

     for i := 1 to NumEntries (FullFQDN, '.') do
         begin
              part := GetEntry (FullFQDN, '.', i);
              len := length (part);

              result[count] := len;
              inc (count);

              for j := 1 to len do
                  begin
                       result[count] := ord (part[j]);
                       inc (count);
                  end;
         end;

     result[count] := 0;
end;

(* -------------------------------------- *)
{ tFQDNPair }
(* -------------------------------------- *)

procedure tFQDNPair.initialize;

begin
     Original.setFQDN('');
     Replacement.setFQDN('');
end;

(* -------------------------------------- *)

procedure tFQDNPair.normalize;

var help : tFQDN;

begin
     if Original > Replacement then
        begin
             help := Original;
             Original := Replacement;
             Replacement := help;
        end;
end;

(* -------------------------------------- *)
{ tNameResolutionAnswer }
(* -------------------------------------- *)

procedure tNameResolutionAnswer.initialize;

begin
     AnswerType := nraUnknown;
     IPAddress.initialize;
     FQDN.initialize;
     TextValue := '';
end;

(* -------------------------------------- *)

procedure tNameResolutionAnswer.initialize (const IPv6Address: tIPv6Address);

begin
     AnswerType := nraIPAddress;
     IPAddress.initialize (IPv6Address);
     FQDN.initialize;
     TextValue := '';
end;

(* -------------------------------------- *)

procedure tNameResolutionAnswer.initialize (const IPv4Address: tIPv4Address);

begin
     AnswerType := nraIPAddress;
     IPAddress.initialize (IPv4Address);
     FQDN.initialize;
     TextValue := '';
end;

(* -------------------------------------- *)

procedure tNameResolutionAnswer.initialize (const NewIPAddress: tIPAddress);

begin
     AnswerType := nraIPAddress;
     IPAddress := NewIPAddress;
     FQDN.initialize;
     TextValue := '';
end;

(* -------------------------------------- *)

procedure tNameResolutionAnswer.initialize (const NewFQDN: tFQDN);

begin
     AnswerType := nraFQDN;
     IPAddress.initialize;
     FQDN := NewFQDN;
     TextValue := '';
end;

(* -------------------------------------- *)

procedure tNameResolutionAnswer.initialize (const NewTextValue: string);

begin
     AnswerType := nraText;
     IPAddress.initialize;
     FQDN.initialize;
     TextValue := NewTextValue;
end;

(* -------------------------------------- *)

function tNameResolutionAnswer.asString: string;

begin
     case AnswerType of
      nraUnknown   : result := '';
      nraIPAddress : result := IPAddress.ValueAsString;
      nraFQDN      : result := FQDN.FullFQDN;
      nraText      : result := TextValue;
     end;
end;

(* -------------------------------------- *)

function tNameResolutionAnswer.DictionaryKey: string;

begin
     result := IntToStr (AnswerTypeToInt (AnswerType)) + ':' + self.asString;
end;

(* -------------------------------------- *)

function isWellKnowServerPort (Port : word) : boolean;

begin
     case Port of
      1025 : result := true;         // MS RPC
      1080 : result := true;         // Socks Proxy
      1194 : result := true;         // OpenVPN
      1433,
      1434 : result := true;         // MSSQL
      1812,
      1813 : result := true;         // Radius
      1701 : result := true;         // L2TP
      1723 : result := true;         // MS PPTP
      1900 : result := true;         // uPNP
      3306 : result := true;         // MySQL
      1494,
      2598 : result := true;         // Citrix
      3128 : result := true;         // Squid
      3389 : result := true;         // RDP
      5480 : result := true;         // Vmware WebConfig
      5432 : result := true;         // PostgreSQL
      5800,
      5900 : result := true;         // UltraVNC
      6000,
      6001 : result := true;         // X11
      8080 : result := true;         // Generic Proxy
      9443 : result := true;         // VMware WebClient
     else
         result := false;
     end;
end;

(* -------------------------------------- *)

function DetermineProbableClientAddress (const SocketPair : tIPSocketPair) : tIPAddress;

begin
     result.initialize;

     if (Socketpair.Source.Port < 1025) and (SocketPair.Destination.Port > 1025) then
        result := Socketpair.Destination.IPAddress
     else
         if (Socketpair.Source.Port > 1025) and (SocketPair.Destination.Port < 1025) then
            result := Socketpair.Source.IPAddress
         else
             begin
                  if isWellKnowServerPort (SocketPair.Source.Port) then
                     result := SocketPair.Destination.IPAddress;

                  if isWellKnowServerPort (SocketPair.Destination.Port) then
                     result := SocketPair.Source.IPAddress;
             end;
end;

(* -------------------------------------- *)

function isValidMACString (MACAddress : string) : boolean;

var help  : string;
    addr  : tMACAddress;
    octet : string;
    dummy : integer;
    test  : integer;
    i     : integer;

begin
     help := stringReplace (MACAddress, '-', '', [rfReplaceAll]);
     help := stringReplace (help, ':', '', [rfReplaceAll]);

     result := length (help) = 12;

     if result then
        begin
             for i := 0 to 5 do
                 begin
                      octet := '$' + copy (help, (i*2) + 1, 2);

                      if octet <> '$00' then
                         begin
                              val (octet, test, dummy);

                              if test = 0 then
                                 result := false;
                         end;
                 end;
        end;
end;

(* -------------------------------------- *)

function isValidMACaddressPair (value : string; const CheckNull : boolean) : boolean;

var help  : string;
    addr  : tMACAddress;

begin
     if numEntries (value, '=') = 2 then
        begin
             // Check first MAC

             help := getEntry (value, '=', 1);

             if isValidMACString (help) then
                begin
                     // Check if source is a zero MAC, if required

                     if CheckNull then
                        begin
                             help := stringReplace (help, '-', ':', [rfReplaceAll]);
                             addr.setfromString (help, ':');

                             result := (addr <> VALUE_ZERO_MAC);
                        end
                     else
                         result := true;

                     // Check second MAC

                     if result then
                        begin
                             help := getEntry (value, '=', 2);

                             result := isValidMACString (help);

                             if result and CheckNull then
                                begin
                                     help := stringReplace (help, '-', ':', [rfReplaceAll]);
                                     addr.setfromString (help, ':');

                                     result := (addr <> VALUE_ZERO_MAC);
                                end
                        end;
                end
             else
                 result := false;
        end
     else
         result := false;
end;

(* -------------------------------------- *)

function isValidIPv4AddressPair (value : string; const CheckNull : boolean) : boolean;

var help  : string;
    addr  : tIPv4Address;

begin
     if numEntries (value, '=') = 2 then
        begin
             // Check first IP address

             help := getEntry (value, '=', 1);

             if isValidIPv4String (help) then
                begin
                     // Check if source is a zero IP address, if required

                     if CheckNull then
                        begin
                             addr.setValueFromString (help);

                             result := not addr.isZeroAddress;
                        end
                     else
                         result := true;

                     // Check second IP address

                     if result then
                        begin
                             help := getEntry (value, '=', 2);

                             result := isValidIPv4String (help);

                             if result and CheckNull then
                                begin
                                     addr.setValueFromString (help);

                                     result := not addr.isZeroAddress;
                                end
                        end;
                end
             else
                 result := false;
        end
     else
         result := false;
end;

(* -------------------------------------- *)

function isValidIPv4SubnetPair (value : string; const CheckNull : boolean) : boolean;

var help  : string;
    addr  : tIPv4Address;

begin
     if numEntries (value, '=') = 2 then
        begin
             // Check first subnet

             help := getEntry (value, '=', 1);

             if isValidIPv4Subnet (help) then
                begin
                     // Check if source is a zero subnet, if required

                     if CheckNull then
                        begin
                             addr.setValueFromCIDRString (help);

                             result := not addr.isZeroAddress;
                        end
                     else
                         result := true;

                     // Check second subnet

                     if result then
                        begin
                             help := getEntry (value, '=', 2);

                             result := isValidIPv4Subnet (help);

                             if result and CheckNull then
                                begin
                                     addr.setValueFromCIDRString (help);

                                     result := not addr.isZeroAddress;
                                end
                        end;
                end
             else
                 result := false;
        end
     else
         result := false;
end;

(* -------------------------------------- *)

function isValidIPv6AddressPair (value : string; const CheckNull : boolean) : boolean;

var help  : string;
    addr  : tIPv6Address;

begin
     if numEntries (value, '=') = 2 then
        begin
             // Check first IP address

             help := getEntry (value, '=', 1);

             if isValidIPv6String (help) then
                begin
                     // Check if source is a zero IP address, if required

                     if CheckNull then
                        begin
                             addr.setValueFromString (help);

                             result := not addr.IsUnspecifiedAddress;
                        end
                     else
                         result := true;

                     // Check second IP address

                     if result then
                        begin
                             help := getEntry (value, '=', 2);

                             result := isValidIPv6String (help);

                             // Check if destination is a zero IP address, if required

                             if result and CheckNull then
                                begin
                                     addr.setValueFromString (help);

                                     result := not addr.IsUnspecifiedAddress;
                                end
                        end;
                end
             else
                 result := false;
        end
     else
         result := false;
end;

(* -------------------------------------- *)

function isValidIPv6SubnetPair (value : string; const CheckNull : boolean) : boolean;

var help  : string;
    addr  : tIPv6Address;

begin
     if numEntries (value, '=') = 2 then
        begin
             // Check first subnet

             help := getEntry (value, '=', 1);

             if isValidIPv6Subnet (help) then
                begin
                     // Check if source is a zero subnet, if required

                     if CheckNull then
                        begin
                             addr.setValueFromStringWithMask (help);

                             result := not addr.IsUnspecifiedAddress;
                        end
                     else
                         result := true;

                     // Check second subnet

                     if result then
                        begin
                             help := getEntry (value, '=', 2);

                             result := isValidIPv6Subnet (help);

                             if result and CheckNull then
                                begin
                                     addr.setValueFromStringWithMask (help);

                                     result := not addr.IsUnspecifiedAddress;
                                end
                        end;
                end
             else
                 result := false;
        end
     else
         result := false;
end;

(* -------------------------------------- *)

function isValidSocketPair (value : string) : boolean;

var help : string;
    ver  : byte;
    sock : tIPSocket;

begin
     if numEntries (value, '=') = 2 then
        begin
             // Check first value for being a socket

             help := getEntry (value, '=', 1);
             result := isValidSocketString (help);

             if result then
                begin
                     // Determine IP version

                     sock.initialize (help);
                     ver := sock.IPAddress.IPVersion;

                     // Check first value for being a socket

                     help := getEntry (value, '=', 2);
                     result := isValidSocketString (help);

                     // Both are valid sockets, so now check if they are both the same IP version

                     if result then
                        begin
                             sock.initialize (help);

                             result := (ver = sock.IPAddress.IPVersion);
                        end;
                end;
        end
     else
         result := false;
end;

(* -------------------------------------- *)

end.

