{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_frameprocessor_baseclass;

interface

uses base_structures,
     class_rawframe,
     class_tracewriter,
     class_simpleframestorage,
     class_parsemaster;

type tOnProcessingError = procedure (Sender : tObject; const ErrorMessage : string) of Object;
     tOnFrameProcessed = procedure (Sender : tObject; Frame : tRawFrame) of Object;
     tOnFrameDiscard = procedure (Sender : tObject; Frame : tRawFrame) of Object;

     tFrameProcessor = class (tObject)

                             /// <summary>
                             ///   Creates a FrameProcessor object
                             /// </summary>
                             /// <param name="Name">
                             ///   Arbitrary name for the processor; used
                             ///   primarily for debugging when analyzing the
                             ///   chain
                             /// </param>
                             /// <param name="TraceWriter">
                             ///   TraceWriter object; only set if this is the
                             ///   last processor in the chain
                             /// </param>
                             /// <param name="MaxFrameCountBeforeWrite">
                             ///   Amount of frames that are stored before
                             ///   moving them to the writer object
                             /// </param>
                             constructor create (const Name : string; const TraceWriter : tTraceWriter = nil; const MaxFrameCountBeforeWrite : integer = 10000);

                             private
                                    priv_nextprocessor   : tFrameProcessor;
                                    priv_framestorage    : tSimpleFrameStorage;
                                    priv_tracewriter     : tTraceWriter;
                                    priv_maxframecount   : integer;

                             protected
                                      prot_OnProcessingError    : tOnProcessingError;
                                      prot_OnFrameProcessed     : tOnFrameProcessed;
                                      prot_OnFrameDiscard       : tOnFrameDiscard;

                                      prot_chainindex           : integer;

                                      prot_framesfiltered       : integer;
                                      prot_framespassed         : integer;

                                      prot_lowesttimestamp      : tFrameTimeStamp;
                                      prot_highesttimestamp     : tFrameTimeStamp;

                                      prot_processorname        : string;
                                      prot_framewasmodified     : boolean;
                                      prot_framesprocessed      : integer;

                                      prot_parsemaster          : tParseMaster;

                                      /// <summary>
                                      ///   This is called when a frame
                                      ///   processor down the chain frees the
                                      ///   Parsemaster, so that all chain
                                      ///   elements update their reference to
                                      ///   NIL, too
                                      /// </summary>
                                      /// <param name="Sender">
                                      ///   frame processor notifying about the
                                      ///   discarding of prot_parsemaster
                                      /// </param>
                                      procedure DiscardFrame (var Frame : tRawFrame);
                                      procedure DiscardParseMaster;
                                      procedure DoFrameStatistics (Frame : tRawFrame);
                                      /// <summary>
                                      ///   Transfers all frames currently in
                                      ///   the queue to the tracewriter object
                                      ///   (which will buffer them again and
                                      ///   flush to disk as soon as its own
                                      ///   threshold has been reached)
                                      /// </summary>
                                      procedure TransferFramesToWriter;

                             public
                                   destructor destroy; override;

                                   property ProcessorName : string read prot_processorname;
                                   property ChainIndex : integer read prot_chainindex write prot_chainindex;
                                   property NextProcessor : tFrameProcessor read priv_nextprocessor write priv_nextprocessor;
                                   property FrameStorage : tSimpleFrameStorage read priv_framestorage write priv_framestorage;
                                   property TraceWriter : tTraceWriter read priv_tracewriter;

                                   property FramesProcessed : integer read prot_framesprocessed;
                                   property FramesFiltered : integer read prot_framesfiltered;
                                   property FramesPassed : integer read prot_framespassed;

                                   property OnProcessingError : tOnProcessingError read prot_OnProcessingError write prot_OnProcessingError;
                                   property OnFrameProcessed : tOnFrameProcessed read prot_OnFrameProcessed  write prot_OnFrameProcessed;
                                   property OnFrameDiscard : tOnFrameDiscard read prot_OnFrameDiscard write prot_OnFrameDiscard;

                                   /// <summary>
                                   ///   Should only be called by the process
                                   ///   chain before adding a frame to the
                                   ///   chain
                                   /// </summary>
                                   procedure InitParseMaster;
{$ifdef debug}
                                   function ParseMasterExists : boolean;
{$endif}
                                   {$REGION 'Documentation'}
                                   /// <summary>
                                   ///   This is the core method of the
                                   ///   processor which works with the
                                   ///   incoming frame and produces an output
                                   ///   frame as result. Usually this routine
                                   ///   is not called directly, but by
                                   ///   AddFrame when the Frameprocessor is
                                   ///   part of a process chain. So only use
                                   ///   this directly if there is just this
                                   ///   one processor.
                                   /// </summary>
                                   /// <param name="Frame">
                                   ///   Frame to process
                                   /// </param>
                                   /// <returns>
                                   ///   Processed frame, or the original if no
                                   ///   processing was necessary/possible
                                   /// </returns>
                                   {$ENDREGION}
                                   function GetProcessedFrame (Frame : tRawFrame) : tRawFrame; virtual;
                                   {$REGION 'Documentation'}
                                   /// <summary>
                                   ///   Main routine to process a frame. This
                                   ///   is called when the frameprocessor is
                                   ///   part of a process chain, and passes
                                   ///   the result down the chain until
                                   ///   complete.
                                   /// </summary>
                                   /// <param name="Frame">
                                   ///   Frame to be processed and passed down
                                   ///   the chain
                                   /// </param>
                                   /// <remarks>
                                   ///   Usually a tTraceWriter object is
                                   ///   receiving the finalized frame at the
                                   ///   end of the process chain, writing it
                                   ///   to the output file.
                                   /// </remarks>
                                   {$ENDREGION}
                                   procedure addFrame (Frame : tRawFrame); virtual;
                       end;

implementation

uses SiAuto,
     SysUtils,
     Math,
     class_interfacelist,
     class_writer_PCAPNG;

(* ------------------------------------------- *)
{ tFrameProcessor }
(* ------------------------------------------- *)

constructor tFrameProcessor.create (const Name : string; const TraceWriter : tTraceWriter = nil; const MaxFrameCountBeforeWrite : integer = 10000);

begin
     inherited create;

     prot_processorname := Name;

     priv_nextprocessor := nil;
     prot_framewasmodified := false;
     prot_framesprocessed := 0;
     prot_framesfiltered  := 0;
     prot_framespassed    := 0;

     prot_parsemaster := nil;

     prot_lowesttimestamp.initialize;
     prot_highesttimestamp.initialize;

     if TraceWriter <> nil then
        begin
             priv_framestorage := tSimpleFrameStorage.create (false);
             priv_tracewriter := TraceWriter;
        end
     else
         begin
             priv_framestorage := nil;
             priv_tracewriter := nil;
         end;

     priv_maxframecount := MaxFrameCountBeforeWrite;
end;

(* ------------------------------------------- *)

destructor tFrameProcessor.destroy;

begin
     if priv_framestorage <> nil then
        begin
             // Transfer remaining frames to writer, if any
             // Flushing and closing the file is done in the task
             // processors, not here, to allow adding stat records etc.

             if priv_framestorage.Count > 0 then
                TransferFramesToWriter;
        end;

     inherited;
end;

(* ------------------------------------------- *)

procedure tFrameProcessor.InitParseMaster;

begin
     prot_ParseMaster := nil;
end;

(* ------------------------------------------- *)

{$ifdef debug}
function tFrameProcessor.ParseMasterExists: boolean;

begin
     result := assigned (prot_ParseMaster);
end;
{$endif}

(* ------------------------------------------- *)

function tFrameProcessor.GetProcessedFrame (Frame: tRawFrame): tRawFrame;

begin
     // Easy: by default, do nothing. Overriding classes will do their main work here instead

     result := Frame;
end;

(* ------------------------------------------- *)

procedure tFrameProcessor.TransferFramesToWriter;

var currentFrame   : tRawFrame;
    Writer         : tPCAPngWriter;
    InterfaceCount : integer;
    InterfaceEntry : tCaptureInterface;
    i              : integer;

begin
     Writer := (priv_tracewriter as tPCAPngWriter);
     InterfaceCount := Writer.InterfaceList.Count;

     // Add frames from framebuffer to the writer object

     for i := 1 to min (priv_framestorage.Count, priv_maxframecount) do
         begin
              currentFrame := priv_framestorage.getFrame(i);

              // Assign the new interface by looking up the writer interface of the same index
              // since the writing process created the new interfaces in the same order

              InterfaceEntry := Writer.InterfaceList.getInterfaceByIndex (CurrentFrame.InterfaceID);
              CurrentFrame.InterfaceLinkID := InterfaceEntry.PacketLinkID;

              priv_tracewriter.addFrameToList(currentFrame);
         end;

     // remove from framebuffer, giving control over the frame to the writer
     // This will not discard the tRawframe. It will just be removed from
     // the framebuffer used for sorting. Can't use clear here because there may
     // be subsets removed

     for i := min (priv_framestorage.Count, priv_maxframecount) downto 1 do
         priv_framestorage.DeleteFrame(i);

     if priv_framestorage.Count > 0 then
        SiMain.LogError('Error: FrameStorage is not empty but should be.');
end;

(* ------------------------------------------- *)

procedure tFrameProcessor.DiscardFrame (var Frame: tRawFrame);

begin
//   siMain.EnterMethod ('tFrameProcessor.DiscardFrame');

     if assigned (prot_OnFrameDiscard) then
        prot_OnFrameDiscard (self, Frame);

     FreeAndNil (Frame);

     if assigned (prot_ParseMaster) then
        begin
//           SiMain.LogDebug ('Discarding ParseMaster');

//           FreeAndNil (prot_parsemaster);
             DiscardParseMaster;
        end;

//   siMain.LeaveMethod ('tFrameProcessor.DiscardFrame');
end;

(* ------------------------------------------- *)

procedure tFrameProcessor.DiscardParseMaster;

begin
     FreeAndNil (prot_parsemaster);
end;

(* ------------------------------------------- *)

procedure tFrameProcessor.DoFrameStatistics (Frame: tRawFrame);

begin
     if prot_lowesttimestamp.datetime = 0 then
        prot_lowesttimestamp := Frame.Timestamp
     else
         if Frame.Timestamp < prot_lowesttimestamp then
            prot_lowesttimestamp := Frame.Timestamp;

     if prot_highesttimestamp.datetime = 0 then
        prot_highesttimestamp := Frame.Timestamp
     else
         if Frame.Timestamp > prot_highesttimestamp then
            prot_highesttimestamp := Frame.Timestamp;
end;

(* ------------------------------------------- *)

procedure tFrameProcessor.addFrame (Frame: tRawFrame);

var NewFrame : tRawFrame;

begin
//   SiMain.EnterMethod ('tFrameProcessor.addFrame');
//   SiMain.LogDebug ('Processing Frame #' + IntToStr (Frame.FileFrameID) + ' in processor "' + prot_processorname + '"');

     // Make sure parsemaster is going to be created if this is the first processor

     if prot_chainindex = 1 then
        prot_parsemaster := nil;

     // Do statistics

     DoFrameStatistics (Frame);

     // Do the processing

     inc (prot_framesprocessed);

     NewFrame := GetProcessedFrame (Frame);

     // If a frame was returned we continue, otherwise the frame was killed in
     // action (maybe by a filter processor that did the Gandalf thing...)

     if NewFrame <> nil then
        begin
             // Pass the frame to the next processor, or the frame storage

             if priv_nextprocessor <> nil then
                begin
                     // Discard parsemaster as it is outdated now

                     if prot_framewasmodified then
                        begin
//                           SiMain.LogDebug ('Frame was modified; discarding parsemaster');

                             FreeAndNil (prot_parsemaster);
                        end;

//                   SiMain.LogDebug ('Pushing frame to next processor ("' + priv_nextprocessor.prot_processorname + '")');

                     priv_nextprocessor.prot_parsemaster := prot_parsemaster;
                     priv_nextprocessor.addFrame (NewFrame)
                end
             else
                 if priv_framestorage <> nil then
                    begin
//                       SiMain.LogDebug ('Pushing frame to next framestorage for write ("' + prot_processorname + '")');

                         priv_framestorage.AddFrame (NewFrame);

                         // If the buffer limit is reached, write it

                         if priv_framestorage.Count >= priv_MaxFrameCount then
                            TransferFramesToWriter;

                         // Get rid of the ParseMaster if it exists

                         if assigned (prot_ParseMaster) then
                            begin
                                 DiscardParseMaster;
                                 FreeAndNil (prot_ParseMaster);
//                               SiMain.LogDebug ('ParseMaster Object destroyed after adding frame to Writer object ("' + prot_processorname + '")');
                            end;
                    end
                 else
                     begin
                          SiMain.LogWarning ('No next processor to push the frame to');

                          if not assigned (prot_OnFrameProcessed) then
                             begin
                                  DiscardFrame (NewFrame);

                                  raise Exception.Create('FrameProcessor failed: no target for frame ("' + prot_processorname + '")');
                             end
                          else
                              prot_OnFrameProcessed (self, NewFrame);
                     end;
        end;

//   SiMain.LeaveMethod ('tFrameProcessor.addFrame');
end;

(* ------------------------------------------- *)

end.
