unit class_frameprocessor_capwap;

interface

uses base_structures,
     class_rawframe,
     class_frameprocessor_baseclass;

type tFrameProcessorCAPWAP = class (tFrameProcessor)

                                    private
                                           function getRawFrameFromFrameChunk (Frame : tRawFrame; PayLoad : tFrameChunk; FrameLengthModifier : integer) : tRawFrame;

                                    public
                                          /// <summary>
                                          ///   Removes the CAPWAP and 802.11
                                          ///   layer from the frame if found,
                                          ///   and moves the tunneled content
                                          ///   to the beginning of the frame.
                                          /// </summary>
                                          /// <param name="Frame">
                                          ///   Frame to be processed
                                          /// </param>
                                          function getProcessedFrame (Frame : tRawFrame) : tRawFrame; override;

                              end;

implementation

uses SiAuto,
     SysUtils,
     Toolkit,
     class_parser_baseclass,
     common_parse_helpers,
     base_checksums,
     class_parser_ethernet,
     class_parser_ipv4,
     class_parser_ipv6,
     class_parser_capwap,
     class_parser_llc,
     class_parsemaster,
     class_assembly_ethernet;

(* ------------------------------------------- *)
{ tFrameProcessorCAPWAP }
(* ------------------------------------------- *)

function tFrameProcessorCAPWAP.getRawFrameFromFrameChunk (Frame: tRawFrame; PayLoad: tFrameChunk; FrameLengthModifier : integer): tRawFrame;

var NewFrame     : tRawFrame;
    TimeDetails  : tTimingDetails;
    WireDetails  : tWireDetails;
    FilePosition : tFileFrameLocation;
    FrameBuffer  : tRawBuffer;
    FinalLen     : integer;
    i            : integer;

begin
     TimeDetails.AbsoluteTime := Frame.Timestamp.datetime;
     TimeDetails.TraceTicks := Frame.Ticks;
     TimeDetails.NanoTicks := (Frame.Timestamp.nanoseconds mod 1000);            // because right now, Nanosecs also contain microsecs

     WireDetails.LinkType := Frame.LinkType;
     WireDetails.WireSize := Frame.WireSize - FrameLengthModifier;
     WireDetails.CaptureSize := Payload.Len;
     WireDetails.InterfaceID := Frame.InterfaceID;
     WireDetails.Dropped := Frame.DropCount;
     WireDetails.FrameFlags := Frame.FrameFlags;
     WireDetails.FrameType := Frame.FrameFileType;

     // Transfer Packet direction

//   WireDetails.PacketDirection := pdUnknown;

     FilePosition.FileIndex := 0;
     FilePosition.FileFrameIndex := prot_framesprocessed;
     FilePosition.FileOffset := 0;

     move (Payload.Data^, FrameBuffer, WireDetails.CaptureSize);

     NewFrame := tRawFrame.create (TimeDetails, WireDetails, FilePosition, FrameBuffer);

     for i := 1 to Frame.Commentlist.Count do
         NewFrame.Commentlist.Add(Frame.Commentlist[i-1]);

     result := NewFrame;
end;

(* ------------------------------------------- *)

function tFrameProcessorCAPWAP.getProcessedFrame (Frame: tRawFrame): tRawFrame;

var NewFrame     : tRawFrame;
    HeaderType   : tHeaderType;
    CAPWAPParser : tCAPWAPParser;
    LLCParser    : tLLCParser;
    IPChunk      : tFrameChunk;
    MergeChunk   : tFrameChunk;
    L2Pseudo     : tFrameChunk;
    found        : boolean;
    i            : integer;

begin
     if not assigned (prot_parsemaster) then
        prot_ParseMaster := tParseMaster.create (Frame, ppmParseAll);

     // Check that we have at least two IP layers

     if prot_ParseMaster.IPLayerCount = 2 then
        begin
             // Find the CAPWAP Header

             i := 0;
             found := false;

             repeat
                   inc (i);
                   HeaderType := prot_ParseMaster.getParserTypeByIndex(i);

                   if HeaderType = NEXT_HEADER_CAPWAP then
                      found := true;

             until found or (i = prot_ParseMaster.ParserCount);

             if found then
                begin
                     CAPWAPParser := prot_ParseMaster.getParserByIndex(i) as tCAPWAPParser;

                     // Check if we have 802.11 and LLC headers following, which we need to skip, too

                     if (cwhfPayloadTypeNative in CAPWAPParser.HeaderFlags) and (i + 1 < prot_ParseMaster.ParserCount) then
                        begin
                             // Check if the headers are as expected, and if so, get the offset of the next header, which would be IP

                             if (prot_ParseMaster.getParserTypeByIndex(i+1) = NEXT_HEADER_802dot11Hdr) and (prot_ParseMaster.getParserTypeByIndex(i+2) = NEXT_HEADER_LLC) then
                                begin
                                     LLCParser := prot_ParseMaster.getParserByIndex(i+2) as tLLCParser;

                                     // Create zeroed Pseudo Ethernet header and patch the ether
                                     // type into the EtherType location at offset 12

                                     L2Pseudo.CreateZeroBuffer (14);
                                     L2Pseudo.Patch (12, word (LLCParser.NextAnchor.HeaderType));

//                                   SiMain.LogDebug ('L2: ' + L2Pseudo.getAsHEXString);

                                     IPChunk := Frame.FrameDataAsChunk (LLCParser.NextAnchor.Offset);
//                                   SiMain.LogDebug ('IP: ' + IPChunk.getAsHEXString);

                                     MergeChunk := mergeFrameChunks (L2Pseudo, IPChunk);
                                     L2Pseudo.free;
                                     IPChunk.free;

                                     NewFrame := getRawFrameFromFrameChunk (Frame, MergeChunk, Frame.FrameSize - MergeChunk.Len);

                                      // Discard Chunk and original Frame

                                      Frame.free;
                                      Frame := nil;

                                      MergeChunk.free;
                                end;
                        end;
{
                     Need a sample trace for this first

                     else
                         begin
                              IPChunk := Frame.FrameDataAsChunk (CAPWAPParser.NextAnchor.Offset);




                         end;
}
                     // Set frame to modified to force discarding the parsemaster object

                     prot_framewasmodified := true;
                end
             else
                 NewFrame := Frame;
        end
     else
         NewFrame := Frame;

     result := NewFrame;
end;

(* ------------------------------------------- *)

end.
