{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_frameprocessor_gre;

interface

uses base_structures,
     class_rawframe,
     class_frameprocessor_baseclass;

type tFrameProcessorGRE = class (tFrameProcessor)

                                private
                                       function getRawFrameFromFrameChunk (Frame : tRawFrame; PayLoad : tFrameChunk; FrameLengthModifier : integer) : tRawFrame;

                                public
                                      {$REGION 'Documentation'}
                                      /// <summary>
                                      ///   Removes a GRE tunnel header from a
                                      ///   frame and moves the tunneled
                                      ///   content to the beginning of the
                                      ///   frame
                                      /// </summary>
                                      /// <param name="Frame">
                                      ///   Frame to be processed
                                      /// </param>
                                      {$ENDREGION}
                                      function getProcessedFrame (Frame : tRawFrame) : tRawFrame; override;

                          end;

implementation

uses class_parser_baseclass,
     common_parse_helpers,
     class_parser_GRE,
     class_parser_ERSPAN,
     class_parser_ethernet,
     class_parser_ipv4,
     class_parser_ipv6,
     class_parsemaster;

(* ------------------------------------------- *)
{ tFrameProcessorGTPU }
(* ------------------------------------------- *)

function tFrameProcessorGRE.getRawFrameFromFrameChunk (Frame: tRawFrame; PayLoad: tFrameChunk; FrameLengthModifier : integer): tRawFrame;

var NewFrame     : tRawFrame;
    TimeDetails  : tTimingDetails;
    WireDetails  : tWireDetails;
    FilePosition : tFileFrameLocation;
    FrameBuffer  : tRawBuffer;
    FinalLen     : integer;
    i            : integer;

begin
     TimeDetails.AbsoluteTime := Frame.Timestamp.datetime;
     TimeDetails.TraceTicks := Frame.Ticks;
     TimeDetails.NanoTicks := (Frame.Timestamp.nanoseconds mod 1000);            // because right now, Nanosecs also contain microsecs


     WireDetails.LinkType := Frame.LinkType;
     WireDetails.WireSize := Frame.WireSize - FrameLengthModifier;
     WireDetails.CaptureSize := Payload.Len;
     WireDetails.InterfaceID := Frame.InterfaceID;
     WireDetails.Dropped := Frame.DropCount;
     WireDetails.FrameFlags := Frame.FrameFlags;
     WireDetails.FrameType := Frame.FrameFileType;

     // Transfer Packet direction

//   WireDetails.PacketDirection := pdUnknown;

     FilePosition.FileIndex := 0;
     FilePosition.FileFrameIndex := prot_framesprocessed;
     FilePosition.FileOffset := 0;

     move (Payload.Data^, FrameBuffer, WireDetails.CaptureSize);

     NewFrame := tRawFrame.create (TimeDetails, WireDetails, FilePosition, FrameBuffer);

     for i := 1 to Frame.Commentlist.Count do
         NewFrame.Commentlist.Add(Frame.Commentlist[i-1]);

     result := NewFrame;
end;

(* ------------------------------------------- *)

function tFrameProcessorGRE.getProcessedFrame (Frame: tRawFrame): tRawFrame;

var ParseEntry   : tParseEntry;
//  ParseMaster  : tParseMaster;
    NewFrame     : tRawFrame;
    HeaderType   : tHeaderType;
    GParser      : tGREParser;
    IPv4Parser   : tIPv4Parser;
    IPv6Parser   : tIPv6Parser;
    ERSPANParser : tERSPANParser;
    Layer2Chunk  : tFrameChunk;
    IPChunk      : tFrameChunk;
    MergeChunk   : tFrameChunk;
    found        : boolean;
    iOffset      : integer;
    i            : integer;

begin
     if not assigned (prot_parsemaster) then
        prot_ParseMaster := tParseMaster.create (Frame, ppmParseAll);

     // Check that we have at least two IP layers

     if prot_ParseMaster.IPLayerCount = 2 then
        begin
             // Find the GRE Header

             i := 0;
             found := false;

             repeat
                   inc (i);
                   HeaderType := prot_ParseMaster.getParserTypeByIndex(i);

                   if HeaderType = NEXT_HEADER_GRE then
                      found := true;

             until found or (i = prot_ParseMaster.ParserCount);

             if found then
                begin
                     GParser := prot_ParseMaster.getParserByIndex(i) as tGREParser;

                     if GParser.NextAnchor.HeaderType = NEXT_HEADER_ERSPAN then
                        begin
                             // ERSPAN header after GRE

                             ERSPANParser := prot_ParseMaster.getParserByIndex(i+1) as tERSPANParser;

                             // Get the part starting right after the GRE Header

                             Layer2Chunk := Frame.FrameDataAsChunk(ERSPANParser.NextAnchor.Offset);

                             // Create Frame

                             NewFrame := getRawFrameFromFrameChunk (Frame, Layer2Chunk, Frame.FrameSize - Layer2Chunk.Len);

                             // Discard Chunk and original Frame

                             Frame.free;
                             Frame := nil;

                             Layer2Chunk.free;
                        end
                     else
                         begin
                              // Get the part of the frame before the GRE header, so lets find an IP header and determine its starting offset

                              iOffset := 0;

                              repeat
                                    dec (i);

                                    ParseEntry := prot_ParseMaster.getParserParseInfoByIndex(i);

                                    case ParseEntry.Anchor.HeaderType of
                                     NEXT_HEADER_IPv4_ENCAPSULATION,
                                     NEXT_HEADER_IPv4,
                                     NEXT_HEADER_TYPE_IPV6,
                                     NEXT_HEADER_IPv6                : iOffset := ParseEntry.Anchor.Offset;
                                    end;

                              until (iOffset > 0) or (i = 1);

                              // If we found the previous IP layer, remove the GRE header

                              if iOffset > 0 then
                                 begin
                                      Layer2Chunk := Frame.FrameDataAsChunk (0, iOffset);

                                      // Get the part starting right after the GRE Header

                                      IPChunk := Frame.FrameDataAsChunk(GParser.NextAnchor.Offset);

                                      // Merge Ethernet and IP

                                      MergeChunk := mergeFrameChunks (Layer2Chunk, IPChunk);

                                      Layer2Chunk.free;
                                      IPChunk.free;

                                      // Create Frame

                                      NewFrame := getRawFrameFromFrameChunk (Frame, MergeChunk, Frame.FrameSize - MergeChunk.Len);

                                      // Discard Chunk and original Frame

                                      Frame.free;
                                      Frame := nil;

                                      MergeChunk.free;
                                 end
                              else
                                  // Give up and pass the frame

                                  NewFrame := Frame;
                         end;

                     // Set frame to modified to force discarding the parsemaster object

                     prot_framewasmodified := true;
                end
             else
                 NewFrame := Frame;
        end
     else
         NewFrame := Frame;

     // Free up the parsemaster

//   ParseMaster.free;

     result := NewFrame;
end;

(* ------------------------------------------- *)

end.
