{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_frameprocessor_gtpu;

interface

uses base_structures,
     class_rawframe,
     class_frameprocessor_baseclass;

type tFrameProcessorGTPU = class (tFrameProcessor)

                                    private
                                           function getRawFrameFromFrameChunk (Frame : tRawFrame; PayLoad : tFrameChunk; FrameLengthModifier : integer) : tRawFrame;

                                    public
                                          {$REGION 'Documentation'}
                                          /// <summary>
                                          ///   Removes a GTP-U tunnel header
                                          ///   from the frame if found, and
                                          ///   moves the tunneled content to
                                          ///   the beginning of the frame.
                                          /// </summary>
                                          /// <param name="Frame">
                                          ///   Frame to be processed
                                          /// </param>
                                          {$ENDREGION}
                                          function getProcessedFrame (Frame : tRawFrame) : tRawFrame; override;

                              end;

implementation

uses SiAuto,
     SysUtils,
     class_parser_baseclass,
     common_parse_helpers,
     class_parser_GTPv1,
     class_parser_ethernet,
     class_parser_linux_sll,
     class_parser_ipv4,
     class_parser_ipv6,
     class_assembly_ethernet,
     class_assembly_linux_sll,
     class_parsemaster;

(* ------------------------------------------- *)
{ tFrameProcessorGTPU }
(* ------------------------------------------- *)

function tFrameProcessorGTPU.getRawFrameFromFrameChunk (Frame: tRawFrame; PayLoad: tFrameChunk; FrameLengthModifier : integer): tRawFrame;

var NewFrame     : tRawFrame;
    TimeDetails  : tTimingDetails;
    WireDetails  : tWireDetails;
    FilePosition : tFileFrameLocation;
    FrameBuffer  : tRawBuffer;
    FinalLen     : integer;
    i            : integer;

begin
     TimeDetails.AbsoluteTime := Frame.Timestamp.datetime;
     TimeDetails.TraceTicks := Frame.Ticks;
     TimeDetails.NanoTicks := (Frame.Timestamp.nanoseconds mod 1000);            // because right now, Nanosecs also contain microsecs

     WireDetails.LinkType := Frame.LinkType;
     WireDetails.WireSize := Frame.WireSize - FrameLengthModifier;
     WireDetails.CaptureSize := Payload.Len;
     WireDetails.InterfaceID := Frame.InterfaceID;
     WireDetails.Dropped := Frame.DropCount;
     WireDetails.FrameFlags := Frame.FrameFlags;
     WireDetails.FrameType := Frame.FrameFileType;

     // Transfer Packet direction

//   WireDetails.PacketDirection := pdUnknown;

     FilePosition.FileIndex := 0;
     FilePosition.FileFrameIndex := prot_framesprocessed;
     FilePosition.FileOffset := 0;

     move (Payload.Data^, FrameBuffer, WireDetails.CaptureSize);

     NewFrame := tRawFrame.create (TimeDetails, WireDetails, FilePosition, FrameBuffer);

     for i := 1 to Frame.Commentlist.Count do
         NewFrame.Commentlist.Add(Frame.Commentlist[i-1]);

     result := NewFrame;
end;

(* ------------------------------------------- *)

function tFrameProcessorGTPU.getProcessedFrame (Frame: tRawFrame): tRawFrame;

var ParseEntry   : tParseEntry;
//  ParseMaster  : tParseMaster;
    NewFrame     : tRawFrame;
    HeaderType   : tHeaderType;
    GParser      : tGTPv1Parser;
    Layer2Chunk  : tFrameChunk;
    EthParser    : tEthernetParser;
    SLLParser    : tLinuxSLLParser;
    EthAssembler : tEthernetAssembler;
    SLLAssembler : tLinuxSLLAssembler;
    IPv4Parser   : tIPv4Parser;
    IPv6Parser   : tIPv6Parser;
    IPChunk      : tFrameChunk;
    MergeChunk   : tFrameChunk;
    found        : boolean;
    iOffset      : integer;
    i            : integer;

begin
//   SiMain.EnterMethod ('tFrameProcessorGTPU.getProcessedFrame');

     if not assigned (prot_ParseMaster) then
        prot_ParseMaster := tParseMaster.create(Frame, ppmParseAll);

     // Check that we have at least two IP layers

     if prot_ParseMaster.IPLayerCount > 1 then
        begin
             // Find the GTP-U Header

             i := 0;
             found := false;
             EthParser := nil;
             SLLParser := nil;

             repeat
                   inc (i);
                   HeaderType := prot_ParseMaster.getParserTypeByIndex(i);

                   // While we're looking for GTP, keep the highest L2 header handy, because the IP protocol transported
                   // may be different than what's carrying the GTP protocol layer

                   if (HeaderType = NEXT_HEADER_ETHERNET) or (HeaderType = NEXT_HEADER_TRANS_ETHER_BRIDGE) then
                      EthParser := prot_ParseMaster.getParserByIndex(i) as tEthernetParser;

                   if (HeaderType = NEXT_HEADER_LINUX_COOKED)  then
                      SLLParser := prot_ParseMaster.getParserByIndex(i) as tLinuxSLLParser;

                   if HeaderType = NEXT_HEADER_GTPv1 then
                      found := true;

             until found or (i = prot_ParseMaster.ParserCount);

             // TODO: warn if no Ethernet header was found

             if found and (assigned (EthParser) or assigned (SLLParser)) then
                begin
//                   SiMain.LogVerbose ('Found GTPv1 layer at index ' + IntToStr (i));

                     GParser := prot_ParseMaster.getParserByIndex(i) as tGTPv1Parser;

                     // Get the part of the frame before the GTP-U header, so lets find an IP header and determine its starting offset

                     iOffset := 0;

                     repeat
                           dec (i);

                           ParseEntry := prot_ParseMaster.getParserParseInfoByIndex(i);

                           if (ParseEntry.Anchor.HeaderType = NEXT_HEADER_IPv4_ENCAPSULATION) or (ParseEntry.Anchor.HeaderType = NEXT_HEADER_IPv4) or
                              (ParseEntry.Anchor.HeaderType = NEXT_HEADER_IPv6) or (ParseEntry.Anchor.HeaderType = NEXT_HEADER_TYPE_IPV6) then
                              iOffset := ParseEntry.Anchor.Offset;

                     until (iOffset > 0) or (i = 1);

                     // If we found the previous IP layer, remove the GRE header

                     if (iOffset > 0) then
                        begin
                             if assigned (EthParser) then
                                begin
                                     // Construct Ethernet Header

                                     EthAssembler := tEthernetAssembler.create;
                                     EthAssembler.CopyFromParser (EthParser);

                                     case GParser.NextAnchor.HeaderType of
                                      NEXT_HEADER_IPv4 : EthAssembler.EtherTypeLength := NEXT_HEADER_IPv4;
                                      NEXT_HEADER_IPv6 : EthAssembler.EtherTypeLength := NEXT_HEADER_TYPE_IPV6;
                                     end;

                                     Layer2Chunk := EthAssembler.GetAssembly;
                                     EthAssembler.Free;
                                end
                             else
                                 if assigned (SLLParser) then
                                    begin
                                         SLLAssembler := tLinuxSLLAssembler.create;
                                         SLLAssembler.CopyFromParser (SLLParser);

                                         case GParser.NextAnchor.HeaderType of
                                          NEXT_HEADER_IPv4 : SLLAssembler.ProtocolType := NEXT_HEADER_IPv4;
                                          NEXT_HEADER_IPv6 : SLLAssembler.ProtocolType := NEXT_HEADER_TYPE_IPV6;
                                         end;

                                         Layer2Chunk := SLLAssembler.GetAssembly;
                                         SLLAssembler.Free;
                                    end;

//                           Layer2Chunk := Frame.FrameDataAsChunk (0, iOffset);

                             // Get the part starting right after the GTPU Header

                             IPChunk := Frame.FrameDataAsChunk(GParser.NextAnchor.Offset);

                             // Merge Ethernet and IP

                             MergeChunk := mergeFrameChunks (Layer2Chunk, IPChunk);

                             Layer2Chunk.free;
                             IPChunk.free;

                             // TODO: handle Ethernet FCS (yup. Lazy me for not doing it right away.)

                             // Create New Frame

                             NewFrame := getRawFrameFromFrameChunk (Frame, MergeChunk, Frame.FrameSize - MergeChunk.Len);

                             // Discard Chunk and old Frame

                             MergeChunk.free;

                             Frame.free;
                             Frame := nil;

                             // Set frame to modified to force discarding the parsemaster object

                             prot_framewasmodified := true;
                        end;
                end
             else
                 // Give up and pass the frame

                 NewFrame := Frame;
        end
     else
         NewFrame := Frame;

     // Free up the parsemaster

//     ParseMaster.free;

     result := NewFrame;

//   SiMain.LeaveMethod ('tFrameProcessorGTPU.getProcessedFrame');
end;

(* ------------------------------------------- *)

end.
