{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_frameprocessor_replaceSLL;

interface

uses classes,
     base_structures,
     definition_ip_addresses,
     class_rawframe,
     class_frameprocessor_baseclass;

type tFrameProcessorReplaceSLL = class (tFrameProcessor)

                                       constructor create (const Name : string);

                                       private
                                              priv_IPtoMACTable : tStringlist;    // Not used yet
                                              priv_ReplaceEtherType : boolean;
                                              priv_OriginalEthertype : integer;
                                              priv_ReplacementEthertype : integer;
                                              priv_ReplacementMACAddress : tMACAddress;
                                              priv_ForceMACReplacement : boolean;

                                              function getRawFrameFromFrameChunk (Frame : tRawFrame; PayLoad : tFrameChunk; FrameLengthModifier : integer) : tRawFrame;

                                       public
                                             property ReplaceEtherType : boolean read priv_ReplaceEtherType write priv_ReplaceEtherType;
                                             property OriginalEtherType : integer read priv_OriginalEthertype write priv_OriginalEthertype;
                                             property ReplacementEtherType : integer read priv_ReplacementEthertype write priv_ReplacementEthertype;
                                             property ReplacementMACAddress : tMACAddress read priv_ReplacementMACAddress write priv_ReplacementMACAddress;
                                             property ForceMACReplacement : boolean read priv_ForceMACReplacement write priv_ForceMACReplacement;

                                             {$REGION 'Documentation'}
                                             /// <summary>
                                             ///   This routine replaces a
                                             ///   Linux cooked header with a
                                             ///   pseudo Ethernet header, and
                                             ///   transfers as much as
                                             ///   possible, e.g. MAC addresses.
                                             /// </summary>
                                             /// <param name="Frame">
                                             ///   Frame to be processed
                                             /// </param>
                                             {$ENDREGION}
                                             function getProcessedFrame (Frame : tRawFrame) : tRawFrame; override;

                                             destructor Destroy; override;

                                 end;

implementation

uses SiAuto,
     SysUtils,
     common_parse_helpers,
     class_parser_baseclass,
     class_parser_ethernet,
     class_assembly_802dot1q,
     class_parser_linux_sll,
     class_parsemaster,
     class_assembly_ethernet;

(* ------------------------------------------- *)
{ tFrameProcessorReplaceSLL }
(* ------------------------------------------- *)

constructor tFrameProcessorReplaceSLL.create (const Name : string);

begin
     inherited create (Name);

     priv_IPtoMACTable := TStringList.Create;
     priv_ReplacementMACAddress.initialize;
end;

(* ------------------------------------------- *)

destructor tFrameProcessorReplaceSLL.Destroy;

begin
     priv_IPtoMACTable.Free;

     inherited;
end;

(* ------------------------------------------- *)

function tFrameProcessorReplaceSLL.getRawFrameFromFrameChunk (Frame: tRawFrame; PayLoad: tFrameChunk; FrameLengthModifier : integer): tRawFrame;

var NewFrame     : tRawFrame;
    TimeDetails  : tTimingDetails;
    WireDetails  : tWireDetails;
    FilePosition : tFileFrameLocation;
    FrameBuffer  : tRawBuffer;
    FinalLen     : integer;
    i            : integer;

begin
     TimeDetails.AbsoluteTime := Frame.Timestamp.datetime;
     TimeDetails.TraceTicks := Frame.Ticks;
     TimeDetails.NanoTicks := (Frame.Timestamp.nanoseconds mod 1000);            // because right now, Nanosecs also contain microsecs

     // Force Ethernet

     WireDetails.LinkType := LINKTYPE_ETHERNET;

     WireDetails.WireSize := Frame.WireSize - FrameLengthModifier;
     WireDetails.CaptureSize := Payload.Len;

     WireDetails.InterfaceID := Frame.InterfaceID;
     WireDetails.Dropped := Frame.DropCount;
     WireDetails.FrameFlags := Frame.FrameFlags;
     WireDetails.FrameType := Frame.FrameFileType;

     // Transfer Packet direction

     FilePosition.FileIndex := 0;
     FilePosition.FileFrameIndex := prot_framesprocessed;
     FilePosition.FileOffset := 0;

     move (Payload.Data^, FrameBuffer, WireDetails.CaptureSize);

     NewFrame := tRawFrame.create (TimeDetails, WireDetails, FilePosition, FrameBuffer);

     for i := 1 to Frame.Commentlist.Count do
         NewFrame.Commentlist.Add(Frame.Commentlist[i-1]);

     result := NewFrame;
end;

(* ------------------------------------------- *)

function tFrameProcessorReplaceSLL.getProcessedFrame (Frame: tRawFrame): tRawFrame;

var
//  ParseMaster    : tParseMaster;
    ParseEntry     : tParseEntry;
    NewFrame       : tRawFrame;
    HeaderType     : tHeaderType;
    EtherChunk     : tFrameChunk;
    Dot1QChunk     : tFrameChunk;
    Layer3Chunk    : tFrameChunk;
    SLLParser      : tLinuxSLLParser;
    EtherAssembler : tEthernetAssembler;
    VLANAssembler  : t802dot1qAssembler;
    HelpChunk      : tFrameChunk;
    MergeChunk     : tFrameChunk;
    found          : boolean;
    iOffset        : integer;
    i              : integer;

begin
//   ParseMaster := tParseMaster.create (Frame, ppmEdit);

//   if ParseMaster.getParserTypeByIndex (1) = NEXT_HEADER_LINUX_COOKED then
     if Frame.LinkType = LINKTYPE_LINUX_SLL then
        begin
             if not assigned (prot_ParseMaster) then
                prot_ParseMaster := tParseMaster.create(Frame, ppmParseAll);

             // Assemble Ethernet from Linux Cooked Header

             SLLParser := prot_ParseMaster.getParserByIndex (1) as tLinuxSLLParser;
             EtherAssembler := tEthernetAssembler.create;

             // TODO: Write down the MAC to IP relationship if we can
{
             if SLLParser.PacketType = sptSentByUs then
                begin
                     // priv_IPtoMACTable.Add ()
                end;
}
             // Take Source MAC if available

             if (SLLParser.LinkLayerAddressLength = 6) and not priv_ForceMACReplacement then
                EtherAssembler.SourceAddress := SLLParser.MACAddress
             else
                 EtherAssembler.SourceAddress := priv_ReplacementMACAddress;

             // Create Destination MAC

             if SLLParser.PacketType = sptBroadcast then
                EtherAssembler.DestinationAddress := VALUE_BROADCAST_DEST
             else
                 EtherAssembler.DestinationAddress := priv_ReplacementMACAddress;

             if SLLParser.ContainsVLANInfo then
                begin
                     // The original frame had malformed VLAN information, so lets keep it

                     VLANAssembler := t802dot1qAssembler.create;

                     VLANAssembler.Priority := SLLParser.VLANInfo.priority;
                     VLANAssembler.CFI := SLLParser.VLANInfo.cfi;
                     VLANAssembler.VLANID := SLLParser.VLANInfo.vlanid;

                     // Set Ethertypes

                     EtherAssembler.EtherTypeLength := NEXT_HEADER_8021_Q;

                     if priv_ReplaceEtherType then
                        begin
                             if SLLParser.ProtocolType = priv_OriginalEthertype then
                                VLANAssembler.EtherType := priv_ReplacementEthertype
                             else
                                 VLANAssembler.EtherType := SLLParser.ProtocolType;
                        end
                     else
                         VLANAssembler.EtherType := SLLParser.ProtocolType;

                     // Create Chunks

                     EtherChunk := EtherAssembler.GetAssembly;
                     EtherAssembler.free;

                     Dot1QChunk := VLANAssembler.GetAssembly;
                     VLANAssembler.Free;

                     Layer3Chunk := Frame.FrameDataAsChunk (SLLParser.NextAnchor.Offset);

                     // Add Ehternet and Dot1Q

                     HelpChunk := mergeFrameChunks (EtherChunk, Dot1QChunk);

                     EtherChunk.free;
                     Dot1QChunk.free;

                     // Add L2 and L3

                     MergeChunk := mergeFrameChunks (HelpChunk, Layer3Chunk);

                     HelpChunk.free;
                     Layer3Chunk.free;
                end
             else
                 begin
                      // Set Ethertype

                      if priv_ReplaceEtherType then
                         begin
                              if SLLParser.ProtocolType = priv_OriginalEthertype then
                                 EtherAssembler.EtherTypeLength := priv_ReplacementEthertype
                              else
                                  EtherAssembler.EtherTypeLength := SLLParser.ProtocolType;
                         end
                      else
                          EtherAssembler.EtherTypeLength := SLLParser.ProtocolType;

                     // Create Chunks

                     EtherChunk := EtherAssembler.GetAssembly;
                     EtherAssembler.free;

                     Layer3Chunk := Frame.FrameDataAsChunk (SLLParser.NextAnchor.Offset);

                     MergeChunk := mergeFrameChunks (EtherChunk, Layer3Chunk);

                     EtherChunk.free;
                     Layer3Chunk.free;
                 end;


             // Create New Frame

             NewFrame := getRawFrameFromFrameChunk (Frame, MergeChunk, Frame.FrameSize - MergeChunk.Len);

             // Discard Chunk and old Frame

             MergeChunk.free;

             Frame.free;
             Frame := nil;

             // Set frame to modified to force discarding the parsemaster object

             prot_framewasmodified := true;
        end
     else
         NewFrame := Frame;

     // Free up the parsemaster

//   ParseMaster.free;

     result := NewFrame;
end;

(* ------------------------------------------- *)

end.
