{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_frameprocessor_vlan;

interface

uses base_structures,
     class_rawframe,
     class_frameprocessor_baseclass;

type tFrameProcessorVLAN = class (tFrameProcessor)

                                    private
                                           function getRawFrameFromFrameChunk (Frame : tRawFrame; PayLoad : tFrameChunk; FrameLengthModifier : integer) : tRawFrame;

                                    public
                                          /// <summary>
                                          ///   Removes all VLANs from the
                                          ///   frame if found, and moves the
                                          ///   tunneled content to the
                                          ///   beginning of the frame.
                                          /// </summary>
                                          /// <param name="Frame">
                                          ///   Frame to be processed
                                          /// </param>
                                          function getProcessedFrame (Frame : tRawFrame) : tRawFrame; override;

                              end;

implementation

uses SiAuto,
     SysUtils,
     class_parser_baseclass,
     common_parse_helpers,
     base_checksums,
     class_parser_802dot1q,
     class_parser_ethernet,
     class_parser_ipv4,
     class_parser_ipv6,
     class_parsemaster,
     class_assembly_ethernet;

(* ------------------------------------------- *)
{ tFrameProcessorVLAN }
(* ------------------------------------------- *)

function tFrameProcessorVLAN.getRawFrameFromFrameChunk (Frame: tRawFrame; PayLoad: tFrameChunk; FrameLengthModifier : integer): tRawFrame;

var NewFrame     : tRawFrame;
    TimeDetails  : tTimingDetails;
    WireDetails  : tWireDetails;
    FilePosition : tFileFrameLocation;
    FrameBuffer  : tRawBuffer;
    FinalLen     : integer;
    i            : integer;

begin
     TimeDetails.AbsoluteTime := Frame.Timestamp.datetime;
     TimeDetails.TraceTicks := Frame.Ticks;
     TimeDetails.NanoTicks := (Frame.Timestamp.nanoseconds mod 1000);            // because right now, Nanosecs also contain microsecs

     WireDetails.LinkType := Frame.LinkType;
     WireDetails.WireSize := Frame.WireSize - FrameLengthModifier;
     WireDetails.CaptureSize := Payload.Len;
     WireDetails.InterfaceID := Frame.InterfaceID;
     WireDetails.Dropped := Frame.DropCount;
     WireDetails.FrameFlags := Frame.FrameFlags;
     WireDetails.FrameType := Frame.FrameFileType;

     // Transfer Packet direction

//   WireDetails.PacketDirection := pdUnknown;

     FilePosition.FileIndex := 0;
     FilePosition.FileFrameIndex := prot_framesprocessed;
     FilePosition.FileOffset := 0;

     move (Payload.Data^, FrameBuffer, WireDetails.CaptureSize);

     NewFrame := tRawFrame.create (TimeDetails, WireDetails, FilePosition, FrameBuffer);

     for i := 1 to Frame.Commentlist.Count do
         NewFrame.Commentlist.Add(Frame.Commentlist[i-1]);

     result := NewFrame;
end;

(* ------------------------------------------- *)

function tFrameProcessorVLAN.getProcessedFrame (Frame: tRawFrame): tRawFrame;

var
    ParseEntry   : tParseEntry;
    NewFrame     : tRawFrame;
    HeaderType   : tHeaderType;
    EthParser    : tEthernetParser;
    EthAssembler : tEthernetAssembler;
    VLANParser   : t802dot1qParser;
    GenParser    : tFrameParser;
    FirstChunk   : tFrameChunk;
    SecondChunk  : tFrameChunk;
    MergeChunk   : tFrameChunk;
    found        : boolean;
    EtherType    : word;
    FirstVLAN    : integer;
    LastVLAN     : integer;
    FirstOffset  : integer;
    LastOffset   : integer;
    i            : integer;

begin
     if not assigned (prot_ParseMaster) then
        prot_ParseMaster := tParseMaster.create(Frame, ppmParseAll);

     // Find the VLAN tags

     FirstVLAN := 0;
     found := false;

     repeat
           inc (FirstVLAN);
           HeaderType := prot_ParseMaster.getParserTypeByIndex (FirstVLAN);

           if (HeaderType = NEXT_HEADER_8021_Q) or (HeaderType = NEXT_HEADER_802_QINQ_Outer_OLD1) or (HeaderType = NEXT_HEADER_802_QINQ_Outer_OLD2) or (HeaderType = NEXT_HEADER_802_QINQ_Outer_OLD3) then
              found := true;

     until found or (FirstVLAN = prot_ParseMaster.ParserCount);

     if found and (FirstVLAN > 1) then
        begin
             // Make sure we have an Ethernet parser

             if (prot_ParseMaster.getParserTypeByIndex (FirstVLAN - 1) = NEXT_HEADER_ETHERNET) or (prot_ParseMaster.getParserTypeByIndex (FirstVLAN - 1) = NEXT_HEADER_TRANS_ETHER_BRIDGE) then
                begin
                     EthParser := prot_ParseMaster.getParserByIndex (FirstVLAN - 1) as tEthernetParser;

                     EthAssembler := tEthernetAssembler.create;
                     EthAssembler.CopyFromParser (EthParser);
                end
             else
                 found := false;
        end
     else
         found := false;

     if found and (FirstVLAN < prot_ParseMaster.ParserCount) then
        begin
             FirstOffset := 0;
             LastOffset := 0;

             // Find last VLAN

             LastVLAN := FirstVLAN;

             while (prot_ParseMaster.getParserTypeByIndex (LastVLAN + 1) = NEXT_HEADER_8021_Q) or (prot_ParseMaster.getParserTypeByIndex (LastVLAN + 1) = NEXT_HEADER_802_QINQ_Outer_OLD1) or
                   (prot_ParseMaster.getParserTypeByIndex (LastVLAN + 1) = NEXT_HEADER_802_QINQ_Outer_OLD2) or (prot_ParseMaster.getParserTypeByIndex (LastVLAN + 1) = NEXT_HEADER_802_QINQ_Outer_OLD3) do
                   inc (LastVLAN);

             if LastVLAN < prot_ParseMaster.ParserCount then
                begin
                     GenParser := prot_ParseMaster.getParserByIndex (LastVLAN + 1);
                     LastOffset := GenParser.FrameOffset;

                     VLANParser := prot_ParseMaster.getParserByIndex (LastVLAN) as t802dot1qParser;
                     Ethertype := VLANParser.EtherType;
                     EthAssembler.EtherTypeLength := Ethertype;

                     // Build new Ethernet chunk

                     FirstChunk := EthAssembler.GetAssembly;
                     EthAssembler.Free;

                     // Get the part starting right after the MPLS Shims

                     SecondChunk := Frame.FrameDataAsChunk (LastOffset);

                     // Merge Ethernet and IP

                     MergeChunk := mergeFrameChunks (FirstChunk, SecondChunk);

                     FirstChunk.free;
                     SecondChunk.free;

                     if prot_ParseMaster.HasEthernetFCS then
                        MergeChunk.AttachCRC32 (not prot_ParseMaster.EthernetFSCCorrect);

                     // Create New Frame

                     NewFrame := getRawFrameFromFrameChunk (Frame, MergeChunk, Frame.FrameSize - MergeChunk.Len);

                     // Discard Chunk and old Frame

                     MergeChunk.free;

                     Frame.free;
                     Frame := nil;

                     // Set frame to modified to force discarding the parsemaster object

                     prot_framewasmodified := true;
                end
             else
                 begin
                      EthAssembler.Free;
                 end;
        end
     else
         // Give up and pass the frame

         NewFrame := Frame;

     result := NewFrame;
end;

(* ------------------------------------------- *)

end.
