{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_loader_enc;

interface

uses base_structures,
     class_rawframe,
     class_traceloader;

type tENCLoader   = class (tTraceLoader)

                          private
                                 priv_PacketLinkID : tGUID;

                                 function ENCTimetoTicks (wTimeLow : word; wTimeMid : word; bTimeHigh : byte; nTimeDays : byte) : int64;   // This converts the ENC timing info to ticks
                                 function ENCDateToTimeStamp (wDate : word; wTime: word) : tDateTime;                                      // This converts the ENC file header timestamp to a Delphi timestamp
                                 function ENCErrorFlagsToGenericFormat (bENCFlags : byte) : tErrorFlagsSet;                                 // Converts the .ENC style error flags to the generic ones

                          protected
                                   function thisFileFormat : tCaptureFileType; override;
                                   function ParseFrameFromOffset (const FrameIndex : integer; var FileOffset: integer; var DataPointer: pByte): tRawFrame; override;

                          public
                                procedure readTraceFileheader; override;
                                procedure readBatchFromDisk; override;
                                function readSingleFrameFromDisk (const FrameIndex : integer) : tRawFrame; override;
                                function readSingleFrameByOffset(const Offset: integer; FrameIndex : integer = -1) : tRawFrame; override;
                    end;


implementation

uses SysUtils,
     Math,
     DateUtils,
     class_interfacelist;


type // ENC File structures (old Sniffer DOS Format)

     tENCTraceDate = packed record
                           wTime  : word;
                           wDate  : word;
                     end;

     pENC_RecordHeader = ^tENC_RecordHeader;
     tENC_RecordHeader = packed record
                               wRecType      : word;
                               wRecLength    : word;
                               wReserved     : word;
                         end;

     pENC_VersionRecord = ^tENC_VersionRecord;
     tENC_VersionRecord = packed record
                                wMajorVersion       : word;
                                wMinorVersion       : word;
                                rdate               : tENCTraceDate;
                                bRecordType         : Byte;
                                bNetwork            : Byte;
                                bFormat             : Byte;
                                bTimeunit           : Byte;
                                bCompressionVersion : Byte;
                                bCompressionLevel   : Byte;
                                aReserved           : array [1..2] of word;
                          end;

     pENC_FrameHeader = ^tENC_FrameHeader;
     tENC_FrameHeader = packed record
                             wTimeLow     : word;
                             wTimeMid     : word;
                             bTimeHigh    : byte;
                             nTimeDays    : byte;
                             wLength      : word;                      // number of bytes actually captured. If slicing is on, this might be shorter than the frame really was
                             bErrorFlags  : byte;
                             bBufferFlags : byte;
                             wWireSize    : word;
                             wReserved    : word;
                        end;

const ENC_VersionRecord     = 1;
      ENC_EOFRecord         = 3;
      ENC_FrameRecord       = 4;
      ENC_FunnyRecord       = 7;

      // ENC Error Flag Values

      ENC_ErrorFlagNone        = $00;
      ENC_ErrorFlagCollision   = $02;
      ENC_ErrorFlagShortRunt   = $08;
      ENC_ErrorFlagOversize    = $10;
      ENC_ErrorLost            = $20;
      ENC_ErrorAlignment       = $40;
      ENC_ErrorBadCRC          = $80;
      ENC_ErrorFragment        = $88;
      ENC_Jabber               = $90;



(* ------------------------------------------- *)
{ tENCLoader }
(* ------------------------------------------- *)

function tENCLoader.thisFileFormat : tCaptureFileType;

begin
     result := cfENC;
end;

(* ------------------------------------------- *)

function tENCLoader.ENCTimetoTicks(wTimeLow, wTimeMid: word; bTimeHigh, nTimeDays: byte): int64;

var
  temp1 : INT64;
  temp2 : INT64;
  temp3 : Extended;

begin
     // These 3 lines are taken from ethereal:
     // t represents the number of ticks since the frame started
     //    t_low = (guint16)(t-(double)((guint32)(t/65536.0))*65536.0);
     //    t_med = (guint16)((guint32)(t/65536.0) % 65536);
     //    t_high = (guint16)(t/4294967296.0);

     // The Documentation for the DOS-Sniffer gives the following record layout:
     //    INT      time_low;              /* Low time in units           */
     //    INT      time_mid;              /* Mid time in units           */
     //    CHAR     time_high;             /* High time in units          */
     //    CHAR     time_day;              /* Time since start of capture */

     // Obviously, Ethereal throws time_high and time_day together into t_high
     // This might cause some problems for traces that span midnight

     // t_low
     temp1 := wTimeLow;

     // t_med
     temp2  := wTimeMid;
     temp1  := temp1 + 65536 * temp2;

     // t_high
     temp2  := bTimeHigh;
     temp1  := temp1 + 4294967296 * temp2;
     temp3  := temp1 * dTraceTimeUnit;

     // number of days
     // each day has (24*3600) seconds, represented in x TimeUnits
     // basically we convert the number of days to a ton of microseconds ...

     temp2 := nTimeDays;
     temp2 := temp2 * MicroSecondsPerDay;      // the last multiplication by 1 Mln is for the microseconds

     // ... and add this to our result
     temp3 := temp3 + temp2;


    // Sniffer files have a capture start date in the file header, and
    // have times relative to the beginning of that day in the packet
    // headers; pick the date of the first packet as the capture start
    // date.

     result := Round(temp3);
end;

(* ------------------------------------------- *)

function tENCLoader.ENCDateToTimeStamp(wDate, wTime: word): tDateTime;

var deltayear, month, day : word;
    second, minute, hour  : word;

begin
     // wFatDate
     // [in] The MS-DOS date. The date is a packed value with the following format.
     // Bits 	Description
     // 0-4 	Day of the month (131)
     // 5-8 	Month (1 = January, 2 = February, and so on)
     // 9-15 	Year offset from 1980 (add 1980 to get actual year)

     deltayear := (wDate and $FE00) shr 9;
     month := (wDate and $01e0) shr 5;
     day   := (wDate and $1f);

     // wFatTime
     // [in] The MS-DOS time. The time is a packed value with the following format.
     // Bits 	Description
     // 0-4 	Second divided by 2
     // 5-10 	Minute (059)
     // 11-15 	Hour (023 on a 24-hour clock)

     hour := (wTime and $F800) shr 11;
     minute := (wTime and $7e0) shr 5;
     second := (wTime and $1f) * 2;

     result := encodeDateTime (1980 + deltaYear, month, day, hour, minute, second, 0);
end;

(* ------------------------------------------- *)

function tENCLoader.ENCErrorFlagsToGenericFormat (bENCFlags: byte): tErrorFlagsSet;

var value : tErrorFlagsSet;

begin
     value := [];

     case bENCFlags of
      ENC_ErrorFlagCollision : value := value + [efCollision];
      ENC_ErrorFlagShortRunt,
      ENC_ErrorFragment      : value := value + [efPacketTooShort];
      ENC_ErrorFlagOversize  : value := value + [efPacketTooLong];
      ENC_ErrorLost          : value := value + [efUnknown];
      ENC_ErrorAlignment     : value := value + [efAlignment];
      ENC_ErrorBadCRC, $84   : value := value + [efBadCRC];
      ENC_Jabber             : value := value + [efBadPreamble];
     end;

     result := value;
end;

(* ------------------------------------------- *)

procedure tENCLoader.readTraceFileheader;

var i              : integer;
    sHelp          : string;
    pHeader        : pENC_RecordHeader;
    pVersionRecord : pENC_VersionRecord;
    InterfaceEntry : tCaptureInterface;

begin
     dTraceTimeUnit := 0.838096;        // Default value to shut up the compiler
     iBatchFileOffset := 0;             // This keeps track of where we are in the whole mapped file

     bChannelCount := 1;                // ENC has no channels, so its always all-in-one

     sHelp := '';
     pBatchPosition := pByte (pData);

     // Check File Length

     if iFileSize < length (ENC_MagicString) then
        raise EInOutError.Create ('Error: ENC file too short (traceloader).');

     // 1. Check if the Magic String has the correct length and matches

     for i := 0 to (length (ENC_Magicstring) - 1) do
         begin
              sHelp := sHelp + chr (pData^[i]);

              inc (iBatchFileOffset);
              inc (pBatchPosition);
         end;

     if sHelp <> ENC_MagicString then
        begin
             raise EInOutError.Create('Error: Not an ENC file (ENCLoader.readTraceFileheader).');

             exit;
        end;

     // Check File Length

     if iFileSize < (iBatchFileOffset + sizeOf (tENC_RecordHeader)) then
        begin
             raise EInOutError.Create ('Error: ENC file too short (traceloader).');

             exit;
        end;

    // 2. Now we read the header of the first record, which should be the header of the version record

    pHeader := pENC_RecordHeader (pBatchPosition);

    inc (iBatchFileOffset, sizeOf (tENC_RecordHeader));
    inc (pBatchPosition, sizeOf (tENC_RecordHeader));

    if (pHeader^.wRecType <> ENC_VersionRecord) then
       begin
            raise EInOutError.Create('Error: first record is not a version record (ENCLoader.readTraceFileheader).');

            exit;
       end;

    // Next, we read the version record

    pVersionRecord := pENC_VersionRecord (pBatchPosition);

    inc (iBatchFileOffset, pHeader^.wRecLength);
    inc (pBatchPosition, pHeader^.wRecLength);

    // 3. Do some checks on the version record to see if we do unterstand the file at all

    // Check Network Topology

    case pVersionRecord^.bNetwork of
     0  : prot_LinkType := LINKTYPE_TOKEN_RING;
     1  : prot_LinkType := LINKTYPE_ETHERNET;
     2  : prot_LinkType := LINKTYPE_ARCNET;
     3  : prot_LinkType := LINKTYPE_NULL;
     4  : prot_LinkType := LINKTYPE_NULL;
     5  : prot_LinkType := LINKTYPE_NULL;
//   6  : priv_LinkType := ntInternetWorkAnalyzer;
     7  : prot_LinkType := LINKTYPE_NULL;
     8  : prot_LinkType := LINKTYPE_NULL;
     9  : prot_LinkType := LINKTYPE_FDDI;
     10 : prot_LinkType :=  LINKTYPE_ATM_RFC1483;
     else
         prot_LinkType := LINKTYPE_NULL;
    end;

    if not ((prot_LinkType = LINKTYPE_ETHERNET) or (prot_LinkType = LINKTYPE_TOKEN_RING)) then
       begin
            raise EInOutError.Create('Error: capture file link type not supported (ENCLoader.readTraceFileheader).');

            exit;
       end;

    // Check Compression level - we can only read uncompressed files as off today

    if ((pVersionRecord^.bCompressionVersion <> 0) or (pVersionRecord^.bCompressionLevel <> 0)) then
       begin
            raise EInOutError.Create('Error: cannot read ENC file containing compressed data (ENCLoader.readTraceFileheader).');

            exit;
       end;

    // Assign and check Timeunit

    case pVersionRecord^.bTimeunit of
     0 : dTraceTimeUnit := 15.0;
     1 : dTraceTimeUnit := 0.838096;
     2 : dTraceTimeUnit := 15.0;
     3 : dTraceTimeUnit := 0.5;
     4 : dTraceTimeUnit := 2.0;
     5 : dTraceTimeUnit := 1.0;
     6 : dTraceTimeUnit := 0.1;
    else
        begin
             raise EInOutError.Create('Error: Unknown timeunit encountered (ENCLoader.readTraceFileheader).');

             exit;
        end
    end;

    TraceDateTime := ENCDateToTimeStamp (pVersionRecord^.rdate.wDate, pVersionRecord^.rdate.wTime);
//  TraceDateTimeString := dateTimeToStr (TraceDateTime);

    // Create a single interface

    InterfaceEntry := tCaptureInterface.create (0, LinkType, ioFromFile, true);

    InterfaceEntry.Name := 'Unknown';
    InterfaceEntry.Description := 'Generic Interface from ENC File';
    InterfaceEntry.SpeedBPS := 100000000;      // 100 Mbit
    InterfaceEntry.LinkType := prot_LinkType;

    priv_PacketLinkID := InterfaceEntry.PacketLinkID;

    InterfaceList.Add (InterfaceEntry);

    boHeaderRead := true;
end;

(* ------------------------------------------- *)

// Helper function: creates and fills a tRawframe object from the given position
// so we don't have to code this twice for ReadBatch and ReadFromOffset methods

function tENCLoader.ParseFrameFromOffset(const FrameIndex : integer; var FileOffset: integer; var DataPointer: pByte): tRawFrame;

var pFrameHeader  : pENC_FrameHeader;
    pFrameBuffer  : pRawBuffer;
//  pHelp         : pointer;
    helpBuf       : tRawBuffer;
    NewFrame      : tRawFrame;
    DeltaTime     : tDeltaTickTime;
    TimingDetails : tTimingDetails;
    WireDetails   : tWireDetails;
    FilePosition  : tFileFrameLocation;
    ErrorFlags    : tErrorFlagsSet;

begin
     // Check File Length if a record header is available

     if iFileSize < (FileOffset + sizeOf (tENC_FrameHeader)) then
        begin
             raise EInOutError.Create ('Error: unexpected end of file (ENCLoader.parseframefromoffset).');

             exit;
        end;

     // 1. Read Frame Header

     pFrameHeader := pENC_FrameHeader (DataPointer);

     if pFrameHeader^.wLength > maxFrameSize then
        begin
             raise EInOutError.Create ('Error: maximum frame size exceeded (ENCLoader.parseframefromoffset).');
             exit;
        end;

     WireDetails.initialize;
     WireDetails.CaptureSize := pFrameHeader^.wLength;

     if pFrameHeader^.wWireSize <> 0 then
        begin
             boFrameSlicing := true;
             WireDetails.WireSize := pFrameHeader^.wWireSize;
        end
     else
         WireDetails.WireSize := pFrameHeader^.wLength;

     WireDetails.LinkType := prot_LinkType;         // Read from file header and stored in object variable

     ErrorFlags := ENCErrorFlagsToGenericFormat (pFrameHeader^.bErrorFlags);

     WireDetails.FrameFlags.SetErrorFlags (ErrorFlags);
     WireDetails.FrameType := tftEnc;
     WireDetails.FrameFlags.Value := 0;
     WireDetails.InterfaceID := 0;
     WireDetails.Dropped := 0;

     // Assign Header Values to Frame Object

     FilePosition.FileIndex := iFileNumber;                                            // Taken from traceloader object - it knows what index the current file has
     FilePosition.FileFrameIndex := iTotalFramesRead;
     FilePosition.FileOffset := FileOffset;

     TimingDetails.AbsoluteTime := TraceDateTime;
     TimingDetails.TraceTicks := ENCTimetoTicks (pFrameHeader^.wTimeLow, pFrameHeader^.wTimeMid, pFrameHeader^.bTimeHigh, pFrameHeader^.nTimeDays);

     // Move the pointer to the Frame Content Start

     inc (FileOffset, sizeOf (tENC_FrameHeader));
     inc (DataPointer, sizeOf (tENC_FrameHeader));

     // 2. Read Frame Content

     // Check File Length if a frame is available

     if iFileSize < (FileOffset + pFrameHeader^.wLength) then
        begin
             raise EInOutError.Create ('Error: unexpected end of file (ENCLoader.parseframefromoffset).');

             exit;
        end;

     pFrameBuffer := pRawBuffer (DataPointer);

     if WireDetails.CaptureSize > maxFrameSize then
        raise EInOutError.Create ('Error: oversized frame encountered (ENCLoader.parseframefromoffset).');

     // To copy the data to a dynamic memory area we first need to move it to a static
     // array. If we do not do this first and copy from pointer to pointer we will slam
     // into a wall as soon as we try to dispose the memory, because the move procedure
     // seems to align pointer locations instead of actually copying the content when called
     // with two pointer types. Took me 2 hours to figure this problem out. Sigh.

     // We hand over a static buffer to the RawFrame.create which will move it to a dynamic buffer

     move (pFrameBuffer^, helpBuf, min (maxFrameSize, WireDetails.CaptureSize));

     // Determine LinkID

     WireDetails.LinkID := priv_PacketLinkID;

     // Create new Frame Object that will hold the data

     NewFrame := tRawFrame.create (TimingDetails, WireDetails, FilePosition, helpBuf);

     // Give it back

     result := NewFrame;
end;

(* ------------------------------------------- *)

procedure tENCLoader.readBatchFromDisk;

var boReadDone    : boolean;
    pRecHeader    : pENC_RecordHeader;
    NewFrame      : tRawFrame;
    StartTime     : tDateTime;
    iDuration     : integer;

begin
     // First we clear the frames we have in memory. but we keep the frame history

     FrameList_ValidFrames.Clear;

     // Some initialisations

     StartTime := now;

     iFramesAvailable := 0;
     iFramesTossed := 0;

     boReadDone := false;

     if boEOF then
        exit;

     // Being here we should have read the file header and perhaps another batch of frames, so the iOffset needs to
     // be greater than 0

     if iBatchFileOffset = 0 then
        begin
             raise EInOutError.Create('Error: invalid file offset - file header not read? (ENCLoader.readbatchfromdisk).');

             exit;
        end;

     repeat
           // Check File Length if a record header is available

           if iFileSize < (iBatchFileOffset + sizeOf (tENC_RecordHeader)) then
              begin
                   raise EInOutError.Create ('Error: unexpected end of file (ENCLoader.readbatchfromdisk).');

                   exit;
              end;

           // Read Header

           pRecHeader := pENC_RecordHeader (pBatchPosition);

           inc (iBatchFileOffset, sizeOf (tENC_RecordHeader));
           inc (pBatchPosition, sizeOf (tENC_RecordHeader));

           // Check Header Data

           case pRecHeader^.wRecType of
            ENC_EOFRecord     : begin
                                     // This indicates that we read the entire file

                                     boEOF := true;
                                     boReadDone := true;
                                end;

            ENC_FrameRecord   : begin
                                     inc (iTotalFramesRead);
                                     inc (iFramesAvailable);

                                     // 1. Call the parsing function to get the frame

                                     NewFrame := ParseFrameFromOffset (iTotalFramesRead, iBatchFileOffset, pBatchPosition);

                                     // 2. Move offsets to next frame

                                     inc (iBatchFileOffset, NewFrame.FrameSize);
                                     inc (pBatchPosition, NewFrame.FrameSize);

                                     // 3. Add new Frame to our list of Frames

                                     self.addRawFrame (NewFrame);

                                     // 4. Check if we have as many frames as we have to read in one batch

                                     if boFrameMemLimited and (iFramesAvailable >= iMaxFramesinMem) then
                                        boReadDone := true;
                                end;

            ENC_FunnyRecord   : begin
                                     // Skip Funny Frame Data - it usually appears once at the start of an ENC file

                                     inc (iBatchFileOffset, pRecHeader^.wRecLength);
                                     inc (pBatchPosition, pRecHeader^.wRecLength);

                                     inc (iFramesSkipped);
                                end;
           else
               inc (iFramesUnknown);
           end;

     until boReadDone;

     iDuration := SecondsBetween (StartTime, now);

     if iDuration > 0 then
        iFPS := iFramesAvailable Div iDuration
     else
         iFPS := iFramesAvailable;
end;

(* ------------------------------------------- *)

// This routine returns a single frame. Index starts at 1, not 0!

function tENCLoader.readSingleFrameFromDisk(const FrameIndex: integer) : tRawFrame;

var pLocalPos    : pByte;
    iLocalOffset : integer;
    iCount       : integer;
    boReadDone   : boolean;
    pRecHeader   : pENC_RecordHeader;

begin
     result := nil;

     if FrameIndex = 0 then
        raise EInOutError.Create ('Error: invalid frame index of 0 (ENCLoader.readSingleFrame).');

     iCount := 0;
     boReadDone := false;

     // First, we skip the header of course

     iLocalOffset := 41;
     pLocalPos := pByte (pData);
     inc (pLocalPos, iLocalOffset);

     // Now pLocalPosition points to the first frame record header

     repeat
          // Check File Length if a record header is available

           if iFileSize < (iLocalOffset + sizeOf (tENC_RecordHeader)) then
              begin
                   raise EInOutError.Create ('Error: unexpected end of file (ENCLoader.readSingleFrame).');

                   exit;
              end;

           // Read Header

           pRecHeader := pENC_RecordHeader (pLocalPos);

           inc (iLocalOffset, sizeOf (tENC_RecordHeader));
           inc (pLocalPos, sizeOf (tENC_RecordHeader));

           case pRecHeader^.wRecType of
            ENC_EOFRecord     : begin
                                     // This indicates that we read the entire file which can't happen as we're looking for just one frame

                                     if iFileSize < (iLocalOffset + sizeOf (tENC_RecordHeader)) then
                                        begin
                                             raise EInOutError.Create ('Error: unexpected end of file (ENCLoader.readSingleFrame).');

                                             exit;
                                        end;
                                end;

            ENC_FrameRecord   : begin
                                     inc (iCount);

                                     if iCount = FrameIndex then
                                        begin
                                             result := readSingleFrameByOffset (iLocalOffset, FrameIndex);
                                             boReadDone := true;
                                        end
                                     else
                                         begin
                                              // Skip FrameHeader and FrameBuffer at once - the Record Header knows the combined length

                                              inc (iLocalOffset, pRecHeader^.wRecLength);
                                              inc (pLocalPos, pRecHeader^.wRecLength);
                                         end;
                                end;

            ENC_FunnyRecord   : begin
                                     // Skip Funny Frame Data - it usually appears once at the start of an ENC file

                                     inc (iLocalOffset, pRecHeader^.wRecLength);
                                     inc (pLocalPos, pRecHeader^.wRecLength);
                                end;
           end;

     until boReadDone;
end;

(* ------------------------------------------- *)

function tENCLoader.readSingleFrameByOffset(const Offset: integer; FrameIndex : integer = -1) : tRawFrame;

var pLocalPos    : pByte;
    NewFrame     : tRawFrame;
    HelpOffset   : integer;

begin
     HelpOffset := Offset; // Necessary because ParseFrameFromOffset will need to modify it, but it will be discarded afterwards anyway

     // Set the local pointer to the required Offset

     pLocalPos := pByte (pData);
     inc (pLocalPos, Offset);

     // Check File Length if a record header is available - this is NOT optional!

     if iFileSize < (Offset + sizeOf (tENC_FrameHeader)) then
        begin
             raise EInOutError.Create ('Error: unexpected end of file (ENCLoader.readSingleFrameByOffset).');

             exit;
        end;

     // Create new Frame Object that will hold the data

     NewFrame := ParseFrameFromOffset (FrameIndex, HelpOffset, pLocalPos);

     // Add new Frame to our list of Frames

     result := NewFrame;
end;

(* ------------------------------------------- *)

end.
