{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_loader_pcap;

interface

uses base_structures,
     class_rawframe,
     class_traceloader;

type
    PCAPLengthSwapType    = (stNotSwapped, stMaybeSwapped, stSwapped);       // Some PCAP specialty with byte swapping for length indicators

    tPCAPLoader  = class (tTraceLoader)

                          private
                                 wMajorVersion        : word;
                                 wMinorVersion        : word;
                                 lSnapLength          : longword;
                                 lNetworkType         : longword;

                                 boModifiedFileFormat : boolean;
                                 boByteSwappedFile    : boolean;
                                 boAIXFile            : boolean;
                                 LengthsSwapped       : PCAPLengthSwapType;

                                 priv_PacketLinkID    : tGUID;

                          protected
                                   function thisFileFormat : tCaptureFileType; override;
                                   function ParseFrameFromOffset (const FrameIndex : integer; var FileOffset: integer; var DataPointer: pByte): tRawFrame;  override;
                          public
                                procedure readTraceFileheader; override;
                                procedure readBatchFromDisk; override;
                                function readSingleFrameFromDisk (const FrameIndex : integer) : tRawFrame; override;
                                function readSingleFrameByOffset (const Offset : integer; FrameIndex : integer = -1) : tRawFrame; override;
                    end;


implementation

uses SiAuto,
     SysUtils,
     Math,
     DateUtils,
     Toolkit,
     class_interfacelist;

type // LibPCAP File structures (TCPDump files)

     pPCAPFileHeader = ^tPCAPFileHeader;
     tPCAPFileHeader = packed record
                             wMajorVersion   : word;                     // major version number
                             wMinorVersion   : word;                     // minor version number
                             iGMTTimeOffset  : longint;                  // GMT to local correction
                             lSigFix         : longword;                 // Accuracy of timestamps
                             lSnapLength     : longword;                 // Max length of captured packets, in octets
                             lNetworkType    : longword;                 // Datalink type
                       end;

     pPCAPDefaultFrameHeader = ^tPCAPDefaultFrameHeader;
     tPCAPDefaultFrameHeader = packed record
                                     lTimeStampSeconds  : longword;      // timestamp seconds
                                     lTimeStampMicroSec : longword;      // timestamp microseconds (nsecs for PCAP_NSEC_MAGIC)
                                     lOctetCount        : longword;      // number of octets of packet saved in file
                                     lOriginalLength    : longword;      // actual length of packet
                               end;

     pPCAPLongFrameHeader    = ^tPCAPLongFrameHeader;
     tPCAPLongFrameHeader    = packed record                                    // "Alexey's patched version"
                                     lTimeStampSeconds  : longword;      // timestamp seconds
                                     lTimeStampMicroSec : longword;      // timestamp microseconds
                                     lOctetCount        : longword;      // number of octets of packet saved in file
                                     lOriginalLength    : longword;      // actual length of packet
                                     lInterfaceIndex    : longword;      // index, in *capturing* machine's list of interfaces, of the interface on which this packet came in.
                                     wProtocol          : word;          // Ethernet Packet Type
                                     bPaketType         : byte;          // broadcast/multicast/etc. indication
                                     bPadding           : byte;          // Pad to a 4-byte boundary
                               end;



(* ------------------------------------------- *)
{ tPCAPLoader }
(* ------------------------------------------- *)


function tPCAPLoader.thisFileFormat : tCaptureFileType;

begin
     result := cfPCAP;
end;

(* ------------------------------------------- *)

procedure tPCAPLoader.readTraceFileheader;

var pMagicIndex          : pCardinal;
    pFileHeader          : pPCAPFileHeader;
    InterfaceEntry       : tCaptureInterface;

begin
     iBatchFileOffset := 0;                  // This keeps track of where we are in the whole mapped file
     pBatchPosition := pByte (pData);

     bChannelCount := 1;                // pCAP has no channels, so its always all-in-one

     // Check File Length

     if iFileSize < PCAP_MagicStringLength then
        raise EInOutError.Create ('Error: PCAP file too short (PCAPLoader.readTraceFileheader).');

     // 1. Check if the Magic String has the correct length and matches

     pMagicIndex := pCardinal (pBatchPosition);
     inc (pBatchPosition, PCAP_MagicStringLength);
     inc (iBatchFileOffset, PCAP_MagicStringLength);

     case pMagicIndex^ of
      PCAP_MagicStandard      : begin
                                     boModifiedFileFormat := false;
                                     boByteSwappedFile := false;
                                     bPrecision := WTAP_FILE_TSPREC_USEC;

                                     ReportMessage (msInfo, 'PCAP Format found: Standard', 'tPCAPLoader.readTraceFileheader');
//                                   StrList_Messages.Add('PCAP Format found: Standard');
                                end;

      PCAP_MagicStandardSwap  : begin
                                     boModifiedFileFormat := false;
                                     boByteSwappedFile := true;
                                     bPrecision := WTAP_FILE_TSPREC_USEC;

                                     ReportMessage (msInfo, 'PCAP Format found: Standard Byte Swapped', 'tPCAPLoader.readTraceFileheader');
//                                   StrList_Messages.Add('PCAP Format found: Standard Byte Swapped');
                                end;

      PCAP_MagicModified      : begin
                                     boModifiedFileFormat := true;
                                     boByteSwappedFile := false;
                                     bPrecision := WTAP_FILE_TSPREC_USEC;

                                     ReportMessage (msInfo, 'PCAP Format found: Modified Standard', 'tPCAPLoader.readTraceFileheader');
//                                   StrList_Messages.Add('PCAP Format found: Modified Standard');
                                end;

      PCAP_MagicModified_Swap : begin
                                     boModifiedFileFormat := true;
                                     boByteSwappedFile := true;
                                     bPrecision := WTAP_FILE_TSPREC_USEC;

                                     ReportMessage (msInfo, 'PCAP Format found: Modified Byte Swapped', 'tPCAPLoader.readTraceFileheader');
//                                   StrList_Messages.Add('PCAP Format found: Modified Byte Swapped');
                                end;
      PCAP_NSEC_MAGIC         : begin
                                     boModifiedFileFormat := false;
                                     boByteSwappedFile := false;
                                     bPrecision := WTAP_FILE_TSPREC_NSEC;

                                     ReportMessage (msInfo, 'PCAP Format found: NanoSecond Standard', 'tPCAPLoader.readTraceFileheader');
//                                   StrList_Messages.Add('PCAP Format found: NanoSecond Standard');
                                end;
      PCAP_SWAPPED_NSEC_MAGIC : begin
                                     boModifiedFileFormat := false;
                                     boByteSwappedFile := true;
                                     bPrecision := WTAP_FILE_TSPREC_NSEC;

                                     ReportMessage (msInfo, 'PCAP Format found: NanoSecond Byte Swapped', 'tPCAPLoader.readTraceFileheader');
//                                   StrList_Messages.Add('PCAP Format found: NanoSecond Byte Swapped');
                                end
      else
          begin
               raise EInOutError.Create ('Error: PCAP file format unknown (PCAPLoader.readTraceFileheader).');
               exit;
          end;
     end;

     // 2. Read File Header

     // Check File Length

     if iFileSize < iBatchFileOffset + sizeOf (tPCAPFileHeader) then
        raise EInOutError.Create ('Error: PCAP file too short (PCAPLoader.readTraceFileheader).');

     pFileHeader := pPCAPFileHeader (pBatchPosition);
     inc (pBatchPosition, sizeOf (tPCAPFileHeader));
     inc (iBatchFileOffset, sizeOf (tPCAPFileHeader));

     if boByteSwappedFile then
        begin
             wMajorVersion := Swap16 (pFileHeader^.wMajorVersion);
             wMinorVersion := Swap16 (pFileHeader^.wMinorVersion);
             lSnapLength   := Swap32 (pFileHeader^.lSnapLength);
             lNetworkType  := Swap32 (pFileHeader^.lNetworkType);
        end
     else
         begin
             wMajorVersion := pFileHeader^.wMajorVersion;
             wMinorVersion := pFileHeader^.wMinorVersion;
             lSnapLength   := pFileHeader^.lSnapLength;
             lNetworkType  := pFileHeader^.lNetworkType;
         end;

     // Check Version

     if wMajorVersion < 2 then
        begin
             raise EInOutError.Create ('Error: Unsupported PCAP version (PCAPLoader.readTraceFileheader).');
             exit;
        end
     else
         begin
              // Check for AIX dump and network type

              if (wMajorVersion = 2) and (wMinorVersion = 2) then
                 begin
                      boAIXFile := true;
                      bPrecision := WTAP_FILE_TSPREC_NSEC;

                      case lNetworkType of
                        6  : prot_LinkType := LINKTYPE_ETHERNET;       // DLT_EN10MB, Ethernet
                        7  : prot_LinkType := LINKTYPE_RAW;            // RAW IP
                        9  : prot_LinkType := LINKTYPE_TOKEN_RING;     // DLT_IEEE802, Token Ring
                       15  : prot_LinkType := LINKTYPE_FDDI;           // DLT_FDDI, FDDI
                      else
                          prot_LinkType := LINKTYPE_NULL;
                      end;
                 end
              else
                  case lNetworkType of
                    1 : prot_LinkType := LINKTYPE_ETHERNET;            // DLT_EN10MB, Ethernet
                    6 : prot_LinkType := LINKTYPE_TOKEN_RING;          // DLT_IEEE802, Token Ring
                    7 : prot_LinkType := LINKTYPE_ARCNET;
                    8 : prot_LinkType := LINKTYPE_SLIP_BSDOS;
                    9 : prot_LinkType := LINKTYPE_PPP_ETHER;
                   10 : prot_LinkType := LINKTYPE_FDDI;
                   12,
                  101 : prot_LinkType := LINKTYPE_RAW;                 // RAW IP
                  113 : prot_LinkType := LINKTYPE_LINUX_SLL;
                  178 : prot_LinkType := LINKTYPE_JUNIPER;
                  else
                     prot_LinkType := LINKTYPE_NULL;              // DLT_NULL, loopback
                  end;
         end;

     // Test for length swap stuff

     case wMajorVersion of
        2 : begin
                 if wMinorVersion < 3 then
                    LengthsSwapped := stSwapped
                 else
                     if wMinorVersion = 3 then
                        LengthsSwapped := stMaybeSwapped
                     else
                         LengthsSwapped := stNotSwapped;
            end;
      543 : begin
                 // hey guys, you need to get your heads fixed... 543 as a MajorVersion? What kind of weed did you guys smoke?!

                 LengthsSwapped := stSwapped;
            end;
      else
          LengthsSwapped := stNotSwapped
     end;
{
     if (LinkType <> LINKTYPE_ETHERNET) and (LinkType <> LINKTYPE_TOKEN_RING) then
        begin
             raise EInOutError.Create ('Error: Unsupported network topology (PCAPLoader.readTraceFileheader).');
             exit;
        end;
}
     if iFileSize <= iBatchFileOffset then
        boEOF := true;

     // Time Stuff

     dTraceTimeUnit := 1.0;

     // Create a single interface

     InterfaceEntry := tCaptureInterface.create (0, LinkType, ioFromFile, true);

     InterfaceEntry.Name := 'Unknown';
     InterfaceEntry.Description := 'Generic Interface from PCAP File';
     InterfaceEntry.SpeedBPS := 100000000;      // 100 Mbit
     InterfaceEntry.LinkType := prot_LinkType;

     priv_PacketLinkID := InterfaceEntry.PacketLinkID;

     InterfaceList.Add (InterfaceEntry);

     boHeaderRead := true;
end;

(* ------------------------------------------- *)

function tPCAPLoader.ParseFrameFromOffset(const FrameIndex: integer; var FileOffset: integer; var DataPointer: pByte): tRawFrame;

var pFrameHeader  : pPCAPDefaultFrameHeader;
    pLongHeader   : pPCAPLongFrameHeader;
    pFrameBuffer  : pRawBuffer;
    helpBuf       : tRawBuffer;
    NewFrame      : tRawFrame;
    DeltaTime     : tDeltaTickTime;
    HelpDate      : tDateTime;
    HelpTicks     : int64;
    TimingDetails : tTimingDetails;
    WireDetails   : tWireDetails;
    FilePosition  : tFileFrameLocation;
    i64Seconds    : int64;

begin
     // Assign File position details

     FilePosition.FileIndex := iFileNumber;
     FilePosition.FileFrameIndex := FrameIndex;
     FilePosition.FileOffset := FileOffset;

     WireDetails.initialize;
     WireDetails.LinkType := prot_LinkType;         // Read from file header and stored in object variable

     // Assign Header Values to Frame Object

     if boModifiedFileFormat then
        begin
             pLongHeader := pPCAPLongFrameHeader (DataPointer);
             inc (DataPointer, sizeOf (tPCAPLongFrameHeader));
             inc (FileOffset, sizeOf (tPCAPLongFrameHeader));

             if boByteSwappedFile then
                begin
                     // If UTCDelta is 0 we calculate the systems offset to UTC and add it, otherwise we use what the user said

                     if iUTCDelta = 0 then
                        begin
                             helpDate := UnixToDateTime (Swap32 (pFrameHeader^.lTimeStampSeconds));

                             i64Seconds := Swap32 (pFrameHeader^.lTimeStampSeconds) + (TimeZoneHandler.GetBiasForDate (helpDate) DIV 60 * -1) * 3600
                        end
                     else
                         i64Seconds := Swap32 (pFrameHeader^.lTimeStampSeconds) + (iUTCDelta * 3600);

                     WireDetails.CaptureSize := Swap32 (pLongHeader^.lOctetCount);        // Capturesize tells what we captured
                     WireDetails.WireSize := Swap32 (pLongHeader^.lOriginalLength);       // Framesize tells the size on the wire
                     TimingDetails.TraceTicks := Swap32 (i64Seconds * 1000 * 1000) + Swap32 (pLongHeader^.lTimeStampMicroSec);
                     TimingDetails.AbsoluteTime := UnixToDateTime (Swap32 (pLongHeader^.lTimeStampSeconds));
                end
             else
                 begin
                      // If UTCDelta is 0 we calculate the systems offset to UTC and add it, otherwise we use what the user said

                      if iUTCDelta = 0 then
                         begin
                              helpDate := UnixToDateTime (pFrameHeader^.lTimeStampSeconds);

                              i64Seconds := pFrameHeader^.lTimeStampSeconds + (TimeZoneHandler.GetBiasForDate (helpDate) DIV 60 * -1) * 3600
                         end
                      else
                          i64Seconds := pFrameHeader^.lTimeStampSeconds + (iUTCDelta * 3600);

                      WireDetails.CaptureSize := pLongHeader^.lOctetCount;                 // Capturesize tells what we captured
                      WireDetails.WireSize := pLongHeader^.lOriginalLength;              // Framesize tells the size on the wire
                      TimingDetails.TraceTicks := (i64Seconds * 1000 * 1000) + Swap32 (pLongHeader^.lTimeStampMicroSec);
                      TimingDetails.AbsoluteTime := UnixToDateTime (pLongHeader^.lTimeStampSeconds);
                 end;
        end
     else
         begin
              pFrameHeader := pPCAPDefaultFrameHeader (DataPointer);
              inc (DataPointer, sizeOf (tPCAPDefaultFrameHeader));
              inc (FileOffset, sizeOf (tPCAPDefaultFrameHeader));

              // Assign Header Values to Frame Object

              if boByteSwappedFile then
                 begin
                      // If UTCDelta is 0 we calculate the systems offset to UTC and add it, otherwise we use what the user said

                      if iUTCDelta = 0 then
                         begin
                              helpDate := UnixToDateTime (Swap32 (pFrameHeader^.lTimeStampSeconds));

                              i64Seconds := Swap32 (pFrameHeader^.lTimeStampSeconds) + (TimeZoneHandler.GetBiasForDate (helpDate) DIV 60 * -1) * 3600
                         end
                      else
                          i64Seconds := Swap32 (pFrameHeader^.lTimeStampSeconds) + (iUTCDelta * 3600);

                      WireDetails.CaptureSize := Swap32 (pFrameHeader^.lOctetCount);        // lOctetCount tells what we captured
                      WireDetails.WireSize := Swap32 (pFrameHeader^.lOriginalLength);     // lOriginalLength tells the size on the wire
                      TimingDetails.TraceTicks := (i64Seconds * 1000 * 1000) + Swap32 (pFrameHeader^.lTimeStampMicroSec);
                      TimingDetails.AbsoluteTime := UnixToDateTime (i64Seconds);
                 end
              else
                  begin
                       // If UTCDelta is 0 we calculate the systems offset to UTC and add it, otherwise we use what the user said

                       if iUTCDelta = 0 then
                          begin
                               helpDate := UnixToDateTime (pFrameHeader^.lTimeStampSeconds);

                               i64Seconds := pFrameHeader^.lTimeStampSeconds + (TimeZoneHandler.GetBiasForDate (helpDate) DIV 60 * -1) * 3600
                          end
                       else
                           i64Seconds := pFrameHeader^.lTimeStampSeconds + (iUTCDelta * 3600);

                       WireDetails.CaptureSize := pFrameHeader^.lOctetCount;                 // lOctetCount tells what we captured
                       WireDetails.WireSize := pFrameHeader^.lOriginalLength;              // lOriginalLength tells the size on the wire
                       TimingDetails.TraceTicks := (i64Seconds * 1000 * 1000) + pFrameHeader^.lTimeStampMicroSec;
                       TimingDetails.AbsoluteTime := UnixToDateTime (i64Seconds);
                  end;
         end;

     // Fix load errors if detected and correction is allowed

     if prot_fixloaderrors then
        begin
             // Capture size cannot exceed wiresize

             if WireDetails.CaptureSize > WireDetails.WireSize then
                begin
                     ReportMessage (msMajor, 'Fixing bad Capture size of ' + IntToStr (WireDetails.CaptureSize) + ' bytes larger than wire size (' + IntTostr (WireDetails.WireSize) + ' bytes) detected for frame #' + IntToStr (FrameIndex), 'tPCAPLoader.ParseFrameFromOffset', FrameIndex);

                     WireDetails.CaptureSize := WireDetails.WireSize;
                end;
        end;

     // Recalculate ticks to only include microseconds since midnight, not the whole daycount, too
     // So we take the absolute TraceDateTime and get Midnight of that day. Convert to ticks. Substract.

     HelpDate := recodeTime (TimingDetails.AbsoluteTime, 0, 0, 0, 0);
     HelpTicks := DateTimeToUnix (HelpDate) * 1000 * 1000;

     TimingDetails.TraceTicks := TimingDetails.TraceTicks - HelpTicks;

     // .pcap files cannot contain multiple channels

     WireDetails.FrameType := tftPCAP;
     WireDetails.InterfaceID := 0;
     WireDetails.FrameFlags.Value := 0;
     WireDetails.Dropped := 0;

     // Determine LinkID

     WireDetails.LinkID := priv_PacketLinkID;

     // Create Buffer

     pFrameBuffer := pRawBuffer (DataPointer);

     if WireDetails.CaptureSize > maxFrameSize then
        begin
             ReportMessage (msCritical, 'capture size exceeds maximum in frame #' + IntToStr (FrameIndex) + ', aborting', 'tPCAPLoader.ParseFrameFromOffset', FrameIndex);

             exit (nil);
        end;
//      raise EInOutError.Create ('Error: oversized frame encountered (PCAPLoader.readbatchfromdisk). Aborting.');

     // To copy the data to a dynamic memory area we first need to move it to a static
     // array. If we do not do this first and copy from pointer to pointer we will slam
     // into a wall as soon as we try to dispose the memory, because the move procedure
     // seems to align pointer locations instead of actually copying the content when called
     // with two pointer types. Took me 2 hours to figure this problem out. Sigh.

     move (pFrameBuffer^, helpBuf, min (maxFrameSize, WireDetails.CaptureSize));

     // Create new Frame Object that will hold the data

     NewFrame := tRawFrame.create (TimingDetails, WireDetails, FilePosition, helpBuf);

     // Give it back

     result := NewFrame;
end;

(* ------------------------------------------- *)

procedure tPCAPLoader.readBatchFromDisk;

var StartTime    : tDateTime;
    boReadDone   : boolean;
    pFrameBuffer : pRawBuffer;
    newFrame     : tRawFrame;
    iDuration    : integer;
    i64Seconds   : int64;
    pHelp        : pointer;
    helpBuf      : tRawBuffer;

begin
     if not boHeaderRead then
        begin
             raise EInOutError.Create('Error: PCAP file header not read before batch load! (PCAPLoader.readBatchFromDisk).');
             exit;
        end;

     // First we clear the frames we have in memory

     FrameList_ValidFrames.Clear;

     // Some initialisations

     StartTime := now;

     iFramesAvailable := 0;
     iFramesTossed := 0;

     boReadDone := false;

     if boEOF then
        exit;

     // Being here we should have read the file header and perhaps another batch of frames, so the iOffset needs to
     // be greater than 0

     if iBatchFileOffset = 0 then
        begin
             raise EInOutError.Create('Error: invalid file offset - PCAP file header not read? (PCAPLoader.readBatchFromDisk).');
             exit;
        end;

     if iBatchFileOffset < iFileSize then
        begin
             // Main Loop

             repeat
                   // Read Header

                   if boModifiedFileFormat then
                      begin
                           // Check File Length if a record header is available

                           if iFileSize < (iBatchFileOffset + sizeOf (tPCAPLongFrameHeader)) then
                              begin
                                   raise EInOutError.Create ('Error: unexpected end of PCAP file (PCAPLoader.readBatchFromDisk).');
                                   exit;
                              end;
                      end
                   else
                       begin
                            // Check File Length if a record header is available

                            if iFileSize < (iBatchFileOffset + sizeOf (tPCAPDefaultFrameHeader)) then
                               begin
                                    raise EInOutError.Create ('Error: unexpected end of PCAP file (PCAPLoader.readBatchFromDisk).');
                                    exit;
                               end;
                       end;

                   try
                      NewFrame := ParseFrameFromOffset (iTotalFramesRead + 1, iBatchFileOffset, pBatchPosition);
                   except
                         SiMain.LogError ('Error while reading frame #' + (iTotalFramesRead + 1).toString + ' from PCAP at offset ' + pointtext (iBatchFileOffset));
                         Exit;
                   end;

                   if NewFrame <> nil then
                      begin
                           inc (iFramesAvailable);
                           inc (iTotalFramesRead);

                           // Create new Frame Object that will hold the data

                           self.addRawFrame (NewFrame);

                           inc (iBatchFileOffset, NewFrame.FrameSize);
                           inc (pBatchPosition, NewFrame.FrameSize);
                      end
                   else
                       begin
                            boReadDone := true;
                            boEof := true;
                            boReadFailed := true;
                            LoaderStatus := lsReadFailed;
                       end;

                   // 4. Check if we have as many frames as we have to read in one batch

                   if (boFrameMemLimited and (iFramesAvailable >= iMaxFramesinMem)) then
                      boReadDone := true;

                   if (iBatchFileOffset >= iFileSize) then
                      begin
                           boReadDone := true;
                           boEOF := true;
                      end;

             until boReadDone;

             iDuration := SecondsBetween (StartTime, now);

             if iDuration > 0 then
                iFPS := iFramesAvailable Div iDuration
             else
                 iFPS := iFramesAvailable;
        end;
end;

(* ------------------------------------------- *)

function tPCAPLoader.readSingleFrameFromDisk (const FrameIndex: integer) : tRawFrame;


var pLocalPos    : pByte;
    iLocalOffset : integer;
    pFrameHeader : pPCAPDefaultFrameHeader;
    pLongHeader  : pPCAPLongFrameHeader;
    pFrameBuffer : pRawBuffer;
    newFrame     : tRawFrame;
    i64Seconds   : int64;
    iCount       : integer;
    SkipBytes    : integer;
    boReadDone   : boolean;

begin
     result := nil;

     if not boHeaderRead then
        begin
             raise EInOutError.Create('Error: PCAP file header not read before load! (PCAPLoader.readSingleFrameByOffset).');

             exit;
        end;

     if FrameIndex = 0 then
        raise EInOutError.Create ('Error: invalid frame index of 0 (PCAPLoader.readSingleFrameFromDisk).');

     iCount := 0;
     boReadDone := false;

     // First we need to skip over the file header to start counting the frames until
     // we find the one we want

     iLocalOffset := PCAP_MagicStringLength + sizeOf (tPCAPFileHeader);
     pLocalPos := pByte (pData);
     inc (pLocalPos, iLocalOffset);

     if iLocalOffset < iFileSize then
        begin
             // Main Loop

             repeat
                   inc (iCount);

                   if iCount = FrameIndex then
                      begin
                           result := readSingleFrameByOffset (iLocalOffset, FrameIndex);
                           boReadDone := true;
                      end
                   else
                       begin
                            if boModifiedFileFormat then
                               begin
                                    if iFileSize < (iLocalOffset + sizeOf (tPCAPLongFrameHeader)) then
                                       begin
                                            raise EInOutError.Create ('Error: unexpected end of PCAP file (PCAPLoader.readBatchFromDisk).');
                                            exit;
                                       end;

                                    // Determine how far we need to skip ahead to get past the current frame

                                    pLongHeader := pPCAPLongFrameHeader (pLocalPos);

                                    if boByteSwappedFile then
                                       SkipBytes := sizeOf (tPCAPLongFrameHeader) + Swap32 (pLongHeader^.lOctetCount)
                                    else
                                        SkipBytes := sizeOf (tPCAPLongFrameHeader) + pLongHeader^.lOctetCount;

                               end
                            else
                                begin
                                     if iFileSize < (iLocalOffset + sizeOf (tPCAPDefaultFrameHeader)) then
                                        begin
                                             raise EInOutError.Create ('Error: unexpected end of PCAP file (PCAPLoader.readBatchFromDisk).');
                                             exit;
                                        end;

                                     // Determine how far we need to skip ahead to get past the current frame

                                     pFrameHeader := pPCAPDefaultFrameHeader (pLocalPos);

                                     if boByteSwappedFile then
                                        SkipBytes := sizeOf (tPCAPDefaultFrameHeader) + Swap32 (pFrameHeader^.lOctetCount)
                                     else
                                         SkipBytes := sizeOf (tPCAPDefaultFrameHeader) + pFrameHeader^.lOctetCount;
                                end;

                            inc (pLocalPos, SkipBytes);
                            inc (iLocalOffset, SkipBytes);

                            if iLocalOffset > iFileSize then
                               boReadDone := true;
                       end;

             until boReadDone;
        end;
end;

(* ------------------------------------------- *)

function tPCAPLoader.readSingleFrameByOffset(const Offset: integer; FrameIndex : integer = -1) : tRawFrame;

var pLocalPos    : pByte;
    NewFrame     : tRawFrame;
    HelpOffset   : integer;

begin
     HelpOffset := OffSet;   // Necessary because ParseFrameFromOffset will need to modify it, but it will be discarded afterwards anyway

     // Set the local pointer to the required Offset

     pLocalPos := pByte (pData);
     inc (pLocalPos, Offset);

     // Check File Length if a record header is available

     if boModifiedFileFormat then
        begin
             if iFileSize < (iBatchFileOffset + sizeOf (tPCAPLongFrameHeader)) then
                begin
                     raise EInOutError.Create ('Error: unexpected end of PCAP file (PCAPLoader.readBatchFromDisk).');
                     exit;
                end;
        end
     else
         begin
              if iFileSize < (iBatchFileOffset + sizeOf (tPCAPDefaultFrameHeader)) then
                 begin
                      raise EInOutError.Create ('Error: unexpected end of PCAP file (PCAPLoader.readBatchFromDisk).');
                      exit;
                 end;
         end;

     // Create new Frame Object that will hold the data

     NewFrame := ParseFrameFromOffset (FrameIndex, HelpOffset, pLocalPos);

     // Add new Frame to our list of Frames

     result := NewFrame;
end;

(* ------------------------------------------- *)

end.
