{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_tracefilelist;

interface

uses Contnrs,
     Classes,
     Generics.Collections,
     VirtualTrees,       // needed to have a node reference property for each file

     // FireDAC

     FireDAC.Stan.Intf, FireDAC.Stan.Option, FireDAC.Stan.Error, FireDAC.UI.Intf,
     FireDAC.Phys.Intf, FireDAC.Stan.Def, FireDAC.Stan.Pool, FireDAC.dAPT,
     FireDAC.Stan.Async, FireDAC.Phys, FireDAC.Stan.ExprFuncs,
     FireDAC.VCLUI.Wait, FireDAC.Comp.UI, FireDAC.Phys.SQLite,
     Data.DB, FireDAC.Comp.Client,

     // Remaining units

     toolkit,
     base_structures,
     class_rawframe,
     pcapng_structures,
     class_diagmessagehandler,
     class_traceloader,
     class_interfacelist,
     class_conversationstorage,
     class_inteldb_manager;

type tOnScanMessage = procedure (Sender : tObject; MessageText : string; out Continue : boolean) of object;
     tOnFullTraceScan = procedure (Sender : tObject; const Filename : string; out Continue : boolean) of object;
     tOnStartScanOfFile = procedure (Sender : tObject; const Status : tProcessingStatus) of object;
     tOnEndScanOfFile = procedure (Sender : tObject; const Status : tProcessingStatus) of object;
     tOnAdvancePercentProcessed = procedure (Sender : tObject; const PercentageParsed, FramesPerSecond : integer; var AbortScan : boolean) of object;
     tOnFileStatusChange = procedure (Sender : tObject; const Status : tProcessingStatus) of object;

     tTracefileEntry  = class (tObject)

                              /// <summary>
                              ///   Creates the a new TraceFile object. Note:
                              ///   IntelDB params are set via property
                              /// </summary>
                              /// <param name="NewFilename">
                              ///   Name of the new file
                              /// </param>
                              /// <param name="AutoScanThreshold">
                              ///   Size Threshold above which a file isn't
                              ///   automatically scanned
                              /// </param>
                              constructor create (const NewFilename : string; const AutoScanThreshold : int64 = 150000000); overload;
                              constructor create (const NewFilename : string; const IntelDBParams : tIntelDBParameters; const AutoScanThreshold : int64 = 150000000); overload;
                              procedure free;

                              private
                                     priv_name                       : string;
                                     priv_tracetype                  : tCaptureFileType;
                                     priv_filesize                   : int64;
                                     priv_filecomment                : string;
                                     priv_fileattributes             : tFileAttributes;
                                     priv_filetimestamp              : tDateTime;
                                     priv_captureduration            : string;
                                     priv_framecount                 : uint64;
                                     priv_channelcount               : integer;
                                     priv_blockchainroot             : pBlockChainEntry;
                                     priv_interfacelist              : tCaptureInterfaceList;
                                     priv_EssentialPacketMode        : boolean;

                                     priv_scanthreshold              : int64;
                                     priv_generalstats               : tGeneralTraceStats;

                                     priv_percentagescanned          : integer;

                                     priv_pcapngstatistics           : tPCAPNGStatistic;
                                     priv_pktcommentlist             : tPCAPNGPacketCommentList;

                                     priv_vstdisplaynode             : pVirtualNode;

                                     priv_inteldbparams              : tIntelDBParameters;
                                     priv_inteldbfilename            : string;
                                     priv_inteldbindex               : integer;
                                     priv_nomeminteldb               : boolean;

                                     priv_abortScan                  : boolean;

                                     // Status flags

                                     priv_processingstatus           : tProcessingStatus;
                                     priv_packetstatsread            : boolean;
                                     priv_pcapngstructread           : boolean;
                                     priv_conversationsscanned       : boolean;
                                     priv_scanduration               : integer;

//                                   priv_intelDBManager             : tIntelDBManager;
                                     priv_conversationstore          : tConversationStorage;
                                     priv_complexity                 : integer;
                                     priv_loadermsgcount             : integer;
                                     priv_loadermessages             : tLoaderMessageList;

                                     EventOnScanMessage              : tOnScanMessage;
                                     EventOnFullScan                 : tOnFullTraceScan;
                                     EventOnAdvancePercentProcessed  : tOnAdvancePercentProcessed;
                                     EventOnStartScanOfFile          : tOnStartScanOfFile;
                                     EventOnEndScanOfFile            : tOnEndScanOfFile;
                                     EventOnAdvancePercentScanned    : tOnAdvancePercentProcessed;
                                     EventOnFileStatusChange         : tOnFileStatusChange;
                                     EventonDiagnosticsMessages      : tEventOnDiagnosticMessages;

                                     function getFileName : string;
                                     function getFilePath : string;
                                     function getTraceType : tCaptureFileType;
                                     function getFilesizeAndTimestamp : int64;
                                     function getFirstFrameTimestamp : tFrameTimeStamp;
                                     {$REGION 'Documentation'}
                                     /// <summary>
                                     ///   Returns the timestamp of the last
                                     ///   frame in the file by scanning the
                                     ///   whole file until the last frame is
                                     ///   reached, unless the scan threshold
                                     ///   prevents it
                                     /// </summary>
                                     /// <returns>
                                     ///   timestamp of the last frame found at
                                     ///   the end of a full file scan
                                     /// </returns>
                                     {$ENDREGION}
                                     function getLastFrameTimestamp : tFrameTimeStamp;
                                     {$REGION 'Documentation'}
                                     /// <summary>
                                     ///   Forces reading the last packet from
                                     ///   PCAPng file, beginning at the end of
                                     ///   the file and scanning backwards.
                                     ///   This can only be done for PCAPng
                                     ///   files.
                                     /// </summary>
                                     /// <returns>
                                     ///   Timestamp of the last frame in the
                                     ///   file
                                     /// </returns>
                                     {$ENDREGION}
                                     function readLastFrameTimestampFromPCAPngFile : tFrameTimeStamp;

                                     procedure CalculateTraceDuration;

                                     procedure readPCAPngStatisticsFromTracefile;
//                                   procedure AddFrameStatistics (Frame : tRawFrame);               now in tracefile_utils
                                     procedure AddToConversationStorage (Frame : tRawFrame);
                                     procedure AddFrameCommentToList (Frame : tRawFrame);
                                     procedure ReportScanProgress (const FrameOffset : int64; const FramesPerSecond : integer);

                                     procedure VerifyAndConnectIntelDB;
                                     procedure DisconnectIntelDB;

                                     {$REGION 'Documentation'}
                                     /// <summary>
                                     ///   Reads the PCAPng file statistics
                                     ///   either from trace file or from
                                     ///   inteldb file if it has already been
                                     ///   scanned before
                                     /// </summary>
                                     /// <param name="AutoScanThreshold">
                                     ///   File size threshold which prevents
                                     ///   doing deep scans if the file size is
                                     ///   large than the threshold. Specified
                                     ///   in MB.
                                     /// </param>
                                     {$ENDREGION}
                                     procedure getPCAPngStatistics (AutoScanThreshold : int64);

                                     procedure setProcessingStatus (Value : tProcessingStatus);
                                     function getProcessingStatus : tProcessingStatus;

                                     procedure ReadFullFrameDetails;

                                     function IsEssentialPacket (const Frame : tRawFrame) : boolean;

                                     procedure onParserHasDiagnostics (Sender : tObject; Messagelist : tDiagnosticMsgList);

                              protected

                              public
                                    property FilePath : string read getFilePath;
                                    property FileName : string read getFileName;
                                    property FileDate : tDatetime read priv_filetimestamp;
                                    property FileComment : string read priv_filecomment;
                                    property FileAttributes : tFileAttributes read priv_fileattributes;
                                    property FullName : string read priv_name;
                                    property TracefileType : tCaptureFileType read getTraceType;
                                    property FileSize : int64 read getFilesizeAndTimestamp;

                                    property FrameCount : uint64 read priv_framecount;
                                    property Channelcount : integer read priv_channelcount;
                                    property BlockChainRoot : pBlockChainEntry read priv_blockchainroot;
                                    property CaptureInterfaces : tCaptureInterfaceList read priv_interfacelist;
                                    property PacketCommentList : tPCAPNGPacketCommentList read priv_pktcommentlist;

                                    property FirstFrameTimestamp : tFrameTimeStamp read getFirstFrameTimestamp;
                                    property LastFrameTimestamp : tFrameTimeStamp read getLastFrameTimestamp;
                                    property CaptureDuration : string read priv_captureduration;
                                    property EssentialPacketSaveMode : boolean read priv_EssentialPacketMode write priv_EssentialPacketMode;

                                    // Scan status and statistics

                                    property ProcessingStatus : tProcessingStatus read getProcessingStatus write setProcessingStatus;
                                    property PacketStatisticsRead : boolean read priv_packetstatsread;
                                    property PCAPngStructRead : boolean read priv_pcapngstructread;
                                    property ConversationsScanned : boolean read priv_conversationsscanned;
                                    property PCAPngStatistics : tPCAPNGStatistic read priv_pcapngstatistics;
                                    property GeneralStatistics : tGeneralTraceStats read priv_generalstats;
                                    property ScanDuration : integer read priv_scanduration;

                                    // IntelDB

                                    property IntelDBParameters : tIntelDBParameters read priv_inteldbparams write priv_inteldbparams;
                                    property IntelDBFileIndex : integer read priv_inteldbindex;
                                    property IntelDBFileName : string read priv_inteldbfilename;

                                    // Conversation Store

                                    property ConversationStore : tConversationStorage read priv_conversationstore;
                                    property OverallComplexity : integer read priv_complexity;

                                    // Settings

                                    property DoNotCacheIntelDBInMemory : boolean read priv_nomeminteldb write priv_nomeminteldb;

                                    // External status connections

                                    property AssociatedTreeNode : PVirtualNode read priv_vstdisplaynode write priv_vstdisplaynode;

                                    property MessageCount : integer read priv_loadermsgcount;
                                    property MessageList : tLoaderMessageList read priv_loadermessages;

                                    property OnScanMessage : tOnScanMessage read EventOnScanMessage write EventOnScanMessage;
                                    property OnFullTraceScan : tOnFullTraceScan read EventOnFullScan write EventOnFullScan;
                                    property OnAdvancePercentageProcessed : tOnAdvancePercentProcessed read EventOnAdvancePercentProcessed write EventOnAdvancePercentProcessed;
                                    property OnStartScanOfFile : tOnStartScanOfFile read EventOnStartScanOfFile write EventOnStartScanOfFile;
                                    property OnEndScanOfFile : tOnEndScanOfFile read EventOnEndScanOfFile write EventOnEndScanOfFile;
                                    property OnAdvancePercentageScanned : tOnAdvancePercentProcessed read EventOnAdvancePercentScanned write EventOnAdvancePercentScanned;
                                    property OnFileStatusChange : tOnFileStatusChange read EventOnFileStatusChange write EventOnFileStatusChange;
                                    property OnDiagnosticsMessager : tEventOnDiagnosticMessages read EventonDiagnosticsMessages write EventonDiagnosticsMessages;

                                    /// <summary>
                                    ///   Reads file details from a capture
                                    ///   file
                                    ///   <list type="number">
                                    ///     <item>
                                    ///       filesize and timestamp on
                                    ///       disk
                                    ///     </item>
                                    ///     <item>
                                    ///       type of capture file
                                    ///     </item>
                                    ///     <item>
                                    ///       reads full file frame details
                                    ///       from IntelDB or from file if
                                    ///       not yet scanned (and scan is
                                    ///       required)
                                    ///     </item>
                                    ///   </list>
                                    /// </summary>
                                    /// <param name="UsePersistentDB">
                                    ///   can be used to force the IntelDB to
                                    ///   behave in a persistent mode to stop
                                    ///   opening and closing it again and
                                    ///   again if it is just the same DB file.
                                    ///   Usually used when adding files to the
                                    ///   file list
                                    /// </param>
                                    procedure ReadFileDetails (const UsePersistentDB : boolean = False);
                                    procedure ForceReadFileDetails;
                                    procedure ForceReadPCAPngDetails;

                                    function SetFiletimeToFirstFrameTime : integer;
                                    function SetFiletimeToLastFrameTime : integer;
                                    function determineRenamedFilename (const Params : tRenameParams) : string;
                                    function rename (const Params : tRenameParams) : boolean;
                                    function getHintText : string;

                                    function IntelDBVersionIsWrong : boolean;
                                    {$REGION 'Documentation'}
                                    /// <summary>
                                    ///   Returns the index of the current file
                                    ///   in the TraceIntel.DB file table, if
                                    ///   using the IntelDB is allowed and the
                                    ///   file is found
                                    /// </summary>
                                    /// <returns>
                                    ///   File Index as found in the IntelDB,
                                    ///   otherwise 0
                                    /// </returns>
                                    {$ENDREGION}
                                    function GetDBFileIndex : integer;
                        end;

     tTracefileList   = class (tObject)

                              constructor create;
                              procedure free;

                              private
                                     priv_filelist           : tObjectList;
                                     priv_inteldb            : TFDConnection;
                                     priv_dirlist            : tDictionary<string, boolean>;         // Keeps track of directory being writeable

                                     priv_renamecounter      : tCounterStruct;

                                     priv_firstframelocation : tFileFrameLocation;
                                     priv_lastframelocation  : tFileFrameLocation;

                                     priv_encfilecount       : integer;
                                     priv_capfilecount       : integer;
                                     priv_pcapfilecount      : integer;
                                     priv_pcapngfilecount    : integer;

                                     priv_alldirswriteable   : boolean;

                                     function getFileCount : integer;

                              protected

                              public
                                    property Count : integer read getFileCount;
                                    property FirstFrameLocation : tFileFrameLocation read priv_firstframelocation;
                                    property LastFrameLocation : tFileFrameLocation read priv_lastframelocation;
                                    property AllDirectoriesAreWriteable : boolean read priv_alldirswriteable;
                                    property ENCFileCount : integer read priv_encfilecount;
                                    property CAPFileCount : integer read priv_capfilecount;
                                    property PCAPFileCount : integer read priv_pcapfilecount;
                                    property PCAPngFileCount : integer read priv_pcapngfilecount;

                                    /// <summary>
                                    ///   Adds a file to the file list and
                                    ///   increases the counters per capture
                                    ///   file type. After that, the first and
                                    ///   last frame locators are updated.
                                    /// </summary>
                                    /// <param name="NewFile">
                                    ///   New file to be added to the file list
                                    /// </param>
                                    procedure addFile (NewFile : tTracefileEntry);
                                    procedure removeFile (index : integer);
                                    function getFile (index : integer) : tTracefileEntry;
                                    procedure changeListposition (Index1, Index2 : integer);
                                    procedure sort (SortType : tTracefileSortType);
                                    /// <summary>
                                    ///   Determines the file and index of the
                                    ///   first and last frame over all files
                                    ///   in the list. The first frame is of
                                    ///   course file 1 frame 1, while the last
                                    ///   may or may not be known
                                    /// </summary>
                                    procedure updateFirstAndLastFrameLocators;
                                    function GetFileNameList : tStringlist;
                                    procedure clear;

                                    destructor Destroy; override;
                        end;

implementation

uses SysUtils,
     DateUtils,
     Math,
     Types,
     Diagnostics,
     SiAuto,
     tracefile_utils,
     class_loader_enc,
     class_loader_naicap,
     class_loader_pcap,
     class_loader_pcapng,
     class_tracewriter,
     class_writer_pcapng,
     class_parsemaster,
     definition_tcp_structures;


(* ------------------------------------------- *)
{ tTracefileEntry }
(* ------------------------------------------- *)

constructor tTracefileEntry.create (const NewFilename : string; const AutoScanThreshold : int64 = 150000000);

begin
     inherited create;

     priv_inteldbparams.initialize;
     priv_name := NewFilename;
     priv_tracetype := cfNone;
     priv_filesize := 0;
     priv_fileattributes := [];
     priv_channelcount := 0;
     priv_blockchainroot := nil;
     priv_EssentialPacketMode := false;

     self.ProcessingStatus := psNewFile;
     priv_packetstatsread := false;
     priv_pcapngstructread := false;
     priv_conversationsscanned := false;

     priv_vstdisplaynode := nil;

     priv_percentagescanned := 0;
     priv_scanduration := 0;

     // Init statistics

     priv_framecount := 0;
     priv_generalstats.initialize;

     // Preset IntelDB values

     priv_inteldbfilename := ExtractFilePath (priv_name) + LocalTraceIntelFilename;
     priv_inteldbindex := 0;

     // Define scan threshold and create the interface list

     priv_scanthreshold := AutoScanThreshold;
     priv_interfacelist := tCaptureInterfaceList.create(true);

     // Create conversation store and assign diagnostics

     priv_conversationstore := tConversationStorage.create (priv_name);
     priv_conversationstore.OnDiagnosticMessageList := onParserHasDiagnostics;

     // Set IntelDB Mode to use mem caching by default

     priv_nomeminteldb := false;

     priv_filecomment := '';
     setLength (priv_pktcommentlist, 0);

     priv_loadermsgcount := 0;
     setLength (priv_loadermessages, priv_loadermsgcount);

//   ReadFileDetails (priv_scanthreshold);    deprecated because it triggers to early before Event handlers could be assigned
end;

(* ------------------------------------------- *)

constructor tTracefileEntry.create(const NewFilename: string; const IntelDBParams: tIntelDBParameters; const AutoScanThreshold: int64);

var filepath  : string;
    filelist  : tStringlist;
    descript  : tStringlist;
    filemap   : string;
    descrfile : string;
    hash      : string;

begin
     create (NewFilename, AutoScanThreshold);

     filepath := ExtractFilePath (priv_name);

     priv_inteldbparams := IntelDBParams;

     if priv_inteldbparams.UseLocalDBFiles or not directoryexists (priv_IntelDBParams.GlobalDBPath) then
        priv_inteldbfilename := filepath + LocalTraceIntelFilename
     else
         begin
              hash := lowercase (getMD5ForString (filepath));
              priv_inteldbfilename := IncludeTrailingPathDelimiter (priv_IntelDBParams.GlobalDBPath) + hash + '.db';

              filelist := tStringlist.create;
              descript := tStringlist.create;

              filemap := IncludeTrailingPathDelimiter (priv_IntelDBParams.GlobalDBPath) + 'pathmapping.txt';
              descrfile := IncludeTrailingPathDelimiter (priv_IntelDBParams.GlobalDBPath) + 'descript.ion';

              if fileexists (filemap) then
                 filelist.LoadFromFile (filemap);

              if fileexists (descrfile) then
                 filelist.LoadFromFile (descrfile);

              // Add the new path if we don't know it yet

              if filelist.IndexOfName (filepath) = -1 then
                 begin
                      filelist.Add (filepath + '=' + hash);

                      try
                         filelist.SaveToFile (filemap);
                      finally
                             filelist.free;
                      end;

                      // Save the Total Commander description file

                      descript.add (hash + '.db ' + 'TraceIntel Database for directory path "' + filepath + '"');

                      FileSetAttr (descrfile, faNormal);

                      try
                         descript.SaveToFile (descrfile);
                      finally
                             descript.free;
                      end;

                      FileSetAttr (descrfile, faHidden);
                 end
              else
                  begin
                       filelist.Free;
                       descript.Free;
                  end;
         end;
end;

(* ------------------------------------------- *)

procedure tTracefileEntry.free;

begin
     setLength (priv_pktcommentlist, 0);
     priv_interfacelist.free;

     priv_conversationstore.free;

     inherited free;
end;

(* ------------------------------------------- *)

function tTracefileEntry.GetDBFileIndex: integer;

var value        : integer;
    SharedDBMode : boolean;

begin
{$ifdef debug}
     SharedDBMode := true;
{$else}
     SharedDBMode := false;
{$endif}

     value := 0;

     if not priv_IntelDBParams.DoNotUseIntelDB then
        begin
             VerifyAndConnectIntelDB;

             assert (assigned (priv_inteldbparams.IntelDB));
             value := priv_inteldbparams.IntelDB.CurrentFileIndex;

             if not priv_inteldbparams.UsePersistentIntelDB then
                DisconnectIntelDB;
        end;

     result := value;
end;

(* ------------------------------------------- *)

procedure tTracefileEntry.onParserHasDiagnostics (Sender: tObject; Messagelist: tDiagnosticMsgList);

begin
     // Forward the messagelist, but replace the sender with this TracefileEntry

     if assigned (EventonDiagnosticsMessages) then
        EventonDiagnosticsMessages (self, Messagelist);
end;

(* ------------------------------------------- *)

function tTracefileEntry.getFileName: string;

begin
     result := extractFilename (priv_name);
end;

(* ------------------------------------------- *)

function tTracefileEntry.getFilePath: string;

begin
     result := extractFilePath (priv_name);
end;

(* ------------------------------------------- *)

function tTracefileEntry.getFilesizeAndTimestamp: int64;

var sr    : TSearchRec;

begin
     // If we already know the file size, return it. Otherwise determine it.

     if priv_filesize = 0 then
        begin
             if FileExists (priv_name) then
                begin
                     if FindFirst(priv_name, faAnyFile, sr ) = 0 then
                        begin
                             priv_filesize := Int64 (sr.FindData.nFileSizeHigh) shl Int64(32) + Int64(sr.FindData.nFileSizeLow);
                             priv_filetimestamp := sr.TimeStamp;

                             // Get File Attributes while were at it

                             if (sr.Attr and faReadOnly)>0 then
                                priv_fileattributes := priv_fileattributes + [atrReadOnly];

                             if (sr.Attr and faHidden)>0 then
                                priv_fileattributes := priv_fileattributes +  [atrHidden];

                             if (sr.Attr and faSysFile)>0 then
                                priv_fileattributes := priv_fileattributes + [atrSysFile];

                             if (sr.Attr and faVolumeId)>0 then
                                priv_fileattributes := priv_fileattributes + [atrVolumeId];

                             if (sr.Attr and faDirectory)>0 then
                                priv_fileattributes := priv_fileattributes + [atrDirectory];

                             if (sr.Attr and faArchive)>0 then
                                priv_fileattributes := priv_fileattributes + [atrArchive];

                             if (sr.Attr and faAnyFile)>0 then
                                priv_fileattributes := priv_fileattributes + [atrAnyFile];
                        end
                     else
                         begin
                              priv_filesize := -1;
                              priv_filetimestamp := 0;
                         end;

                     FindClose(sr) ;
                end;
        end;

     result := priv_filesize;
end;

(* ------------------------------------------- *)

// Gets the trace file type it is not already determined

function tTracefileEntry.getTraceType : tCaptureFileType;

begin
     priv_tracetype := tracefile_utils.getTraceFileType (priv_name);

     result := priv_tracetype;
end;

(* ------------------------------------------- *)

// Gets the time of the first frame when it is not already determined

function tTracefileEntry.getFirstFrameTimestamp : tFrameTimeStamp;

var testloader : tTraceLoader;
    boUnknown  : boolean;
    firstframe : tRawFrame;
    i          : integer;

begin
     if (priv_generalstats.FirstFrameTime.datetime = 0) then
        begin
             SiMain.LogMessage ('Determining timestamp of first frame in file');

             if fileexists (priv_name) then
                begin
                     boUnknown := false;

                     // See what type of file we have

                     case TracefileType of
                      cfENC    : testloader := tENCLoader.create (priv_name);
                      cfCAP    : testloader := tCAPLoader.create (priv_name);
                      cfPCAP   : testloader := tPCAPLoader.create (priv_name);
                      cfPCAPng : testloader := tPCAPNGLoader.create (priv_name);
                     else
                         boUnknown := true;
                     end;

                     if not boUnknown then
                        begin
                             testloader.openFile;
                             testloader.readTraceFileheader;
                             firstFrame := testloader.readSingleFrameFromDisk(1);

                             if firstframe <> nil then
                                begin
                                     // Report the Date/Time the trace file header tells us

                                     priv_generalstats.FirstFrameTime := firstframe.Timestamp;
                                     priv_generalstats.FirstFrameOffset := firstframe.FileOffset;
                                end
                             else
                                 begin
                                      priv_generalstats.FirstFrameTime.datetime := 0;
                                      priv_generalstats.FirstFrameTime.nanoseconds := 0;
                                      priv_generalstats.FirstFrameOffset := 0;
                                 end;

                             // While we're at it, lets get the interface list and other things, too

                             case TraceFileType of
                              cfPCAPng : begin
                                              testloader.CaptureInterfaces.copyTo (priv_interfacelist);

                                              priv_filecomment := '';

                                              for i := 1 to testloader.HeaderComments.Count do
                                                  begin
                                                       if i < testloader.HeaderComments.Count then
                                                          priv_filecomment := priv_filecomment + testloader.HeaderComments[i-1] + CRLF
                                                       else
                                                           priv_filecomment := priv_filecomment + testloader.HeaderComments[i-1];
                                                  end;
                                         end;
                              cfENC,
                              cfCAP,
                              cfPCAP   : begin
                                              testloader.CaptureInterfaces.copyTo (priv_interfacelist);
                                         end;
                                else
                                    begin
                                         // Add single pseudo interface

                                         priv_interfacelist.AddPseudoInterface (LINKTYPE_ETHERNET);

                                         // No comment

                                         priv_filecomment := '';
                                    end;
                             end;

                             firstFrame.free;

                             // Close the file

                             testloader.closeFile;
                             testloader.free;
                        end;
                end;
        end;

     result := priv_generalstats.FirstFrameTime;
end;

(* ------------------------------------------- *)

procedure tTracefileEntry.ReadFullFrameDetails;

var testloader     : tTraceLoader;
    eWriter        : tPCAPNGWriter;
    eFilename      : string;
    TraceType      : tCaptureFileType;
    boUnknown      : boolean;
    currentframe   : tRawFrame;
    essentialFrame : tRawFrame;
    FpS            : integer;
//  FileIndex      : integer;
    i              : integer;
    help           : integer;
    ScanStart      : tDateTime;
    StopWatch      : tStopWatch;
    BatchSize      : integer;
    SharedDBMode   : boolean;

begin
     SiMain.EnterMethod('tTracefileEntry.ReadFullFrameDetails');
{$ifdef debug}
     SharedDBMode := true;
{$else}
     SharedDBMode := false;
{$endif}

     if fileexists (priv_name) then
        begin
             boUnknown := false;
             ScanStart := now;

             // See what type of file we have

             TraceType := getTraceType;

             case TraceType of
              cfENC    : testloader := tENCLoader.create (priv_name);
              cfCAP    : testloader := tCAPLoader.create (priv_name);
              cfPCAP   : testloader := tPCAPLoader.create (priv_name);
              cfPCAPng : testloader := tPCAPNGLoader.create (priv_name);
             else
                 boUnknown := true;
             end;

             if not boUnknown then
                begin
                     StopWatch := tStopWatch.Create;

                     priv_abortScan := false;

                     // Signal the start of the scan process for GUI setup

                     if assigned (EventOnStartScanOfFile) then
                        EventOnStartScanOfFile (self, psScanningBasics);

                     // Init statistics (not all, because that would also void firstframetime)

                     priv_framecount := 0;
                     priv_generalstats.FrameCount := 0;
                     priv_generalstats.TotalFrameBytes := 0;
                     priv_generalstats.MinFrameSize := 100000;
                     priv_generalstats.MaxFrameSize := 0;
                     priv_generalstats.AverageFrameSize := 0;
                     priv_generalstats.TotalWireBytes := 0;
                     priv_generalstats.isSliced := false;
                     priv_generalstats.SliceSizeConstant := true;
                     priv_generalstats.SliceSize := 0;
                     priv_generalstats.TimeOrderCorrect := true;
                     priv_generalstats.OutOfTimeFrames := 0;
                     priv_generalstats.LastFrameTime.initialize;

                     // Do the scan

                     testloader.openFile;
                     testloader.MaxFramesInMem := 10000;
                     testloader.readTraceFileheader;

                     // If essential packets are to be extracted, prepare the writer

                     if priv_EssentialPacketMode then
                        begin
                             eFilename := extractFilename (priv_name);
                             eFilename := extractFilePath (priv_name) + changeFileExt (eFilename, '_essential.pcapng');

                             eWriter := tPCAPNGWriter.create (eFilename);

                             try
                                eWriter.Open;

                                case TracefileType of
                                 cfENC   : eWriter.WriteGenericHeader (testloader.LinkType);
                                 cfCAP   : ewriter.WriteGenericHeader ((testloader as tCAPLoader).CaptureInterfaces);
                                 cfPCAP  : eWriter.WriteGenericHeader (testloader.LinkType);
                                 cfPCAPng: eWriter.CopyPCAPngHeaderDetailsFromLoader (testloader as tPCAPngLoader);
                                end;

                             except
                                   priv_EssentialPacketMode := false;                // disable on error
                             end;
                        end;

                     // Tell the Conversation storage that we opened a new file

                     priv_inteldbindex := GetDBFileIndex;
                     priv_conversationstore.CurrentFileIndex := priv_inteldbindex;

                     // Now we load all frames until we find the last one

                     fps := 0;

                     SiMain.LogDebug ('Start reading frames to build ConversationStorage');

                     StopWatch.Reset;
                     StopWatch.Start;

                     repeat
                           testloader.readBatchFromDisk;
                           BatchSize := testloader.FramesAvailable;

                           priv_framecount := priv_framecount + BatchSize;

                           i := 0;

//                         for i := 1 to testloader.FramesAvailable do
                           while (i < BatchSize) and not priv_abortscan do
                                 begin
                                      inc (i);

                                      currentframe := testloader.getRawFrame(i);

                                      // Do frame statistics

                                      AddFrameStatistics (priv_generalstats, currentFrame);

                                      // Do endpoint and conversation scan

                                      try
                                         AddToConversationStorage (currentFrame);
                                      except
                                            SiMain.LogError ('Adding frame #' + currentFrame.FileFrameID.ToString + ' to ConversationStorage failed with exception');
                                      end;

                                      if currentframe.Commentlist.Count > 0 then
                                         AddFrameCommentToList (currentframe);
{
                                      // record the timestamp and offset of the lastest frame

                                      priv_generalstats.LastFrameTime := currentframe.Timestamp;
                                      priv_generalstats.LastFrameOffset := currentframe.FileOffset;
}
                                      // Do display stuff

                                      ReportScanProgress (currentframe.FileOffset, FpS);

                                      // If in "Essential Packet save mode", check it

                                      if priv_EssentialPacketMode then
                                         begin
                                              if IsEssentialPacket (currentframe) then
                                                 begin
                                                      essentialFrame := tRawFrame.create (currentframe);
                                                      essentialFrame.InterfaceLinkID := currentframe.InterfaceLinkID;
                                                      eWriter.addFrameToList (essentialframe);

                                                 end;
                                         end;

                                      // Discard the frame

                                      currentframe.free;
                                 end;

                           if StopWatch.Elapsed.Seconds > 0 then
                              FpS := (priv_framecount div StopWatch.Elapsed.Seconds)
                           else
                               FpS := 0;

                     until testloader.EOF or priv_abortscan;

                     StopWatch.Stop;

                     SiMain.LogDebug ('Done reading frames to build ConversationStorage');

                     // Transfer messages
                     // TODO: Store to IntelDB

                     if testloader.MessageCount > 0 then
                        AddMessagesToList (testloader.MessageList, testloader.MessageCount, priv_loadermessages);

                     testloader.closeFile;
                     testloader.free;

                     if priv_EssentialPacketMode then
                        begin
                             eWriter.Flush;
                             eWriter.Close;
                             eWriter.free;
                        end;

                     if not priv_abortscan then
                        begin
                             // Note that we've done it all

                             priv_packetstatsread := true;
                             priv_conversationsscanned := true;

                             // Finalize Statistics

                             priv_generalstats.FrameCount := priv_framecount;
{
                             if priv_framecount > 0 then
                                priv_generalstats.AverageFrameSize := priv_generalstats.TotalFrameBytes div priv_framecount
                             else
                                 priv_generalstats.AverageFrameSize := 0;
}
                             FinalizeFrameStatistics (priv_generalstats);

                             CalculateTraceDuration;

                             // Store to database

//                           if priv_useinteldb then
                             if not priv_IntelDBParams.DoNotUseIntelDB then
                                begin
                                     VerifyAndConnectIntelDB;

                                     // Assign statistics, which will write them to the database

                                     priv_inteldbparams.IntelDB.GeneralStatistics := priv_generalstats;

                                     if length (priv_pktcommentlist) > 0 then
                                        priv_inteldbparams.IntelDB.WriteFrameComments (priv_pktcommentlist);

                                     // Store Conversation Scans

                                     priv_conversationstore.StoreConversationsToDatabase (priv_inteldbparams.IntelDB.Database, priv_name, priv_filesize, priv_filetimestamp, priv_generalstats.FirstFrameTime, priv_generalstats.LastFrameTime);

                                     // Set the database to dirty, because the conversation store accessed a table directly and can't set it

                                     priv_inteldbparams.IntelDB.DatabaseDirty := true;

                                     // Disconnect database if not persistent

                                     if not priv_inteldbparams.UsePersistentIntelDB then
                                        DisconnectIntelDB
                                     else
                                         begin
                                              // Write to disk as a precaution against crashes

                                              priv_inteldbparams.IntelDB.WriteMemDBtoDisk;

                                             if priv_inteldbparams.HideIntelDBFile then
                                                FileSetAttr (priv_inteldbfilename, faHidden);
                                         end;
                                end;

                             priv_scanduration := secondsbetween (now, ScanStart);

                             // Signal scan is complete

                             if assigned (EventOnEndScanOfFile) then
                                EventOnEndScanOfFile (self, psScanningBasics);
                        end
                     else
                         begin
                              // Handle abort of scan

                              priv_processingstatus := psScanAborted;
                         end;
                end;
        end;

     SiMain.LeaveMethod('tTracefileEntry.ReadFullFrameDetails');
end;

(* ------------------------------------------- *)

procedure tTracefileEntry.ForceReadFileDetails;

begin
     ReadFullFrameDetails;
end;

(* ------------------------------------------- *)

procedure tTracefileEntry.ForceReadPCAPngDetails;

begin
     getPCAPngStatistics (0);
end;

(* ------------------------------------------- *)

function tTracefileEntry.getLastFrameTimestamp : tFrameTimeStamp;

begin
      if (priv_generalstats.LastFrameTime.datetime = 0) and (priv_filesize < priv_scanthreshold) then
         ReadFullFrameDetails;

     // Determine trace duration now that we know the end time

     if priv_generalstats.LastFrameTime.datetime <> 0 then
        CalculateTraceDuration;

     // return the result

     result := priv_generalstats.LastFrameTime;
end;

(* ------------------------------------------- *)

function tTracefileEntry.readLastFrameTimestampFromPCAPngFile : tFrameTimeStamp;

var testloader : tPCAPNGLoader;
    lastframe  : tRawFrame;
    lasttime   : tFrameTimeStamp;

begin
     testloader := tPCAPNGLoader.create (priv_name);
     testloader.openFile;
     testloader.readTraceFileheader;
     lastFrame := testloader.readLastFrameFromDisk;
     testloader.closeFile;
     testloader.free;

     if lastframe <> nil then
        begin
             lasttime := lastframe.Timestamp;
             lastframe.free;
        end
     else
         lasttime.initialize;

     result := lasttime;
end;

(* ------------------------------------------- *)
{
procedure tTracefileEntry.AddFrameStatistics (Frame: tRawFrame);

begin
     if Frame.WireSize < priv_generalstats.MinFrameSize then
        priv_generalstats.MinFrameSize := Frame.WireSize;

     if Frame.WireSize > priv_generalstats.MaxFrameSize then
        priv_generalstats.MaxFrameSize := Frame.WireSize;

     if Frame.WireSize > Frame.FrameSize then
        begin
             if priv_generalstats.isSliced then
                begin
                     // Check if we have different slice sizes

                     if priv_generalstats.SliceSize <> Frame.FrameSize then
                        begin
                             priv_generalstats.SliceSizeConstant := false;
                             priv_generalstats.SliceSize := 0;
                        end;
                end
             else
                 begin
                      priv_generalstats.isSliced := true;
                      priv_generalstats.SliceSizeConstant := true;
                      priv_generalstats.SliceSize := frame.FrameSize;
                 end;
        end
     else
         begin
              // Check if we have a full frame with a large size than the slice size we know

              if priv_generalstats.isSliced and (frame.WireSize > priv_generalstats.SliceSize) then
                 priv_generalstats.SliceSizeConstant := false;
         end;

     if (Frame.FileFrameID > 1) and (Frame.Timestamp < priv_generalstats.LastFrameTime) then
        begin
             priv_generalstats.TimeOrderCorrect := false;

             inc (priv_generalstats.OutOfTimeFrames);
        end;

     priv_generalstats.TotalWireBytes := priv_generalstats.TotalWireBytes + Frame.WireSize;
     priv_generalstats.TotalFrameBytes := priv_generalstats.TotalFrameBytes + Frame.FrameSize;
end;
}
(* ------------------------------------------- *)

procedure tTracefileEntry.AddToConversationStorage (Frame: tRawFrame);

begin
     priv_conversationstore.AddFrame (Frame);

     priv_complexity := priv_conversationstore.OverallComplexity;
end;

(* ------------------------------------------- *)

procedure tTracefileEntry.AddFrameCommentToList(Frame: tRawFrame);

var len : integer;
    i   : integer;

begin
     len := length (priv_pktcommentlist);

     setlength (priv_pktcommentlist, len + Frame.Commentlist.Count);

     for i := 0 to Frame.Commentlist.Count - 1 do
         begin
              priv_pktcommentlist[len + i].FrameNo := Frame.FileFrameID;
              priv_pktcommentlist[len + i].Comment := Frame.Commentlist[i];
         end;
end;

(* ------------------------------------------- *)

procedure tTracefileEntry.CalculateTraceDuration;

begin
     priv_captureduration := CaptureDurationAsString (priv_generalstats.LastFrameTime, priv_generalstats.FirstFrameTime);
end;

(* ------------------------------------------- *)

function tTracefileEntry.IsEssentialPacket (const Frame: tRawFrame): boolean;

var ParseMaster : tParseMaster;
    value       : boolean;

begin
     value := false;

     // Parse up to layer 4

     ParseMaster := tParseMaster.create (Frame, ppmUpToLayer4);

     // Check for TCP SYN/FIN/RST flags

     if (ppTCP in ParseMaster.ParsedProtocols) then
        value := (tcpSYN in ParseMaster.TCPFlags) or
                 (tcpFIN in ParseMaster.TCPFlags) or
                 (tcpRST in ParseMaster.TCPFlags);

     // Check for DNS

     if Not value and (ppUDP in ParseMaster.ParsedProtocols) then
        value := (ParseMaster.IPv4SocketPair.Source.Port = 53) or (ParseMaster.IPv4SocketPair.Destination.Port = 53) or
                 (ParseMaster.IPv6SocketPair.Source.Port = 53) or (ParseMaster.IPv6SocketPair.Destination.Port = 53);

     // Check for ICMP

     if (ppICMPv4 in ParseMaster.ParsedProtocols) or (ppICMPv6 in ParseMaster.ParsedProtocols) then
        value := true;

     ParseMaster.free;

     result := value;
end;

(* ------------------------------------------- *)

function tTracefileEntry.SetFiletimeToFirstFrameTime : integer;

var timestamp : tFrameTimeStamp;
    rcode     : integer;

begin
     timestamp := getFirstFrameTimestamp;

     if timestamp.datetime <> 0 then
        begin
             rcode := FileSetDate (priv_name, DateTimeToFileDate(timestamp.datetime));
        end
     else
         rcode := NoFrameTimeAvailable;

     result := rcode;
end;

(* ------------------------------------------- *)

function tTracefileEntry.SetFiletimeToLastFrameTime: integer;

var timestamp : tFrameTimeStamp;
    rcode     : integer;

begin
     timestamp := getLastFrameTimestamp;

     if timestamp.datetime <> 0 then
        begin
             rcode := FileSetDate (priv_name, DateTimeToFileDate(timestamp.datetime));
        end
     else
         rcode := NoFrameTimeAvailable;


     result := rcode;
end;

(* ------------------------------------------- *)

function tTracefileEntry.determineRenamedFilename (const Params : tRenameParams) : string;

var timestamp : tFrametimeStamp;
    path      : string;
    name      : string;
    ext       : string;
    countstr  : string;
    value     : string;
    timestr   : string;

begin
     // gather the details we need

     path := ExtractFilePath (priv_name);

     name := ExtractFileName (priv_name);
     name := changeFileExt (name, '');

     ext := ExtractFileExt (priv_name);
     ext := copy (ext, 2, length (ext) - 1);

     timestamp.initialize;

     case Params.TimestampMode of
      tsmOriginalFile : timestamp.datetime := priv_filetimestamp;
      tsmFirstFrame   : timestamp := getFirstFrameTimestamp;
      tsmLastFrame    : timestamp := getLastFrameTimestamp;
     end;

     // Build timestamp

     case Params.TimestampPatternIndex of
      0 : DateTimeToString (timestr, 'yyyyddmmhhnnss', timestamp.datetime);
      1 : DateTimeToString (timestr, 'yyyymmddhhnnss', timestamp.datetime);
     end;

     // Counter

     countstr := Format ('%.' + IntToStr (Params.Counter.Digits) + 'd', [Params.Counter.Current]);

     // Replace according to pattern

     value := Params.FilePattern;

//   value := StringReplace (value, '<path>', path, [rfReplaceAll, rfIgnoreCase]);                  disabled; this wreaks havoc on IntelDB changes
     value := StringReplace (value, '<filename>', name, [rfReplaceAll, rfIgnoreCase]);
     value := StringReplace (value, '<extension>', ext, [rfReplaceAll, rfIgnoreCase]);
     value := StringReplace (value, '<counter>', countstr, [rfReplaceAll, rfIgnoreCase]);
     value := StringReplace (value, '<timestamp>', TimeStr, [rfReplaceAll, rfIgnoreCase]);

     result := value;
end;

(* ------------------------------------------- *)

function tTracefileEntry.rename (const Params: tRenameParams): boolean;

var newName      : string;
    fullNewName  : string;
    path         : string;
    boOkay       : boolean;
//  SharedDBMode : boolean;

begin
     newName := self.determineRenamedFilename (Params);
     path := self.getFilePath;

     fullNewName := path + newName;

     // Rename if the target name does not already exist

     if not FileExists (fullNewName) then
        begin
             boOkay := RenameFile (self.priv_name, fullNewName);

             if boOkay and not priv_IntelDBParams.DoNotUseIntelDB then
                begin
                     // Need to adjust intelDB

                     VerifyAndConnectIntelDB;

                     boOkay := priv_inteldbparams.IntelDB.rename (newName);

                     if not priv_inteldbparams.UsePersistentIntelDB then
                        DisconnectIntelDB;

                     // Adjust entry

                     self.priv_name := FullNewName;
                end;
        end
     else
         boOkay := False;

     result := boOkay;
end;

(* ------------------------------------------- *)

procedure tTracefileEntry.setProcessingStatus (Value: tProcessingStatus);

begin
//   SiMain.LogDebug('Setting fileentry processing status of file "' + priv_name + '" to "' + getStatusDescription (Value) + '"');

     priv_processingstatus := Value;

     if assigned (EventOnFileStatusChange) then
        EventOnFileStatusChange (self, priv_processingstatus);
end;

(* ------------------------------------------- *)

function tTracefileEntry.getProcessingStatus: tProcessingStatus;

begin
//   SiMain.LogDebug('Retrieving fileentry processing status of file "' + priv_name + '", which is "' + getStatusDescription (priv_processingstatus) + '"');

     result := priv_processingstatus;
end;

(* ------------------------------------------- *)

procedure tTracefileEntry.readPCAPngStatisticsFromTracefile;

var OpenResult     : boolean;
    FtPCAPNGLoader : tPCAPNGLoader;

begin
     FtPCAPNGLoader := tPCAPNGLoader.Create (priv_name);

     if FtPCAPNGLoader.openFile then
        begin
             FtPCAPNGLoader.readTraceFileheader;
             priv_pcapngstatistics := FtPCAPNGLoader.readFileStatistics;
        end;

     FtPCAPNGLoader.closeFile;
     FtPCAPNGLoader.Free;
     FtPCAPNGLoader := nil;
end;

(* ------------------------------------------- *)

procedure tTracefileEntry.ReportScanProgress (const FrameOffset: int64; const FramesPerSecond : integer);

var percentagescanned : integer;
    iHelp             : integer;
    dHelp             : double;
    boAbort           : boolean;

begin
     boAbort := false;

     if (priv_filesize > 0) and assigned (EventOnAdvancePercentScanned) then
        begin
             dHelp := 100 * frameoffset / priv_filesize;
             iHelp := trunc (dHelp);

             if iHelp > priv_percentagescanned then
                begin
                     priv_percentagescanned := iHelp;

                     EventOnAdvancePercentScanned (self, priv_percentagescanned, FramesPerSecond, boAbort);

                     if boAbort then
                        priv_abortScan := true;
                end;
        end;
end;

(* ------------------------------------------- *)

procedure tTracefileEntry.VerifyAndConnectIntelDB;

var OpenDBFile   : boolean;
    SharedDBMode : boolean;

begin
     SiMain.EnterMethod ('tTracefileEntry.VerifyAndConnectIntelDB');

{$ifdef debug}
     SharedDBMode := true;
{$else}
     SharedDBMode := false;
{$endif}

     if priv_IntelDBParams.UsePersistentIntelDB then
        begin
             SiMain.LogDebug ('IntelDB is in persistent mode');

             // Check if there is an open DB, and if so, if it's the right one

             if priv_inteldbparams.IntelDB = nil then
                begin
                     OpenDBFile := True;
                     SiMain.LogDebug ('IntelDB in parameters is NIL');
                end
             else
                 begin
                      OpenDBFile := not (priv_inteldbfilename = ExtractFilePath (priv_name) + LocalTraceIntelFilename);

                      if OpenDBFile then
                         SiMain.LogDebug ('Path of existing IntelDB doesn''t match current file path');

                      // If we're going to open a different file, make sure we close the current one (if any)

                      if OpenDBFile and assigned (priv_inteldbparams.IntelDB) then
                         begin
                              DisconnectIntelDB;
{
                              priv_inteldbparams.IntelDB.Disconnect;
                              priv_inteldbparams.IntelDB.free;
                              priv_inteldbparams.IntelDB := nil;
}
                         end;
                 end;
        end
     else
         begin
              SiMain.LogDebug ('IntelDB is NOT in persistent mode');

              assert (not assigned (priv_inteldbparams.IntelDB));

              OpenDBFile := true;
         end;

     // Either open the DB file, or point it to the file we are working on

     if OpenDBFile then
        begin
             SiMain.LogVerbose ('Creating new IntelDB Object, accessing "' + priv_inteldbfilename + '" for file "' + priv_name + '"');

             priv_inteldbparams.IntelDB := tIntelDBManager.create (priv_inteldbfilename, priv_name, priv_filesize, priv_filetimestamp, priv_generalstats.FirstFrameTime, priv_generalstats.LastFrameTime, priv_nomeminteldb, SharedDBMode);
        end
     else
         begin
              if priv_name <> priv_inteldbparams.IntelDB.CurrentTraceFilename then
                 begin
                      // Transfer the pointer to allow access to the persistent DB

                      SiMain.LogVerbose ('Changing trace file assignment of IntelDB Object "' + priv_inteldbfilename + '" to new file "' + priv_name + '"');

                      priv_inteldbparams.IntelDB.AssignTraceFile (priv_name, priv_filesize, priv_filetimestamp, priv_generalstats.FirstFrameTime, priv_generalstats.LastFrameTime);
                 end;
         end;

     SiMain.LeaveMethod ('tTracefileEntry.VerifyAndConnectIntelDB');
end;

(* ------------------------------------------- *)

procedure tTracefileEntry.DisconnectIntelDB;

begin
     SiMain.EnterMethod ('tTracefileEntry.DisconnectIntelDB');

     assert (assigned (priv_inteldbparams.IntelDB));

     // Disconnect database

     priv_inteldbparams.IntelDB.Disconnect;
     priv_inteldbparams.IntelDB.free;
     priv_inteldbparams.IntelDB := nil;

     // TODO: clear existing hidden flag if applicable; using NOT faHidden doesn't help since it will set all other attributes

     if priv_IntelDBParams.HideIntelDBFile then
        FileSetAttr (priv_inteldbfilename, faHidden);

     SiMain.LeaveMethod ('tTracefileEntry.DisconnectIntelDB');
end;

(* ------------------------------------------- *)

procedure tTracefileEntry.getPCAPngStatistics (AutoScanThreshold : int64);

var ReadFromFile : boolean;
//  SharedDBMode : boolean;

begin
     SiMain.EnterMethod ('tTracefileEntry.getPCAPngStatistics');

//{$ifdef debug}
//     SharedDBMode := true;
//{$else}
//     SharedDBMode := false;
//{$endif}
     ReadFromFile := false;

     // See if we have statistics in a intel db file

     if not priv_inteldbparams.DoNotUseIntelDB then
        begin
             if FileExists (priv_inteldbfilename) then
                begin
                     VerifyAndConnectIntelDB;

                     if priv_inteldbparams.IntelDB.PCAPNGStructureScanned then
                        begin
                             priv_pcapngstatistics := priv_inteldbparams.IntelDB.readPCAPngStatisticsFromDB;
                             priv_pcapngstructread := true;
                        end
                     else
                         ReadFromFile := true;

                     if not priv_inteldbparams.UsePersistentIntelDB then
                        DisconnectIntelDB;
                end
             else
                 ReadFromFile := true;
        end
     else
         ReadFromFile := true;

     // Otherwise read from pcap file (unless we have to consider the Threshold)

     if ReadFromFile and (priv_filesize <= AutoScanThreshold) or (AutoScanThreshold = 0) then
        begin
             SiMain.LogVerbose ('Scanning PCAPng structure for file "' + priv_name + '" (' + IntToStr (priv_filesize) + ' bytes)');

             // Signal the start of the scan process for GUI setup

             if assigned (EventOnStartScanOfFile) then
                EventOnStartScanOfFile (self, psScanningPCAPngStruct);

             readPCAPngStatisticsFromTracefile;
             priv_pcapngstructread := true;

             if assigned (EventOnEndScanOfFile) then
                EventOnEndScanOfFile (self, psScanningPCAPngStruct);

             // Write to DB

             if not priv_IntelDBParams.DoNotUseIntelDB then
                begin
                     VerifyAndConnectIntelDB;
                     priv_inteldbparams.IntelDB.writePCAPngStatisticsToDB (priv_pcapngstatistics);

                     if not priv_inteldbparams.UsePersistentIntelDB then
                        DisconnectIntelDB;

                     priv_pcapngstructread := true;
                end;
        end;

     // We keep the full statistics, but we also assign some of the values for easier access
     // and compatibility reasons (other file formats)

     if priv_pcapngstructread then
        begin
             priv_framecount := priv_pcapngstatistics.FilePacketCount;
             priv_channelcount := priv_pcapngstatistics.FileInterfaceCount;
             priv_blockchainroot := priv_pcapngstatistics.BlockChain;
        end
     else
         begin
              // Set some defaults

              priv_framecount := priv_generalstats.FrameCount;

              priv_channelcount := 1;
         end;

     SiMain.LeaveMethod ('tTracefileEntry.getPCAPngStatistics');
end;

(* ------------------------------------------- *)

procedure tTracefileEntry.ReadFileDetails (const UsePersistentDB : boolean = False);

var DoFullFileStatistics : boolean;
    DoFileStatistics     : boolean;

begin
     SiMain.EnterMethod ('tTracefileEntry.ReadFileDetails');
{
     // If we are supposed to use a persistent IntelDB, prepare for it

     if IntelDBParams.UsePersistentIntelDB then
        priv_IntelDBParams := IntelDBParams
     else
         priv_inteldbparams.initialize;
}
     // Easy things first: filesize, tracetype, filedate

     getFilesizeAndTimestamp;
     getTraceType;

     // If we can, get the first frame time stamp

     if priv_tracetype <> cfUnknown then
        begin
             // Its safe to assume we have at least one channel

             priv_channelcount := 1;

             getFirstFrameTimeStamp;

             // If it is a PCAPng file, we determine the last frame timestamp as well

             if priv_tracetype = cfPCAPng then
                priv_generalstats.LastFrameTime := readLastFrameTimestampFromPCAPngFile;

             DoFullFileStatistics := (priv_filesize <= priv_scanthreshold) or (priv_scanthreshold = 0);

             // Next, the more complex stuff - get it from the intel DB if available

             if not priv_IntelDBParams.DoNotUseIntelDB then
                begin
                     // Connect DB

                     VerifyAndConnectIntelDB;
                     priv_inteldbindex := priv_IntelDBParams.IntelDB.CurrentFileIndex;

                     setProcessingStatus (psReadingDatabase);

                     // Check if we already scanned the full file

                     if priv_IntelDBParams.IntelDB.FullFileScanned then
                        begin
                             DoFullFileStatistics := false;
                             priv_packetstatsread := true;
                             priv_conversationsscanned := true;

                             priv_generalstats := priv_IntelDBParams.IntelDB.GeneralStatistics;
                             priv_pktcommentlist := priv_IntelDBParams.IntelDB.ReadFrameComments;
                             priv_conversationstore.LoadConversationsFromDatabase (priv_IntelDBParams.IntelDB.Database, priv_name, priv_filesize, priv_filetimestamp, priv_generalstats.FirstFrameTime, priv_generalstats.LastFrameTime);

                             if priv_generalstats.LastFrameTime.datetime <> 0 then
                                CalculateTraceDuration;

                             if (priv_framecount = 0) then
                                priv_framecount := priv_generalstats.FrameCount;
                        end
                     else
                         DoFileStatistics := true;

                     // While we're at it, check if we have PCAPng statistics, and if so, read them (to avoid two database connection setup/destroy runs)

                     if priv_IntelDBParams.IntelDB.PCAPngStatisticsRecordsExistInDB then
                        begin
                             priv_pcapngstatistics := priv_IntelDBParams.IntelDB.readPCAPngStatisticsFromDB;
                             priv_pcapngstructread := true;

                             priv_framecount := priv_pcapngstatistics.FilePacketCount;
                             priv_channelcount := priv_pcapngstatistics.FileInterfaceCount;
                             priv_blockchainroot := priv_pcapngstatistics.BlockChain;
                        end;

                     // Disconnect the DB if not in persistent mode

                     if not priv_IntelDBParams.UsePersistentIntelDB then
                        DisconnectIntelDB;
                end;

             if DoFullFileStatistics then
                begin
                     // If possible, ask the user if a fullscan is okay if the threshold is keeping us from it

                     if assigned (EventOnFullScan) and not DoFullFileStatistics then
                        EventOnFullScan (self, priv_name, DoFullFileStatistics);

                     // If okay, we scan the full trace until we find the last frame, plus, we're collecting other statistics

                     ReadFullFrameDetails;
                end;

             if (priv_tracetype = cfPCAPng) and not priv_pcapngstructread then
                getPCAPngStatistics (priv_scanthreshold);
        end;

     // Set the status to no task assigned. This will be adjusted in the GUI if there are tasks in the list

     setProcessingStatus (psNoTaskAssigned);

     // Transfer the IntelDB to the parameter set again to update the status for the calling routine

     SiMain.LeaveMethod ('tTracefileEntry.ReadFileDetails');
end;

(* ------------------------------------------- *)

function tTracefileEntry.IntelDBVersionIsWrong : boolean;

begin
     VerifyAndConnectIntelDB;
     result := priv_inteldbparams.IntelDB.DBVersionIsWrong;

     if not priv_inteldbparams.UsePersistentIntelDB then
        DisconnectIntelDB;
end;

(* ------------------------------------------- *)

function tTracefileEntry.getHintText : string;

var hinttext : string;
    node     : pBlockChainEntry;

begin
     hinttext := priv_name;

     result := hinttext;
end;

(* ------------------------------------------- *)
{ tTracefileList }
(* ------------------------------------------- *)

function SortByName(Item1, Item2: Pointer): Integer;

var file1, file2 : tTracefileEntry;

begin
     file1 := Item1;
     file2 := Item2;

     result := CompareStr (file1.Filename, file2.Filename);
end;

(* ------------------------------------------- *)

function SortByFrameTime(Item1, Item2: Pointer): Integer;

var file1, file2 : tTracefileEntry;

begin
     file1 := Item1;
     file2 := Item2;

     result := CompareValue (file1.FirstFrameTimestamp, file2.FirstFrameTimestamp);
end;

(* ------------------------------------------- *)

constructor tTracefileList.create;

begin
     inherited create;

     priv_filelist := tObjectList.Create (false);
     priv_dirlist := tDictionary<string, boolean>.create;
     priv_alldirswriteable := true;

     priv_encfilecount      := 0;
     priv_capfilecount      := 0;
     priv_pcapfilecount     := 0;
     priv_pcapngfilecount   := 0;

     priv_renamecounter.initialize;
end;

(* ------------------------------------------- *)

procedure tTracefileList.free;

begin
     inherited free;
end;

(* ------------------------------------------- *)

destructor tTracefileList.Destroy;

var entry : tTraceFileEntry;
    i     : integer;

begin
     for i := 1 to priv_filelist.Count do
         begin
              // Always get the first, the rest will slide up

              entry := GetFile (1);
              priv_filelist.Delete(0);

              entry.free;
              entry := nil;
         end;

     priv_filelist.Free;
     priv_dirlist.Free;

     inherited;
end;

(* ------------------------------------------- *)

procedure tTracefileList.addFile (NewFile: tTracefileEntry);

var path        : string;
    boWriteable : boolean;

begin
     priv_filelist.Add(NewFile);

     case NewFile.TracefileType of
      cfENC    : inc (priv_encfilecount);
      cfCAP    : inc (priv_capfilecount);
      cfPCAP   : inc (priv_pcapfilecount);
      cfPCAPng : inc (priv_pcapngfilecount);
     end;

     updateFirstAndLastFrameLocators;

     // Check directory being writeable

     path := NewFile.getFilePath;

     if not priv_dirlist.ContainsKey (path) then
        begin
             boWriteable := isDirectoryWriteable (path);

             priv_dirlist.Add (path, boWriteable);
        end;

     // Check if all dirs are writeable

     priv_alldirswriteable := true;

     for path in priv_dirlist.Keys do
         begin
              if not priv_dirlist.Items[path] then
                 priv_alldirswriteable := false;
         end;
end;

(* ------------------------------------------- *)

function tTracefileList.getFile (index: integer): tTracefileEntry;

begin
     if (index > 0) and (index <= priv_filelist.Count) then
        result := priv_filelist[index -1] as tTraceFileEntry
     else
         raise EInOutError.Create ('Error: file index out of bounds (tTracefileList.getFile).');
end;

(* ------------------------------------------- *)

function tTracefileList.getFileCount: integer;

begin
     result := priv_filelist.Count;
end;

(* ------------------------------------------- *)

procedure tTracefileList.removeFile (index: integer);

var entry       : tTracefileEntry;
    path        : string;
    i           : integer;
    boWriteable : boolean;

begin
     if (index > 0) and (index <= priv_filelist.Count) then
        begin
             entry := priv_filelist[index-1] as tTracefileEntry;

             case entry.TracefileType of
              cfENC    : dec (priv_encfilecount);
              cfCAP    : dec (priv_capfilecount);
              cfPCAP   : dec (priv_pcapfilecount);
              cfPCAPng : dec (priv_pcapngfilecount);
             end;

             priv_filelist.Delete(index - 1)
        end
     else
         raise EInOutError.Create ('Error: file index out of bounds (tTracefileList.removeFile).');

     updateFirstAndLastFrameLocators;

     // Update directory list

     priv_dirlist.Clear;
     priv_alldirswriteable := true;

     for i := 1 to priv_filelist.Count do
         begin
              entry := priv_filelist[i-1] as tTracefileEntry;
              path := entry.getFilePath;

              if not priv_dirlist.ContainsKey (path) then
                 begin
                      boWriteable := isDirectoryWriteable (path);

                      if not boWriteable then
                         priv_alldirswriteable := false;

                      priv_dirlist.Add (path, boWriteable);
                 end;
         end;
end;

(* ------------------------------------------- *)

procedure tTracefileList.sort(SortType: tTracefileSortType);

begin
     case SortType of
      stFilename       : begin
                              priv_filelist.Sort(SortByName);
                         end;
      stFirstFrameTime : begin
                              priv_filelist.Sort(SortByFrameTime);
                         end;
     end;
end;

(* ------------------------------------------- *)

procedure tTracefileList.changeListposition(Index1, Index2: integer);

begin
     priv_filelist.Exchange(Index1, Index2);
end;

(* ------------------------------------------- *)

procedure tTracefileList.clear;

begin
     priv_filelist.Clear;
end;

(* ------------------------------------------- *)

function tTracefileList.GetFileNameList: tStringlist;

var list : tStringlist;
    i    : integer;

begin
     list := tStringlist.Create;

     for i := 1 to priv_filelist.Count do
         list.Add ((priv_filelist[i-1] as tTraceFileEntry).FullName);

     result := list;
end;

(* ------------------------------------------- *)

procedure tTracefileList.updateFirstAndLastFrameLocators;

var entry : tTraceFileEntry;

begin
     if priv_filelist.Count = 0 then
        begin
             priv_firstframelocation.initialize;
             priv_lastframelocation.initialize;
        end
     else
         begin
              entry := getFile (1);

              priv_firstframelocation.FileIndex := 1;
              priv_firstframelocation.FileFrameIndex := 1;
              priv_firstframelocation.FileOffset := entry.priv_generalstats.FirstFrameOffset;

              // Get the last file in the list

              entry := getFile (priv_filelist.Count);

              // Determine Last Location if we can (if the file had been scanned)

              priv_lastframelocation.FileIndex := priv_filelist.Count;

              if entry.FrameCount > 0 then
                 begin
                      priv_lastframelocation.FileFrameIndex := entry.FrameCount;
                      priv_lastframelocation.FileOffset := entry.priv_generalstats.LastFrameOffset;
                 end
              else
                  begin
                       priv_lastframelocation.FileFrameIndex := int64.MaxValue;
                       priv_lastframelocation.FileOffset := int64.MaxValue;
                  end;
         end;
end;

(* ------------------------------------------- *)

end.
