{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit tracefile_utils;

interface

uses base_structures,
     class_traceloader,
     class_rawframe;


         function getTracefileType (const Filename : string) : tCaptureFileType;
         function createTraceLoader (const Filename : string; const TracefileIndex : integer = 0; const FileType : tCaptureFileType = cfUnknown) : tTraceLoader;
         procedure AddFrameStatistics (var GeneralStatistics : tGeneralTraceStats; Frame : tRawFrame);
         procedure FinalizeFrameStatistics (var GeneralStatistics : tGeneralTraceStats);

implementation

uses Math,
     SysUtils,
     class_loader_enc,
     class_loader_naicap,
     class_loader_pcap,
     class_loader_pcapng;

(* ------------------------------------------- *)

// Gets the trace file type it is not already determined

function getTracefileType (const Filename : string) : tCaptureFileType;

var fHandle    : File of byte;
    aBuffer    : array [1..20] of Byte;
    sMagicText : string;
    pMagicHelp : pCardinal;
    iReadBytes : integer;
    i          : integer;
    openResult : integer;
    boMatch    : boolean;
    value      : tCaptureFileType;

begin
     value := cfUnknown;

     if FileExists (Filename) then
        begin
              assignfile (fHandle, Filename);
              FileMode := fmOpenRead;

              try
                 reset (fHandle);
              except

              end;

              openResult := IOResult;

              if openResult = 0 then
                 begin
                      // Determine the capture File Type

                      BlockRead (fHandle, aBuffer, 16, iReadBytes);

                      // check for ENC format

                      boMatch := true;

                       for i := 1 to min (ENC_MagicStringLength, iReadBytes) do
                           begin
                                if ENC_MagicString[i] <> char (aBuffer[i]) then
                                   boMatch := false;
                           end;

                       if boMatch then
                          value := cfENC
                       else
                           begin
                                boMatch := true;

                                // Check for Sniffer CAP format

                                for i := 1 to min (CAP_MagicStringLength, iReadBytes) do
                                    begin
                                         if CAP_MagicString[i] <> char (aBuffer[i]) then
                                            boMatch := false;
                                    end;

                                if boMatch then
                                   value := cfCAP
                                else
                                    begin
                                         // Check for PCAP and PCAP-ng format

                                         pMagicHelp := @aBuffer;

                                          case pMagicHelp^ of
                                           PCAP_MagicStandard,
                                           PCAP_MagicStandardSwap,
                                           PCAP_MagicModified,
                                           PCAP_MagicModified_Swap : value := cfPCAP;
                                           PCAPNG_MagicString      : value := cfPCAPng;
                                         end;
                                    end;
                           end;

                       closeFile (fHandle);
                  end;
        end
     else
         result := cfUnknown;

     result := value;
end;

(* ------------------------------------------- *)

function createTraceLoader (const Filename : string; const TracefileIndex : integer = 0; const FileType : tCaptureFileType = cfUnknown) : tTraceLoader;

var tracetype : tCaptureFileType;
    loader    : tTraceLoader;

begin
     loader := nil;

     if FileType = cfUnknown then
        tracetype := getTraceFileType (Filename)
     else
         tracetype := FileType;

     case tracetype of
      cfENC    : begin
                      loader := tENCLoader.create (Filename, TracefileIndex, true);
                 end;
      cfCAP    : begin
                      loader := tCAPLoader.create (Filename, TracefileIndex, true);
                 end;
      cfPCAP   : begin
                      loader := tPCAPLoader.create (Filename, TracefileIndex, true);
                 end;
      cfPCAPng : begin
                      loader := tPCAPNGLoader.create (Filename, TracefileIndex, true);
                 end;
     end;

     result := loader;
end;

(* ------------------------------------------- *)

procedure AddFrameStatistics (var GeneralStatistics : tGeneralTraceStats; Frame: tRawFrame);

begin
     if Generalstatistics.FirstFrameOffset = 0 then
        begin
             GeneralStatistics.FirstFrameTime := Frame.Timestamp;
             GeneralStatistics.FirstFrameOffset := Frame.FileOffset;
        end;

     // record the timestamp and offset of the lastest frame

     GeneralStatistics.LastFrameTime := Frame.Timestamp;
     GeneralStatistics.LastFrameOffset := Frame.FileOffset;

     if Frame.WireSize < GeneralStatistics.MinFrameSize then
        GeneralStatistics.MinFrameSize := Frame.WireSize;

     if Frame.WireSize > GeneralStatistics.MaxFrameSize then
        GeneralStatistics.MaxFrameSize := Frame.WireSize;

     if Frame.WireSize > Frame.FrameSize then
        begin
             if GeneralStatistics.isSliced then
                begin
                     // Check if we have different slice sizes

                     if GeneralStatistics.SliceSize <> Frame.FrameSize then
                        begin
                             GeneralStatistics.SliceSizeConstant := false;
                             GeneralStatistics.SliceSize := 0;
                        end;
                end
             else
                 begin
                      GeneralStatistics.isSliced := true;
                      GeneralStatistics.SliceSizeConstant := true;
                      GeneralStatistics.SliceSize := frame.FrameSize;
                 end;
        end
     else
         begin
              // Check if we have a full frame with a large size than the slice size we know

              if GeneralStatistics.isSliced and (frame.WireSize > GeneralStatistics.SliceSize) then
                 GeneralStatistics.SliceSizeConstant := false;
         end;

     if (Frame.FileFrameID > 1) and (Frame.Timestamp < GeneralStatistics.LastFrameTime) then
        begin
             GeneralStatistics.TimeOrderCorrect := false;

             inc (GeneralStatistics.OutOfTimeFrames);
        end;

     GeneralStatistics.TotalWireBytes := GeneralStatistics.TotalWireBytes + Frame.WireSize;
     GeneralStatistics.TotalFrameBytes := GeneralStatistics.TotalFrameBytes + Frame.FrameSize;
end;

(* ------------------------------------------- *)

procedure FinalizeFrameStatistics (var GeneralStatistics : tGeneralTraceStats);

begin
     if GeneralStatistics.FrameCount > 0 then
        GeneralStatistics.AverageFrameSize := GeneralStatistics.TotalFrameBytes div GeneralStatistics.FrameCount
     else
         GeneralStatistics.AverageFrameSize := 0;
end;

(* ------------------------------------------- *)

end.
