{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_parser_802dot11management;

interface

uses base_structures,
     common_parse_helpers,
     class_parser_baseclass;

type t802dot11MgmtCapability = (capESS, capIBSS, capCFPollable, capCFPollRequest, capPrivacy, capShortPreamble, capPBCC, capChannelAgility, capSpectrumManagement, capShortSlotTime, capAutoPowerSafe, capRadioMeasurement, capDSSSOFDM, capDelayedBlockAck, capImmediateBlockAck);
     t802dot11MgmtCapabilitiesSet = set of t802dot11MgmtCapability;

     tBeaconFixedParams = packed record
                                       Timestamp            : uint64;
                                       BeaconInterval       : word;
                                       Capabilities         : word;
                          end;
     pBeaconFixedParams = ^tBeaconFixedParams;

     t802dot11TrafficIndicationMap = packed record
                                                  DTIMCount        : byte;
                                                  DTIMPeriod       : byte;
                                                  BitmapControl    : byte;
                                                  PartialvBitmap   : byte;

                                                  procedure initialize;
                                            end;
     p802dot11TrafficIndicationMap = ^t802dot11TrafficIndicationMap;

     t802dot11CountryInformation = packed record
                                                CountryCode        : word;
                                                Environment        : byte;
                                                FirstChannel       : byte;
                                                NumberOfChannels   : byte;
                                                MaximumPowerLevel  : shortint;

                                                procedure initialize;
                                          end;
     p802dot11CountryInformation = ^t802dot11CountryInformation;

     t802dot11TaggedParam = record
                                  ParamType                 : byte;
                                  ParamLen                  : byte;
                                  SSID                      : string;
                                  SupportedRates            : array of byte;
                                  CurrentChannel            : byte;
                                  TrafficIndicationMap      : t802dot11TrafficIndicationMap;
                                  CountryInformation        : t802dot11CountryInformation;
                                  ERPInformation            : byte;
                                  ExtendedSupportedRates    : array of byte;

                                  isUnknown                 : boolean;
                                  UnknownParam              : array of byte;

                                  procedure initialize;
                            end;
     t802dot11TaggedParamList = array of t802dot11TaggedParam;

     t802dot11ManagementParser = class (tFrameParser)

                                       constructor create (const SubType : t802dot11FrameSubType; FrameData : pointer; const FrameOffset, DataLength, WireLength : integer; const FrameIndex : integer = 0); reintroduce;

                                       private
                                              priv_SubType          : t802dot11FrameSubType;

                                              priv_Timestamp        : uint64;
                                              priv_BeaconInterval   : word;                                 // in milliseconds, usually 100
                                              priv_Capabilities     : t802dot11MgmtCapabilitiesSet;

                                              priv_TaggedParams     : t802dot11TaggedParamList;

                                              function ParseCapabilities (Value : Word) : t802dot11MgmtCapabilitiesSet;
                                              procedure ParseTaggedParameters (var LocalOffset : integer);
                                              function ParseBeaconFrame : tParseResult;

                                       public
                                             property SubType : t802dot11FrameSubType read priv_SubType;
                                             property Timestamp : uint64 read priv_Timestamp;
                                             property BeaconInterval : word read priv_BeaconInterval;
                                             property Capabilities : t802dot11MgmtCapabilitiesSet read priv_Capabilities;
                                             property TaggedParameters : t802dot11TaggedParamList read priv_TaggedParams;

                                             function checkIsPlausible : boolean; override;
                                             function parse : tParseResult; override;
                                 end;


const cSSIDParameterSet = 0;
      cSupportedRates  = 1;
      cDSParameterSet = 3;

      cTrafficIndicationMap = 5;
      cCountryInformation = 7;

      cERPInformation = 42;

      cExtendedSupportedRates = 50;

implementation

(* ------------------------------------------- *)
{ t802dot11TrafficIndicationMap }
(* ------------------------------------------- *)

procedure t802dot11TrafficIndicationMap.initialize;

begin
     DTIMCount := 0;
     DTIMPeriod := 0;
     BitmapControl := 0;
     PartialvBitmap := 0;
end;

(* ------------------------------------------- *)
{ t802dot11CountryInformation }
(* ------------------------------------------- *)

procedure t802dot11CountryInformation.initialize;

begin
     CountryCode := 0;
     Environment := 0;
     FirstChannel := 0;
     NumberOfChannels := 0;
     MaximumPowerLevel := 0;
end;

(* ------------------------------------------- *)
{ t802dot11TaggedParam }
(* ------------------------------------------- *)

procedure t802dot11TaggedParam.initialize;

begin
     ParamType := 0;
     ParamLen := 0;

     SSID := '';
     setLength (SupportedRates, 0);
     CurrentChannel := 0;

     TrafficIndicationMap.initialize;
     CountryInformation.initialize;

     ERPInformation := 0;

     setLength (ExtendedSupportedRates, 0);

     isUnknown := false;
     setLength (UnknownParam, 0);
end;

(* ------------------------------------------- *)
{ t802dot11ManagementParser }
(* ------------------------------------------- *)

constructor t802dot11ManagementParser.create (const SubType: t802dot11FrameSubType; FrameData: pointer; const FrameOffset, DataLength, WireLength, FrameIndex: integer);

begin
     inherited create (FrameData, FrameOffset, DataLength, Wirelength, FrameIndex);

     priv_SubType := SubType;

     priv_Timestamp := 0;
     priv_BeaconInterval := 0;
     priv_Capabilities := [];

     setLength (priv_TaggedParams, 0)
end;

(* ------------------------------------------- *)

function t802dot11ManagementParser.checkIsPlausible: boolean;

begin
     // if we get here, it's by 802.11 frame type decision, so its okay
     // but we may not be able to parse all things yet

     result := false;

     case priv_SubType of
      mgmtftAssocReq    : ;
      mgmtftAssocResp   : ;
      mgmtftReassocReq  : ;
      mgmtftReassocResp : ;
      mgmtftProbeReq    : ;
      mgmtftProbeResp   : ;
      mgmtftBeacon      : result := true;
      mgmtftATIM        : ;
      mgmtftDisassoc    : ;
      mgmtftAuth        : ;
      mgmtftDeAuth      : ;
      mgmtftAction      : ;
      mgmtftActionNoAck : ;
     end;
end;

(* ------------------------------------------- *)

function t802dot11ManagementParser.ParseCapabilities (Value: Word) : t802dot11MgmtCapabilitiesSet;

var flag    : t802dot11MgmtCapability;
    flagset : t802dot11MgmtCapabilitiesSet;
    isSet   : boolean;
    i       : integer;

begin
     flagset := [];

     for i := 0 to 15 do
         begin
              isSet := BitIsSet (Value, i);

              if isSet then
                 begin
                      case i of
                        0 : flag := capESS;
                        1 : flag := capIBSS;
                        2 : flag := capCFPollable;
                        3 : flag := capCFPollRequest;
                        4 : flag := capPrivacy;
                        5 : flag := capShortPreamble;
                        6 : flag := capPBCC;
                        7 : flag := capChannelAgility;
                        8 : flag := capSpectrumManagement;
                        // 9 is reserved
                        10 : flag := capShortSlotTime;
                        11 : flag := capAutoPowerSafe;
                        12 : flag := capRadioMeasurement;
                        13 : flag := capDSSSOFDM;
                        14 : flag := capDelayedBlockAck;
                        15 : flag := capImmediateBlockAck;
                      end;

                      flagset := flagset + [flag];
                 end;
         end;

     result := flagset;
end;

(* ------------------------------------------- *)

procedure t802dot11ManagementParser.ParseTaggedParameters (var LocalOffset: integer);

var OptionValue  : t802dot11TaggedParam;
    optioncount  : integer;
    i            : Integer;
    valueoffset  : integer;
    len          : integer;

begin
     optioncount := 0;

     // Read options as long as there are enough bytes (minus 4 for the FCS)

     while (priv_framelength - 4) > (priv_frameoffset + LocalOffset + 2) do
           begin
                OptionValue.initialize;

                // Read code, length, value

                OptionValue.ParamType := pByte (priv_framedata + priv_frameoffset + LocalOffset)^;
                OptionValue.ParamLen := pByte (priv_framedata + priv_frameoffset + LocalOffset + 1)^;

                inc (LocalOffset, 2);

                // parse if code, len, value are covered by enough bytes in the frame

                if (priv_framelength - 4) >= (priv_frameoffset + LocalOffset + OptionValue.ParamLen) then
                   begin
                        valueoffset := priv_frameoffset + LocalOffset;

                        case OptionValue.ParamType of
                         cSSIDParameterSet        : begin
                                                         OptionValue.SSID := ParseStringOption (priv_framedata + valueoffset, OptionValue.ParamLen);
                                                    end;
                         cSupportedRates          : begin
                                                         setLength (OptionValue.SupportedRates, OptionValue.ParamLen);

                                                         for i := 0 to OptionValue.ParamLen - 1 do
                                                             OptionValue.SupportedRates[i] := ParseByteOption (priv_framedata + valueOffset + i);
                                                    end;
                         cDSParameterSet          : begin
                                                         OptionValue.CurrentChannel := ParseByteOption (priv_framedata + valueOffset);
                                                    end;
                         cTrafficIndicationMap    : begin
                                                         OptionValue.TrafficIndicationMap := p802dot11TrafficIndicationMap (priv_framedata + valueOffset)^;
                                                    end;

                         cCountryInformation      : begin
                                                         OptionValue.CountryInformation := p802dot11CountryInformation (priv_framedata + valueoffset)^;
                                                    end;
                         cERPInformation          : begin
                                                         OptionValue.ERPInformation := ParseByteOption (priv_framedata + valueOffset);
                                                    end;
                         cExtendedSupportedRates  : begin
                                                         setLength (OptionValue.ExtendedSupportedRates, OptionValue.ParamLen);

                                                         for i := 0 to OptionValue.ParamLen - 1 do
                                                             OptionValue.ExtendedSupportedRates[i] := ParseByteOption (priv_framedata + valueOffset + i);
                                                    end;
                         else
                              begin
                                   // since we may not know how to parse all parameters, copy those we
                                   // don't into a simple unparsed byte array

                                   OptionValue.isUnknown := true;

                                   setLength (OptionValue.UnknownParam, OptionValue.ParamLen);

                                   for i := 0 to OptionValue.ParamLen - 1 do
                                       OptionValue.UnknownParam[i] := ParseByteOption (priv_framedata + valueOffset + i);
                              end;
                        end
                   end;

                // If the option was parsed successfully, add it

                if OptionValue.ParamLen > 0 then
                   begin
                        len := length (priv_TaggedParams);

                        setLength (priv_TaggedParams, len + 1);
                        priv_TaggedParams[len] := OptionValue;
                   end;

                // Move the offset along

                inc (LocalOffset, OptionValue.ParamLen);
           end;
end;

(* ------------------------------------------- *)

function t802dot11ManagementParser.ParseBeaconFrame : tParseResult;

var pData : pBeaconFixedParams;

begin
     if priv_framelength < priv_frameoffset + 4 then
        result := prFailUndersize
     else
         begin
              pData := pBeaconFixedParams (priv_framedata + priv_frameoffset);

              priv_Timestamp := pData^.Timestamp;
              priv_BeaconInterval := pData^.BeaconInterval;
              priv_Capabilities := ParseCapabilities (pData^.Capabilities);

              result := prSuccess;
         end;
end;

(* ------------------------------------------- *)

function t802dot11ManagementParser.parse : tParseResult;

var status    : tParseResult;
    pData     : pByte;
    offset    : integer;

begin
     priv_osilayer := 1;     // Set Layer this parser is for

     status := prSuccess;

     if priv_framelength < priv_frameoffset + 4 then
        status := prFailUndersize
     else
         begin
              offset := 0;

              case priv_SubType of
               mgmtftAssocReq    : ;
               mgmtftAssocResp   : ;
               mgmtftReassocReq  : ;
               mgmtftReassocResp : ;
               mgmtftProbeReq    : ;
               mgmtftProbeResp   : ;
               mgmtftBeacon      : begin
                                        status := ParseBeaconFrame;

                                        offset := sizeOf (tBeaconFixedParams);
                                   end;
               mgmtftATIM        : ;
               mgmtftDisassoc    : ;
               mgmtftAuth        : ;
               mgmtftDeAuth      : ;
               mgmtftAction      : ;
               mgmtftActionNoAck : ;
              end;

            if offset > 0 then
               ParseTaggedParameters (offset);

            prot_Layersize := offset;

            // Determine next anchor

            priv_nextAnchor.HeaderType := NEXT_HEADER_NONE;
            priv_nextAnchor.Offset := priv_frameoffset + prot_Layersize;
         end;

     result := status;
end;

(* ------------------------------------------- *)



end.
