{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_parser_capwap;

{
   CAPWAP Header (RFC 5415 section 4.3):

        0                   1                   2                   3
        0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
       |CAPWAP Preamble|  HLEN   |   RID   | WBID    |T|F|L|W|M|K|Flags|
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
       |          Fragment ID          |     Frag Offset         |Rsvd |
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
       |                 (optional) Radio MAC Address                  |
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
       |            (optional) Wireless Specific Information           |
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
       |                        Payload ....                           |
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
}

interface

uses base_structures,
     common_parse_helpers,
     class_parser_baseclass;

type tCAPWAPHdrFlags = (cwhfPayloadTypeNative, cwhfFragment, cwhfLastFragment, cwhfWirelessHeader, cwhfRadioMACHeader, cwhfKeepAlive);
     tCAPWAPHdrFlagSet = set of tCAPWAPHdrFlags;

     tCAPWAPHdrStruct = packed record
                                     VTLenRadioIDFlags : longword;
                                     FragmentID        : word;
                                     FragmentOffset    : word;
                        end;
     pCAPWAPHdrStruct = ^tCAPWAPHdrStruct;

     tCAPWAPWirelessData = packed record
                                        WirelessRSSI         : shortint;
                                        WirelessSNR          : byte;
                                        WirelessDataRateMbps : byte;
                           end;

     pCAPWAPWirelessData = ^tCAPWAPWirelessData;

     tCAPWAPParser = class (tFrameParser)

                           private
                                  priv_Version           : byte;
                                  priv_Type              : byte;
                                  priv_Headerlength      : byte;
                                  priv_RadioID           : byte;
                                  priv_WirelessBindingID : byte;
                                  priv_HeaderFlags       : tCAPWAPHdrFlagSet;
                                  priv_FragmentID        : word;
                                  priv_FragmentOffset    : word;

                                  priv_OptWirelessData   : tCAPWAPWirelessData;

                                  procedure initialize;

                           public
                                 property Version : byte read priv_Version;
                                 property CAPWAPType : byte read priv_type;
                                 property HeaderLength : byte read priv_HeaderLength;
                                 property RadioID : byte read priv_RadioID;
                                 property WirelessBindingID : byte read priv_WirelessBindingID;
                                 property HeaderFlags : tCAPWAPHdrFlagSet read priv_HeaderFlags;
                                 property FragmentID : word read priv_FragmentID;
                                 property FragmentOffset : word read priv_FragmentOffset;
                                 property OptionalWirelessData : tCAPWAPWirelessData read priv_OptWirelessData;

                                 function checkIsPlausible : boolean; override;
                                 function parse : tParseResult; override;
                     end;

implementation

uses Math,
     Toolkit;

(* ------------------------------------------- *)
{ tCAPWAPParser }
(* ------------------------------------------- *)

procedure tCAPWAPParser.initialize;

begin
     priv_Version := 0;
     priv_Type := 0;
     priv_Headerlength := 0;
     priv_RadioID := 0;
     priv_WirelessBindingID := 0;
     priv_HeaderFlags := [];
     priv_FragmentID := 0;
     priv_FragmentOffset := 0;

     priv_OptWirelessData.WirelessRSSI := 0;
     priv_OptWirelessData.WirelessSNR := 0;
     priv_OptWirelessData.WirelessDataRateMbps := 0;
end;

(* ------------------------------------------- *)

function tCAPWAPParser.checkIsPlausible: boolean;

var value     : boolean;
    pPreamble : pByte;
    version   : byte;
    cwtype    : byte;

begin
     // Need preamble and hdr

     if priv_framelength < priv_frameoffset + sizeOf (tCAPWAPHdrStruct) then
        value := false
     else
         begin
              pPreamble := pByte (priv_framedata + priv_frameoffset);

              version := GetNibble (pPreamble^, true);
              cwtype := GetNibble (pPreamble^, false);

              value := (version = 0) and (cwtype = 0);
         end;

     result := value;
end;

(* ------------------------------------------- *)

function tCAPWAPParser.parse: tParseResult;

var status       : tParseResult;
    pPreamble    : pByte;
    pHeader      : pCAPWAPHdrStruct;
    pOptWireless : pCAPWAPWirelessData;
    lHelp        : longword;
    plHelp       : pLongword;
    lFlags       : longword;
    pLen         : pByte;
    bHelp        : byte;
    offset       : integer;

begin
     priv_osilayer := 3;     // Set Layer this parser is for
     prot_layersize := sizeOf (tCAPWAPHdrStruct);

     initialize;

     status := prSuccess;

     if priv_framelength < priv_frameoffset + sizeOf (tCAPWAPHdrStruct) then
        status := prFailUndersize
     else
         try
            // Parse Version and type directly via byte pointer

            pPreamble := pByte (priv_framedata + priv_frameoffset);

            priv_Version := GetNibble (pPreamble^, true);
            priv_Type := GetNibble (pPreamble^, false);

            // Get the complete 32 bit structure and process it

            pHeader := pCAPWAPHdrStruct (priv_framedata + priv_frameoffset);
            lHelp := swap32 (pHeader^.VTLenRadioIDFlags);

            // Parse Length, Radio ID, Wireless Binding, Flags

            // Length

            priv_headerlength := ExtractBitRange (lHelp, 19, 5);

            // Radio ID

            priv_radioid := ExtractBitRange (lHelp, 14, 5);

            // Wireless Binding ID

            priv_WirelessBindingID := ExtractBitRange (lHelp, 9, 5);

            // Header Flags

            lFlags := ExtractBitRange (lHelp, 0, 9);

            if BitIsSet (lFlags, 8) then
               priv_HeaderFlags := priv_HeaderFlags + [cwhfPayloadTypeNative];

            if BitIsSet (lFlags, 7) then
               priv_HeaderFlags := priv_HeaderFlags + [cwhfFragment];

            if BitIsSet (lFlags, 6) then
               priv_HeaderFlags := priv_HeaderFlags + [cwhfLastFragment];

            if BitIsSet (lFlags, 5) then
               priv_HeaderFlags := priv_HeaderFlags + [cwhfWirelessHeader];

            if BitIsSet (lFlags, 4) then
               priv_HeaderFlags := priv_HeaderFlags + [cwhfRadioMACHeader];

            if BitIsSet (lFlags, 3) then
               priv_HeaderFlags := priv_HeaderFlags + [cwhfKeepAlive];

            priv_FragmentID := swap16 (pHeader^.FragmentID);
            priv_FragmentOffset := swap16 (pHeader^.FragmentOffset);
            priv_FragmentOffset := (priv_FragmentOffset and $fff8) shr 3;

            offset := priv_frameoffset + sizeOf (tCAPWAPHdrStruct);

            // Parse optional stuff

            if cwhfRadioMACHeader in priv_HeaderFlags then
               begin
                    pLen := pByte (priv_framedata + offset);

                    // I have no sample PCAP, so right now this is ignored





                    // Update layer size by 32bit aligned size (len + 1 for the lenght byte)

                    prot_layersize := prot_layersize + Get32BitAlignment (pLen^ + 1);
               end;

            if cwhfWirelessHeader in priv_HeaderFlags then
               begin
                    pLen := pByte (priv_framedata + offset);

                    if pLen^ = 4 then
                       begin
                            pOptWireless := pCAPWAPWirelessData (priv_framedata + offset + 1);
                            priv_OptWirelessData := pOptWireless^;
                       end;

                    // Update layer size by 32bit aligned size (len + 1 for the lenght byte)

                    prot_layersize := prot_layersize + Get32BitAlignment (pLen^ + 1);
               end;

            // Determine next anchor

            if cwhfPayloadTypeNative in priv_HeaderFlags then
               priv_nextAnchor.HeaderType := NEXT_HEADER_802dot11Hdr
            else
                priv_nextAnchor.HeaderType := NEXT_HEADER_ETHERNET;

            priv_nextAnchor.Offset := priv_frameoffset + prot_layersize;

         except
               status := prFailInvalidValues;
         end;

     result := status;
end;

(* ------------------------------------------- *)

end.
