{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_parser_hsrp;

interface

uses base_structures,
     base_checksums,
     common_parse_helpers,
     definition_ip_addresses,
     definition_tcp_structures,
     class_parser_baseclass;

type tHSRPAuthData = array[0..7] of byte;

     tHSRPOpCode = (hoHello, hoCoup, hoResign, hoAdvertise, hoUnknown);
     tHSRPState = (hsInitial, hsLearn, hsListen, hsSpeak, hsStandby, hsActive, hsUnknown);

     pHSRPVersionCode = ^tHSRPVersionCode;
     tHSRPVersionCode = packed record
                                     Version      : byte;
                                     OpCode       : byte;
                               end;

     pHSRPStruct = ^tHSRPStruct;
     tHSRPStruct = packed record
                                Version      : byte;
                                OpCode       : byte;
                                State        : byte;
                                HelloTime    : byte;
                                HoldTime     : byte;
                                Priority     : byte;
                                Group        : byte;
                                Reserved     : byte;
                                AuthData     : tHSRPAuthData;
                                VirtualIP    : longword;
                           end;

     pHSRPAdvStruct = ^tHSRPAdvStruct;
     tHSRPAdvStruct = packed record
                                   Version           : byte;
                                   OpCode            : byte;
                                   TLVType           : word;
                                   TLVLength         : word;
                                   State             : byte;
                                   Reserved1         : byte;
                                   ActiveGroupCount  : word;
                                   PassiveGroupCount : word;
                                   Reserved2         : longword;
                             end;

     tHSRPParser = class (tFrameParser)

                         private
                                priv_version    : byte;
                                priv_opcode     : tHSRPOpCode;
                                priv_state      : tHSRPState;
                                priv_hellotime  : byte;
                                priv_holdtime   : byte;
                                priv_priority   : byte;
                                priv_group      : byte;
                                priv_authdata   : tHSRPAuthData;
                                priv_virtualip  : tIPv4Address;

                                // Advertisement stuff

                                priv_AdvTLVType           : word;
                                priv_AdvTLVLength         : word;
                                priv_AdvActiveGroupCount  : word;
                                priv_AdvPassiveGroupCount : word;

                         public
                               property Version : byte read priv_version;
                               property OpCode : tHSRPOpCode read priv_opcode;
                               property State : tHSRPState read priv_state;
                               property HelloTime : byte read priv_hellotime;
                               property HoldTime : byte read priv_holdtime;
                               property Priority : byte read priv_priority;
                               property Group : byte read priv_group;
                               property AuthData : tHSRPAuthData read priv_authdata;
                               property VirtualIP : tIPv4Address read priv_virtualip;

                               property AdvTLVType : word read priv_AdvTLVType;
                               property AdvTLVLength : word read priv_AdvTLVLength;
                               property AdvActiveGroupCount : word read priv_AdvActiveGroupCount;
                               property AdvPassiveGroupCount : word read priv_AdvPassiveGroupCount;

                               function checkIsPlausible : boolean; override;
                               function parse : tParseResult; override;
                   end;

const DefaultAuthData : tHSRPAuthData = ($63,$69,$73,$63,$6F,0,0,0);    // Taken from RFC 2281, spells "cisco"

implementation

uses Toolkit;

(* ------------------------------------------- *)
{ tHSRPParser }
(* ------------------------------------------- *)

function tHSRPParser.checkIsPlausible: boolean;

var value : boolean;
    pData : pHSRPVersionCode;

begin
     value := true;

     if priv_framelength < priv_frameoffset + 2 then
        value := false
     else
         begin
              pData := pHSRPVersionCode (priv_framedata + priv_frameoffset);

              // OpCode can be 0, 1 or 2

              if (pData^.Version > 0) or ((pData^.OpCode) > 3) then
                 value := false;
         end;

     result := value;
end;

(* ------------------------------------------- *)

function tHSRPParser.parse: tParseResult;

var pVerCode : pHSRPVersionCode;
    pData    : pHSRPStruct;
    pAdvData : pHSRPAdvStruct;
    status   : tParseResult;
    CalcCRC  : boolean;
    help     : word;
    iCount   : integer;


begin
     priv_osilayer := 5;     // Set Layer this parser is for
     status := prSuccess;

     if priv_framelength < priv_frameoffset + 2 then
        status := prFailUndersize
     else
         begin
              pVerCode := pHSRPVersionCode (priv_framedata + priv_frameoffset);

              case pVerCode^.OpCode of
               0..2 : begin
                           if priv_framelength < priv_frameoffset + sizeOf (tHSRPStruct) then
                              status := prFailUndersize
                      end;
                  3 : begin
                           if priv_framelength < priv_frameoffset + sizeOf (tHSRPAdvStruct) then
                              status := prFailUndersize
                      end;
              end;
         end;

     if status = prSuccess then
        try
           pVerCode := pHSRPVersionCode (priv_framedata + priv_frameoffset);

           case pVerCode^.OpCode of
            0..2 : begin
                        prot_layersize := sizeOf (tHSRPStruct);

                        pData := pHSRPStruct (priv_framedata + priv_frameoffset);

                        // Assign data

                        priv_version := pData^.Version;

                        case pData^.OpCode of
                         0 : priv_opcode := hoHello;
                         1 : priv_opcode := hoCoup;
                         2 : priv_opcode := hoResign;
                        else
                            priv_opcode := hoUnknown;
                        end;

                        case pData^.State of
                          0 : priv_state := hsInitial;
                          1 : priv_state := hsLearn;
                          2 : priv_state := hsListen;
                          4 : priv_state := hsSpeak;
                          8 : priv_state := hsStandby;
                         16 : priv_state := hsActive;
                        else
                            priv_state := hsUnknown;
                        end;

                        priv_hellotime := pData^.HelloTime;
                        priv_holdtime := pData^.HoldTime;
                        priv_priority := pData^.Priority;
                        priv_group := pData^.Group;
                        priv_authdata := pData^.AuthData;
                        priv_virtualip.ValueAsLong := swap32 (pData^.VirtualIP);
                   end;
               3 : begin
                        prot_layersize := sizeOf (tHSRPAdvStruct);

                        pAdvData := pHSRPAdvStruct (priv_framedata + priv_frameoffset);

                        priv_opcode := hoAdvertise;

                        case pAdvData^.State of
                          0 : priv_state := hsInitial;
                          1 : priv_state := hsLearn;
                          2 : priv_state := hsListen;
                          4 : priv_state := hsSpeak;
                          8 : priv_state := hsStandby;
                         16 : priv_state := hsActive;
                        else
                            priv_state := hsUnknown;
                        end;

                        priv_AdvTLVType := swap16 (pAdvData^.TLVType);
                        priv_AdvTLVLength := swap16 (pAdvData^.TLVLength);
                        priv_AdvActiveGroupCount := swap16 (pAdvData^.ActiveGroupCount);
                        priv_AdvPassiveGroupCount := swap16 (pAdvData^.PassiveGroupCount);
                   end;
           end;

           // Prepare next anchor

           priv_nextAnchor.Offset := priv_frameoffset + prot_layersize;
           priv_nextAnchor.HeaderType := NEXT_HEADER_NONE;
        except
              status := prFailInvalidValues;
        end;

     result := status;
end;

(* ------------------------------------------- *)

end.
