{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_parser_icmpv6;

interface

uses base_structures,
     definition_ip_addresses,
     common_parse_helpers,
     class_parser_baseclass;

type tIPv6PrefixOptionFlags = (poOnLink, poAutonomous, poRouterAddress);
     tIPv6PrefixOptionFlagSet = set of tIPv6PrefixOptionFlags;
     tICMPv6RtrAdvtFlags = (rfManaged, rfOther, rfHomeAgent, rfProxy, rfReserved);
     tICMPv6RtrAdvtFlagSet = set of tICMPv6RtrAdvtFlags;
     tICMPv6RtrPreference = (rpNone, rpLow, rpMedium, rpHigh, rpReserved);
     tICMPv6NDAdvFlags = (afRouter, afSolicited, afOverride);
     tICMPv6NDAdvFlagSet = set of tICMPv6NDAdvFlags;

     pICMPv6Struct = ^tICMPv6Struct;
     tICMPv6Struct = packed record
                                  icmp_type     : byte;
                                  icmp_code     : byte;
                                  checksum      : word;
                            end;

     tICMPv6NDOption = record
                             option_type         : byte;
                             option_len          : byte;
                             MACAddress          : tMACAddress;
                             PrefixLength        : byte;
                             PrefixFlags         : tIPv6PrefixOptionFlagSet;
                             PrefixReserved      : byte;
                             PrefixValidLifetime : longword;
                             PrefixPrefLifetime  : longword;
                             PrefixAddress       : tIPv6Address;
                             DNSLifeTime         : longword;
                             DNSServerList       : array of tIPv6Address;
                             DNSDomainNameList   : array of tFQDN;
                             MTU                 : longword;

                             procedure initialize;
                       end;

     tICMPv6NDOptionList = array of tICMPv6NDOption;

     tICMPv6Parser = class (tFrameParser)

                           /// <summary>
                           ///   Constructor of the ICMPv6 parser class
                           /// </summary>
                           /// <param name="FrameData">
                           ///   Pointer to the Buffer holding the Frame data
                           /// </param>
                           /// <param name="FrameOffset">
                           ///   Offset of the TCP layer starting
                           /// </param>
                           /// <param name="DataLength">
                           ///   Total Size of the Frame data in the buffer.
                           ///   This is less or equal to Wirelength.
                           /// </param>
                           /// <param name="WireLength">
                           ///   Size of the frame on the wire
                           /// </param>
                           /// <param name="ICMPSectionLength">
                           ///   Length of the ICMP section, usually the IP
                           ///   payload length
                           /// </param>
                           /// <param name="FrameIndex">
                           ///   Number of the frame in the file
                           /// </param>
                           /// <example>
                           ///   tTCPParser.create (Frame.Data,
                           ///   ProtocolAnchor.Offset, Frame.FrameSize,
                           ///   TempValues.IPPayloadLen, Frame.FileFrameID);
                           /// </example>
                           constructor create (FrameData : pointer; const FrameOffset, DataLength, WireLength, ICMPSectionLength  : integer; const FrameIndex : integer = 0); reintroduce;

                           private
                                  priv_sectionlen    : integer;
                                  priv_lastoffset    : integer;     // the last available byte, either from sectionlength or framelength (whatever is less)

                                  priv_type          : byte;
                                  priv_code          : byte;
                                  priv_checksum      : word;
                                  priv_payload       : tRawBuffer;
                                  priv_payloadlength : integer;
                                  priv_payloadoffset : integer;

                                  // Optional values

                                  priv_echo_identifier         : word;
                                  priv_echo_sequence           : word;
                                  priv_next_hop_mtu            : longword;
                                  priv_problem_offset          : longword;
                                  priv_source_linklayer_addr   : tMACAddress;
                                  priv_target_linklayer_addr   : tMACAddress;
                                  priv_nd_target_address       : tIPv6Address;
                                  priv_redirect_target_address : tIPv6Address;
                                  priv_redirect_dest_address   : tIPv6Address;
                                  priv_router_hoplimit         : byte;
                                  priv_routeradv_flags         : tICMPv6RtrAdvtFlagSet;
                                  priv_router_preference       : tICMPv6RtrPreference;
                                  priv_router_lifetime         : word;
                                  priv_router_reachtime        : longword;
                                  priv_router_retranstime      : longword;
                                  priv_router_mtu              : longword;
                                  priv_neighboradv_flags       : tICMPv6NDAdvFlagSet;
                                  priv_icmpsectionwords        : tBufferWords;
                                  priv_optionlist              : tICMPv6NDOptionList;

                                  function getPayloadString : string;
                                  function getPayLoadWordBuffer : tBufferWords;
                                  function parseNDOptions (Offset : integer) : tProtocolAnchor;
                                  function determineNextAnchor (Offset : integer) : tProtocolAnchor;

                           public
                                 property ICMPType : byte read priv_type;
                                 property ICMPCode : byte read priv_code;
                                 property Checksum : word read priv_checksum;
                                 property EchoIdentifier : word read priv_echo_identifier;
                                 property EchoSequence : word read priv_echo_sequence;
                                 property NextHopMTU : longword read priv_next_hop_mtu;
                                 property ParameterProblemOffset : longword read priv_problem_offset;
                                 property SourceLinkLayerAddress : tMACAddress read priv_source_linklayer_addr;
                                 property TargetLinkLayerAddress : tMACAddress read priv_target_linklayer_addr;
                                 property NDTargetAddress : tIPv6Address read priv_nd_target_address;
                                 property RedirectTargetAddress : tIPv6Address read priv_redirect_target_address;
                                 property RedirectDestinationAddress : tIPv6Address read priv_redirect_dest_address;
                                 property RAHopLimit : byte read priv_router_hoplimit;
                                 property RAFlags : tICMPv6RtrAdvtFlagSet read priv_routeradv_flags;
                                 property RAPreference : tICMPv6RtrPreference read priv_router_preference;
                                 property RALifeTime : word read priv_router_lifetime;
                                 property RAReachableTime : longword read priv_router_reachtime;
                                 property RARetransTimer : longword read priv_router_retranstime;
                                 property RAMTU : longword read priv_router_mtu;
                                 property NeighborAdvertisementFlags : tICMPv6NDAdvFlagSet read priv_neighboradv_flags;

                                 property OptionList : tICMPv6NDOptionList read priv_optionlist;
                                 property Payload  : tRawBuffer read priv_payload;
                                 property PayloadAsString : string read getPayloadString;
                                 property PayloadAsWordBuffer : tBufferWords read getPayLoadWordBuffer;
                                 property PayloadLength : integer read priv_payloadlength;
                                 property HeaderAsWordBuffer : tBufferWords read priv_icmpsectionwords;

                                 function checkIsPlausible : boolean; override;
                                 function parse : tParseResult; override;
                     end;

implementation

uses SysUtils,
     base_messages,
     toolkit;

(* ------------------------------------------- *)
{ tICMPv6NDOption }
(* ------------------------------------------- *)

procedure tICMPv6NDOption.initialize;

begin
     option_type := 0;
     option_len := 0;
     MACAddress.initialize;
     PrefixLength := 0;
     PrefixFlags := [];
     PrefixReserved := 0;
     PrefixValidLifetime := 0;
     PrefixPrefLifetime := 0;
     PrefixAddress.initialize;
     MTU := 0;

     DNSLifeTime := 0;
     setLength (DNSServerList, 0);
     setLength (DNSDomainNameList, 0);
end;

(* ------------------------------------------- *)
{ tICMPv6Parser}
(* ------------------------------------------- *)

constructor tICMPv6Parser.create (FrameData: pointer; const FrameOffset, DataLength, WireLength, ICMPSectionLength, FrameIndex: integer);

begin
     inherited create (FrameData, FrameOffset, DataLength, Wirelength, FrameIndex);

     priv_type := 0;
     priv_code := 0;
     priv_checksum := 0;
     priv_payloadlength := 0;
     priv_payloadoffset := 0;

     priv_sectionlen := ICMPSectionLength;

     // calculate the end offset we can parse up to

     if (priv_frameoffset + priv_sectionlen) > priv_framelength  then
        priv_lastoffset := priv_framelength
     else
         priv_lastoffset := priv_frameoffset + priv_sectionlen - 1;

     setlength (priv_OptionList, 0)
end;

(* ------------------------------------------- *)

function tICMPv6Parser.checkIsPlausible: boolean;

begin
     result := true;
end;

(* ------------------------------------------- *)

function tICMPv6Parser.parseNDOptions (Offset : integer) : tProtocolAnchor;

var Option     : tICMPv6NDOption;
    bHelp      : pByte;
    plHelp     : pLongword;
    pMACAddr   : pMACBytes;
    pIPAddr    : pIPv6AddressByteStruct;
    FQDNResult : tFQDNResult;
    ParseDone  : boolean;
    boAbort    : boolean;
    len        : integer;
    optcount   : integer;
    help       : integer;
    offsetmod  : integer;
    i          : integer;

begin
     ParseDone := false;
     result.HeaderType := NEXT_HEADER_UNKNOWN;
     optcount := 0;
     boAbort := false;

     repeat
           // Length Check for option header (officially testing "offset + 2 > priv_lastoffset + 1")

           if (Offset + 1) > priv_lastoffset then
              begin
                   boAbort := true;

                   addDiagnosticMsg (msWarning, MSG_PARSER_OUT_OF_DATA, 'ICMPv6', offset.toString, priv_lastoffset.ToString);
              end
           else
               begin
                    Option.initialize;

                    Option.option_type := pByte (priv_framedata + Offset)^;
                    Option.option_len := pByte (priv_framedata + Offset + 1)^ * 8;

                    if (Offset + Option.option_len) > priv_lastoffset then
                       begin
                            boAbort := true;

                            addDiagnosticMsg (msWarning, MSG_PARSER_OUT_OF_DATA_FOR_LENGTH, 'ICMPv6', offset.toString, Option.option_len.ToString, priv_lastoffset.ToString);
                       end;
               end;

           if not boAbort then
              begin
                   case Option.option_type of
                    1,2 : begin
                               // Source/Target link layer address

                               pMACAddr := pMACBytes (priv_framedata + Offset + 2);

                               Option.MACAddress.RangeBytes := pMACAddr^;

                               priv_source_linklayer_addr := Option.MACAddress;
                          end;
                    3   : begin
                               // Prefix information

                               // Length

                               bHelp := pByte (priv_framedata + Offset + 2);
                               Option.PrefixLength := bHelp^;

                               // Flags

                               bHelp := pByte (priv_framedata + Offset + 3);

                               if (bHelp^ and $80) = $80 then
                                  Option.PrefixFlags := [poOnLink]
                               else
                                   Option.PrefixFlags := [];

                               if (bHelp^ and $40) = $40 then
                                  Option.PrefixFlags := Option.PrefixFlags + [poAutonomous];

                               if (bHelp^ and $20) = $20 then
                                  Option.PrefixFlags := Option.PrefixFlags + [poRouterAddress];

                               Option.PrefixReserved := (bHelp^ and $1f);

                               // Lifetimes

                               plHelp := pLongword (priv_framedata + Offset + 4);
                               Option.PrefixValidLifetime := Swap32 (plHelp^);

                               plHelp := pLongword (priv_framedata + Offset + 8);
                               Option.PrefixPrefLifetime := Swap32 (plHelp^);

                               // Prefix address

                               pIPAddr := pIPv6AddressByteStruct (priv_framedata + Offset + 16);
                               Option.PrefixAddress.SetValueFromByteStruct(pIPAddr^);
                               Option.PrefixAddress.MaskBitCount := Option.PrefixLength;
                          end;
                    4   : begin
                               // Redirect - nothing here except the option length
                               // which includes the quote, so we need to be careful on assembly

                               result.HeaderType := NEXT_HEADER_IPv6;
                          end;
                    5   : begin
                               // MTU (with 2 reserved bytes in front of it, making the modifier 2 + 2)

                               plHelp := pLongword (priv_framedata + Offset + 4);
                               Option.MTU := Swap32 (plHelp^);

                               priv_router_mtu := Option.MTU;
                          end;
                    25  : begin
                               // Recursive DNS option (with 2 reserved bytes in front of it, making the modifier 2 + 2)

                               plHelp := pLongword (priv_framedata + Offset + 4);
                               Option.DNSLifeTime := Swap32 (plHelp^);

                               // The first eight bytes are type, length, reserved, lifetime. The rest are addresses

                               help := (Option.option_len - 8) div 16;

                               setlength (Option.DNSServerList, help);

                               for i := 0 to help - 1 do
                                   begin
                                        pIPAddr := pIPv6AddressByteStruct (priv_framedata + Offset + 8 + (i * 16));

                                        Option.DNSServerList[i].SetValueFromByteStruct (pIPAddr^);
                                        Option.DNSServerList[i].MaskBitCount := 64;
                                   end;
                          end;
                    31  : begin
                               // DNS Search List option (with 2 reserved bytes in front of it, making the modifier 2 + 2)

                               plHelp := pLongword (priv_framedata + Offset + 4);
                               Option.DNSLifeTime := Swap32 (plHelp^);

                               help := 0;
                               offsetmod := 8;

                               repeat
                                     FQDNResult := readFQDNFromOffset (priv_framedata, priv_sectionlen, Offset, Offset + offsetmod);

                                     setLength (Option.DNSDomainNameList, help + 1);
                                     OPtion.DNSDomainNameList[help] := FQDNResult.Name;

                                     inc (help);
                                     inc (offsetmod, FQDNResult.OffsetIncrement);

                                     // Read the next byte after the FQDN to tell if it is padding

                                     bHelp := pByte (priv_framedata + Offset + offsetmod);

                               until (offsetmod >= Option.option_len) or (bHelp^ = 0);
                          end;
                   end;

                   inc (len, Option.option_len);
                   inc (Offset, Option.option_len);

                   // Check if we're done, either by reaching end of the frame, or by
                   // exceeding IPv6 payload length (in case of Ethernet trailers)

                   if (Offset >= priv_framelength) or ((Offset - priv_frameoffset) >= priv_sectionlen) then
                      ParseDone := true;

                   // Add Option to list

                   inc (optcount);
                   setlength (priv_optionlist, optcount);
                   priv_optionlist[optcount-1] := Option;
              end
           else
               ParseDone := true;

     until ParseDone;

     result.Offset := offset;
end;

(* ------------------------------------------- *)

function tICMPv6Parser.determineNextAnchor (Offset : integer) : tProtocolAnchor;

var anchor     : tProtocolAnchor;
    optanchor  : tProtocolAnchor;
    pbHelp     : pByte;
    bHelp      : Byte;
    pwHelp     : pWord;
    wHelp      : word;
    plHelp     : pLongword;
    lHelp      : longword;
    optionlen  : integer;
    pAddr      : pIPv6AddressByteStruct;

begin
     // Init values

     priv_echo_identifier := 0;
     priv_echo_sequence := 0;
     priv_next_hop_mtu := 0;
     priv_nd_target_address.initialize;
     priv_redirect_target_address.initialize;
     priv_redirect_dest_address.initialize;
     priv_router_hoplimit := 0;
     priv_routeradv_flags := [];
     priv_router_preference := rpNone;
     priv_router_lifetime := 0;
     priv_router_reachtime := 0;
     priv_router_retranstime := 0;
     priv_router_mtu := 0;
     setlength (priv_optionlist, 0);

     // Since there are always 4 bytes following the ICMP header, the next anchor starts at least at +4

     anchor.Offset := offset + 4;

     // Determine values depending on type

     case priv_type of
        1 : begin
                 // Destination Unreachable

                anchor.HeaderType := NEXT_HEADER_IPv6;
            end;
        2 : begin
                 // Packet too big

                 plHelp := pLongword (priv_framedata + Offset);
                 priv_next_hop_mtu := Swap32 (plHelp^);

                 anchor.HeaderType := NEXT_HEADER_IPv6;
            end;
        3 : begin
                 // Time Exceeded

                 anchor.HeaderType := NEXT_HEADER_IPv6;
            end;
        4 : begin
                 // Parameter Problem

                 plHelp := pLongword (priv_framedata + Offset);
                 priv_problem_offset := Swap32 (plHelp^);

                 anchor.HeaderType := NEXT_HEADER_IPv6;
            end;
      128,
      129 : begin
                 // Echo Request/Reply

                 pwHelp := pWord (priv_framedata + Offset);
                 priv_echo_identifier := Swap16 (pwHelp^);

                 pwHelp := pWord (priv_framedata + Offset + 2);
                 priv_echo_sequence := Swap16 (pwHelp^);

                 // Set correct next anchor offset and payload length

                 priv_payloadoffset := anchor.Offset;
                 priv_payloadlength := priv_framelength - priv_payloadoffset;

                 // Set correct next anchor offset

                 anchor.Offset := Offset + 4 + priv_payloadlength;
                 anchor.HeaderType := NEXT_HEADER_NONE;
            end;
      133 : begin
                 // Router Solicitation

                 anchor.HeaderType := NEXT_HEADER_NONE;

                 // Parse additional TLVs, at offset + 4 because of the 4 reserved bytes

                 optanchor := parseNDOptions (Offset + 4);

                 // When the options determined a new anchor, take it, otherwise keep the one we have

                 if optanchor.HeaderType <> NEXT_HEADER_UNKNOWN then
                    anchor.HeaderType := optanchor.HeaderType;

                 anchor.Offset := optanchor.Offset;
            end;
      134 : begin
                 // Router Advertisement

                 // HopLimit

                 pbHelp := pByte (priv_framedata +  Offset);
                 priv_router_hoplimit := pbHelp^;

                 // Flags

                 bHelp := pByte (priv_framedata +  Offset + 1)^;

                 if (bHelp and $80) = $80 then
                    priv_routeradv_flags := [rfManaged]
                 else
                     priv_routeradv_flags := [];

                 if (bHelp and $40) = $40 then
                    priv_routeradv_flags := priv_routeradv_flags + [rfOther];

                 if (bHelp and $20) = $20 then
                    priv_routeradv_flags := priv_routeradv_flags + [rfHomeAgent];

                 if (bHelp and $4) = $4 then
                    priv_routeradv_flags := priv_routeradv_flags + [rfProxy];

                 if (bHelp and $2) = $2 then
                    priv_routeradv_flags := priv_routeradv_flags + [rfReserved];

                 // Router Preference

                 bHelp := (bHelp and $18) shr 3;

                 case bHelp of
                  0 : priv_router_preference := rpMedium;
                  1 : priv_router_preference := rpHigh;
                  2 : priv_router_preference := rpReserved;
                  3 : priv_router_preference := rpLow;
                 end;

                 inc (Offset, 2);

                 // Timers

                 pwHelp := pWord (priv_framedata + Offset);
                 priv_router_lifetime := swap16 (pwHelp^);

                 plHelp := pLongword (priv_framedata + Offset + 2);
                 priv_router_reachtime := swap32 (plHelp^);

                 plHelp := pLongword (priv_framedata + Offset + 6);
                 priv_router_retranstime := swap32 (plHelp^);

                 inc (Offset, 10);

                 anchor.HeaderType := NEXT_HEADER_NONE;

                 // Parse additional option TLVs

                 optanchor := parseNDOptions (Offset);

                 // When the options determined a new anchor, take it, otherwise keep the one we have

                 if optanchor.HeaderType <> NEXT_HEADER_UNKNOWN then
                    anchor := optanchor;

                 anchor.Offset := optanchor.Offset;
            end;
      135 : begin
                 // Neighbor Solicitation

                 inc (Offset, 4);
                 pAddr := pIPv6AddressByteStruct (priv_framedata + Offset);

                 priv_nd_target_address.SetValueFromByteStruct (pAddr^);
                 priv_nd_target_address.MaskBitCount := 64;

                 inc (Offset, 16);

                 // Parse additional option TLVs

                 if offset < priv_lastoffset then
                    begin
                         optanchor := parseNDOptions (Offset);

                         // When the options determined a new anchor, take it, otherwise keep the one we have

                         if optanchor.HeaderType <> NEXT_HEADER_UNKNOWN then
                            anchor.HeaderType := optanchor.HeaderType;

                         anchor.Offset := optanchor.Offset;
                    end
                 else
                     begin
                          anchor.HeaderType :=  NEXT_HEADER_NONE;
                          anchor.Offset := offset;
                     end;
            end;
      136 : begin
                 // Neighbor Advertisement

                 plHelp := pLongword (priv_framedata + Offset);
                 lHelp := Swap32 (plHelp^);

                 // Flags

                 bHelp := lHelp shr 29;

                 if (bHelp and $4) = $4 then
                    priv_neighboradv_flags := [afRouter]
                 else
                     priv_neighboradv_flags := [];

                 if (bHelp and $2) = $2 then
                    priv_neighboradv_flags := priv_neighboradv_flags + [afSolicited];

                 if (bHelp and $1) = $1 then
                    priv_neighboradv_flags := priv_neighboradv_flags + [afOverride];

                 inc (Offset, 4);

                 pAddr := pIPv6AddressByteStruct (priv_framedata + Offset);
                 priv_nd_target_address.SetValueFromByteStruct(pAddr^);
                 priv_nd_target_address.MaskBitCount := 64;

                 inc (Offset, 16);

                 anchor.HeaderType := NEXT_HEADER_NONE;

                 // Parse additional option TLVs

                 optanchor := parseNDOptions (Offset);

                 // When the options determined a new anchor, take it, otherwise keep the one we have

                 if optanchor.HeaderType <> NEXT_HEADER_UNKNOWN then
                    anchor.HeaderType := optanchor.HeaderType;

                 anchor.Offset := optanchor.Offset;
            end;
      137 : begin
                 // Redirect

                 inc (Offset, 4);

                 pAddr := pIPv6AddressByteStruct (priv_framedata + Offset);

                 priv_redirect_target_address.SetValueFromByteStruct (pAddr^);
                 priv_redirect_target_address.MaskBitCount := 64;

                 inc (Offset, 16);

                 pAddr := pIPv6AddressByteStruct (priv_framedata + Offset);

                 priv_redirect_dest_address.SetValueFromByteStruct (pAddr^);
                 priv_redirect_dest_address.MaskBitCount := 64;

                 inc (Offset, 16);

                 optanchor := parseNDOptions (Offset);

                 // Since the redirect option holds the total length of all quoted layers it is much too large.
                 // So we set it manually instead, pointing at the next layer, so we can parse it instead of
                 // skipping over it

                 anchor.Offset := Offset + 8;
                 anchor.HeaderType := NEXT_HEADER_IPv6;
            end;
     else
         anchor.HeaderType := NEXT_HEADER_UNKNOWN;

//       // determine remaining payload length

//       priv_payloadlength := priv_framelength - priv_payloadoffset;
     end;

     result := anchor;
end;

(* ------------------------------------------- *)

function tICMPv6Parser.getPayLoadWordBuffer: tBufferWords;

begin
     result := GetRawBufferAsWordBuffer (priv_payload, priv_payloadlength);
end;

(* ------------------------------------------- *)

function tICMPv6Parser.getPayloadString: string;

var i     : integer;
    value : string;

begin
     value := '';

     for i := 1 to priv_payloadlength do
         value := value + chr (priv_payload[i]);

     result := value;
end;

(* ------------------------------------------- *)

function tICMPv6Parser.parse: tParseResult;

var pData    : pICMPv6Struct;
    status   : tParseResult;
    help     : word;
    pPayload : pByte;
    CRCsize  : integer;
    calcCRC  : boolean;

begin
     priv_osilayer := 4;     // Set Layer this parser is for

     status := prSuccess;

     if priv_framelength < priv_frameoffset + sizeOf (tICMPv6Struct) then
        status := prFailUndersize
     else
         try
            pData := pICMPv6Struct (priv_framedata + priv_frameoffset);

            priv_type := pData^.icmp_type;
            priv_code := pData^.icmp_code;

            help := pData^.checksum;

            priv_checksum := swap16 (help);

            // Determine substructures and then the next anchor and payload length

            priv_nextAnchor := determineNextAnchor (priv_frameoffset + sizeOf (tICMPv6Struct));

            // Get the payload if any

            if priv_payloadlength > 0 then
               begin
                    pPayload := pByte (priv_framedata + priv_payloadoffset);

                    move (pPayload^, priv_payload, priv_payloadlength);
               end;

            // Determine LayerSize

            prot_layersize := priv_sectionlen;
{
            case priv_type of
             128,
             129  : prot_layersize := sizeOf (tICMPv6Struct) + 4 + priv_payloadlength;
            else
//              prot_layersize := sizeOf (tICMPv6Struct) + 4;
            end;
}
            // Prepare the word array for crc calculation if possible - frame must be long enough
            // Note: the actual CRC verification is done in parsemaster, not here, because of the required Pseudoheader

            CalcCRC := (priv_framelength = priv_wirelength);

            // But not too long - segmentation offloading seems to be the death to getWordsStructFromOffset, so we skip it in that case

            if priv_framelength > 9018 then
               CalcCRC := false;               // TODO: see if we can manage ICMPv6 checksum calculation for large segments in a more elegant way

            if CalcCRC then
               begin
                    CRCsize := sizeOf (tICMPv6Struct) + 4 + priv_payloadlength;

//                  priv_icmpsectionwords := getWordsStructFromOffset (priv_FrameData, priv_frameOffset, sizeOf (tICMPv6Struct) + 4);
                    priv_icmpsectionwords := getWordsStructFromOffset (priv_FrameData, priv_frameOffset, priv_sectionlen);
               end
            else
                begin
                     status := prFailPayloadTruncated;
                     setlength (priv_icmpsectionwords, 0);
                end;

         except
               status := prFailInvalidValues;
         end;

     result := status;
end;

(* ------------------------------------------- *)

end.
