{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_parser_ipv4;

interface

uses base_structures,
     base_checksums,
     definition_ip_addresses,
     common_parse_helpers,
     class_parser_baseclass;


type tIPv4OptionClass = (ocControl, ocReserved1, ocDebugging, ocReserved2);

     pIPv4TimestampRecord = ^tIPv4TimestampRecord;
     tIPv4TimestampRecord = record
                                  address   : longword;
                                  timestamp : longword;
                            end;

     tIPv4Timestamp = record
                            address   : tIPv4Address;
                            timestamp : longword;
                      end;

     tIPv4Option = record
                         CopyOnFragment   : boolean;
                         OptionClass      : tIPv4OptionClass;
                         OptionNumber     : byte;
                         OptionPointer    : byte;
                         OptionOverflow   : byte;
                         OptionFlag       : byte;
                         OptionLength     : byte;
                         OptionWord       : word;
                         RouteRecords     : array of tIPv4Address;
                         TimestampRecords : array of tIPv4Timestamp;

                         procedure setOptionType (Value : byte);
                         function getOptionType : byte;
                         procedure copyFrom (Source : tIPv4Option);
                   end;

     tIPv4OptionList = array of tIPv4Option;

     tIPv4Parser  = class (tFrameParser)

                          {$REGION 'Documentation'}
                          ///	<summary>
                          ///	  Returns the IPv4 header as word array, as
                          ///	  needed for checksum calculation.
                          ///	</summary>
                          ///	<param name="Buffer">
                          ///	  Byte Pointer to the frame data
                          ///	</param>
                          ///	<param name="Offset">
                          ///	  Offset, at which the IPv4header starts
                          ///	</param>
                          ///	<param name="IHL">
                          ///	  IPv4 header length, including IPv4 options
                          ///	</param>
                          ///	<returns>
                          ///	  Array containing all IPv4 header bytes as
                          ///	  BigEndian formatted tIPv4HeaderWords
                          ///	</returns>
                          {$ENDREGION}
                          class function getHeaderWordsFromOffset (Buffer : pByte; Offset, IHL : integer) : tIPv4HeaderWords;

                          constructor create (FrameData : pointer; const FrameOffset, DataLength, WireLength : integer; isICMPQuote : boolean = false; const FrameIndex : integer = 0); reintroduce;

                          private
                                 priv_version         : byte;
                                 priv_ihl             : byte;
                                 priv_tos             : byte;
                                 priv_totallength     : word;
                                 priv_identification  : word;
                                 priv_flags           : tIPFlagSet;
                                 priv_offset          : word;
                                 priv_isFragment      : boolean;
                                 priv_ttl             : byte;
                                 priv_protocol        : byte;
                                 priv_checksum        : word;
                                 priv_source          : tIPv4Address;
                                 priv_destination     : tIPv4Address;
                                 priv_currentroute    : tIPv4Address;
                                 priv_hasOptions      : boolean;
                                 priv_headerwords     : tIPv4HeaderWords;
                                 priv_optionlist      : tIPv4OptionList;

                                 priv_segmentlenzero  : boolean;
                                 priv_payloaddeltalen : integer;
                                 priv_totaldeltalen   : integer;

                                 priv_isICMPQuote     : boolean;

                                 function DetermineIfLastFragment : boolean;

                                 function getIPv4Flags (FlagBits : word) : tIPFlagSet;
                                 function getIPPayloadLength: integer;
                                 function getIPv4PseudoHeader: tIPv4PseudoHeader;
                                 procedure ParseIPv4Options (Offset, OptionBytes : integer);

                          public
                                property Version : byte read priv_version;
                                property IHL : byte read priv_ihl;
                                property TOS : byte read priv_tos;
                                property TotalLength : word read priv_totallength;
                                property Identification : word read priv_identification;
                                property Flags : tIPFlagSet read priv_flags;
                                property Offset : word read priv_offset;
                                property IsFragment : boolean read priv_isFragment;
                                property IsLastFragment : boolean read DetermineIfLastFragment;
                                property TTL : byte read priv_ttl;
                                property Protocol : byte read priv_protocol;
                                property Checksum : word read priv_checksum;
                                property Source : tIPv4Address read priv_source;
                                property Destination : tIPv4Address read priv_destination;
                                property CurrentRoute : tIPv4Address read priv_currentroute;
                                property HasOptions : boolean read priv_hasOptions;
                                property isICMPQuote : boolean read priv_isICMPQuote;
                                property OptionList : tIPv4OptionList read priv_optionlist;
                                property IPPayloadLength : integer read getIPPayloadLength;
                                property HeaderAsWordBuffer : tIPv4HeaderWords read priv_headerwords;
                                property PseudoHeader : tIPv4PseudoHeader read getIPv4PseudoHeader;
                                /// <summary>
                                ///   True if the IP total length was zero,
                                ///   caused by segment offloading
                                /// </summary>
                                property SegmentOffloadLenZero : boolean read priv_segmentlenzero;

                                /// <summary>
                                ///   Tells how many bytes the IP payload is
                                ///   greater than IP total length (which means
                                ///   there's an Ethernet trailer), or is
                                ///   shorter than IP total length ( meaning,
                                ///   the frame was sliced). Normally zero when
                                ///   IP total length and remaining frame
                                ///   length match.
                                /// </summary>
                                property FramePayloadLengthDelta : integer read priv_payloaddeltalen;
                                /// <summary>
                                ///   Tells how many bytes the frame exceeds IP
                                ///   total length (which means there's an
                                ///   Ethernet trailer), or is shorter than IP
                                ///   total length ( meaning, the frame was
                                ///   sliced). Normally zero when IP total
                                ///   length and remaining frame length match.
                                /// </summary>
                                property FrameTotalLengthDelta : integer read priv_totaldeltalen;

                                function checkIsPlausible : boolean; override;
                                function parse : tParseResult; override;
                    end;

     pIPv4Struct = ^tIPv4Struct;
     tIPv4Struct = packed record
                                VersionIHL     : byte;
                                TOS            : byte;
                                TotalLength    : word;
                                Identification : word;
                                FlagsOffset    : word;
                                TTL            : byte;
                                Protocol       : byte;
                                Checksum       : word;
                                source         : longword;
                                destination    : longword;
                          end;

const IPV4_OPTION_EOL = 0;
      IPV4_OPTION_NOP = 1;
      IPV4_OPTION_SECURITY = 2;
      IPV4_OPTION_LOOSE_SRC_ROUTE = 3;
      IPV4_OPTION_TIMESTAMP = 4;
      IPV4_OPTION_EXTENTED_SECUCITY = 5;
      IPV4_OPTION_COMMERCIAL_SECURITY = 6;
      IPV4_OPTION_RECORD_ROUTE = 7;
      IPV4_OPTION_STREAM_IDENTIFIER = 8;
      IPV4_OPTION_STRICT_SRC_ROUTE = 9;
      IPV4_OPTION_EXPERIMENTAL_MEASUREMENT = 10;
      IPV4_OPTION_MTU_PROBE = 11;
      IPV4_OPTION_MTU_REPLY = 12;
      IPV4_OPTION_EXPERIMENTAL_FLOW_CONTROL = 13;
      IPV4_OPTION_EXPERIMENTAL_ACCESS_CONTROL = 14;
      IPV4_OPTION_ENCODE = 15;
      IPV4_OPTION_IMI_TRAFFIC_DESCRIPTIOR = 16;
      IPV4_OPTION_EXTENDED_INTERNET_PROTOCOL = 17;
      IPV4_OPTION_TRACEROUTE = 18;
      IPV4_OPTION_ADDRESS_EXTENSION = 19;
      IPV4_OPTION_ROUTER_ALERT = 20;
      IPV4_OPTION_SELECTIVE_DIRECTED_BROADCAST_MODE = 21;
      IPV4_OPTION_DYNAMIC_PACKET_STATE = 23;
      IPV4_OPTION_UPSTREAM_MULTICAST_PACKET = 24;
      IPV4_OPTION_QUICK_START = 25;

implementation

uses Toolkit,
     base_messages;

(* ------------------------------------------- *)
{ tIPv4Option }
(* ------------------------------------------- *)

function tIPv4Option.getOptionType: byte;

var value : byte;
    help  : byte;

begin
     if CopyOnFragment then
        value := 128
     else
         value := 0;

     // Set Class as 2 bit value at bits 6 and 7. Control is 0, so we can ignore it.

     case OptionClass of
      ocReserved1 : value := (value or 32);
      ocDebugging : value := (value or 64);
      ocReserved2: value := (value or 96);
     end;

     // Set Number

     value := value or OptionNumber;
     result := value;
end;

(* ------------------------------------------- *)

procedure tIPv4Option.setOptionType (Value: byte);

var help : byte;

begin
     CopyOnFragment := ((Value and $80) = $80);

     help := (Value and $60) shr 5;

     case help of
      0 : OptionClass := ocControl;
      1 : OptionClass := ocReserved1;
      2 : OptionClass := ocDebugging;
      3 : OptionClass := ocReserved2;
     end;

     OptionNumber := (Value and $1f);

     // Init arrays

     setlength (RouteRecords, 0);
     setlength (TimestampRecords, 0);
end;

(* ------------------------------------------- *)

procedure tIPv4Option.copyFrom(Source: tIPv4Option);

var i   : integer;

begin
     CopyOnFragment := Source.CopyOnFragment;
     OptionClass := Source.OptionClass;
     OptionNumber := Source.OptionNumber;
     OptionPointer := Source.OptionPointer;
     OptionOverflow := Source.OptionOverflow;
     OptionFlag := Source.OptionFlag;
     OptionLength := Source.OptionLength;

     setlength (RouteRecords, length (Source.RouteRecords));

     for i := 1 to Length (RouteRecords) do
         begin
              RouteRecords[i-1].ValueAsLong := Source.RouteRecords[i-1].ValueAsLong;
              RouteRecords[i-1].MaskBitCount := Source.RouteRecords[i-1].MaskBitCount;
         end;

     setlength (TimestampRecords, length (Source.TimestampRecords));

     for i := 1 to Length (TimestampRecords) do
         begin
              TimestampRecords[i-1].address.ValueAsLong := Source.TimestampRecords[i-1].address.ValueAsLong;
              TimestampRecords[i-1].address.MaskBitCount := Source.TimestampRecords[i-1].address.MaskBitCount;
              TimestampRecords[i-1].timestamp := Source.TimestampRecords[i-1].timestamp;
         end;
end;

(* ------------------------------------------- *)
{ IPv4Parser }
(* ------------------------------------------- *)

function tIPv4Parser.checkIsPlausible: boolean;

var pData   : pIPv4Struct;
    bhelp   : byte;
    wHelp   : word;

begin
     if priv_framelength >= priv_frameoffset + sizeOf (tIPv4Struct) then
        begin
             result := true;

             pData := pIPv4Struct (priv_framedata + priv_frameoffset);

            // Get Version

            bHelp := pData^.VersionIHL and $f0;
            bHelp := bHelp shr 4;

            if bHelp <> 4 then
               exit (false);

            // Get IHL - specified as number of 32bit (=4 Byte) values in the header

            bhelp := pData^.VersionIHL and $0f;
            bhelp := bhelp * 4;

            if bHelp < 20 then
               exit (false);

            // Get total Length (0 can be caused by offloading, but other values less than 20 are not good)

            whelp := pData^.TotalLength;
            whelp := swap16 (whelp);

            result := not ((whelp > 0) and (whelp < 20));
        end
     else
         result := false;
end;

(* ------------------------------------------- *)

constructor tIPv4Parser.create (FrameData: pointer; const FrameOffset, DataLength, WireLength: integer; isICMPQuote : boolean; const FrameIndex: integer);

begin
     inherited create (FrameData, FrameOffset, DataLength, WireLength, FrameIndex);

     priv_isICMPQuote := isICMPQuote;
end;

(* ------------------------------------------- *)

function tIPv4Parser.DetermineIfLastFragment: boolean;

begin
     result := not (ipfMoreFragments in priv_flags) and (priv_offset > 0);
end;

(* ------------------------------------------- *)

class function tIPv4Parser.getHeaderWordsFromOffset (Buffer : pByte; Offset, IHL: integer): tIPv4HeaderWords;

var helpbuf : pIPv4HeaderWords;
    i       : integer;

begin
     helpbuf := pIPv4HeaderWords (Buffer + Offset);
     result := helpbuf^;

     for i := 1 to 30 do    // 15 * 32 bit = 30 * 16 bit
         begin
              // Make sure to set all unneeded words to zero, and swap the others

              if i > (IHL div 2) then
                 result[i] := 0
              else
                  result[i] := swap16 (result[i]);
         end;
end;

(* ------------------------------------------- *)

function tIPv4Parser.getIPv4Flags (FlagBits : word) : tIPFlagSet;

var flagset : tIPFlagSet;

begin
     if (FlagBits and $8000) = $8000 then
        flagset := [ipfReserved]
     else
         flagset := [];

     if (FlagBits and $4000) = $4000 then
        flagset := flagset + [ipfDontFragment];

     if (FlagBits and $2000) = $2000 then
        flagset := flagset + [ipfMoreFragments];

     result := flagset;
end;

(* ------------------------------------------- *)

function tIPv4Parser.getIPPayloadLength: integer;

begin
     result := priv_totallength - priv_ihl;
end;

(* ------------------------------------------- *)

procedure tIPv4Parser.ParseIPv4Options (Offset, OptionBytes : integer);

var pOptionType  : pByte;
    pOptionLen   : pByte;
    pHelp        : pByte;
    Help         : Byte;
    pHelpLong    : pLongword;
    LocalOffset  : integer;
    count        : integer;
    reccount     : integer;
    i            : integer;
    timerec      : pIPv4TimestampRecord;
    OptionEntry  : tIPv4Option;
    boDone       : boolean;
    boUnknown    : boolean;

begin
     LocalOffset := Offset;
     setlength (priv_optionlist, 0);
     count := 0;
     boDone := false;

     repeat
           // Get basic option details

           pOptionType := pByte (priv_frameData + LocalOffset);

           if pOptionType^ > 1 then
              begin
                   boUnknown := false;          // tracks if we can handle the option

                   pOptionLen := pByte (priv_frameData + LocalOffset + 1);

                   OptionEntry.OptionLength := pOptionLen^;
                   OptionEntry.setOptionType (pOptionType^);

                   // Parse option values

                   case OptionEntry.OptionNumber of
                    IPV4_OPTION_EOL              : begin
                                                        boDone := true;
                                                   end;
                    IPV4_OPTION_LOOSE_SRC_ROUTE,
                    IPV4_OPTION_STRICT_SRC_ROUTE : begin
                                                        // Option Pointer

                                                        pHelp := pByte (priv_frameData + LocalOffset + 2);
                                                        OptionEntry.OptionPointer := pHelp^;

                                                        // Records

                                                        reccount := (pOptionLen^ - 3) div 4;

                                                        with OptionEntry do
                                                             begin
                                                                  setLength (RouteRecords, reccount);

                                                                  for i := 1 to reccount do
                                                                      begin
                                                                           pHelpLong := pLongword (priv_frameData + LocalOffset + 3 + (4 * (i-1)));

                                                                           RouteRecords[i-1].ValueAsLong := pHelpLong^;
                                                                           RouteRecords[i-1].ValueAsLong := swap32 (RouteRecords[i-1].ValueAsLong);
                                                                           RouteRecords[i-1].MaskBitCount := 32;
                                                                      end;

                                                                  // Update IP addresses, since the last IP in the list is the real destination, and
                                                                  // the one we have as destination is only the next hop

                                                                  priv_currentroute := priv_destination;
                                                                  priv_destination := RouteRecords[reccount-1];
                                                             end;
                                                   end;
                    IPV4_OPTION_TIMESTAMP        : begin
                                                        // Option Pointer

                                                        pHelp := pByte (priv_frameData + LocalOffset + 2);
                                                        OptionEntry.OptionPointer := pHelp^;

                                                        // Option Overflow and Flag

                                                        pHelp := pByte (priv_frameData + LocalOffset + 3);
                                                        Help := pHelp^;

                                                        OptionEntry.OptionOverflow := (Help and $f0) shr 4;
                                                        OptionEntry.OptionFlag := (Help and $f);

                                                        // Records

                                                        reccount := (pOptionLen^ - 4) div 8;

                                                        with OptionEntry do
                                                             begin
                                                                  setLength (TimestampRecords, reccount);

                                                                  for i := 1 to reccount do
                                                                      begin
                                                                           timerec := pIPv4TimestampRecord (priv_frameData + LocalOffset + 4 + (8 * (i-1)));

                                                                           TimestampRecords[i-1].address.ValueAsLong := timerec^.address;
                                                                           TimestampRecords[i-1].timestamp := timerec^.timestamp;

                                                                           TimestampRecords[i-1].address.valueaslong := swap32 (TimestampRecords[i-1].address.valueaslong);
                                                                           TimestampRecords[i-1].timestamp := swap32 (TimestampRecords[i-1].timestamp);
                                                                      end;
                                                             end;
                                                   end;
                    IPV4_OPTION_RECORD_ROUTE     : begin
                                                        // Option Pointer

                                                        pHelp := pByte (priv_frameData + LocalOffset + 2);
                                                        OptionEntry.OptionPointer := pHelp^;

                                                        // Records

                                                        reccount := (pOptionLen^ - 3) div 4;

                                                        with OptionEntry do
                                                             begin
                                                                  setLength (RouteRecords, reccount);

                                                                  for i := 1 to reccount do
                                                                      begin
                                                                           pHelpLong := pLongword (priv_frameData + LocalOffset + 3 + (4 * (i-1)));

                                                                           RouteRecords[i-1].ValueAsLong := pHelpLong^;
                                                                           RouteRecords[i-1].ValueAsLong := swap32 (RouteRecords[i-1].ValueAsLong);
                                                                           RouteRecords[i-1].MaskBitCount := 32;
                                                                      end;
                                                             end;
                                                   end;
                    IPV4_OPTION_ROUTER_ALERT     : begin
                                                        // Option Pointer

                                                        pHelp := pByte (priv_frameData + LocalOffset + 2);
                                                        OptionEntry.OptionPointer := pHelp^;

                                                        OptionEntry.OptionWord := pWord (priv_frameData + LocalOffset + 2)^;
                                                   end
                   else
                       boUnknown := true;
                   end;

                   if not boUnknown then
                      begin
                           inc (count);
                           setLength (priv_optionlist, count);
                           priv_optionlist[count-1] := OptionEntry;
                      end;

                   inc (LocalOffset, pOptionLen^);

                   if (LocalOffset - Offset) >= OptionBytes then
                      boDone := true;
              end
           else
               inc (LocalOffset);

     until boDone;
end;

(* ------------------------------------------- *)

function tIPv4Parser.parse: tParseResult;

var pData   : pIPv4Struct;
    status  : tParseResult;
    help    : word;
    lhelp   : longword;
    prio    : integer;
    id      : integer;
    remain  : integer;
//  helpbuf : pIPv4HeaderWords;
    i       : integer;

begin
     priv_osilayer := 3;     // Set Layer this parser is for

     status := prSuccess;

     if priv_framelength < priv_frameoffset + sizeOf (tIPv4Struct) then
        status := prFailUndersize
     else
         try
            pData := pIPv4Struct (priv_framedata + priv_frameoffset);

            // Get Version

            priv_version := pData^.VersionIHL and $f0;
            priv_version := priv_version shr 4;

            // Get IHL - specified as number of 32bit (=4 Byte) values in the header

            help := pData^.VersionIHL and $0f;
            priv_ihl := help * 4;

            // See if there are options

            priv_hasOptions := (priv_ihl > 20);

            // Get TOS

            priv_tos := pData^.TOS;

            // Get Length

            help := pData^.TotalLength;
            priv_totallength := swap16 (help);

            // sometimes total length is reported as 0, when segmentation offloading
            // was active, so we need to get around that

            if priv_totallength = 0 then
               begin
                    priv_totallength := (priv_framelength - priv_frameoffset);
                    priv_payloaddeltalen := 0;
                    priv_totaldeltalen := 0;
                    priv_segmentlenzero := true;

                    addDiagnosticMsg (msWarning, MSG_IP_TOTAL_LENGTH_ZERO);
               end
            else
                begin
                     priv_segmentlenzero := false;

                     if not priv_isICMPQuote then
                        begin
                             priv_totaldeltalen := (priv_wirelength - priv_frameoffset) - priv_totallength;

                             // Check if we have truncated frames or trailing bytes
                             // So calculcate how many IP payload bytes we have and then
                             // substract how many we should have

                             remain := (priv_framelength - (priv_frameoffset + priv_ihl));
                             priv_payloaddeltalen := (remain - (priv_totallength - priv_ihl));
                        end
                     else
                         begin
                              priv_totaldeltalen := 0;
                              priv_payloaddeltalen := 0;
                         end;
                end;

            // Get IP ID

            help := pData^.Identification;
            priv_identification := swap16 (help);

            // Get Flags and Offset

            help := pData^.FlagsOffset;
            help := swap16 (help);

            priv_offset := (help and $1fff) * 8;     // measured in units of 8 octets

            help := help and $e000;

            priv_flags := getIPv4Flags (help);

            // Determine if this is a fragment

            if not priv_isICMPQuote then
               priv_isFragment := (ipfMoreFragments in priv_flags) or (priv_offset > 0)
            else
                priv_isFragment := false;

            // TTL and Protocol

            priv_ttl := pData^.TTL;
            priv_protocol := pData^.Protocol;

            // Determine next anchor

            priv_nextAnchor.Offset := priv_frameoffset + priv_ihl;

            if (priv_offset = 0) then
               priv_nextAnchor.HeaderType := priv_protocol
            else
                begin
                     // this is a fragment, and it's not the first one, so we don't know what follows after the header

                     priv_nextAnchor.HeaderType := NEXT_HEADER_PAYLOAD;
                end;

            // Checksum

            help := pData^.Checksum;
            priv_checksum := swap16 (help);

            // Source and Destination Addresses

            lhelp := pData^.source;
            priv_source.ValueAsLong := swap32 (lhelp);
            priv_source.MaskBitCount := 32;

            lhelp := pData^.destination;
            priv_destination.ValueAsLong := swap32 (lHelp);
            priv_destination.MaskBitCount := 32;

            // Get word structure of header (needed for checksum calculation)

            priv_headerwords := getHeaderWordsFromOffset (priv_framedata, priv_frameoffset, priv_IHL);

            if priv_hasOptions then
               ParseIPv4Options (priv_frameoffset + 20, (priv_ihl - 20));

         except
               status := prFailInvalidValues;
         end;

     prot_layersize := priv_ihl;

     result := status;
end;

(* ------------------------------------------- *)

function tIPv4Parser.getIPv4PseudoHeader: tIPv4PseudoHeader;

var temp : tIPv4PseudoHeader;

begin
     temp.Source := priv_source.ValueAsLong;
     temp.Destination := priv_destination.ValueAsLong;
     temp.Protocol := priv_protocol;
     temp.L4Length := getIPPayloadLength;

     result := temp;
end;

(* ------------------------------------------- *)

end.
