{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_parser_ipv6;

interface

uses base_structures,
     definition_ip_addresses,
     base_checksums,
     common_parse_helpers,
     class_parser_baseclass;

type pIPv6Struct = ^tIPv6Struct;
     tIPv6Struct = packed record
                                VersionClassLabel    : longword;
                                PayloadLength        : word;
                                NextHeader           : byte;
                                HopLimit             : byte;
                                SourceAddress        : tIPv6AddressByteStruct;
                                DestinationAddress   : tIPv6AddressByteStruct;
                          end;

     tIPv6Parser  = class (tFrameParser)

                          private
                                 priv_version       : byte;
                                 priv_trafficclass  : byte;
                                 priv_flowlabel     : longword;
                                 priv_payloadlength : word;
                                 priv_nextheader    : byte;
                                 priv_hoplimit      : byte;
                                 priv_source        : tIPv6Address;
                                 priv_destination   : tIPv6Address;

                                 {$REGION 'Documentation'}
                                 ///	<summary>
                                 ///	  Helper function to swap the IPv6 word
                                 ///	  values to BigEndian
                                 ///	</summary>
                                 ///	<param name="Address">
                                 ///	  IPv6 Address
                                 ///	</param>
                                 {$ENDREGION}
                                 function swapAddressBlocks (Address : tIPv6AddressWordStruct) : tIPv6AddressWordStruct;
                                 function getIPv6PseudoHeader: tIPv6PseudoHeader;

                          public
                                property Version : byte read priv_version;
                                property TrafficClass : byte read priv_trafficclass;
                                property FlowLabel : longword read priv_flowlabel;
                                property PayloadLength : word read priv_payloadlength;
                                property NextHeader : byte read priv_nextheader;
                                property HopLimit : byte read priv_hoplimit;
                                property Source : tIPv6Address read priv_source;
                                property Destination : tIPv6Address read priv_destination;
                                property PseudoHeader : tIPv6PseudoHeader read getIPv6PseudoHeader;

                                function checkIsPlausible : boolean; override;
                                function parse : tParseResult; override;
                    end;


implementation

uses Toolkit;

(* ------------------------------------------- *)
{ IPv6Parser }
(* ------------------------------------------- *)

function tIPv6Parser.checkIsPlausible: boolean;

begin
     result := true;
end;

(* ------------------------------------------- *)

function tIPv6Parser.swapAddressBlocks (Address: tIPv6AddressWordStruct): tIPv6AddressWordStruct;

var i    : integer;
    addr : tIPv6AddressWordStruct;

begin
     for i := 0 to 7 do
         addr[i] := swap16 (Address[i]);

     result := addr;
end;

(* ------------------------------------------- *)

function tIPv6Parser.parse: tParseResult;

var pData  : pIPv6Struct;
    status : tParseResult;
    help   : word;
    vcl    : longword;
    lhelp  : longword;
    addr   : tIPv6AddressByteStruct;

begin
     priv_osilayer := 3;     // Set Layer this parser is for
     prot_layersize := 40;

     status := prSuccess;

     if priv_framelength < priv_frameoffset + sizeOf (tIPv6AddressWordStruct) then
        status := prFailUndersize
     else
         try
            pData := pIPv6Struct (priv_framedata + priv_frameoffset);

            vcl := pData^.VersionClassLabel;
            vcl := swap32 (vcl);

            // Get Version

            lHelp := vcl and $f0000000;
            lHelp := lHelp shr 28;
            priv_version := lHelp;

            // Get Traffic Class

            lHelp := vcl and $ff00000;
            lHelp := lHelp shr 20;
            priv_trafficclass := lHelp;

            // Get Flow Label

            lHelp := vcl and $fffff;
            priv_flowlabel := lHelp;

            // Payload Length

            help := pData^.PayloadLength;
            priv_payloadlength := swap16 (help);

            // Next Header and hop limit

            priv_nextheader := pData^.NextHeader;
            priv_hoplimit := pData^.HopLimit;

            // Addresses

            priv_source.SetValueFromByteStruct (pData^.SourceAddress);
            priv_source.MaskBitCount := 64;

            priv_destination.SetValueFromByteStruct (pData^.DestinationAddress);
            priv_destination.MaskBitCount := 64;

            // Anchor

            priv_nextAnchor.HeaderType := priv_nextheader;
            priv_nextAnchor.Offset := priv_frameoffset + sizeOf (tIPv6Struct);

         except
               status := prFailInvalidValues;
         end;

     result := status;
end;

(* ------------------------------------------- *)

function tIPv6Parser.getIPv6PseudoHeader: tIPv6PseudoHeader;

var temp : tIPv6PseudoHeader;

begin
     temp.Source := priv_source.ValueAsWordStruct;
     temp.Destination := priv_destination.ValueAsWordStruct;
     temp.L4Length := priv_payloadlength;

     if (priv_nextheader = NEXT_HEADER_TCP) or (priv_nextheader = NEXT_HEADER_UDP)
        or (priv_nextheader = NEXT_HEADER_IPV6_ICMP) or (priv_nextheader = NEXT_HEADER_IPV6_FRAGMENTATION) then
        temp.NextHeader := priv_nextheader
     else
         temp.NextHeader := 0;            // signaling, that extension headers need to be checked

     result := temp;
end;

(* ------------------------------------------- *)

end.
