{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_parser_juniper;

interface

uses base_structures,
     common_parse_helpers,
     class_parser_baseclass;

type pJuniperStruct = ^tJuniperStruct;
     tJuniperStruct = packed record
                                   MagicNumberFlags  : longword;
                                   ExtensionTotalLen : word;
                      end;

     tJuniperExtension = record
                               ExtType    : byte;
                               ExtLen     : byte;
                               ExtValue32 : longword;

                               procedure initialize;
                         end;

     tJuniperExtensionList = array of tJuniperExtension;

     tJuniperParser  = class (tFrameParser)

                                 private
                                        priv_MagicNumber       : longword;
                                        priv_Direction         : tPacketDirection;
                                        priv_L2HeaderStripped  : boolean;
                                        priv_ExtensionsPresent : boolean;
                                        priv_ExtensionTotalLen : word;

                                        priv_ExtensionList : tJuniperExtensionList;

                                        procedure ParseExtensions (Offset, TotalLen : integer);

                                 public
                                       property MagicNumber : longword read priv_MagicNumber;
                                       property Direction : tPacketDirection read priv_Direction;
                                       property Layer2HeaderStripped : boolean read priv_L2HeaderStripped;
                                       property ExtensionsPresent : boolean read priv_ExtensionsPresent;
                                       property ExtensionTotalLength : word read priv_ExtensionTotalLen;

                                       property ExtensionList : tJuniperExtensionList read priv_ExtensionList;

                                       function checkIsPlausible : boolean; override;
                                       function parse : tParseResult; override;
                           end;


implementation

uses toolkit;

(* ------------------------------------------- *)
{ tJuniperExtension }
(* ------------------------------------------- *)

procedure tJuniperExtension.initialize;

begin
     ExtType := 0;
     ExtLen := 0;
     ExtValue32 := 0;
end;

(* ------------------------------------------- *)
{ tJuniperMetaParser }
(* ------------------------------------------- *)

function tJuniperParser.checkIsPlausible: boolean;

var pHelp : pLongword;
    value : boolean;

begin
     value := (priv_framelength >= (priv_frameoffset + sizeOf (tJuniperStruct)));

     if value then
        begin
             // Check magic

             pHelp := pLongword (priv_framedata + priv_frameoffset);

             value := (swap32 (pHelp^) and $ffffff00) = $4d474300;
        end;

     result := value;
end;

(* ------------------------------------------- *)

procedure tJuniperParser.ParseExtensions (Offset, TotalLen: integer);

var Extension : tJuniperExtension;
    HelpByte  : pByte;
    HelpLong  : pLongword;
    boDone    : boolean;
    EndOffset : integer;
    len       : integer;

begin
     boDone := false;
     EndOffset := Offset + TotalLen;

     repeat
           HelpByte := pByte (priv_framedata + Offset);
           Extension.ExtType := HelpByte^;

           HelpByte := pByte (priv_framedata + Offset + 1);
           Extension.ExtLen := HelpByte^;

           case Extension.ExtType of
            1 : begin
                     // Device Interface Index
                end;
            2 : begin
                     // Device Interface Name
                end;
            3 : begin
                     // Device Media Type
                end;
            4 : begin
                     // Logical Interface Index

                     HelpLong := pLongword (priv_framedata + Offset + 2);

                     Extension.ExtValue32 := HelpLong^;
                end;
            5 : begin
                     // Logical Unit Number
                end;
            6 : begin
                     // Logical Interface Encapsulation
                end;
            7 : begin
                     // TTP derived Device Media Type
                end;
            8 : begin
                     // TTP derived Logical Interface Encapsulation
                end;
           end;

           // If we parsed something, add it to the list

           if Extension.ExtType > 0 then
              begin
                   len := length (priv_ExtensionList) + 1;
                   setLength (priv_ExtensionList, len);

                   priv_ExtensionList[len-1] := Extension;
              end;

           inc (Offset, 2 + Extension.ExtLen);

           boDone := (Offset >= EndOffset);

     until boDone;

end;

(* ------------------------------------------- *)

function tJuniperParser.parse: tParseResult;

var pData    : pJuniperStruct;
    status   : tParseResult;
    HelpWord : word;
    HelpLong : longword;

begin
     priv_osilayer := 1;     // Set Layer this parser is for
     prot_layersize := sizeOf (tJuniperStruct);

     // Preset the next anchor type as Ethernet

     priv_nextAnchor.HeaderType := NEXT_HEADER_ETHERNET;

     status := prSuccess;

     if priv_framelength < (priv_frameoffset + sizeOf (tJuniperStruct)) then
        status := prFailUndersize
     else
         try
            pData := pJuniperStruct (priv_framedata + priv_frameoffset);

            HelpLong := swap32 (pData^.MagicNumberFlags);
            priv_MagicNumber := HelpLong shr 8;

            HelpLong := HelpLong and $ff;

            // Direction

            if (HelpLong and 1) = 1 then
               priv_Direction := pdInbound
            else
                priv_Direction := pdOutbound;

            // L2 Header Present or stripped

            priv_L2HeaderStripped := ((HelpLong and 2) = 2);

            if priv_L2HeaderStripped then
               priv_nextAnchor.HeaderType := NEXT_HEADER_IPv4;

            // Extensions

            priv_ExtensionsPresent := ((HelpLong and $80) = $80);

            if priv_ExtensionsPresent then
               begin
                    priv_ExtensionTotalLen := swap16 (pData^.ExtensionTotalLen);

                    // If the there is no L2 header, add 4 bytes. This is taken from the Wireshark dissector, but
                    // it had no explanation why these 4 bytes had to be added. Maybe padding of some sort.

                    if priv_L2HeaderStripped then
                       prot_layersize := sizeOf (tJuniperStruct) + priv_ExtensionTotalLen + 4
                    else
                        prot_layersize := sizeOf (tJuniperStruct) + priv_ExtensionTotalLen;

                    // Parse Extensions

                    ParseExtensions (priv_frameoffset + sizeof (tJuniperStruct), priv_ExtensionTotalLen);
               end
            else
                priv_ExtensionTotalLen := 0;

            // Determine next Anchor Offset

            priv_nextAnchor.Offset := priv_frameoffset + prot_layersize;

         except
               status := prFailInvalidValues;
         end;

     result := status;
end;

(* ------------------------------------------- *)


end.
