{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_parser_linux_sll;

interface

uses base_structures,
     definition_ip_addresses,
     common_parse_helpers,
     class_parser_baseclass;


// http://www.tcpdump.org/linktypes/LINKTYPE_LINUX_SLL.html

type tSLLPacketType = (sptUnicastToUs, sptBroadcast, sptMulticast, sptUnicastToOtherHost, sptSentByUs);
     pSLLStruct = ^tSLLStruct;
     tSLLStruct = packed record
                               PacketType        : word;
                               ARPHdrType        : word;
                               LinkLayerAddrLen  : word;
                               LinkLayerAddr     : uint64;
                               ProtocolType      : word;
                  end;

     tLinuxSLLParser  = class (tFrameParser)

                              private
                                     priv_packettype         : tSLLPacketType;
                                     priv_ARPHdrType         : word;
                                     priv_LinkLayerAddrLen   : word;
                                     priv_MACAddress         : tMACAddress;
                                     priv_OtherLinkLayerAddr : uint64;
                                     priv_ProtocolType       : word;

                                     priv_containsVLANInfo   : boolean;
                                     priv_vlandata           : t802dot1qData;

                              public
                                    property PacketType : tSLLPacketType read priv_packettype;
                                    property ARPHeaderType : word read priv_ARPHdrType;
                                    property LinkLayerAddressLength : word read priv_LinkLayerAddrLen;
                                    property MACAddress : tMACAddress read priv_MACAddress;
                                    property LinkLayerAddress : uint64 read priv_OtherLinkLayerAddr;
                                    property ProtocolType : word read priv_ProtocolType;
                                    property ContainsVLANInfo : boolean read priv_containsVLANInfo;
                                    property VLANInfo : t802dot1qData read priv_vlandata;

                                    function checkIsPlausible : boolean; override;
                                    function parse : tParseResult; override;
                        end;


implementation

uses Math,
     Toolkit;

(* ------------------------------------------- *)
{ tLinuxSLLParser }
(* ------------------------------------------- *)

function tLinuxSLLParser.checkIsPlausible: boolean;

begin
     result := true;
end;

(* ------------------------------------------- *)

function tLinuxSLLParser.parse: tParseResult;

var pData    : pSLLStruct;
    status   : tParseResult;
    i        : integer;
    help16   : word;

begin
     priv_osilayer := 2;     // Set Layer this parser is for
     prot_layersize := sizeOf (tSLLStruct);

     status := prSuccess;

     if priv_framelength < priv_frameoffset + sizeOf (tSLLStruct) then
        status := prFailUndersize
     else
         try
            pData := pSLLStruct (priv_framedata + priv_frameoffset);

            help16 := swap16 (pData^.PacketType);

            case help16 of
             0 : priv_packettype := sptUnicastToUs;
             1 : priv_packettype := sptBroadcast;
             2 : priv_packettype := sptMulticast;
             3 : priv_packettype := sptUnicastToOtherHost;
             4 : priv_packettype := sptSentByUs;
            end;

            priv_ARPHdrType := swap16 (pData^.ARPHdrType);
            priv_LinkLayerAddrLen := Swap16 (pData^.LinkLayerAddrLen);
            priv_OtherLinkLayerAddr := Swap64 (pData^.LinkLayerAddr);

            // In case of a 6 byte long address, we assume it's a MAC and make it easier to access it

            if priv_LinkLayerAddrLen = 6 then
               begin
                    for i := 0 to 5 do
                        priv_MACAddress.RangeBytes[i+1] := pByte (priv_framedata + priv_frameoffset + 6 + i)^
               end
            else
                priv_MACAddress := VALUE_ZERO_MAC;

            // Check for modified SLL using the 2 remaining address bytes as VLAN indicator

            help16 := swap16 (pWord (priv_framedata + priv_frameoffset + 12)^);

            if (help16 = $8100) then
               begin
                    priv_containsVLANInfo := true;

                    // Need to parse VLAN info manually here, as there is two trailing bytes
                    // otherwise we could have just used the 802.1q parser

                    help16 := swap16 (pWord (priv_framedata + priv_frameoffset + 14)^);
                    priv_vlandata := Parse802dot1qData (help16);

                    // The next protocol is now at a different place, too

                    priv_ProtocolType := Swap16 (pWord (priv_framedata + priv_frameoffset + 18)^);

                    // Prepare special Anchor

                    priv_nextAnchor.HeaderType := priv_ProtocolType;
                    priv_nextAnchor.Offset := priv_frameoffset + sizeOf (tSLLStruct) + 4;
               end
            else
                begin
                     priv_containsVLANInfo := false;
                     priv_vlandata.initialize;

                     priv_ProtocolType := Swap16 (pData^.ProtocolType);

                     // Prepare normal Anchor

                     priv_nextAnchor.HeaderType := priv_ProtocolType;
                     priv_nextAnchor.Offset := priv_frameoffset + sizeOf (tSLLStruct);
                end;

         except
               status := prFailInvalidValues;
         end;

     result := status;
end;

(* ------------------------------------------- *)


end.
