{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_parser_netflow;

interface

uses base_structures,
     base_checksums,
     common_parse_helpers,
     definition_ip_addresses,
     definition_tcp_structures,
     class_parser_baseclass;


type tNetFlowSamplingMode = (nfsmNone, nfsmPacketInterval, nfsmRandom);

     pNetFlowV5HdrStruct = ^tNetFlowV5HdrStruct;
     tNetFlowV5HdrStruct = packed record
                                        version           : word;
                                        count             : word;
                                        sys_uptime        : longword;
                                        unix_secs         : longword;
                                        unix_nsecs        : longword;
                                        flow_sequence     : longword;
                                        engine_type       : byte;
                                        engine_id         : byte;
                                        sampling_interval : word;

                                        procedure initalize;
                                   end;

     pNetFlowV5RecordStruct = ^tNetFlowV5RecordStruct;
     tNetFlowV5RecordStruct = packed record
                                           sourceaddr     : longword;
                                           destaddr       : longword;
                                           nexthop        : longword;
                                           input          : word;
                                           output         : word;
                                           dpkts          : longword;
                                           doctets        : longword;
                                           first          : longword;
                                           last           : longword;
                                           srcport        : word;
                                           dstport        : word;
                                           pad1           : byte;
                                           tcp_flags      : byte;
                                           prot           : byte;
                                           tos            : byte;
                                           src_as         : word;
                                           dst_as         : word;
                                           src_mask       : byte;
                                           dst_mask       : byte;
                                           pad2           : word;

                                           procedure initalize;
                                      end;
{
     pNetFlowV9HdrStruct = ^tNetFlowV9HdrStruct;
     tNetFlowV9HdrStruct = packed record
                                        version           : word;
                                        count             : word;
                                        sys_uptime        : longword;
                                        unix_secs         : longword;
                                        flow_sequence     : longword;
                                        source_id         : longword;
                                   end;

     pNetFlowV9TemplateStruct = ^tNetFlowV9TemplateStruct;
     tNetFlowV9TemplateStruct = packed record
                                             FlowSetID    : word;
                                             Length       : word;
                                             TemplateID   : word;
                                             FieldCount   : word;
                                      end;
}
     tNetFlowV5Header       = record
                                    Version               : word;
                                    RecordCount           : word;
                                    SystemUptime          : longword;
                                    UnixSeconds           : longword;
                                    UnixNanoSeconds       : longword;
                                    FlowSequence          : longword;
                                    EngineType            : byte;
                                    EngineID              : byte;
                                    SamplingMode          : tNetFlowSamplingMode;
                                    SamplingInterval      : word;

                                    procedure initalize;
                              end;

     tNetFlowV5Record       = record
                                    SourceAddress         : tIPv4Address;
                                    DestinationAddress    : tIPv4Address;
                                    NextHop               : tIPv4Address;
                                    SNMPInputInterfaceId  : word;
                                    SNMPOutputInterfaceId : word;
                                    FlowPacketCount       : longword;
                                    FlowPacketOctets      : longword;
                                    FlowFirstSysUptime    : longword;
                                    FlowLastSysUptime     : longword;
                                    SourcePort            : word;
                                    DestinationPort       : word;
                                    TCPFlags              : tTCPFlagSet;
                                    IPProtocolType        : byte;
                                    IPToS                 : byte;
                                    SourceAS              : word;
                                    DestinationAS         : word;
                                    SourceMask            : byte;
                                    DestinationMask       : byte;

                                    procedure initalize;
                              end;

     tNetFlowV5Array = array of tNetFlowV5Record;

     tNetFlowParser   = class (tFrameParser)

                              private
                                     priv_NetFlowVersion   : word;
                                     priv_NetFlowV5Header  : tNetFlowV5Header;
                                     priv_NetFlowV5Records : tNetFlowV5Array;

                                     function ParseNetFlowV5Packet (Offset : integer) : integer;
                                     function ParseNetFlowV9Packet (Offset : integer) : integer;
                              public
                                    property NetFlowVersion : word read priv_netFlowVersion;
                                    property NetFlowV5Header : tNetFlowV5Header read priv_NetFlowV5Header;
                                    property NetFlowV5Records : tNetFlowV5Array read priv_NetFlowV5Records;

                                    function checkIsPlausible : boolean; override;
                                    function parse : tParseResult; override;
                        end;


implementation

uses Math,
     Toolkit;

(* ------------------------------------------- *)
{ tNetFlowV5HdrStruct }
(* ------------------------------------------- *)

procedure tNetFlowV5HdrStruct.initalize;

begin
     version := 0;
     count := 0;
     sys_uptime  := 0;
     unix_secs := 0;
     unix_nsecs := 0;
     flow_sequence := 0;
     engine_type := 0;
     engine_id := 0;
     sampling_interval := 0;
end;

(* ------------------------------------------- *)
{ tNetFlowV5RecordStruct }
(* ------------------------------------------- *)

procedure tNetFlowV5RecordStruct.initalize;

begin
     sourceaddr := 0;
     destaddr := 0;
     nexthop := 0;
     input := 0;
     output := 0;
     dpkts := 0;
     doctets := 0;
     first := 0;
     last := 0;
     srcport := 0;
     dstport := 0;
     pad1 := 0;
     tcp_flags := 0;
     prot := 0;
     tos := 0;
     src_as := 0;
     dst_as := 0;
     src_mask := 0;
     dst_mask := 0;
     pad2 := 0;
end;

(* ------------------------------------------- *)
{ tNetFlowV5Header }
(* ------------------------------------------- *)

procedure tNetFlowV5Header.initalize;

begin
     Version := 0;
     RecordCount := 0;
     SystemUptime := 0;
     UnixSeconds := 0;
     UnixNanoSeconds := 0;
     FlowSequence := 0;
     EngineType := 0;
     EngineID := 0;
     SamplingMode := nfsmNone;
     SamplingInterval := 0;

end;

(* ------------------------------------------- *)
{ tNetFlowV5Record }
(* ------------------------------------------- *)

procedure tNetFlowV5Record.initalize;

begin
     SourceAddress.initialize;
     DestinationAddress.initialize;
     NextHop.initialize;
     SNMPInputInterfaceId := 0;
     SNMPOutputInterfaceId := 0;
     FlowPacketCount := 0;
     FlowPacketOctets := 0;
     FlowFirstSysUptime := 0;
     FlowLastSysUptime := 0;
     SourcePort := 0;
     DestinationPort := 0;
     TCPFlags := [];
     IPProtocolType := 0;
     IPToS := 0;
     SourceAS := 0;
     DestinationAS := 0;
     SourceMask := 0;
     DestinationMask := 0;
end;

(* ------------------------------------------- *)

function tNetFlowParser.checkIsPlausible: boolean;

var value     : boolean;
    pHelpWord : pWord;
    HelpWord  : word;

begin
     value := true;

     if priv_framelength < priv_frameoffset + 2 then
        value := false
     else
         begin
              pHelpWord := pWord (priv_framedata + priv_frameoffset);
              HelpWord := swap16 (pHelpWord^);

//            if (HelpWord = 0) or (HelpWord > 9) then
              if (HelpWord <> 5) then
                 value := false;
         end;

     result := value;
end;

(* ------------------------------------------- *)

function tNetFlowParser.ParseNetFlowV5Packet (Offset : integer) : integer;

var pHdrData    : pNetFlowV5HdrStruct;
    pRecData    : pNetFlowV5RecordStruct;
    LocalOffset : integer;
    RecordSize  : integer;
    help        : word;
    mode        : word;
    maxrecs     : integer;
    i           : integer;
    FlowRecord  : pNetFlowV5RecordStruct;

begin
     LocalOffset := Offset;

     // read the V5 Header first

     pHdrData := pNetFlowV5HdrStruct (priv_framedata + Offset);

     priv_NetFlowV5Header.version := swap16 (pHdrData^.version);
     priv_NetFlowV5Header.RecordCount := swap16 (pHdrData^.count);
     priv_NetFlowV5Header.SystemUptime := swap32 (pHdrData^.sys_uptime);
     priv_NetFlowV5Header.UnixSeconds := swap32 (pHdrData^.unix_secs);
     priv_NetFlowV5Header.UnixNanoSeconds := swap32 (pHdrData^.unix_nsecs);
     priv_NetFlowV5Header.FlowSequence := swap32 (pHdrData^.flow_sequence);
     priv_NetFlowV5Header.EngineType := pHdrData^.engine_type;
     priv_NetFlowV5Header.EngineID := pHdrData^.engine_id;

     help := swap16 (pHdrData^.sampling_interval);
     mode := (help and $c000) shr 14;

     priv_NetFlowV5Header.SamplingMode := nfsmNone;

     case mode of
      1 : priv_NetFlowV5Header.SamplingMode := nfsmPacketInterval;
      2 : priv_NetFlowV5Header.SamplingMode := nfsmRandom;
     end;

     priv_NetFlowV5Header.SamplingInterval := (help and $3fff);

     LocalOffset := LocalOffset + sizeOf (tNetFlowV5HdrStruct);
     RecordSize := priv_NetFlowV5Header.RecordCount * sizeOf (tNetFlowV5RecordStruct);

     if priv_framelength < (LocalOffset + RecordSize) then
        result := -1
     else
         begin
              // Check how much space we have for records and parse what we can

              maxrecs := (priv_framelength - Localoffset) div sizeOf (tNetFlowV5RecordStruct);
              maxrecs := min (maxrecs, priv_NetFlowV5Header.RecordCount);

              setLength (priv_NetFlowV5Records, priv_NetFlowV5Header.RecordCount);

              // Parse Records

              for i := 0 to (priv_NetFlowV5Header.RecordCount - 1) do
                  begin
                       FlowRecord := pNetFlowV5RecordStruct (priv_framedata + LocalOffset);

                       priv_NetFlowV5Records[i].SourceAddress.ValueAsLong := swap32 (FlowRecord^.sourceaddr);
                       priv_NetFlowV5Records[i].DestinationAddress.ValueAsLong := swap32 (FlowRecord^.destaddr);
                       priv_NetFlowV5Records[i].NextHop.ValueAsLong := swap32 (FlowRecord^.nexthop);
                       priv_NetFlowV5Records[i].SNMPInputInterfaceId := swap16 (FlowRecord^.input);
                       priv_NetFlowV5Records[i].SNMPOutputInterfaceId := swap16 (FlowRecord^.output);
                       priv_NetFlowV5Records[i].FlowPacketCount := swap32 (FlowRecord^.dpkts);
                       priv_NetFlowV5Records[i].FlowPacketOctets := swap32 (FlowRecord^.doctets);
                       priv_NetFlowV5Records[i].FlowFirstSysUptime := swap32 (FlowRecord^.first);
                       priv_NetFlowV5Records[i].FlowLastSysUptime := swap32 (FlowRecord^.last);
                       priv_NetFlowV5Records[i].SourcePort := swap16 (FlowRecord^.srcport);
                       priv_NetFlowV5Records[i].DestinationPort := swap16 (FlowRecord^.dstport);
                       priv_NetFlowV5Records[i].TCPFlags := getTCPFlagSetFromWord (FlowRecord^.tcp_flags);
                       priv_NetFlowV5Records[i].IPProtocolType := FlowRecord^.prot;
                       priv_NetFlowV5Records[i].IPToS := FlowRecord^.tos;
                       priv_NetFlowV5Records[i].SourceAS := swap16 (FlowRecord^.src_as);
                       priv_NetFlowV5Records[i].DestinationAS := swap16 (FlowRecord^.dst_as);
                       priv_NetFlowV5Records[i].SourceMask := FlowRecord^.src_mask;
                       priv_NetFlowV5Records[i].DestinationMask := FlowRecord^.dst_mask;

                       inc (LocalOffset, sizeOf (tNetFlowV5RecordStruct));
                  end;

              result := LocalOffset - Offset;
         end;
end;

(* ------------------------------------------- *)

function tNetFlowParser.ParseNetFlowV9Packet (Offset : integer) : integer;

begin
     // yeah... uh. no. not yet.
end;

(* ------------------------------------------- *)
function tNetFlowParser.parse: tParseResult;

var status    : tParseResult;
    pHelpWord : pWord;
    HelpWord  : Word;

begin
     priv_osilayer := 5;     // Set Layer this parser is for
     prot_layersize := 0;

     setLength (priv_NetFlowV5Records, 0);

     status := prSuccess;

     if priv_framelength < priv_frameoffset + 2 then
        status := prFailUndersize
     else
         try
            pHelpWord := pWord (priv_framedata + priv_frameoffset);
            priv_NetFlowVersion := swap16 (pHelpWord^);

            case priv_NetFlowVersion of
             5 : begin
                      if priv_frameLength < (priv_frameoffset + sizeOf (tNetFlowV5HdrStruct)) then
                         status := prFailUnderSize
                      else
                          prot_layersize := ParseNetFlowV5Packet (priv_frameoffset);
                 end;
             9 : begin
                      setLength (priv_NetFlowV5Records, 0);

//                    prot_layersize := ParseNetFlowV9Packet (priv_frameoffset);
                 end;
            end;

            // Prepare next anchor

            priv_nextAnchor.Offset := priv_frameoffset + prot_layersize;
            priv_nextAnchor.HeaderType := NEXT_HEADER_NONE;

         except
               status := prFailInvalidValues;
         end;

     result := status;
end;

(* ------------------------------------------- *)

end.
