{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_parser_tcp;

interface

uses base_structures,
     base_checksums,
     definition_ip_addresses,
     definition_tcp_structures,
     common_parse_helpers,
     class_parser_baseclass;

type tTCPMultiPathOption = record
                                 optionlength   : integer;
                                 subType        : tTCPMPathSubType;
                                 TCPVersion     : byte;
                                 IPVersion      : byte;
                                 MPathFlags     : tTCPMpathFlagSet;
                                 SendersKey     : uint64;
                                 ReceiversKey   : uint64;
                                 AddressID      : byte;
                                 ReceiversToken : longword;
                                 SendersRndNr   : longword;
                                 ACKNumber      : uint64;
                                 SEQNumber      : uint64;
                                 IPv4Address    : tIPv4Address;
                                 IPv6Address    : tIPv6Address;
                                 TCPPort        : word;
                                 DataSequenceNr : uint64;

                                 procedure initialize;
                                 procedure CopyFrom (Source : tTCPMultiPathOption);
                           end;

     tTCPOptionPad = array of byte;

     tTCPMPathOptionList = array of tTCPMultiPathOption;

     tFastOpenCookie = array[0..15] of byte;
     tTCPFastOpenOption = record
                                CookieLength    : byte;
                                Cookie          : tFastOpenCookie;

                                procedure initialize;
                                procedure CopyFrom (Source : tTCPFastOpenOption);
                          end;
     tTCPFastOpenOptionList = array of tTCPFastOpenOption;

     tTCPRiverbedOption  = record
                                 OptionLength    : integer;
                                 OptionType      : byte;
                                 OptionVersion   : byte;
                                 IPversion       : byte;
                                 AppVersion      : byte;
                                 IPv4Address     : tIPv4Address;

                                 procedure initialize;
                           end;

     tRiverbedOptionList = array of tTCPRiverbedOption;

{
             0                   1                   2                   3
             0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
            |          Source Port          |       Destination Port        |
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
            |                        Sequence Number                        |
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
            |                    Acknowledgment Number                      |
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
            |  Data |           |U|A|P|R|S|F|                               |
            | Offset| Reserved  |R|C|S|S|Y|I|            Window             |
            |       |           |G|K|H|T|N|N|                               |
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
            |           Checksum            |         Urgent Pointer        |
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
            |                    Options                    |    Padding    |
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
            |                             data                              |
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
}

     pTCPStruct = ^tTCPStruct;
     tTCPStruct = packed record
                               Sourceport      : word;
                               Destinationport : word;
                               Sequence        : longword;
                               Acknowledge     : longword;
                               DataOffsetFlags : word;
                               Window          : word;
                               CRC             : word;
                               UrgentPointer   : word;
                          end;

     tTCPParser   = class (tFrameParser)

                          {$REGION 'Documentation'}
                          ///	<summary>
                          ///	  Constructor of the TCP parser class
                          ///	</summary>
                          ///	<param name="FrameData">
                          ///	  Pointer to the Buffer holding the Frame
                          ///	  data
                          ///	</param>
                          ///	<param name="FrameOffset">
                          ///	  Offset of the TCP layer starting
                          ///	</param>
                          ///	<param name="DataLength">
                          ///	  Total Size of the Frame data in the
                          ///	  buffer. This is less or equal to
                          ///	  Wirelength.
                          ///	</param>
                          ///	<param name="WireLength">
                          ///	  Size of the frame on the wire
                          ///	</param>
                          ///	<param name="TCPSectionLength">
                          ///	  Length of the TCP section, usually the IP
                          ///	  payload length
                          ///	</param>
                          ///	<param name="FrameIndex">
                          ///	  Number of the frame in the file
                          ///	</param>
                          ///	<example>
                          ///	  tTCPParser.create (Frame.Data,
                          ///	  ProtocolAnchor.Offset, Frame.FrameSize,
                          ///	  TempValues.IPPayloadLen,
                          ///	  Frame.FileFrameID);
                          ///	</example>
                          {$ENDREGION}
                          constructor create (FrameData : pointer; const FrameOffset, DataLength, WireLength, TCPSectionLength : integer; const FrameIndex : integer = 0); reintroduce;
                          procedure free;

                          private
                                 priv_sourceport           : word;
                                 priv_destinationport      : word;
                                 priv_sequence             : tTCPSequence;
                                 priv_acknowledgement      : tTCPSequence;
                                 priv_nextexpectedsequence : tTCPSequence;
                                 priv_headerlength         : byte;
                                 priv_flags                : tTCPFlagSet;
                                 priv_windowsize           : word;
                                 priv_checksum             : word;
                                 priv_urgentpointer        : word;

                                 priv_optionset            : tTCPOptionSet;          // normal enumeration set, without duplicates
                                 priv_exactoptionlist      : tTCPOptionList;         // exact order of options, including duplicates
                                 priv_mss                  : word;
                                 priv_scalefactor          : tTCPScaleFactorOption;
                                 priv_timestamp            : tTCPTimeStampOption;

                                 priv_sectionlen           : integer;
                                 priv_lastoffset           : integer;

                                 priv_payloadlen           : integer;
                                 priv_tcpsectionwords      : tBufferWords;
                                 priv_mpathoptionlist      : tTCPMPathOptionList;
                                 priv_fastopenoptionlist   : tTCPFastOpenOptionList;
                                 priv_riverbedoptionlist   : tRiverbedOptionList;

                                 priv_sackranges           : tTCPSackRanges;

                                 priv_paddinglen           : integer;
                                 priv_optionpad            : tTCPOptionPad;
                                 priv_totaloptionlen       : integer;

                                 {$REGION 'Documentation'}
                                 ///	<summary>
                                 ///	  Returns the actual length of the TCP
                                 ///	  header
                                 ///	</summary>
                                 ///	<param name="Value">
                                 ///	  Byte from the frame containing the 4
                                 ///	  "offset" bits
                                 ///	</param>
                                 {$ENDREGION}
                                 function getTCPHeaderLength (Value : word) : byte;
//                               function getTCPFlagSet (Value : word) : tTCPFlagSet;

                                 function parseMultiPathOption (Offset, OptionLen : integer) : tTCPMultiPathOption;
                                 {$REGION 'Documentation'}
                                 ///	<summary>
                                 ///	  Parses the TCP option bytes
                                 ///	</summary>
                                 ///	<param name="Offset">
                                 ///	  Offset at which the TCP option block
                                 ///	  starts
                                 ///	</param>
                                 {$ENDREGION}
                                 function parseTCPOptions (Offset : integer) : tParseResult;

                                 function getIsSYNPacket : boolean;
                                 function getIsSYNACKPacket : boolean;

                          public
                                property SourcePort : word read priv_sourceport;
                                property DestinationPort : word read priv_destinationport;
                                property SequenceNumber : tTCPSequence read priv_sequence;
                                property NextExpectedSequenceNumber : tTCPSequence read priv_nextexpectedsequence;
                                property AcknowledgeNumber : tTCPSequence read priv_acknowledgement;
                                property HeaderLength : byte read priv_headerlength;
                                property Flags : tTCPFlagSet read priv_flags;
                                property WindowSize : word read priv_windowsize;
                                property Checksum : word read priv_checksum;
                                property UrgentPointer : word read priv_urgentpointer;
                                property MSS : word read priv_mss;
                                property TCPOptionsSet : tTCPOptionSet read priv_optionset;
                                property ExactOptionList : tTCPOptionList read priv_exactoptionlist;
                                property SACKRanges : tTCPSACKRanges read priv_sackranges;
                                property Scalefactor : tTCPScaleFactorOption read priv_scalefactor;
                                property Timestamp : tTCPTimeStampOption read priv_timestamp;
                                property MultiPathOptionList : tTCPMPathOptionList read priv_mpathoptionlist;
                                property FastOpenOptionList : tTCPFastOpenOptionList read priv_fastopenoptionlist;
                                property PaddingLength : integer read priv_paddinglen;
                                property PaddingBytes : tTCPOptionPad read priv_optionpad;
                                property PayloadLength : integer read priv_payloadlen;
                                property isSYNPacket : boolean read getIsSYNPacket;
                                property isSYNACKPacket : boolean read getIsSYNACKPacket;
                                property TCPSectionAsWordBuffer : tBufferWords read priv_tcpsectionwords;

                                function PortPair : tPortPair;

                                function GetFlowStateRecord : tTCPFlowState;

                                function checkIsPlausible : boolean; override;
                                function parse : tParseResult; override;
                    end;

implementation

uses SysUtils,
     Math,
     base_messages,
     SiAuto,
     Toolkit;

(* ------------------------------------------- *)
{ tTCPMultiPathOption }
(* ------------------------------------------- *)

procedure tTCPMultiPathOption.initialize;

begin
     optionlength := 0;
     subType := mtcpNone;
     TCPVersion := 0;
     IPversion := 0;
     MPathFlags := [];
     SendersKey := 0;
     ReceiversKey := 0;
     AddressID := 0;

     ReceiversToken := 0;
     SendersRndNr := 0;

     ACKNumber := 0;
     SEQNumber := 0;
     IPv4Address.initialize;
     IPv6Address.initialize;

     TCPPort := 0;
     DataSequenceNr := 0;
end;

(* ------------------------------------------- *)

procedure tTCPMultiPathOption.CopyFrom(Source: tTCPMultiPathOption);

begin
     optionlength := Source.optionlength;
     subType := Source.subType;
     TCPVersion := Source.TCPVersion;
     IPversion := Source.IPVersion;
     MPathFlags := Source.MPathFlags;
     SendersKey := Source.SendersKey;
     ReceiversKey := Source.ReceiversKey;
     AddressID := Source.AddressID;

     ReceiversToken := Source.ReceiversToken;
     SendersRndNr := Source.SendersRndNr;

     ACKNumber := Source.ACKNumber;
     SEQNumber := Source.SEQNumber;
     IPv4Address := Source.IPv4Address;
     IPv6Address := Source.IPv6Address;

     TCPPort := Source.TCPPort;
     DataSequenceNr := Source.DataSequenceNr;
end;

(* ------------------------------------------- *)
{ tTCPFastOpenOption }
(* ------------------------------------------- *)

procedure tTCPFastOpenOption.initialize;

begin
     CookieLength := 0;
     FillChar (Cookie[0], length(Cookie), 0);
end;

(* ------------------------------------------- *)

procedure tTCPFastOpenOption.CopyFrom (Source: tTCPFastOpenOption);

begin
     CookieLength := Source.CookieLength;
     Cookie := Source.Cookie;
end;

(* ------------------------------------------- *)
{ tTCPRiverbedOption }
(* ------------------------------------------- *)

procedure tTCPRiverbedOption.initialize;

begin
     OptionLength := 0;
     OptionType := 0;
     OptionVersion := 0;
     IPversion := 4;
     AppVersion := 0;
     IPv4Address.initialize;
end;

(* ------------------------------------------- *)
{ TCPParser }
(* ------------------------------------------- *)

constructor tTCPParser.create(FrameData: pointer; const FrameOffset, DataLength, WireLength, TCPSectionLength, FrameIndex: integer);

begin
     inherited create (FrameData, FrameOffset, DataLength, Wirelength, FrameIndex);

     priv_sectionlen := TCPSectionLength;

     priv_mss := 0;
     priv_timestamp.tsval := 0;
     priv_timestamp.tsecho := 0;
     setlength (priv_sackranges, 0);
     priv_scalefactor.ShiftCount := 0;
     priv_scalefactor.Multiplier := 1;

     setlength (priv_mpathoptionlist, 0);
     setlength (priv_riverbedoptionlist, 0);
     setLength (priv_optionpad, 0);
     priv_totaloptionlen := 0;

     // calculate the end offset we can parse up to

     if (priv_frameoffset + priv_sectionlen) > priv_framelength  then
        priv_lastoffset := priv_framelength
     else
         priv_lastoffset := priv_frameoffset + priv_sectionlen - 1;
end;

(* ------------------------------------------- *)

procedure tTCPParser.free;

begin
     setlength (priv_tcpsectionwords, 0);

     inherited;
end;

(* ------------------------------------------- *)

function tTCPParser.checkIsPlausible: boolean;

begin
     result := true;
end;

(* ------------------------------------------- *)

function tTCPParser.getIsSYNPacket: boolean;

var help : tTCPFlagSet;

begin
     // Remove all valid flags for a SYN packet (we're allowing anything that isn't clearly a deal breaker)

     help := priv_flags - [tcpSYN, tcpECN, tcpCWR, tcpNON];

     // Check if the remainder is empty, and the important flags exist

     result := ((tcpSYN in priv_flags) and (help = []));
end;

(* ------------------------------------------- *)

function tTCPParser.getIsSYNACKPacket: boolean;

var help : tTCPFlagSet;

begin
     // Remove all valid flags for a SYN/ACK packet (we're allowing anything that isn't clearly a deal breader)

     help := priv_flags - [tcpSYN, tcpACK, tcpPSH, tcpECN, tcpCWR,tcpNON];

     // Check if the remainder is empty, and the important flags exist

     result := ((tcpSYN in priv_flags) and (tcpACK in priv_flags) and (help = []));
end;

(* ------------------------------------------- *)

function tTCPParser.getTCPHeaderLength (Value : word) : byte;

var count : word;

begin
     count := Value and $f000;           // get the high nibble of the higher byte
     count := count shr 12;              // Move to lower byte

     result := count * 4;                // 32bit values a.k.a. 4 bytes
end;

(* ------------------------------------------- *)

function tTCPParser.GetFlowStateRecord : tTCPFlowState;

var value : tTCPFlowState;
    i     : Integer;

begin
     value.initialize (priv_sequence, priv_payloadlen, priv_windowsize, priv_flags);

     if tcpScalefactor in priv_optionset then
        value.WindowMultiplier := priv_scalefactor.Multiplier;

     if tcpMSS in priv_optionset then
        value.MaxSegmentSize := priv_mss;

     setlength (value.SackRanges, length (priv_sackranges));

     for i := 1 to length (priv_sackranges) do
         value.SackRanges[i-1] := priv_sackranges[i-1];

     result := value;
end;

(* ------------------------------------------- *)

function tTCPParser.parseMultiPathOption (Offset, OptionLen : integer): tTCPMultiPathOption;

var pHelpByte    : pByte;
    HelpByte     : byte;
    pHelpWord    : pWord;
    pHelpLong    : pLongword;
    pHelpUint64  : pUINT64;
    pHelpIPv6Adr : pIPv6AddressByteStruct;
    Value        : tTCPMultiPathOption;
    HelpOffset   : integer;

begin
     Value.initialize;
     Value.optionlength := OptionLen;

     HelpOffset := 0;

     // Subtype and Version

     pHelpByte := pByte (priv_framedata + Offset);

     HelpByte := (pHelpByte^ shr 4);

     case HelpByte of
      0  : begin
                // Subtype: SetCapable

                Value.subType := mtcpCapable;

                // Version

                Value.TCPVersion := (pHelpByte^ and $f);

                // Flags

                pHelpByte := pByte (priv_framedata + Offset + 1);

                if (pHelpByte^ and $80) = $80 then
                   Value.MPathFlags := [tmpChecksumReq];

                if (pHelpByte^ and $40) = $40 then
                   Value.MPathFlags := Value.MPathFlags + [tmpExtensibility];

                if (pHelpByte^ and $1) = $1 then
                   Value.MPathFlags := Value.MPathFlags + [tmpCryptoUseHMAC];

                // Senders Key

                pHelpUint64 := pUINT64 (priv_framedata + Offset + 2);

                Value.SendersKey := Swap64 (pHelpUint64^);

                if OptionLen = 20 then
                   begin
                        pHelpUint64 := pUINT64 (priv_framedata + Offset + 10);

                        Value.ReceiversKey := Swap64 (pHelpUint64^);
                   end;
           end;
      1  : begin
                // Subtype: Join Connection

                Value.subType := mtcpJoinConnection;

                // Flags

                if (pHelpByte^ and $1) = $1 then
                   Value.MPathFlags := [tmpBackup];

                // Address ID

                pHelpByte := pByte (priv_framedata + Offset + 1);
                Value.AddressID := pHelpByte^;

                // Tokens

                pHelpLong := pLongword (priv_framedata + Offset + 2);
                Value.ReceiversToken := swap32 (pHelpLong^);

                pHelpLong := pLongword (priv_framedata + Offset + 6);
                Value.SendersRndNr := swap32 (pHelpLong^);

           end;
      2  : begin
                // Subtype: Data Sequence Signal

                Value.subType := mtcpDataSequenceSignal;

                // Flags

                pHelpByte := pByte (priv_framedata + Offset + 1);

                if (pHelpByte^ and $10) = $10 then
                   Value.MPathFlags := [tmpDataFin];

                if (pHelpByte^ and $8) = $8 then
                   Value.MPathFlags := Value.MPathFlags + [tmpDataSeq8Octets];

                if (pHelpByte^ and $4) = $4 then
                   Value.MPathFlags := Value.MPathFlags + [tmpDNSetcPresent];

                if (pHelpByte^ and $2) = $2 then
                   Value.MPathFlags := Value.MPathFlags + [tmpDataACK8Octets];

                if (pHelpByte^ and $1) = $1 then
                   Value.MPathFlags := Value.MPathFlags + [tmpDataACKpresent];

                // Read ACK and Sequence

                HelpOffset := Offset + 2;

                if tmpDataACKpresent in Value.MPathFlags then
                   begin
                        if tmpDataACK8Octets in Value.MPathFlags then
                           begin
                                pHelpUint64 := pUINT64 (priv_framedata + HelpOffset);
                                value.ACKNumber := Swap64 (pHelpUInt64^);

                                inc (HelpOffset, 8);
                           end
                        else
                            begin
                                 pHelpLong := pLongword (priv_framedata + HelpOffset);
                                 value.ACKNumber := swap32 (pHelpLong^);

                                 inc (HelpOffset, 4);
                            end;
                   end;

                if tmpDNSetcPresent in Value.MPathFlags then
                   begin
                        if tmpDataSeq8Octets in Value.MPathFlags then
                           begin
                                pHelpUint64 := pUINT64 (priv_framedata + HelpOffset);
                                value.SEQNumber := Swap64 (pHelpUInt64^);
                           end
                        else
                            begin
                                 pHelpLong := pLongword (priv_framedata + HelpOffset);
                                 value.SEQNumber := swap32 (pHelpLong^);
                            end;

                        value.ACKNumber := pHelpLong^;
                   end;
           end;
      3  : begin
                // Subtype: Add Address

                Value.subType := mtcpAddAddress;

                // Version

                Value.IPVersion := (pHelpByte^ and $f);

                // Address ID

                pHelpByte := pByte (priv_framedata + Offset + 1);
                Value.AddressID := pHelpByte^;

                // Address

                case Value.IPVersion of
                 4 : begin
                          pHelpLong := pLongword (priv_framedata + Offset + 2);
                          value.IPv4Address.ValueAsLong := swap32 (pHelpLong^);

                          // Read optional port number if available (option length not being 8 bytes)

                          if OptionLen = 10 then
                             begin
                                  pHelpWord := pWord (priv_framedata + Offset + 6);

                                  value.TCPPort := swap16 (pHelpWord^);
                             end;
                     end;
                 6 : begin
                          pHelpIPv6Adr := pIPv6AddressByteStruct (priv_framedata + Offset + 2);
                          value.IPv6Address.SetValueFromByteStruct (pHelpIPv6Adr^);

                          // Read optional port number if available (option length not being 20 bytes)

                          if OptionLen = 22 then
                             begin
                                  pHelpWord := pWord (priv_framedata + Offset + 18);

                                  value.TCPPort := swap16 (pHelpWord^);
                             end;
                     end;
                end;
           end;
      4 : begin
               // Subtype: Remove Address

               value.subType := mtcpRemovAddress;

                // Address ID

                pHelpByte := pByte (priv_framedata + Offset + 1);
                Value.AddressID := pHelpByte^;

                // TODO: there could be more than one address ID in this option
          end;
      5 : begin
               // Subtype: Subflow Priority Policy

               value.subType := mtcpSubflowPrio;

                // Flags

                if (pHelpByte^ and $1) = $1 then
                   Value.MPathFlags := [tmpBackup];
          end;
      6 : begin
               // Subtype: Fallback

               value.subType := mtcpFallback;

               // Data Sequence Number

                pHelpUint64 := pUINT64 (priv_framedata + Offset + 2);

                Value.DataSequenceNr := Swap64 (pHelpUint64^);
          end;
      7 : begin
               // Subtype: FastClose

               value.subType := mtcpFastClose;

               // Read receivers key

                pHelpUint64 := pUINT64 (priv_framedata + Offset + 2);

                Value.ReceiversKey := Swap64 (pHelpUint64^);
          end;
     end;

     result := Value;
end;

(* ------------------------------------------- *)

function tTCPParser.parseTCPOptions (Offset : integer) : tParseResult;

var OptionCode    : Byte;
    OptionLen     : Byte;
    OptionCount   : integer;
    pHelpWord     : pWord;
    pHelpLong     : pLongword;
    pHelpByte     : pByte;
    helpWord      : word;
    helpLong      : longword;
    HelpDouble    : double;
    sackcount     : integer;
    MPathOpt      : tTCPMultiPathOption;
    pSackOpt      : pTCPSackOption;
    MPathCount    : integer;
    FastOpenOpt   : tTCPFastOpenOption;
    FastOpenCount : integer;
    RBedOption    : tTCPRiverbedOption;
    RBedCount     : integer;
    i             : integer;
    EOLOption     : boolean;
    boWarned      : boolean;
    parseresult   : tParseResult;

begin
     // Initialize

     EOLOption := false;
     parseresult := prSuccess;

     priv_optionset := [];

     // set ordered option list to 32, and trim it later

     setlength (priv_exactoptionlist, 32);
     OptionCount := 0;
     priv_totaloptionlen := 0;

     setlength (priv_sackranges, 0);

     with priv_scalefactor do
          begin
               ShiftCount := 0;
               Multiplier := 1;
          end;

     with priv_timestamp do
          begin
               priv_timestamp.tsval := 0;
               priv_timestamp.tsecho := 0;
          end;

     // Parse

     repeat
           // Check if we have at least one byte left to read. Since the offset starts at 0 we use >=

           if (offset >= priv_lastoffset) then
              begin
                   setlength (priv_exactoptionlist, 0);

                   exit (prOptionsTruncated);
              end;

           OptionCode := pByte (priv_framedata + Offset)^;
           OptionLen := 0;

           boWarned := false;

           // Check if we need to read an option length, for which we may need to be able to read an additional byte
           // The option length is for the whole option, including code and len bytes

           if (OptionCode > 1) and (offset + 1 > priv_lastoffset) then
              exit (prOptionsTruncated);

           case OptionCode of
             0 : begin
                      // EOL

                      OptionLen := 1;

                      if tcpEOL in priv_optionset then
                         addDiagnosticMsg (msWarning, MSG_TCP_MULTIPLE_EOL_OPTION_SPECIFIED);

                      priv_optionset := priv_optionset + [tcpEOL];

                      inc (OptionCount);
                      priv_exactoptionlist[OptionCount-1] := tcpEOL;

                      // Check if remaining bytes are 0, which would be padding

                      priv_paddinglen := (priv_frameoffset + priv_headerlength) - (Offset + 1);
                      setlength (priv_optionpad, priv_paddinglen);

                      for i := 1 to priv_paddinglen do
                          begin
                               pHelpByte := pByte (priv_framedata + Offset + i);
                               priv_optionpad[i-1] := pHelpByte^;

                               // If byte is zero we consider it to be padding, otherwise complain :-)

                               if (pHelpByte^ <> 0) and not boWarned then
                                  begin
                                       addDiagnosticMsg (msWarning, MSG_TCP_OPTION_DATA_AFTER_EOL);
                                       boWarned := true;
                                  end
                               else
                                   inc (OptionLen);
                          end;

                      EOLOption := true;
                 end;
             1 : begin
                      // NOP. Let's be lazy and only increase the offset

                      OptionLen := 1;

                      priv_optionset := priv_optionset + [tcpNOP];

                      inc (OptionCount);
                      priv_exactoptionlist[OptionCount-1] := tcpNOP;
                 end;
             2 : begin
                      // MSS Size

                      OptionLen := pByte (priv_framedata + Offset + 1)^;

                      // Check for enough bytes to parse (option len includes the type we already read, so -1)

                      if (offset + OptionLen - 1 > priv_lastoffset) then
                         exit (prOptionsTruncated);

                      if tcpMSS in priv_optionset then
                         addDiagnosticMsg (msWarning, MSG_TCP_MULTIPLE_MSS_OPTIONS_SPECIFIED);

                      if (OptionLen = 4) then
                         begin
                              priv_optionset := priv_optionset + [tcpMSS];

                              inc (OptionCount);
                              priv_exactoptionlist[OptionCount-1] := tcpMSS;

                              HelpWord := pWord (priv_framedata + Offset + 2)^;

                              priv_mss := Swap16 (helpWord);
                         end
                      else
                          addDiagnosticMsg (msWarning, MSG_TCP_BAD_MSS_OPTION_LEN);
                 end;
             3 : begin
                      // Window Scaling

                      OptionLen := pByte (priv_framedata + Offset + 1)^;

                      // Check for enough bytes to parse (option len includes the type we already read, so -1)

                      if (offset + OptionLen - 1 > priv_lastoffset) then
                         exit (prOptionsTruncated);

                      if tcpScalefactor in priv_optionset then
                         addDiagnosticMsg (msWarning, MSG_TCP_MULTIPLE_SCALE_OPTIONS_SPECIFIED);

                      if OptionLen = 3 then
                         begin
                              priv_optionset := priv_optionset + [tcpScalefactor];

                              inc (OptionCount);
                              priv_exactoptionlist[OptionCount-1] := tcpScalefactor;

                              pHelpByte := pByte (priv_framedata + Offset + 2);

                              with priv_scalefactor do
                                   begin
                                        ShiftCount := pHelpByte^;

                                        HelpDouble := power (2, priv_scalefactor.ShiftCount);
                                        Multiplier := round (HelpDouble);
                                   end;
                         end
                      else
                          addDiagnosticMsg (msWarning, MSG_TCP_BAD_MSS_OPTION_LEN, '3', OptionLen.ToString);
                 end;
             4 : begin
                      // SACK Permitted

                      OptionLen := pByte (priv_framedata + Offset + 1)^;

                      // Check for enough bytes to parse (option len includes the type we already read, so -1)

                      if (offset + OptionLen - 1 > priv_lastoffset) then
                         exit (prOptionsTruncated);

                      if OptionLen = 2 then
                         begin
                              if tcpSACKPermitted in priv_optionset then
                                 addDiagnosticMsg (msWarning, MSG_TCP_MULTIPLE_SACK_OPTIONS_SPECIFIED);

                              priv_optionset := priv_optionset + [tcpSACKPermitted];

                              inc (OptionCount);
                              priv_exactoptionlist[OptionCount-1] := tcpSACKPermitted;
                         end
                      else
                          addDiagnosticMsg (msWarning, MSG_TCP_BAD_SACK_OPTION_LEN, '2', optionlen.ToString);
                 end;
             5 : begin
                      // SACK block

                      OptionLen := pByte (priv_framedata + Offset + 1)^;

                      // Check for enough bytes to parse (option len includes the type we already read, so -1)

                      if (offset + OptionLen - 1 > priv_lastoffset) then
                         exit (prOptionsTruncated);

                      sackcount := (OptionLen - 2) div 8;   // 4 Bytes left edge, 4 bytes right edge
                      setlength (priv_sackranges, sackcount);

                      priv_optionset := priv_optionset + [tcpSACK];

                      inc (OptionCount);
                      priv_exactoptionlist[OptionCount-1] := tcpSACK;

                      for i := 0 to sackcount - 1 do
                          begin
                               pSackOpt := pTCPSackOption (priv_framedata + Offset + 2 + (i * 8));

                               priv_sackranges[i].leftEdge.initialize (swap32 (pSackOpt^.leftEdge));
                               priv_sackranges[i].rightEdge.initialize (swap32 (pSackOpt^.rightEdge));
                          end;
                 end;
             8 : begin
                      // Timestamp

                      OptionLen := pByte (priv_framedata + Offset + 1)^;

                      // Check for enough bytes to parse (option len includes the type we already read, so -1)

                      if (offset + OptionLen - 1 > priv_lastoffset) then
                         exit (prOptionsTruncated);

                      if tcpTimestamp in priv_optionset then
                         addDiagnosticMsg (msWarning, MSG_TCP_MULTIPLE_TS_OPTIONS_SPECIFIED);

                      priv_optionset := priv_optionset + [tcpTimestamp];

                      inc (OptionCount);
                      priv_exactoptionlist[OptionCount-1] := tcpTimestamp;

                      pHelpLong := pLongword (priv_framedata + Offset + 2);
                      helpLong := pHelpLong^;

                      priv_timestamp.tsval := Swap32 (helpLong);

                      pHelpLong := pLongword (priv_framedata + Offset + 6);
                      helpLong := pHelpLong^;

                      priv_timestamp.tsecho := Swap32 (helpLong);
                 end;
            11 : begin
                      //  CC Option, RFC 1644

                      OptionLen := pByte (priv_framedata + Offset + 1)^;

                      // Check for enough bytes to parse (option len includes the type we already read, so -1)

                      if (offset + OptionLen - 1 > priv_lastoffset) then
                         exit (prOptionsTruncated);

                      priv_optionset := priv_optionset + [tcpCC];

                      inc (OptionCount);
                      priv_exactoptionlist[OptionCount-1] := tcpCC;
                 end;

            12 : begin
                      // CC.NEW Option, RFC 1644

                      OptionLen := pByte (priv_framedata + Offset + 1)^;

                      // Check for enough bytes to parse (option len includes the type we already read, so -1)

                      if (offset + OptionLen - 1 > priv_lastoffset) then
                         exit (prOptionsTruncated);

                      priv_optionset := priv_optionset + [tcpCCNEW];

                      inc (OptionCount);
                      priv_exactoptionlist[OptionCount-1] := tcpCCNEW;
                 end;
            13 : begin
                      // CC.ECHO Option, RFC 1644

                      OptionLen := pByte (priv_framedata + Offset + 1)^;

                      // Check for enough bytes to parse (option len includes the type we already read, so -1)

                      if (offset + OptionLen - 1 > priv_lastoffset) then
                         exit (prOptionsTruncated);

                      priv_optionset := priv_optionset + [tcpCCECHO];

                      inc (OptionCount);
                      priv_exactoptionlist[OptionCount-1] := tcpCCECHO;
                 end;
            30 : begin
                      // MultiPath TCP, RFC 6824

                      OptionLen := pByte (priv_framedata + Offset + 1)^;

                      // Check for enough bytes to parse (option len includes the type we already read, so -1)

                      if (offset + OptionLen - 1 > priv_lastoffset) then
                         exit (prOptionsTruncated);

                      MPathOpt := parseMultiPathOption (Offset + 2, OptionLen);

                      // Add option to list

                      MPathCount := length (priv_mpathoptionlist);
                      inc (MPathCount);
                      setlength (priv_mpathoptionlist, MPathCount);

                      priv_mpathoptionlist[MPathCount-1] := MPathOpt;

                      // Add to optionset

                      priv_optionset := priv_optionset + [tcpMPath];

                      inc (OptionCount);
                      priv_exactoptionlist[OptionCount-1] := tcpMPath;
                 end;
            34 : begin
                      // TCP Fast Open, RFC 7413

                      OptionLen := pByte (priv_framedata + Offset + 1)^;

                      // Check for enough bytes to parse (option len includes the type we already read, so -1)

                      if (offset + OptionLen - 1 > priv_lastoffset) then
                         exit (prOptionsTruncated);

                      if (OptionLen > 1) and (OptionLen < 19) then
                         begin
                              FastOpenOpt.initialize;

                              // copy cookie, if any

                              if Optionlen > 2 then
                                 begin
                                      FastOpenOpt.CookieLength := OptionLen - 2;

                                      for i := 0 to FastOpenOpt.CookieLength - 1 do
                                          FastOpenOpt.Cookie[i] := (priv_framedata + Offset + 2 + i)^;
                                 end;

                              // Add option to list

                              FastOpenCount := length (priv_fastopenoptionlist);
                              inc (FastOpenCount);
                              setlength (priv_fastopenoptionlist, FastOpenCount);

                              priv_fastopenoptionlist[FastOpenCount-1] := FastOpenOpt;

                              // Add to optionset

                              priv_optionset := priv_optionset + [tcpFastOpen];

                              inc (OptionCount);
                              priv_exactoptionlist[OptionCount-1] := tcpFastOpen;

                         end
                      else
                          addDiagnosticMsg (msWarning, MSG_TCP_BAD_FASTOPEN_OPTION_LEN, '2-18', optionlen.ToString);
                 end;
{
            76 : begin
                      // Riverbed probe

                      // TODO: need more info first

                      pOptionLen := pByte (priv_framedata + Offset + 1);
                      OptionLen := pOptionLen^;

                      if (offset + OptionLen > priv_framelength) then
                         exit (prFailTruncated);

                      RBedOption.initialize;
                      RBedOption.OptionLength := OptionLen;

                      // Type & Version

                      pHelpByte := pByte (priv_framedata + Offset + 2);

                      if (pHelpByte^ and $1) = $1 then
                         begin
                              RBedOption.OptionVersion := 1;
                              RBedOption.OptionType := (pHelpByte^ and $f0) shr 4;
                         end
                      else
                          begin
                               RBedOption.OptionVersion := 2;
                               RBedOption.OptionType := (pHelpByte^ and $fe) shr 1;
                          end;








                      RBedCount := length (priv_riverbedoptionlist);
                      inc (RBedCount);
                      setlength (priv_riverbedoptionlist, RBedCount);

                      priv_riverbedoptionlist[RBedCount-1] := RBedOption;

                      // Add to optionset

                      priv_optionset := priv_optionset + [tcpRiverbed];

                 end
}
            else
                begin
                     // Unknown option, so we read the length to skip it

                     OptionLen := pByte (priv_framedata + Offset + 1)^;

                      // Check for enough bytes to parse (option len includes the type we already read, so -1)

                      if (offset + OptionLen - 1 > priv_lastoffset) then
                         exit (prOptionsTruncated);

                     if OptionCode > 18 then
                        addDiagnosticMsg (msWarning, MSG_TCP_OPTION_CODE_INVALID, OptionCode.ToString)
                     else
                         addDiagnosticMsg (msWarning, MSG_TCP_OPTION_NOT_IMPLEMENTED, OptionCode.ToString);
                end;
           end;

           // Check if the option length is zero (which is not possible because it always has to
           // include at least the option type byte)

           if OptionLen = 0 then
              exit (prFailParsingLoop);

           inc (Offset, OptionLen);

           // Track the total byte length of all options to be able to calculate the protocol layer size value

           inc (priv_totaloptionlen, OptionLen);

     until ((Offset >= (priv_frameoffset + priv_headerlength)) or (Offset > priv_lastoffset)) or EOLOption;

     if Offset > (priv_frameoffset + priv_headerlength) then
        addDiagnosticMsg (msMajor, MSG_TCP_OPTION_SIZE_EXCEEDS_HEADER_LENGTH, (offset - priv_frameoffset).ToString, priv_headerlength.ToString);

     if parseResult <> prSuccess then
        SiMain.LogWarning ('Failed to parse TCP options for frame #' + IntToStr (priv_frameindex));

     // Trim exact option list to actual number of options found

     setlength (priv_exactoptionlist, OptionCount);

     result := parseresult;
end;

(* ------------------------------------------- *)

function tTCPParser.parse: tParseResult;

var pData    : pTCPStruct;
    status   : tParseResult;
    help     : word;
    helplong : longword;
    tcplen   : integer;
    iCount   : integer;
    i        : integer;
    phelp    : pWord;
    pHelp2   : pByte;
    CalcCRC  : boolean;

begin
     status := prSuccess;

     priv_osilayer := 4;     // Set Layer this parser is for

     help := sizeOf (tTCPStruct);

     if priv_frameoffset + help > priv_lastoffset + 1 then
        begin
             prot_layersize := 0;
             status := prFailUndersize;
        end
     else
         try
            pData := pTCPStruct (priv_framedata + priv_frameoffset);

            help := pData^.Sourceport;
            priv_sourceport := Swap16 (help);

            help := pData^.Destinationport;
            priv_destinationport := Swap16 (help);

            helplong := pData^.Sequence;
            priv_sequence.initialize (Swap32 (helplong));

            helplong := pData^.Acknowledge;
            priv_acknowledgement.initialize (Swap32 (helplong));

            help := pData^.DataOffsetFlags;
            help := Swap16 (help);

            // Parse Headerlength and Flags from value

            priv_headerlength := getTCPHeaderLength (help);
//          priv_flags := getTCPFlagSet (help);
            priv_flags := getTCPFlagSetFromWord (help);

            help := pData^.Window;
            priv_windowsize := Swap16 (help);

            help := pData^.CRC;
            priv_checksum := Swap16 (help);

            help := pData^.UrgentPointer;
            priv_urgentpointer := Swap16 (help);

            // Parse TCP options if there are any

            if priv_headerlength > 20 then
               begin
                    status :=  parseTCPOptions (priv_frameoffset + sizeOf (tTCPStruct));

                    if status = prOptionsTruncated then
                       priv_optionset := priv_optionset + [tcpOptionsTruncated];

                    if (status = prFailTruncated) or (status = prFailParsingLoop) then
                       exit (status);
               end;

            if status = prSuccess then
               begin
                    // Calculate Payload length. priv_sectionlen is a value that is set on create, usually as a result of determining it from IP layer payload length

                    priv_payloadlen := priv_sectionlen - priv_headerlength;
               end
            else
                priv_payloadlen := 0;

            // Calculate Next expected Sequence Number

            if (tcpSYN in priv_flags) or (tcpFIN in priv_flags)  then
               priv_nextexpectedsequence := priv_sequence + priv_payloadlen + 1
            else
                priv_nextexpectedsequence := priv_sequence + priv_payloadlen;

            // Determine protocol layer size

            if not (status = prOptionsTruncated) then
               prot_layersize := priv_headerlength
            else
                prot_layersize := 20 + priv_totaloptionlen;

            // Prepare the word array for crc calculation if possible - frame must be long enough. So we check
            // if the frame was sliced, but if it was, we also check if it's maybe still possible to calculcate
            // the CRC (e.g. when the slicing only took away Ethernet padding bytes)

            CalcCRC := (priv_framelength = priv_wirelength);

            if not CalcCRC then
               CalcCRC := (status = prSuccess);

            // But not too long - segmentation offloading seems to be the death to getWordsStructFromOffset, so we skip it in that case

            if priv_framelength > 9018 then
               CalcCRC := false;               // TODO: see if we can manage TCP checksum calculation for large segments in a more elegant way

            if CalcCRC then
               priv_tcpsectionwords := getWordsStructFromOffset (priv_FrameData, priv_frameOffset, priv_sectionlen)
            else
                begin
                     status := prFailPayloadTruncated;
                     setlength (priv_tcpsectionwords, 0);
                end;

            // Prepare next anchor

            priv_nextAnchor.Offset := priv_frameoffset + priv_headerlength;

            if (priv_payloadlen > 0) then
               begin
                    if status = prFailPayloadTruncated then
                       priv_nextAnchor.HeaderType := NEXT_HEADER_PAYLOAD
                    else
                        priv_nextAnchor.HeaderType := NEXT_HEADER_DYNAMIC;
               end
            else
                priv_nextAnchor.HeaderType := NEXT_HEADER_NONE;

         except
               status := prFailInvalidValues;
         end;

     result := status;
end;

(* ------------------------------------------- *)

function tTCPParser.PortPair: tPortPair;

begin
     result.Value1 := priv_sourceport;
     result.Value2 := priv_destinationport;
end;

(* ------------------------------------------- *)


end.
