{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_parser_tzsp;

interface

uses base_structures,
     common_parse_helpers,
     class_parser_baseclass;

type tTZSPHdrStruct = packed record
                                   version       : byte;
                                   TZSPType      : byte;
                                   Protocol      : word;
                      end;
     pTZSPHdrStruct = ^tTZSPHdrStruct;

     tTZSPType = (tzspReceivedTagList, tzspPacketForTransmit, tzspReserved, tzspConfiguration, tzspKeepalive, tzspPortOpener);

     tTZSPTaggedField = (tzfPadding, tzfEnd, tzfRawRSSI, tzfSNR, tzfDataRate, tzfTimestamp, tzfContentionFree, tzfDecrypted, tzfFCSError, tzfRXChannel, tzfPacketCount, tzfRXFrameLength, tzfWLANRadioHDRSerial);
     tTZSPTaggedFieldSet = set of tTZSPTaggedField;

     tTZSPSerialArray = array of byte;

     tTZSPParser  = class (tFrameParser)

                            private
                                   priv_Version                : byte;
                                   priv_TZSPType               : tTZSPType;
                                   priv_EncapsulatedProtocol   : longword;

                                   priv_TaggedFieldTotalLength : integer;
                                   priv_TaggedFields           : tTZSPTaggedFieldSet;

                                   priv_RawRSSI                : smallint;
                                   priv_SNR                    : smallint;
                                   priv_DataRate               : double;
                                   priv_Timestamp              : longword;
                                   priv_ContentionFree         : boolean;
                                   priv_Decrypted              : boolean;
                                   priv_FCSError               : boolean;
                                   priv_RXChannel              : shortint;
                                   priv_PacketCount            : longword;
                                   priv_RXFrameLength          : word;
                                   priv_WLANRadioHDRSerial     : tTZSPSerialArray;

                                   procedure initializeTaggedFields;
                                   function ParseTaggedFields (const Offset : integer) : tParseResult;

                            public
                                  property Version : byte read priv_Version;
                                  property TZSPType : tTZSPType read priv_TZSPType;
                                  property EncapsulatedProtocol : longword read priv_EncapsulatedProtocol;
                                  property TaggedFields : tTZSPTaggedFieldSet read priv_TaggedFields;

                                  property RawRSSI : smallint read priv_RawRSSI;
                                  property SNR : smallint read priv_SNR;
                                  property DataRateMBps : double read priv_DataRate;
                                  property Timestamp : longword read priv_Timestamp;
                                  property ContentionFree : boolean read priv_ContentionFree;
                                  property Decrypted : boolean read priv_Decrypted;
                                  property FCSError : boolean read priv_FCSError;
                                  property RXChannel : shortint read priv_RXChannel;
                                  property PacketCount : longword read priv_PacketCount;
                                  property RXFrameLength : word read priv_RXFrameLength;
                                  property WLANRadioHdrSerial : tTZSPSerialArray read priv_WLANRadioHdrSerial;

                                  function checkIsPlausible : boolean; override;
                                  function parse : tParseResult; override;
                        end;

implementation

uses Math,
     Toolkit;

const cMaxWLANRadioHdrSerialBytes = 256;

(* ------------------------------------------- *)
{ tTZSPParser }
(* ------------------------------------------- *)

function tTZSPParser.checkIsPlausible: boolean;

var value : boolean;
    pData : pTZSPHdrStruct;
    help  : word;

begin
     value := true;

     if priv_framelength < priv_frameoffset + sizeOf (tTZSPHdrStruct) then
        value := false
     else
         begin
              pData := pTZSPHdrStruct (priv_framedata + priv_frameoffset);

              help := swap16 (pData^.Protocol);

              // Check if we have version 1, Types less than 6 and Ethernet or IEEE 802.11

              if (pData^.version <> 1) or (pData^.TZSPType > 5) or ((help <> $1) and (help <> $12)) then
                 value := false;
         end;

     result := value;
end;

(* ------------------------------------------- *)

procedure tTZSPParser.initializeTaggedFields;

begin
     priv_TaggedFieldTotalLength := 0;
     priv_TaggedFields := [];

     priv_RawRSSI := 0;
     priv_SNR := 0;
     priv_DataRate := 0;
     priv_Timestamp := 0;
     priv_ContentionFree := false;
     priv_Decrypted := false;
     priv_FCSError := false;
     priv_RXChannel := 0;
     priv_PacketCount := 0;
     priv_RXFrameLength := 0;
     setLength (priv_WLANRadioHDRSerial, 0);
end;

(* ------------------------------------------- *)

function tTZSPParser.ParseTaggedFields (const Offset: integer) : tParseResult;

var pTagType    : pByte;
    boDone      : boolean;
    localoffset : integer;
    taglength   : integer;
    pHelpByte   : pByte;
    pHelpWord   : pWord;
    pHelpInt8   : pShortInt;
    pHelpInt16  : pSmallInt;
    pHelpLong   : pLongword;
    status      : tParseResult;
    i           : integer;

begin
     result := prSuccess;
     boDone := false;
     localoffset := offset;

     initializeTaggedFields;

     repeat
           taglength := 0;

           // Check if we can read one tag type byte

           if priv_framelength < localoffset + 1 then
              status := prFailUndersize
           else
               begin
                    pTagType := pByte (priv_framedata + localoffset);

                    // tzfPadding, tzfEnd, tzfRawRSSI, tzfSNR, tzfDataRate, tzfTimestamp, tzfContentionFree, tzfDecrypted, tzfFCSError, tzfRXChannel, tzfPacketCount, tzfRXFrameLength, tzfWLANRadioHDRSerial

                    if pTagType^ > 1 then
                       begin
                            // Check if we can read the length byte

                            if priv_framelength < localoffset + 2 then
                               exit (prFailUndersize);

                            TagLength := pByte (priv_framedata + localoffset + 1)^;

                            // Check if we can read the tag itself

                            if priv_framelength < localoffset + 2 + TagLength then
                               exit (prFailUndersize);
                       end;

                    case pTagType^ of
                      0 : begin
                               // Padding, ignore

                               priv_TaggedFields := priv_TaggedFields + [tzfPadding];

                               tagLength := 1;
                          end;
                      1 : begin
                               // End

                               priv_TaggedFields := priv_TaggedFields + [tzfEnd];

                               tagLength := 1;
                               boDone := true;
                          end;
                     10 : begin
                               // RAW RSSI

                               priv_TaggedFields := priv_TaggedFields + [tzfRawRSSI];

                               case TagLength of
                                1 : begin
                                         pHelpInt8 := pShortInt (priv_framedata + localoffset + 2);
                                         priv_RawRSSI := pHelpInt8^;
                                    end;
                                2 : begin
                                         pHelpInt16 := pSmallInt (priv_framedata + localoffset + 2);
                                         priv_RawRSSI := swap16 (pHelpInt16^);
                                    end;
                               else
                                   status := prFailInvalidValues;
                               end;

                               inc (Taglength, 2);
                          end;
                     11 : begin
                               // SNR

                               priv_TaggedFields := priv_TaggedFields + [tzfSNR];

                               case TagLength of
                                1 : begin
                                         pHelpInt8 := pShortInt (priv_framedata + localoffset + 2);
                                         priv_SNR := pHelpInt8^;
                                    end;
                                2 : begin
                                         pHelpInt16 := pSmallInt (priv_framedata + localoffset + 2);
                                         priv_SNR := swap16 (pHelpInt16^);
                                    end;
                               else
                                   status := prFailInvalidValues;
                               end;

                               inc (Taglength, 2);
                          end;
                     12 : begin
                               // DATA RATE

                               priv_TaggedFields := priv_TaggedFields + [tzfDataRate];

                               if TagLength <> 1 then
                                  status := prFailInvalidValues
                               else
                                   begin
                                        pHelpInt8 := pShortInt (priv_framedata + localoffset + 2);

                                        // get MB/s Data rate from known index values

                                        case pHelpInt8^ of
                                           2 : priv_DataRate := 1;
                                           4 : priv_DataRate := 2;
                                          10 : priv_DataRate := 1;    // deprecated
                                          11 : priv_DataRate := 5.5;
                                          12 : priv_DataRate := 6;
                                          18 : priv_DataRate := 9;
                                          20 : priv_DataRate := 2;    // deprecated
                                          22 : priv_DataRate := 11;
                                          24 : priv_DataRate := 12;
                                          36 : priv_DataRate := 18;
                                          44 : priv_DataRate := 22;
                                          48 : priv_DataRate := 24;
                                          55 : priv_DataRate := 5.5;  // deprecated
                                          66 : priv_DataRate := 33;
                                          72 : priv_DataRate := 36;
                                          96 : priv_DataRate := 48;
                                         108 : priv_DataRate := 54;
                                         110 : priv_DataRate := 11;   // deprecated
                                        end;
                                   end;

                               inc (Taglength, 2);
                          end;
                     13 : begin
                               // TIMESTAMP

                               priv_TaggedFields := priv_TaggedFields + [tzfTimestamp];

                               if TagLength = 4 then
                                  begin
                                       pHelpLong := pLongword (priv_framedata + localoffset + 2);

                                       priv_Timestamp := swap32 (pHelpLong^);
                                  end
                               else
                                   status := prFailInvalidValues;

                               inc (Taglength, 2);
                          end;
                     15 : begin
                               // CONTENTION FREE

                               priv_TaggedFields := priv_TaggedFields + [tzfContentionFree];

                               if TagLength = 1 then
                                  begin
                                       pHelpByte := (priv_framedata + localoffset + 2);

                                       priv_ContentionFree := (pHelpByte^ = 0);
                                  end
                               else
                                   status := prFailInvalidValues;

                               inc (Taglength, 2);
                          end;
                     16 : begin
                               // DECRYPTED

                               priv_TaggedFields := priv_TaggedFields + [tzfDecrypted];

                               if TagLength = 1 then
                                  begin
                                       pHelpByte := (priv_framedata + localoffset + 2);

                                       priv_Decrypted := (pHelpByte^ = 0);
                                  end
                               else
                                   status := prFailInvalidValues;

                               inc (Taglength, 2);
                          end;
                     17 : begin
                               // FCS ERROR

                               priv_TaggedFields := priv_TaggedFields + [tzfFCSError];

                               if TagLength = 1 then
                                  begin
                                       pHelpByte := (priv_framedata + localoffset + 2);

                                       priv_FCSError := not (pHelpByte^ = 0);
                                  end
                               else
                                   status := prFailInvalidValues;

                               inc (Taglength, 2);
                          end;
                     18 : begin
                               // RX CHANNEL

                               priv_TaggedFields := priv_TaggedFields + [tzfRXChannel];

                               if TagLength = 1 then
                                  begin
                                       pHelpByte := (priv_framedata + localoffset + 2);

                                       priv_RXChannel := pHelpByte^;
                                  end
                               else
                                   status := prFailInvalidValues;

                               inc (Taglength, 2);
                          end;
                     40 : begin
                               // PACKET COUNT

                               priv_TaggedFields := priv_TaggedFields + [tzfPacketCount];

                               if TagLength = 4 then
                                  begin
                                       pHelpLong := pLongword (priv_framedata + localoffset + 2);

                                       priv_PacketCount := swap32 (pHelpLong^);
                                  end
                               else
                                   status := prFailInvalidValues;

                               inc (Taglength, 2);
                          end;
                     41 : begin
                               // RX_FRAME_LENGTH

                               priv_TaggedFields := priv_TaggedFields + [tzfRXFrameLength];

                               if TagLength = 2 then
                                  begin
                                       pHelpWord := pWord (priv_framedata + localoffset + 2);

                                       priv_RXFrameLength := swap16 (pHelpWord^);
                                  end
                               else
                                   status := prFailInvalidValues;

                               inc (Taglength, 2);
                          end;
                     60 : begin
                               // WLAN RADIO HDR SERIAL

                               priv_TaggedFields := priv_TaggedFields + [tzfWLANRadioHdrSerial];

                               // Set an arbitrary max to avoid crazy numbers killing us

                               setLength (priv_WLANRadioHDRSerial, min (TagLength, cMaxWLANRadioHdrSerialBytes));

                               // Read the stuff in a loop

                               for i := 0 to length (priv_WLANRadioHDRSerial) - 1 do
                                   begin
                                        pHelpByte := (priv_framedata + localoffset + 2 + i);

                                        priv_WLANRadioHDRSerial[i] := pHelpByte^;
                                   end;

                               inc (Taglength, 2);
                          end;
                    end;

                    assert (taglength > 0, 'Invalid tag length (tTZSPParser.ParseTaggedFields)');
                    inc (localoffset, tagLength);
               end;

     until boDone or (status <> prSuccess);

     priv_TaggedFieldTotalLength := localoffset - offset;

     result := status;
end;

(* ------------------------------------------- *)

function tTZSPParser.parse: tParseResult;

var status    : tParseResult;
    pData     : pTZSPHdrStruct;
    wHelp     : word;

begin
     priv_osilayer := 3;     // Set Layer this parser is for
     prot_layersize := sizeOf (pTZSPHdrStruct);

     status := prSuccess;

     if priv_framelength < priv_frameoffset + sizeOf (tTZSPHdrStruct) then
        status := prFailUndersize
     else
         try
            pData := pTZSPHdrStruct (priv_framedata + priv_frameoffset);

            // Version and Length (length as 32 bit value count)

            priv_Version := pData^.version;

            case pData^.TZSPType of
             0 : priv_TZSPType := tzspReceivedTagList;
             1 : priv_TZSPType := tzspPacketForTransmit;
             2 : priv_TZSPType := tzspReserved;
             3 : priv_TZSPType := tzspConfiguration;
             4 : priv_TZSPType := tzspKeepalive;
             5 : priv_TZSPType := tzspPortOpener;
            end;

            wHelp := swap16 (pData^.Protocol);

            case wHelp of
              1 : priv_EncapsulatedProtocol := NEXT_HEADER_ETHERNET;
             18 : priv_EncapsulatedProtocol := NEXT_HEADER_802dot11Hdr;
            end;

            // Read tagged fields

            ParseTaggedFields (priv_frameoffset + sizeOf (tTZSPHdrStruct));


            // Determine next anchor

            priv_nextAnchor.HeaderType := priv_EncapsulatedProtocol;
            priv_nextAnchor.Offset := priv_frameoffset + sizeOf (tTZSPHdrStruct) + priv_TaggedFieldTotalLength;

         except
               status := prFailInvalidValues;
         end;

     result := status;
end;

(* ------------------------------------------- *)

end.
