{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_trace_preparser;

interface

uses Contnrs,
     base_structures,
     base_determine_protocols,
     definition_ip_addresses,
     class_rawframe,
     class_loader_pcapng,
     class_tracefilelist,
     class_tracefilesetloader,
     class_parsemaster,
     class_inteldb_manager,
     class_parser_baseclass,
     class_parser_tcp,
     class_parser_udp,
     class_parser_dhcpv4;

type tOnStartScanOfFile = procedure (const FileID : integer; FileEntry : tTracefileEntry) of object;
     tOnAdvancePercentScanned = procedure (const FileID : integer; FileEntry : tTracefileEntry; const PercentageParsed : integer) of object;
     tOnPreparseFrame = procedure (const FileID : integer; FileEntry : tTracefileEntry; const FrameIndex : integer) of object;
     tOnErrorParsingFrame = procedure (const FileID : integer; FileEntry : tTracefileEntry; const FrameIndex : uint64) of object;
     tOnCompleteScanOfFile = procedure (const FileID : integer; FileEntry : tTracefileEntry; const ErrorCount : integer; const CacheHitRatio : string) of object;

     tTracePreparser = class (tObject)

                             {$REGION 'Documentation'}
                             ///	<summary>
                             ///	  Creates a TracePreparser object
                             ///	</summary>
                             ///	<param name="TraceIntelDB">
                             ///	  Trace IntelDB Manager, must be already
                             ///	  created and connected to database, and
                             ///	  freeing must be handled afterwards, too
                             ///	</param>
                             {$ENDREGION}
//                           constructor Create (TraceIntelDB : tIntelDBManager);
                             constructor Create (TraceSetLoader : tTracefileSetLoader; UseDiskDatabase : boolean);
                             destructor Destroy; override;

                             private
                                    EventOnStartScanOfFile       : tOnStartScanOfFile;
                                    EventOnAdvancePercentScanned : tOnAdvancePercentScanned;
                                    EventOnPreparseFrame         : tOnPreparseFrame;
                                    EventOnErrorParsingFrame     : tOnErrorParsingFrame;
                                    EventOnCompleteScanOfFile    : tOnCompleteScanOfFile;

                                    priv_usediskdatabase         : boolean;
                                    priv_TotalFramesScanned      : integer;
                                    priv_ParseErrorCount         : integer;

                                    priv_TraceSetLoader          : tTracefileSetLoader;
                                    priv_inteldb                 : tIntelDBManager;
                                    priv_parsemaster             : tParseMaster;

                                    procedure preparsePCAPngHeaders (Traceloader : tPCAPNGLoader);
                                    procedure PreparseFile (Index : integer; FileEntry : tTracefileEntry);
                                    procedure AddFrame (Frame : tRawFrame);

                             public
                                   property onStartScanOfFile : tOnStartScanOfFile read EventOnStartScanOfFile write EventOnStartScanOfFile;
                                   property onCompleteScanOfFile : tOnCompleteScanOfFile read EventOnCompleteScanOfFile write EventOnCompleteScanOfFile;
                                   property onPreparseFrame : tOnPreparseFrame read EventOnPreparseFrame write EventOnPreparseFrame;
                                   property onErrorParsingFrame : tOnErrorParsingFrame read EventOnErrorParsingFrame write EventOnErrorParsingFrame;
                                   property onAdvancePercentScanned : tOnAdvancePercentScanned read EventOnAdvancePercentScanned write EventOnAdvancePercentScanned;
                                   property ParseErrorCount : integer read priv_ParseErrorCount;

                                   procedure run;

                       end;

implementation

uses DateUtils,
     SiAuto,
     base_checksums,
     definition_simple_nameresolution,

     // Loader

     class_traceloader,
     class_loader_enc,
     class_loader_naicap,
     class_loader_pcap,

     // Parser

     class_parser_ethernet,
     class_parser_tokenring,
     class_parser_802dot1q,
     class_parser_mpls,
     class_parser_arp,
     class_parser_ipv4,
     class_parser_ipv6,
     class_parser_icmpv4;

(* ------------------------------------------- *)

// constructor tTracePreparser.Create (TraceIntelDB : tIntelDBManager);
constructor tTracePreparser.Create (TraceSetLoader : tTraceFileSetLoader; UseDiskDatabase : boolean);

begin
     inherited Create;

     priv_usediskdatabase := UseDiskDatabase;
     priv_TraceSetLoader := TraceSetLoader;
end;

(* ------------------------------------------- *)

destructor tTracePreparser.Destroy;

begin
     inherited;
end;

(* ------------------------------------------- *)

procedure tTracePreparser.AddFrame (Frame: tRawFrame);

begin
     // Create a parsemaster that will dissect all the layers for us

     priv_parsemaster := tParseMaster.create (Frame, ppmParseAll);

     // If we have simple ethernet - ip layers (no IPinIP) we can just add the details to the DB

     if priv_parsemaster.EtherLayerCount = 1 then
        begin
             priv_inteldb.addEtherConversationDetails (priv_parsemaster.MACAddressPair, Frame.FileFrameID, Frame.Timestamp.datetime, priv_parsemaster.VLANID);
        end
     else
         begin
              // TODO: handle multiple Ethernet layers

         end;

     if priv_parsemaster.IPLayerCount = 1 then
        begin
             case priv_parsemaster.LastIPLayerVersion of
              4 : priv_inteldb.addIPv4ConversationDetails (priv_parsemaster.IPv4SocketPair, priv_parsemaster.Layer4Protocol, Frame.FileFrameID, Frame.Timestamp.datetime, priv_parsemaster.VLANID);
              6 : priv_inteldb.addIPv6ConversationDetails (priv_parsemaster.IPv6SocketPair, priv_parsemaster.Layer4Protocol, Frame.FileFrameID, Frame.Timestamp.datetime, priv_parsemaster.VLANID);
             end;
        end
     else
         begin
              // TODO: handle multiple IP layers

         end;

     // Drop the parsemaster object, we're done here

     priv_parsemaster.free;
end;

(* ------------------------------------------- *)

procedure tTracePreparser.preparsePCAPngHeaders (Traceloader: tPCAPNGLoader);

var NREntryIPv4 : tSimpleIPv4NameResolutionEntry;
    NREntryIPv6 : tSimpleIPv6NameResolutionEntry;
    i           : integer;

begin
     for i := 1 to Traceloader.NameResolutionManagerIPv4.count do
         begin
              NREntryIPv4 := Traceloader.NameResolutionManagerIPv4.GetEntry(i);

              priv_inteldb.addIPv4AddressFromNRB (NREntryIPv4.IPv4Address);
         end;

     for i := 1 to Traceloader.NameResolutionManagerIPv6.count do
         begin
              NREntryIPv6 := Traceloader.NameResolutionManagerIPv6.GetEntry(i);

              priv_inteldb.addIPv6AddressFromNRB (NREntryIPv6.IPv6Address);
         end;
end;

(* ------------------------------------------- *)

procedure tTracePreparser.PreparseFile (Index : integer; FileEntry: tTracefileEntry);

var fileframesscanned : integer;
    percentagescanned : integer;
    traceloader       : tTraceLoader;
    frame             : tRawFrame;
    dHelp             : double;
    iHelp             : integer;
    i                 : integer;

begin
     percentagescanned := 0;
     fileframesscanned := 0;

     // Mark as scan active

     priv_inteldb.setFileProcessingStatus (psScanningConversations);

     // Get the right parser

     case fileentry.TracefileType of
      cfENC    : traceloader := tENCLoader.create (FileEntry.FullName, Index, true);
      cfCAP    : traceloader := tCAPLoader.create (FileEntry.FullName, Index, true);
      cfPCAP   : traceloader := tPCAPLoader.create (FileEntry.FullName, Index, true);
      cfPCAPng : traceloader := tPCAPNGLoader.create (FileEntry.FullName, Index, true);
     end;

     if assigned (EventOnStartScanOfFile) then
        EventOnStartScanOfFile (i, fileentry);

     SiMain.LogDebug ('Preparsing File "' + fileentry.FullName + '" for conversations etc.');

     // Open the file

     traceloader.openFile;
     traceloader.readTraceFileheader;

     if fileentry.TracefileType = cfPCAPng then
        preparsePCAPngHeaders (traceloader as tPCAPNGLoader);

     // Read frames

     repeat
           traceloader.readBatchFromDisk;

           for i := 1 to traceloader.FramesAvailable do
               begin
                    frame := traceloader.getRawFrame(i);

                    if assigned (EventOnPreparseFrame) then
                       EventOnPreparseFrame (Index, fileentry, fileframesscanned);

                    try
                       AddFrame (frame);
                    except
                          if assigned (onErrorParsingFrame) then
                             onErrorParsingFrame (Index, fileentry, frame.FileFrameID);

                          inc (priv_ParseErrorCount);
                    end;

                    // Drop the frame, we're done with it

                    frame.free;
                    frame := nil;

                    inc (fileframesscanned);
                    inc (priv_TotalFramesScanned);

                    // Trigger report if possible

                    if (fileentry.FrameCount > 0) and assigned (EventOnAdvancePercentScanned) then
                       begin
                            dHelp := 100 * fileframesscanned / fileentry.FrameCount;
                            iHelp := trunc (dHelp);

                            if iHelp > percentagescanned then
                               begin
                                    percentagescanned := iHelp;

                                    EventOnAdvancePercentScanned (i, fileentry, percentagescanned);
                               end;
                       end;
               end;

     until traceloader.EOF;

     priv_inteldb.FlushCaches;

     // Write global values to DB

     priv_inteldb.writeAdditionalFileDetails;

     // Mark as processed

     fileentry.ProcessingStatus := psTaskPending;

     priv_inteldb.setFileProcessingStatus (fileentry.ProcessingStatus);

     if assigned (EventOnCompleteScanOfFile) then
        EventOnCompleteScanOfFile (i, fileentry, priv_ParseErrorCount, priv_inteldb.CacheHitRatioSummary);

     // Cleanup

     priv_inteldb.free;

     traceloader.closeFile;
     traceloader.free;

     SiMain.LogSeparator;
end;

(* ------------------------------------------- *)

procedure tTracePreparser.run;

var fileentry         : tTracefileEntry;
    inteldbfile       : string;
    i,j               : integer;

begin
     SiMain.EnterMethod ('tTracePreparser.Run');

     priv_TraceSetLoader.reset;

     priv_TotalFramesScanned := 0;

     for i := 1 to priv_TraceSetLoader.FileCount do
         begin
              fileentry := priv_TraceSetLoader.getFile(i);
              priv_ParseErrorCount := 0;

              inteldbfile := fileentry.FilePath + 'traceintel.db';

{$ifdef debug}
              priv_inteldb := tIntelDBManager.create (inteldbfile, fileentry.FullName, fileentry.FileSize, fileentry.FileDate, fileentry.FirstFrameTimestamp, priv_usediskdatabase, true);
{$else}
              priv_inteldb := tIntelDBManager.create (inteldbfile, fileentry.FullName, fileentry.FileSize, fileentry.FileDate, fileentry.FirstFrameTimestamp, priv_usediskdatabase, false);
{$endif}

              PreparseFile (i, fileentry);
         end;

     SiMain.LeaveMethod ('tTracePreparser.Run');
end;

(* ------------------------------------------- *)

end.

