{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit common_parse_helpers;

interface

uses class_diagmessagehandler,
     definition_ip_addresses,
     definition_5tuple;


type tParseResult = (prSuccess, prOptionsTruncated, prFailUnknown, prFailUndersize, prFailInvalidValues, prFailTruncated, prFailPayloadTruncated, prFailOutOfBounds, prFailParsingLoop, prFailPlausibleCheck, prFailSegmentedPayload, prNotParsed);
     tClientServerOrder = (csoNone, csoClientIsSource, csoClientIsDestination);

     tHeaderType = integer;

     tTLVByteByteHeader = packed record
                                       OptionType   : byte;
                                       OPtionLength : byte;
                          end;
     pTLVByteByteHeader = ^tTLVByteByteHeader;

     tTLVByteWordHeader = packed record
                                       OptionType   : byte;
                                       OPtionLength : word;
                          end;
     pTLVByteWordHeader = ^tTLVByteWordHeader;

     tTLVWordWordHeader = packed record
                                       OptionType   : word;
                                       OPtionLength : word;
                          end;
     pTLVWordWordHeader = ^tTLVWordWordHeader;

     tProtocolAnchor = packed record
                             HeaderType  : tHeaderType;             // Protocol ID
                             Offset      : integer;                 // Offset of the start of the protocol header in the frame buffer
                       end;

     tFQDNResult = record
                         Name            : tFQDN;
                         OffsetIncrement : integer;
                         ParseResult     : tParseResult;
                         Msg             : tDiagnosticMsg;
                   end;

     t802dot1qData    = record
                              priority    : byte;
                              cfi         : boolean;
                              vlanid      : integer;

                              procedure initialize;
                        end;

         {$REGION 'Documentation'}
         /// <summary>
         ///   Checks if a bit in a longword is set.
         /// </summary>
         /// <param name="value">
         ///   Set of bits to be tested
         /// </param>
         /// <param name="BitToTest">
         ///   The bit to be tested. Index starts at zero!
         /// </param>
         {$ENDREGION}
         function BitIsSet (const value : longword; const BitToTest : byte) : boolean; inline;
         {$REGION 'Documentation'}
         /// <summary>
         ///   Sets a bit in a longword
         /// </summary>
         /// <param name="value">
         ///   longword value to be modified
         /// </param>
         /// <param name="BitToSet">
         ///   Bit that is going to be set. Index starts at zero!
         /// </param>
         {$ENDREGION}
         function SetBit (const value : longword; const BitToSet : byte) : longword; inline;
         {$REGION 'Documentation'}
         /// <summary>
         ///   Clears a bit in a longword
         /// </summary>
         /// <param name="value">
         ///   longword value to be modified
         /// </param>
         /// <param name="BitToSet">
         ///   Bit that is going to be cleared. Index starts at zero!
         /// </param>
         {$ENDREGION}
         function ClearBit (const value : longword; const BitToSet : byte) : longword; inline;
         /// <summary>
         ///   This function returns a value extracted from a longword, which
         ///   is often required to parse details from protocols where values
         ///   are stored as subsets of bytes, e.g. 5 bits forming a value.
         /// </summary>
         /// <param name="value">
         ///   Longword containing the value to be parsed
         /// </param>
         /// <param name="startbit">
         ///   position of the first bit
         /// </param>
         /// <param name="count">
         ///   number of bits that make up the value
         /// </param>
         /// <returns>
         ///   extracted value
         /// </returns>
         function ExtractBitRange (const value : longword; const startbit, count : byte) : longword; inline;
         /// <summary>
         ///   Returns either the high or the low nibble of a byte
         /// </summary>
         /// <param name="ByteValue">
         ///   Original Byte
         /// </param>
         /// <param name="HighNibble">
         ///   if true, the high nibble is returned
         /// </param>
         function GetNibble (const ByteValue : byte; HighNibble : boolean) : byte; inline;

         function ParseByteOption (Data : pByte) : byte;
         function ParseMACAddressOption (Data : pByte) : tMACAddress;
         function ParseStringOption (Data : pByte; Length : integer) : string;

         {$REGION 'Documentation'}
         /// <summary>
         ///   Parses FQDN details from frames, including following pointers
         ///   when necessary
         /// </summary>
         /// <param name="FrameData">
         ///   Pointer to the frame data
         /// </param>
         /// <param name="SectionSize">
         ///   Size of the available data section, for range checks
         /// </param>
         /// <param name="PointerBaseOffset">
         ///   Base offset where the current parse should jump from when
         ///   following pointers. Usually the start offset of the DNS layer
         /// </param>
         /// <param name="ReadAtOffset">
         ///   Offset at which the FQDN shall be parsed from
         /// </param>
         /// <param name="Depth">
         ///   Counter for following pointers, to avoid endless pointer loops
         /// </param>
         /// <remarks>
         ///   When parsing the FQDN runs into trouble, the tFQDNResult will
         ///   have an error code greater than 0
         /// </remarks>
         {$ENDREGION}
         function readFQDNFromOffset (const FrameData : pByte; const SectionSize, PointerBaseOffset, ReadAtOffset: integer; const Depth : integer = 0): tFQDNResult;
//       function readFQDNFromOffset (const FrameData : pByte; const FrameSize, PointerBaseOffset, ReadAtOffset: integer; const Depth : integer = 0): tFQDNResult;

         /// <summary>
         ///   Parses the VLAN information from a Word
         /// </summary>
         /// <param name="Value">
         ///   raw bytes containing the VLAN information
         /// </param>
         function Parse802dot1qData (const Value : word) : t802dot1qData; inline;

         /// <summary>
         ///   Tries to determine client and server from ports. May not be 100%
         ///   correct, but it can be a start.
         /// </summary>
         function DetermineClientOfTuple (const Tuple : t5Tuple) : tClientServerOrder;

         function ExtractLongWordFromByteArray (const buffer : array of byte; const offset : integer) : longword;
         procedure PatchLongwordToByteArray (var buffer : array of byte; const offset : integer; const value : longword);

const ParseStatusNoError = 0;
      FQDNParseErrorPointerLoopExceeded = 1;
      FQDNParseErrorParsingOutOfBounds = 2;
      DNSParseErrorParsingOutOfBounds = 3;
      ParseStatusNotParsed = 255;


implementation

uses SysUtils,
     Math,
     toolkit,
     base_structures,
     base_messages;

const MaxPointerLoopCount = 5;

(* ------------------------------------------- *)
{ t802dot1qData }
(* ------------------------------------------- *)

procedure t802dot1qData.initialize;

begin
     priority := 0;
     cfi := false;
     vlanid := 0;
end;

(* ------------------------------------------- *)

function BitIsSet (const value : longword; const BitToTest : byte) : boolean; inline;

begin
     result := (value and (1 shl BitToTest)) <> 0;
end;

(* ------------------------------------------- *)

function SetBit (const value : longword; const BitToSet : byte) : longword; inline;

begin
     result := value or (1 shl BitToSet);
end;

(* ------------------------------------------- *)

function ClearBit (const value : longword; const BitToSet : byte) : longword; inline;

begin
     result := value and not (1 shl BitToSet);
end;

(* ------------------------------------------- *)

function ExtractBitRange (const value : longword; const startbit, count : byte) : longword; inline;

var extract : longword;
    mask    : longword;

begin
     extract := 0;

     // Step 1: create mask

     mask := trunc (power (2, count) - 1);

     if startbit > 0 then
        mask := mask shl (startbit - 1);

     // Step 2: apply mask

     extract := (value and mask);

     // Step 3: shift results

     if startbit > 0 then
        extract := extract shr startbit;

     result := extract;
end;

(* ------------------------------------------- *)

function GetNibble (const ByteValue : byte; HighNibble : boolean) : byte; inline;

begin
     if HighNibble then
        result := (ByteValue and $f0) shr 4
     else
         result := ByteValue and $0f;
end;

(* ------------------------------------------- *)

function ParseByteOption (Data : pByte) : byte;

begin
     result := Data^;
end;

(* ------------------------------------------- *)

function ParseMACAddressOption (Data : pByte) : tMACAddress;

begin
     result.RangeBytes := pMACBytes (Data)^;
end;

(* ------------------------------------------- *)

function ParseStringOption (Data : pByte; Length : integer) : string;

var help  : pANSIChar;
    value : string;

begin
     Help := pANSIChar (Data);
     value := string (Help);

     result := copy (value, 1, Length);
end;

(* ------------------------------------------- *)

function readFQDNFromOffset (const FrameData : pByte; const SectionSize, PointerBaseOffset, ReadAtOffset: integer; const Depth : integer = 0): tFQDNResult;

var RecordLength  : Byte;
    NewOffset     : pWord;
    HelpWord      : word;
    HelpChar      : pANSIChar;
    HelpString    : string;
    FQDN          : string;
    ResultSet     : tFQDNResult;
    HelpSet       : tFQDNResult;
    PointerOffset : integer;
    Offset        : integer;
    boParseDone   : boolean;
    LastOffset    : integer;
    boAbort       : boolean;

begin
     ResultSet.Name.initialize;
     ResultSet.OffsetIncrement := 0;
     ResultSet.ParseResult := prSuccess;
     ResultSet.Msg.initialize;
     boAbort := false;

     Offset := ReadAtOffset;
     LastOffset := PointerBaseOffset + SectionSize - 1;

     // Check if we're trying to parse before or after the valid range (start and end of the DNS layer)

     if (Offset < PointerBaseOffset) or (Offset > LastOffset) then
        begin
             ResultSet.ParseResult := prFailOutOfBounds;
             exit (ResultSet);
        end
     else
         ResultSet.ParseResult := prSuccess;

     // Recursion attack defense - we do not follow pointers into more than one layer of depth or we might loop to death

     if Depth > MaxPointerLoopCount then
        begin
             ResultSet.ParseResult := prFailParsingLoop;
             ResultSet.Msg.initialize (msMajor, MSG_DNS_POINTER_LOOP_EXCEEDED, 'FQDNs', Depth.toString, MaxPointerLoopCount.ToString);

             Exit (ResultSet);
        end;

     boParseDone := false;

     // Read Name block by block until done (zero terminated domain parts and/or pointers)

     repeat
           HelpString := '';

           RecordLength := pByte (FrameData + Offset)^;

           case RecordLength of
            $0       : begin
                            // Zero length - we're done

                            inc (ResultSet.OffsetIncrement);
                            boParseDone := true;
                       end;
            $c0..$ff : begin
                            // when the two highest bits of the length are set, we got a pointer

                            if (Offset + 1) <= lastoffset then
                               begin
                                    // the rest of the bits are the relative offset to the label

                                    NewOffset := pWord (FrameData + Offset);
                                    HelpWord := NewOffset^;
                                    HelpWord := Swap16 (HelpWord);
                                    HelpWord := HelpWord AND $3fff;      // remove the two highest bits, which are the pointer indicator. The rest is the offset to jump to

                                    PointerOffset := PointerBaseOffset + HelpWord;

                                    // Make sure that the pointer points to something valid, meaning it must point to
                                    // offsets inside the DNS payload

                                    if (PointerOffset >= PointerBaseOffset) and (PointerOffset < LastOffset) then
                                       begin
                                            HelpSet := readFQDNFromOffset (FrameData, SectionSize, PointerBaseOffset, PointerOffset, Depth + 1);

                                            if FQDN = '' then
                                               FQDN := HelpSet.Name.FullFQDN
                                            else
                                                FQDN := FQDN + '.' + HelpSet.Name.FullFQDN;

                                            inc (ResultSet.OffsetIncrement, 2);     // for the pointer and pointer offset bytes
                                       end
                                    else
                                        begin
                                             ResultSet.ParseResult := prFailOutOfBounds;
                                             ResultSet.Msg.initialize (msWarning, MSG_PARSER_OUT_OF_DATA, 'FQDN Pointers', PointerOffset.ToString, LastOffset.ToString);
                                        end;

                                    boParseDone := true;
                               end
                            else
                                begin
                                     ResultSet.ParseResult := prFailOutOfBounds;
                                     ResultSet.Msg.initialize (msWarning, MSG_PARSER_OUT_OF_DATA_FOR_LENGTH, 'FQDN Pointers', PointerOffset.ToString, '2', LastOffset.ToString);

                                     boAbort := true;
                                end;
                       end;
           else
               begin
                    // Normal length - parse a label (maximum of 63 characters)

                    if (Offset + RecordLength) <= lastoffset then
                       begin
                            inc (Offset);
                            inc (ResultSet.OffsetIncrement);

                            // get domain substring and set/add it to the FQDN accordingly

                            HelpChar := pANSIChar (FrameData + Offset);
                            HelpString := string (HelpChar);
                            HelpString := copy (HelpString, 1, RecordLength);

                            if FQDN = '' then
                               FQDN := HelpString
                            else
                                FQDN := FQDN + '.' + HelpString;

                            // Move offset

                            inc (ResultSet.OffsetIncrement, RecordLength);
                            inc (Offset, RecordLength);
                       end
                    else
                        begin
                             boAbort := true;

                             ResultSet.ParseResult := prFailOutOfBounds;
                             ResultSet.Msg.initialize (msWarning, MSG_PARSER_OUT_OF_DATA_FOR_LENGTH, 'FQDN Elements', Offset.ToString, RecordLength.ToString, LastOffset.ToString);
                        end;
               end;
           end;

     until boParseDone or boAbort;

     if boAbort then
        begin
             ResultSet.Name.initialize;
             ResultSet.ParseResult := prFailOutOfBounds;
        end
     else
         ResultSet.Name.setFQDN(FQDN);

     result := ResultSet;
end;

(* ------------------------------------------- *)

function Parse802dot1qData (const Value : word) : t802dot1qData; inline;

var prio   : integer;
    id     : integer;
    data   : t802dot1qData;

begin
     // Get Priority

     prio := value;
     prio := prio and $e000;
     prio := prio shr 13;

     data.priority := prio;

     // Get CFI

     data.cfi := (value and $1000) = $1000;

     // Get VLAN ID

     id := value and $0fff;

     data.vlanid := id;

     result := data;
end;

(* ------------------------------------------- *)

function DetermineClientOfTuple (const Tuple : t5Tuple) : tClientServerOrder;

var SourcePort      : word;
    DestinationPort : word;
    value           : tClientServerOrder;

begin
     value := csoNone;

     case Tuple.IPVersion of
       4 : begin
                SourcePort := Tuple.IPv4SocketPair.Source.Port;
                DestinationPort := Tuple.IPv4SocketPair.Destination.Port;
           end;
       6 : begin
                SourcePort := Tuple.IPv6SocketPair.Source.Port;
                DestinationPort := Tuple.IPv6SocketPair.Destination.Port;
           end;
     end;

     // Use ports to guess who is client and who is server
     // First, do the ephemeral port vs. well known thing

     if (SourcePort < 1025) and (DestinationPort > 1024) then
        value := csoClientIsDestination
     else
         if (DestinationPort < 1025) and (SourcePort > 1024) then
            value := csoClientIsSource
         else
             begin
                  // Things are more complicated
                  // Let's got for typical ephemeral ports of modern clients

                  if (SourcePort > 49151) and (DestinationPort < 49152) then
                     value := csoClientIsSource
                  else
                      if (SourcePort < 49152) and (DestinationPort > 49151) then
                         value := csoClientIsDestination;
             end;

     result := value;
end;

(* ------------------------------------------- *)

function ExtractLongWordFromByteArray (const buffer : array of byte; const offset : integer) : longword;

var value : longword;

begin
     assert ((offset >= 0) and (offset < length (buffer)));

     value := buffer[offset] shl 24 + buffer[offset+1] shl 16 + buffer[offset+2] shl 8 + buffer[offset+3];

     result := value;
end;

(* ------------------------------------------- *)

procedure PatchLongwordToByteArray (var buffer : array of byte; const offset : integer; const value : longword);

begin
     assert ((offset >= 0) and (offset < length (buffer)));

     buffer[offset] := (value and $ff000000) shr 24;
     buffer[offset+1] := (value and $ff0000) shr 16;
     buffer[offset+2] := (value and $ff00) shr 8;
     buffer[offset+3] := (value and $ff);
end;

(* ------------------------------------------- *)

end.
