{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_beaconfile;

interface

uses class_rawframe,
     definition_filter_structures;

type tOnAdvanceBeaconsPercentProcessed = procedure (Sender : tObject; const FileSize : uint64; const FiltersAdded : integer; const PercentageParsed : integer) of object;

     tBeaconFile = class (tObject)

                         constructor create (const BeaconFileName : string);

                         private
                                EventOnBeaconsPercentProcessed : tOnAdvanceBeaconsPercentProcessed;

                                priv_filename            : string;
                                priv_filesize            : uint64;
                                priv_percentageprocessed : integer;

                                priv_filterlist          : tFilterSet;

                                priv_ignoreSYN           : boolean;
                                priv_ignoreICMP          : boolean;
                                priv_commentAsLabel      : boolean;

                                procedure ReportProgress (const FrameOffset: uint64);
                                function getFilterForFrame (const Frame : tRawFrame) : tFrameFilter;
//                              procedure AddFilterToList (Filter : tFrameFilter);

                         public
                               property OnBeaconsPercentProcessed : tOnAdvanceBeaconsPercentProcessed read EventOnBeaconsPercentProcessed write EventOnBeaconsPercentProcessed;
                               property FilterList : tFilterSet read priv_filterlist;
                               property IgnoreSYN : boolean read priv_ignoreSYN write priv_ignoreSYN;
                               property IgnoreICMP : boolean read priv_ignoreICMP write priv_ignoreICMP;
                               property UseFrameCommentAsFilterLabel : boolean read priv_commentAsLabel write priv_commentAsLabel;

                               function isValidFile : boolean;
                               procedure DetermineFilterList;
                               destructor Destroy; override;

                   end;

implementation

uses base_structures,
     definition_ip_addresses,
     definition_tcp_structures,
     class_traceloader,
     tracefile_utils,
     class_parsemaster;

(* ------------------------------------------- *)
{ tBeaconFile }
(* ------------------------------------------- *)

constructor tBeaconFile.create (const BeaconFileName: string);

begin
     inherited create;

     priv_filename := BeaconFileName;
//   setLength (priv_filterlist, 0);
     priv_filterlist := tFilterSet.create;

     // Set defaults

     priv_filesize := 0;

     priv_ignoreSYN  := true;
     priv_ignoreICMP := false;
     priv_commentAsLabel := true;
end;

(* ------------------------------------------- *)

destructor tBeaconFile.Destroy;

begin
//   setLength (priv_filterlist, 0);
     priv_filterlist.Free;

     inherited;
end;

(* ------------------------------------------- *)

procedure tBeaconFile.ReportProgress (const FrameOffset: uint64);

var percentagescanned : integer;
    iHelp             : integer;
    dHelp             : double;

begin
     if (priv_filesize > 0) and assigned (EventOnBeaconsPercentProcessed) then
        begin
             dHelp := 100 * frameoffset / priv_filesize;
             iHelp := trunc (dHelp);

             if iHelp > priv_percentageprocessed then
                begin
                     priv_percentageprocessed := iHelp;

                     if assigned (EventOnBeaconsPercentProcessed) then
                        EventOnBeaconsPercentProcessed (self, priv_filesize, priv_filterlist.count, priv_percentageprocessed);
                end;
        end;
end;

(* ------------------------------------------- *)

function tBeaconFile.getFilterForFrame (const Frame: tRawFrame) : tFrameFilter;

var filter      : tFrameFilter;
    parser      : tParseMaster;
    socketpair4 : tIPv4SocketPair;
    socketpair6 : tIPv6SocketPair;
    goAhead     : boolean;

begin
     goAhead := true;

     filter.initialize;
     parser := tParseMaster.create (Frame, ppmUpToLayer4);

     if priv_ignoreSYN and (tcpSYN in parser.TCPFlags) then
        goAhead := false;

     if priv_ignoreICMP and ((ppICMPv4 in parser.ParsedProtocols) or (ppICMPv6 in parser.ParsedProtocols)) then
        goAhead := false;

     if priv_commentAsLabel then
        filter.FilterLabel := copy (Frame.Commentlist.Text, 1, MaxFilterLabelSize);

     if goAhead then
        begin
             case parser.LastIPLayerVersion of
              4 : begin
                       socketpair4 := parser.IPv4SocketPair;
                       socketpair4.normalize;

                       if ppUDP in parser.ParsedProtocols then
                          begin
                               filter.IPv4Filter.AddressFilterType := ftSingle;
                               filter.IPv4Filter.AddressScope := fsBoth;
                               filter.IPv4Filter.SourceAddress := socketpair4.Source.IPAddress;
                               filter.IPv4Filter.DestinationAddress := socketpair4.Destination.IPAddress;

                               filter.PortFilter.Layer4Protocol := l4UDP;
                               filter.PortFilter.PortScope := fsBoth;
                               filter.PortFilter.PortFilterType := ftSingle;
                               filter.PortFilter.SourcePort := socketpair4.Source.Port;
                               filter.PortFilter.DestinationPort := socketpair4.Destination.Port;
                          end;

                       if ppTCP in parser.ParsedProtocols then
                          begin
                               filter.IPv4Filter.AddressFilterType := ftSingle;
                               filter.IPv4Filter.AddressScope := fsBoth;
                               filter.IPv4Filter.SourceAddress := socketpair4.Source.IPAddress;
                               filter.IPv4Filter.DestinationAddress := socketpair4.Destination.IPAddress;

                               filter.PortFilter.Layer4Protocol := l4TCP;
                               filter.PortFilter.PortScope := fsBoth;
                               filter.PortFilter.PortFilterType := ftSingle;
                               filter.PortFilter.SourcePort := socketpair4.Source.Port;
                               filter.PortFilter.DestinationPort := socketpair4.Destination.Port;
                          end;

                       if ppICMPv4 in parser.ParsedProtocols then
                          begin
                               filter.IPv4Filter.AddressFilterType := ftSingle;
                               filter.IPv4Filter.AddressScope := fsBoth;
                               filter.IPv4Filter.SourceAddress := socketpair4.Source.IPAddress;
                               filter.IPv4Filter.DestinationAddress := socketpair4.Destination.IPAddress;

                               filter.PortFilter.Layer4Protocol := l4ICMP;
                               filter.PortFilter.PortScope := fsAny;
                               filter.PortFilter.PortFilterType := ftSingle;
                          end;
                  end;
              6 : begin
                       socketpair6 := parser.IPv6SocketPair;
                       socketpair6.normalize;

                       if ppUDP in parser.ParsedProtocols then
                          begin
                               filter.IPv6Filter.AddressFilterType := ftSingle;
                               filter.IPv6Filter.AddressScope := fsBoth;
                               filter.IPv6Filter.SourceAddress := socketpair6.Source.IPAddress;
                               filter.IPv6Filter.DestinationAddress := socketpair6.Destination.IPAddress;

                               filter.PortFilter.Layer4Protocol := l4UDP;
                               filter.PortFilter.PortScope := fsBoth;
                               filter.PortFilter.PortFilterType := ftSingle;
                               filter.PortFilter.SourcePort := socketpair6.Source.Port;
                               filter.PortFilter.DestinationPort := socketpair6.Destination.Port;
                          end;

                       if ppTCP in parser.ParsedProtocols then
                          begin
                               filter.IPv6Filter.AddressFilterType := ftSingle;
                               filter.IPv6Filter.AddressScope := fsBoth;
                               filter.IPv6Filter.SourceAddress := socketpair6.Source.IPAddress;
                               filter.IPv6Filter.DestinationAddress := socketpair6.Destination.IPAddress;

                               filter.PortFilter.Layer4Protocol := l4TCP;
                               filter.PortFilter.PortScope := fsBoth;
                               filter.PortFilter.PortFilterType := ftSingle;
                               filter.PortFilter.SourcePort := socketpair6.Source.Port;
                               filter.PortFilter.DestinationPort := socketpair6.Destination.Port;
                          end;

                       if ppICMPv6 in parser.ParsedProtocols then
                          begin
                               filter.IPv6Filter.AddressFilterType := ftSingle;
                               filter.IPv6Filter.AddressScope := fsBoth;
                               filter.IPv6Filter.SourceAddress := socketpair6.Source.IPAddress;
                               filter.IPv6Filter.DestinationAddress := socketpair6.Destination.IPAddress;

                               filter.PortFilter.Layer4Protocol := l4ICMP;
                               filter.PortFilter.PortScope := fsAny;
                               filter.PortFilter.PortFilterType := ftSingle;
                          end;
                  end;
             end;
     end;

     parser.free;
     result := filter;
end;

(* ------------------------------------------- *)
{
procedure tBeaconFile.AddFilterToList (Filter: tFrameFilter);


var found : boolean;
    len   : integer;
    i     : integer;

begin
     found := false;
     len := length (priv_filterlist);

     if len > 0 then
        begin
             i := 0;

             repeat
                   if filter = priv_filterlist[i] then
                      found := true;

                   inc (i);

             until found or (i > len);
        end;

     if not found then
        begin
             setLength (priv_filterlist, len + 1);
             priv_filterlist[len] := Filter;
        end;
end;
}

(* ------------------------------------------- *)

function tBeaconFile.isValidFile: boolean;

var loader : tTraceLoader;

begin
     loader := createTraceLoader (priv_filename);

     if loader = nil then
        result := false
     else
         begin
              result := true;

              loader.free;
         end;
end;

(* ------------------------------------------- *)

procedure tBeaconFile.DetermineFilterList;

var loader : tTraceLoader;
    frame  : tRawFrame;
    filter : tFrameFilter;
    i      : Integer;
{$ifdef debug}
    test   : string;
{$endif}

begin
     loader := createTraceLoader (priv_filename);

     if assigned (loader) and loader.openFile then
        begin
             priv_filesize := loader.Filesize;

             loader.readTraceFileheader;

             // Read frames

             loader.readBatchFromDisk;

             repeat
                   for i := 1 to loader.FramesAvailable do
                       begin
                            frame := loader.getRawFrame (i);

                            filter := getFilterForFrame (frame);
{$ifdef debug}
                            test := filter.asSocketString;
{$endif}
                            // the filter may be unspecified if the frame is not to be filtered for,
                            // e.g. when excluding TCP SYN or ICMP

                            if not filter.isUnspecified then
                               priv_filterlist.AddFilter (Filter);

                            ReportProgress (frame.FileOffset);

                            frame.free;
                            frame := nil;
                       end;

                   loader.readBatchFromDisk;

             until loader.FramesAvailable = 0;

             // Cleanup

             loader.closeFile;
             loader.free;
        end;

     priv_filterlist.Trim;
end;

(* ------------------------------------------- *)

end.
