{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_interfacemapper;

interface

uses // FireDAC

     FireDAC.Stan.Intf, FireDAC.Stan.Option, FireDAC.Stan.Error, FireDAC.UI.Intf,
     FireDAC.Phys.Intf, FireDAC.Stan.Def, FireDAC.Stan.Pool, FireDAC.dAPT,
     FireDAC.Stan.Async, FireDAC.Phys, FireDAC.Stan.ExprFuncs,
     FireDAC.VCLUI.Wait, FireDAC.Comp.UI, FireDAC.Phys.SQLite, FireDAC.Phys.SQLiteWrapper,
     Data.DB, FireDAC.Comp.Client, FireDAC.Stan.Param,

     definition_interface_mapping_types,
     class_interfacelist,
     class_tracefilelist;

type tInterfaceMapMode = (immMerge, immDistinctIFace, immSingleIFace);
     tInterfaceTransferScope = (itsKeepAll, itsKeepActive);

     tInterfaceMapper = class (tObject)

                              constructor create (TraceFileList : tTraceFileList; const MappingMode : tInterfaceMapMode = immMerge);
                              procedure free;

                              private
                                     priv_filelist           : tTraceFilelist;
                                     priv_targetinterfaces   : tCaptureInterfaceList;
                                     priv_relevantinterfaces : tInterfaceFileIDList;

                                     priv_mappingmode        : tInterfaceMapMode;

                                     {$REGION 'Documentation'}
                                     /// <summary>
                                     ///   Builds a list of interfaces by
                                     ///   creating copies of each interface
                                     ///   that is not already in the list. If
                                     ///   an existing target interface is
                                     ///   found it is simply assigned as
                                     ///   mapping target to the original
                                     ///   interface. Otherwise the original is
                                     ///   copied and then assigned.
                                     /// </summary>
                                     /// <param name="OriginalEntry">
                                     ///   Capture interface coming from the
                                     ///   original trace file
                                     /// </param>
                                     {$ENDREGION}
                                     procedure MergeInterfaceToList (var OriginalEntry : tCaptureInterface);
                                     procedure CreateUniqueInterfaceList (var OriginalEntry : tCaptureInterface);
                                     procedure MergeInterfacesToSingle (var OriginalEntry : tCaptureInterface);

                                     {$REGION 'Documentation'}
                                     /// <summary>
                                     ///   Creates a new interface object from
                                     ///   the Query fields. The object needs
                                     ///   to be discarded as soon as no longer
                                     ///   in use.
                                     /// </summary>
                                     /// <param name="Query">
                                     ///   Query with fields containing the
                                     ///   values for the interface
                                     /// </param>
                                     {$ENDREGION}
                                     function GetInterfaceFromQuery (Query : tFDQuery) : tCaptureInterface;
                                     {$REGION 'Documentation'}
                                     /// <summary>
                                     ///   Determines if an interface entry
                                     ///   coming from a file in the file list
                                     ///   is relevant for the output file
                                     /// </summary>
                                     /// <param name="FileIndex">
                                     ///   DB Index of the file the interface
                                     ///   originate from
                                     /// </param>
                                     /// <param name="InterfaceEntry">
                                     ///   Interface entry to be checked
                                     /// </param>
                                     /// <returns>
                                     ///   true if the interface needs to be
                                     ///   written to the IDB of the target file
                                     /// </returns>
                                     {$ENDREGION}
                                     function IncludeInterfaceInMapping (const FileIndex : integer; const InterfaceEntry : tCaptureInterface) : boolean;

                              public
                                    property TraceFileList : tTraceFileList read priv_filelist write priv_filelist;
                                    property TargetInterfaceList : tCaptureInterfaceList read priv_targetinterfaces write priv_targetinterfaces;
                                    property MappingMode : tInterfaceMapMode read priv_mappingmode write priv_mappingmode;
                                    {$REGION 'Documentation'}
                                    /// <summary>
                                    ///   This list can be used to tell the
                                    ///   interface mapper to only pay
                                    ///   attention to the interfaces it
                                    ///   contains. Any interface not found in
                                    ///   the list will not be mapped to a
                                    ///   target interface. This is useful when
                                    ///   extracting packets from a file that
                                    ///   has multiple interfaces and not all
                                    ///   of them are relevant for them.
                                    /// </summary>
                                    {$ENDREGION}
                                    property RelevantInterfaceList : tInterfaceFileIDList read priv_relevantinterfaces write priv_relevantinterfaces;

                                    {$REGION 'Documentation'}
                                    /// <summary>
                                    ///   Removes all entries from the
                                    ///   interface target list, and removes
                                    ///   the mapping for all original
                                    ///   interface entries in the file list
                                    /// </summary>
                                    {$ENDREGION}
                                    procedure clearMappings;
                                    {$REGION 'Documentation'}
                                    /// <summary>
                                    ///   Maps the interfaces of all files
                                    ///   found in the file list to an output
                                    ///   interface, depending on the mapping
                                    ///   mode
                                    /// </summary>
                                    {$ENDREGION}
                                    procedure doAutoMapping;
                                    {$REGION 'Documentation'}
                                    /// <summary>
                                    ///   Returns a list of all capture
                                    ///   interfaces that the original
                                    ///   interfaces have been mapped to. At
                                    ///   the same time the indexes of the
                                    ///   mapped interfaces will be updated to
                                    ///   match the order of the interfaces in
                                    ///   the output file
                                    /// </summary>
                                    /// <param name="InterfaceList">
                                    ///   Initialized empty list to be filled
                                    ///   with entries
                                    /// </param>
                                    {$ENDREGION}
                                    procedure getTargetInterfaceList (var InterfaceList : tCaptureInterfaceList);

                                    {$REGION 'Documentation'}
                                    /// <summary>
                                    ///   Stores the interface mapping settings
                                    ///   to the task database, which needs to
                                    ///   be connected before calling this
                                    ///   method.
                                    /// </summary>
                                    /// <param name="Database">
                                    ///   Task settings file
                                    /// </param>
                                    {$ENDREGION}
                                    procedure StoreToDatabase (Database : tFDConnection);
                                    {$REGION 'Documentation'}
                                    /// <summary>
                                    ///   Loads the interface mapping settings
                                    ///   from the task database, which needs
                                    ///   to be connected before calling this
                                    ///   method.
                                    /// </summary>
                                    /// <param name="Database">
                                    ///   Task settings file
                                    /// </param>
                                    {$ENDREGION}
                                    procedure LoadFromDatabase (Database : tFDConnection);
                        end;

implementation

uses SiAuto,
     SysUtils,
     pcapng_structures;

(* ------------------------------------------- *)
{ tInterfaceMapper }
(* ------------------------------------------- *)

constructor tInterfaceMapper.create (TraceFileList: tTraceFileList; const MappingMode : tInterfaceMapMode = immMerge);

begin
     inherited create;

     priv_filelist := Tracefilelist;
     priv_targetinterfaces := tCaptureInterfaceList.create (true);

     priv_mappingmode := MappingMode;

     setlength (priv_relevantinterfaces, 0);
end;

(* ------------------------------------------- *)

procedure tInterfaceMapper.free;

begin
     priv_targetinterfaces.free;

     inherited free;
end;

(* ------------------------------------------- *)

procedure tInterfaceMapper.clearMappings;

var entry : tTracefileEntry;
    IFace : tCaptureInterface;
    i,j   : integer;

begin
//   SiMain.EnterMethod ('tInterfaceMapper.clearMappings');

//   SiMain.LogVerbose ('Clearing interface mappings for ' + intToStr (priv_filelist.Count) + ' files');

     // Go through all source interfaces and clear their mapping

     for i := 1 to priv_filelist.Count do
         begin
              entry := priv_filelist.getFile (i);

              for j := 1 to entry.CaptureInterfaces.Count do
                  begin
                       IFace := entry.CaptureInterfaces.getInterfaceByIndex(j-1);

                       IFace.MappedToInterface := nil;
                  end;
         end;

     priv_targetinterfaces.clear;

//   SiMain.LeaveMethod ('tInterfaceMapper.clearMappings');
end;

(* ------------------------------------------- *)

function tInterfaceMapper.IncludeInterfaceInMapping (const FileIndex : integer; const InterfaceEntry : tCaptureInterface) : boolean;

var entry : tInterfaceFileIDEntry;
    found : boolean;
    test  : integer;
    len   : integer;
    i     : integer;

begin
     len := length (priv_relevantinterfaces);

     // If the list has any entries at all, determine if we keep the current Interface,
     // otherwise accepted just like that

     if len > 0 then
        begin
             found := false;
             i := 0;

             // check if the interface entry is in the list

             while (i < len) and not found do
                 begin
                      entry := priv_relevantinterfaces[i];

                      if (entry.FileIndex = FileIndex) then
                         begin
                              test := 1 shl InterfaceEntry.InterfaceID;

                              if  (entry.InterfaceFlags and test) = test then
                                  found := true
                              else
                                  inc (i);
                         end
                      else
                          inc (i);
                 end;
        end
     else
         found := true;

     result := found;
end;

(* ------------------------------------------- *)

procedure tInterfaceMapper.MergeInterfaceToList (var OriginalEntry: tCaptureInterface);

var boFound     : boolean;
    TargetEntry : tCaptureInterface;
    i           : integer;

begin
     SiMain.EnterMethod ('tInterfaceMapper.MergeInterfaceToList');
     SiMain.LogDebug ('Determining mapping for interface "' + OriginalEntry.Name + '" (' + OriginalEntry.PacketLinkID.ToString + ')');

     boFound := false;
     i := 0;

     // Try to find a matching Interface in the list of existing target interfaces

     while (i < priv_targetinterfaces.Count) and not boFound do
           begin
                TargetEntry := priv_targetinterfaces.getInterfaceByIndex (i);
                inc (i);

                boFound := InterfacesAreEqual (OriginalEntry, TargetEntry);
           end;

     if boFound then
        begin
             SiMain.LogVerbose ('Mapping Interface #' + IntToStr (OriginalEntry.InterfaceID) + ' "' + OriginalEntry.Name + '" to existing target interface #' + IntToStr (TargetEntry.InterfaceID));

             OriginalEntry.MappedToInterface := TargetEntry
        end
     else
         begin
              SiMain.LogVerbose ('No matching interface found, creating a new interface copy of the original');

              // New Entry needs to be added to the list, so we create a copy of the original to have two separate objects for the same interface

              TargetEntry := tCaptureInterface.create (OriginalEntry.InterfaceID, OriginalEntry.LinkType, OriginalEntry.Origin, true);
              OriginalEntry.copyTo (TargetEntry);
              TargetEntry.MappedToInterface := nil;

              // Set the Interface index and add it to the target list

              TargetEntry.OriginalID := priv_targetinterfaces.Count;
              TargetEntry.InterfaceID := TargetEntry.OriginalID;

              priv_targetInterfaces.Add (TargetEntry);

              // Finally, assign the new interface object as mapping target to the original interface

//            SiMain.LogVerbose ('Mapping Interface #' + IntToStr (OriginalEntry.InterfaceID) + ' "' + OriginalEntry.Name + '" to new target interface #' + IntToStr (TargetEntry.InterfaceID));
//            SiMain.LogVerbose ('Mapping Interface #' + IntToStr (OriginalEntry.InterfaceID) + ' "' + OriginalEntry.PacketLinkID.ToString + '" to new target interface #' + IntToStr (TargetEntry.InterfaceID) + '(' + TargetEntry.PacketLinkID.ToString + ')');

              OriginalEntry.MappedToInterface := TargetEntry;
         end;

     SiMain.LeaveMethod ('tInterfaceMapper.MergeInterfaceToList');
end;

(* ------------------------------------------- *)

procedure tInterfaceMapper.CreateUniqueInterfaceList (var OriginalEntry: tCaptureInterface);

var TargetEntry : tCaptureInterface;
    Guid        : tGUID;

begin
     SiMain.EnterMethod ('tInterfaceMapper.CreateUniqueInterfaceList');

     // New Entry needs to be added to the list

     TargetEntry := tCaptureInterface.create (OriginalEntry.InterfaceID, OriginalEntry.LinkType, ioAutogenerated, true);
     OriginalEntry.copyTo (TargetEntry, false);

     // create unique names if possible

     if pos ('\Device\NPF', TargetEntry.Name) > 0 then
        begin
             CreateGUID (Guid);

             TargetEntry.Name := '\Device\NPF_' + GUIDToString (Guid);
        end;

     if pos ('eth', TargetEntry.Name) = 1 then
        TargetEntry.Name := 'eth' + IntToStr (priv_targetInterfaces.Count);

     if pos ('em', TargetEntry.Name) = 1 then
        TargetEntry.Name := 'em' + IntToStr (priv_targetInterfaces.Count);

     if pos ('wlan', TargetEntry.Name) = 1 then
        TargetEntry.Name := 'wlan' + IntToStr (priv_targetInterfaces.Count);

     // Set the Interface index and add it

     TargetEntry.OriginalID := priv_targetinterfaces.Count;
     TargetEntry.InterfaceID := TargetEntry.OriginalID;

     priv_targetInterfaces.Add (TargetEntry);

     OriginalEntry.MappedToInterface := TargetEntry;

     SiMain.LeaveMethod ('tInterfaceMapper.CreateUniqueInterfaceList');
end;

(* ------------------------------------------- *)

procedure tInterfaceMapper.MergeInterfacesToSingle (var OriginalEntry: tCaptureInterface);

var TargetEntry : tCaptureInterface;
    Guid        : tGUID;

begin
     SiMain.EnterMethod ('tInterfaceMapper.MergeInterfacesToSingle');

     if priv_targetinterfaces.Count = 0 then
        begin
             // New Entry needs to be added to the list

             TargetEntry := tCaptureInterface.create (OriginalEntry.InterfaceID, OriginalEntry.LinkType, ioAutogenerated, true);
             OriginalEntry.copyTo (TargetEntry, false);

             // create unique names if possible

             if pos ('\Device\NPF', TargetEntry.Name) > 0 then
                begin
                     CreateGUID (Guid);

                     TargetEntry.Name := '\Device\NPF_' + GUIDToString (Guid);
                end;

             if pos ('eth', TargetEntry.Name) = 1 then
                TargetEntry.Name := 'eth' + IntToStr (priv_targetInterfaces.Count);

             if pos ('em', TargetEntry.Name) = 1 then
                TargetEntry.Name := 'em' + IntToStr (priv_targetInterfaces.Count);

             if pos ('wlan', TargetEntry.Name) = 1 then
                TargetEntry.Name := 'wlan' + IntToStr (priv_targetInterfaces.Count);

             priv_targetInterfaces.Add (TargetEntry);
        end
     else
         TargetEntry := priv_targetInterfaces.getInterfaceByIndex(0);

     OriginalEntry.MappedToInterface := TargetEntry;

     SiMain.LeaveMethod ('tInterfaceMapper.MergeInterfacesToSingle');
end;

(* ------------------------------------------- *)

procedure tInterfaceMapper.getTargetInterfaceList (var InterfaceList : tCaptureInterfaceList);

var fileentry : tTracefileEntry;
    ifentry   : tCaptureInterface;
    ifxentry  : tCaptureInterface;
    i,j,k     : integer;
    index     : integer;
    boFound   : boolean;

begin
     SiMain.EnterMethod ('tInterfaceMapper.getTargetInterfaceList');

     index := InterfaceList.Count;

     for i := 1 to priv_filelist.Count do
         begin
              FileEntry := priv_filelist.getFile(i);

              SiMain.LogVerbose ('Processing interfaces of file ' + fileEntry.FileName);

              // get each interface of the file

              for j := 1 to FileEntry.CaptureInterfaces.Count do
                  begin
                      ifEntry := FileEntry.CaptureInterfaces.getInterfaceByIndex(j-1);

                      if ifEntry.MappedToInterface <> nil then
                         begin
                              SiMain.LogVerbose ('Interface with index #' + IntToStr (IfEntry.InterfaceID) + ' to be included in target list');

                              boFound := false;
                              k := 0;

                              // Find the Entry in the existing list. Since we're usually mapping multiple source
                              // interfaces (from multiple files) to the same target interface we need to see if
                              // we already have it in our list

                              while (k < InterfaceList.Count) and not boFound do
                                    begin
                                         ifxentry := InterfaceList.getInterfaceByIndex (k);
                                         inc (k);

                                         boFound := InterfacesAreEqual (ifEntry, ifXEntry);
                                    end;

                              // If found, use the existing interface indexes, otherwise create a new one

                              if boFound then
                                 begin
                                      SiMain.LogVerbose ('Target Interface already exists in the target list, mapping');

                                      // Set index of the mapped interface to the index in the list, because
                                      // now we know the order they will be written to file

                                      ifEntry.MappedToInterface.InterfaceID := ifXEntry.InterfaceID;
                                      ifEntry.MappedToInterface.OriginalID := ifXEntry.OriginalID;
                                 end
                              else
                                  begin
                                       SiMain.LogVerbose ('Adding new Target Interface to the target list');

                                       // Create new output interface; the "index" value is the place it will appear in
                                       // the interface list in the output file.

                                       ifxentry := tCaptureinterface.create (index, ifEntry.MappedToInterface.LinkType, ifEntry.MappedToInterface.Origin, true);

                                       // Copy the details, put keep the index intact

                                       ifEntry.MappedToInterface.copyTo (ifxentry, true);

                                       // store in final list

                                       InterfaceList.Add (ifxentry);

                                       inc (index);
                                  end;
                         end
                      else
//                        SiMain.LogVerbose ('Interface with index #' + IntToStr (IfEntry.InterfaceID) + ' ignored');
                  end;
         end;

     SiMain.LeaveMethod ('tInterfaceMapper.getTargetInterfaceList');
end;

(* ------------------------------------------- *)

procedure tInterfaceMapper.doAutoMapping;

var entry : tTracefileEntry;
    IFace : tCaptureInterface;
    i,j   : integer;

begin
     SiMain.EnterMethod ('tInterfaceMapper.doAutoMapping');

     clearMappings;

     // Go through all source interfaces and map them

     for i := 1 to priv_filelist.Count do
         begin
              entry := priv_filelist.getFile (i);

              SiMain.LogDebug ('Merging Interfaces of file "' + entry.FileName + '"');

              for j := 0 to entry.CaptureInterfaces.Count - 1 do
                  begin
                       IFace := entry.CaptureInterfaces.getInterfaceByIndex(j);

                       // Map if allowed by the "relevant interfaces" list (or if the list is empty)

                       if IncludeInterfaceInMapping (i, IFace) then
                          begin
                               case priv_mappingmode of
                                immMerge         : MergeInterfaceToList (IFace);
                                immDistinctIFace : CreateUniqueInterfaceList (IFace);
                                immSingleIFace   : MergeInterfacesToSingle (IFace);
                               end;
                          end;
                  end;
         end;

     SiMain.LeaveMethod ('tInterfaceMapper.doAutoMapping');
end;

(* ------------------------------------------- *)

procedure tInterfaceMapper.StoreToDatabase (Database: tFDConnection);

var Query     : tFDQuery;
    FileEntry : tTraceFileEntry;
    IfEntry   : tCaptureInterface;
    TargetIf  : tCaptureInterface;
    i,j,k     : integer;
    MapID     : integer;

begin
     SiMain.EnterMethod ('tInterfaceMapper.StoreToDatabase');

     Query := tFDQuery.Create(nil);
     Query.Connection := Database;

     // Write mapping targets first

     for i := 1 to priv_targetinterfaces.Count do
         begin
              // Write mapped interface, starting at index 1.

              IfEntry := priv_targetinterfaces.getInterfaceByIndex(i-1);

              Query.SQL.Text := 'INSERT INTO formTargetInterfaceList (interfaceid,name,description,linklayertype,snaplen,tsresolution,timezone,capturefilter,operatingsystem,origin,packetlinkid) VALUES' +
                                                                   ' (:interfaceid,:name,:description,:linklayertype,:snaplen,:tsresolution,:timezone,:capturefilter,:operatingsystem,:origin,:packetlinkid);';
              Query.ParamByName ('interfaceid').AsInteger := i;
              Query.ParamByName ('name').AsString := IfEntry.Name;
              Query.ParamByName ('description').AsString := IfEntry.Description;
              Query.ParamByName ('linklayertype').AsInteger := IfEntry.LinkType;
              Query.ParamByName ('snaplen').AsInteger := IfEntry.SnapLength;
              Query.ParamByName ('tsresolution').AsInteger := IfEntry.TimeStampResolution;
              Query.ParamByName ('timezone').AsInteger := IfEntry.TimeZone;
              Query.ParamByName ('capturefilter').AsString := IfEntry.CaptureFilter.FilterString;
              Query.ParamByName ('operatingsystem').AsString := IfEntry.OperatingSystem;
              Query.ParamByName ('origin').AsInteger := ord (IfEntry.Origin);
              Query.ParamByName ('packetlinkid').AsString := IfEntry.PacketLinkID.ToString;

              Query.ExecSQL;
         end;

     // Write files and their interfaces

     for i := 1 to priv_filelist.Count do
         begin
              FileEntry := priv_filelist.getFile(i);

              SiMain.LogDebug ('Storing Interfaces of file "' + FileEntry.FileName + '" to database');

              // Write File Entry

              Query.SQL.Text := 'INSERT INTO formFileList (id, filename,filesize,firstframetimestamp) VALUES (:id,:filename,:filesize,:firstframetimestamp);';
              Query.ParamByName ('id').AsInteger := i;
              Query.ParamByName ('filename').AsString := FileEntry.FilePath + FileEntry.FileName;
              Query.ParamByName ('filesize').AsLargeInt := FileEntry.FileSize;

              // Workaround for empty files: set first frame time to zero to avoid rereading the file for nothing, and losing the interface list

              if FileEntry.FrameCount > 0 then
                 Query.ParamByName ('firstframetimestamp').AsDateTime := FileEntry.FirstFrameTimestamp.datetime
              else
                  Query.ParamByName('firstframetimestamp').AsDateTime := 0;

              Query.ExecSQL;

              for j := 1 to FileEntry.CaptureInterfaces.Count do
                  begin
                       IfEntry := FileEntry.CaptureInterfaces.getInterfaceByIndex(j-1);

                       SiMain.LogDebug ('Storing Interface #' + IntToStr (IfEntry.InterfaceID) + ' "' + IfEntry.Name + '" (' + IfEntry.PacketLinkID.ToString + ') to database');

                       if IfEntry.MappedToInterface = nil then
                          raise Exception.Create('Error: Interface has no mapped target interface (tInterfaceMapper.StoreToDatabase)');

                       MapID := 0;

                       // Find the mapped interface index (just by scanning through the list. Slow but who cares :-))

                       for k := 1 to priv_targetinterfaces.Count do
                           begin
                                TargetIf := priv_targetinterfaces.getInterfaceByIndex(k-1);

                                if TargetIF.PacketLinkID = IfEntry.MappedToInterface.PacketLinkID then
                                   MapID := k;
                           end;

                       // Write Original Interface

                       Query.SQL.Text := 'INSERT INTO formFileInterfaces (fileid,interfaceid,name,description,linklayertype,snaplen,tsresolution,timezone,capturefilter,operatingsystem,packetlinkid,origin,mappedtoid) VALUES' +
                                                                       ' (:fileid,:interfaceid,:name,:description,:linklayertype,:snaplen,:tsresolution,:timezone,:capturefilter,:operatingsystem,:packetlinkid,:origin,:mappedtoid);';

                       Query.ParamByName ('fileid').AsInteger := i;
                       Query.ParamByName ('interfaceid').AsInteger := j;
                       Query.ParamByName ('name').AsString := IfEntry.Name;
                       Query.ParamByName ('description').AsString := IfEntry.Description;
                       Query.ParamByName ('linklayertype').AsInteger := IfEntry.LinkType;
                       Query.ParamByName ('snaplen').AsInteger := IfEntry.SnapLength;
                       Query.ParamByName ('tsresolution').AsInteger := IfEntry.TimeStampResolution;
                       Query.ParamByName ('timezone').AsInteger := IfEntry.TimeZone;
                       Query.ParamByName ('capturefilter').AsString := IfEntry.CaptureFilter.FilterString;
                       Query.ParamByName ('operatingsystem').AsString := IfEntry.OperatingSystem;
                       Query.ParamByName ('packetlinkid').AsString := IfEntry.PacketLinkID.ToString;
                       Query.ParamByName ('origin').AsInteger := ord (IfEntry.Origin);
                       Query.ParamByName ('mappedtoid').AsInteger := MapID;

                       Query.ExecSQL;
                  end;
         end;

     Query.Free;

     SiMain.LeaveMethod ('tInterfaceMapper.StoreToDatabase');
end;

(* ------------------------------------------- *)

function tInterfaceMapper.GetInterfaceFromQuery (Query: tFDQuery): tCaptureInterface;

var entry    : tCaptureInterface;
    index    : integer;
    mappedto : integer;
    linktype : integer;
    origin   : tInterfaceOrigin;
    help     : integer;
    filter   : tPCAPNGCaptureFilter;

begin
     SiMain.EnterMethod ('tInterfaceMapper.GetInterfaceFromQuery');

     index := Query.FieldByName ('interfaceid').AsInteger - 1;              // Interfaces are stored starting with index 1
     linktype := Query.FieldByName ('linklayertype').AsInteger;
     help := Query.FieldByName ('origin').AsInteger;

     case help of
      0 : origin := ioUnknown;
      1 : origin := ioFromFile;
      2 : origin := ioAutogenerated;
      3 : origin := ioUserCreated;
     end;

     entry := tCaptureInterface.create (index, linktype, origin, true);

     entry.Name := Query.FieldByName ('name').AsString;
     entry.Description := Query.FieldByName ('description').AsString;
     entry.SnapLength := Query.FieldByName ('snaplen').AsInteger;
     entry.TimeStampResolution := Query.FieldByName ('tsresolution').AsInteger;
     entry.TimeZone := Query.FieldByName ('timezone').AsInteger;

     filter.FilterString := Query.FieldByName ('capturefilter').AsString;

     if filter.FilterString <> '' then
        filter.FilterType := cftLibPCAP
     else
         filter.FilterType := cftUnknown;

     entry.CaptureFilter := filter;

     entry.OperatingSystem := Query.FieldByName ('operatingsystem').AsString;
     entry.PacketLinkID := StringToGUID (Query.FieldByName ('packetlinkid').AsString);
     entry.Origin := origin;

     result := entry;

     SiMain.LeaveMethod ('tInterfaceMapper.GetInterfaceFromQuery');
end;

(* ------------------------------------------- *)

procedure tInterfaceMapper.LoadFromDatabase (Database: tFDConnection);

var Query       : tFDQuery;
    TargetQuery : tFDQuery;
    FileEntry   : tTraceFileEntry;
    IfEntry     : tCaptureInterface;
    DBIfEntry   : tCaptureInterface;
    TargetIf    : tCaptureInterface;
    FileIndex   : integer;
    IfIndex     : integer;
    i,j,k       : integer;

begin
     SiMain.EnterMethod ('tInterfaceMapper.LoadFromDatabase');

     Query := tFDQuery.Create(nil);
     Query.Connection := Database;

     TargetQuery := tFDQuery.Create(nil);
     TargetQuery.Connection := Database;

     // Clean up

     clearMappings;

     // Get all target interfaces first

     TargetQuery.SQL.Text := 'SELECT * FROM formTargetInterfaceList ORDER BY interfaceid;';

     TargetQuery.Open;
     TargetQuery.FetchAll;

     for i := 1 to TargetQuery.RecordCount do
         begin
              TargetIf := GetInterfaceFromQuery (TargetQuery);

              // Add to target list

              priv_targetinterfaces.Add (TargetIf);

              TargetQuery.Next;
         end;

     // Load files entries that exist in the current filelist

     for i := 1 to priv_filelist.Count do
         begin
              // Get files one by one from the DB

              FileEntry := priv_filelist.getFile(i);

              Query.SQL.Text := 'SELECT * FROM formFileList WHERE filename=:filename AND filesize=:filesize AND firstframetimestamp=:firstframetimestamp;';

              Query.ParamByName ('filename').AsString := FileEntry.FilePath + FileEntry.FileName;
              Query.ParamByName ('filesize').AsLargeInt := FileEntry.FileSize;
              Query.ParamByName ('firstframetimestamp').AsDateTime := FileEntry.FirstFrameTimestamp.datetime;

              Query.Open;

              if Query.RecordCount > 0 then
                 begin
                      // We have found a file that matches the on in our list

                      FileIndex := Query.FieldByName('id').AsInteger;

                      // Find target interfaces for each interface of the file

                      Query.SQL.Text := 'SELECT * FROM formFileInterfaces WHERE fileid=:fileid ORDER BY interfaceid ASC;';
                      Query.ParamByName ('fileid').AsInteger := FileIndex;

                      Query.Open;

                      // we should have as many interfaces as the current file has

                      if Query.RecordCount <> FileEntry.CaptureInterfaces.Count then
                         raise Exception.Create('Error: Stored interface count ' + IntToStr (Query.RecordCount) + ' does not match real file interface count: ' + IntToStr (FileEntry.CaptureInterfaces.Count) + ' (tInterfaceMapper.LoadFromDatabase)');

                      for j := 1 to Query.RecordCount do
                          begin
                               DBIfEntry := FileEntry.CaptureInterfaces.getInterfaceByIndex(j-1);

                               // Find get the interface and the target interface from the DB

                               IfEntry := GetInterfaceFromQuery (Query);

                               if not InterfacesAreEqual (DBIfEntry, IfEntry) then
                                  raise Exception.Create('Error: Stored interface does not match real file interface (tInterfaceMapper.LoadFromDatabase)');

                               IfIndex := Query.FieldByName('mappedtoid').AsInteger;

                               // Map the target interface

                               DBIfEntry.MappedToInterface := priv_targetinterfaces.getInterfaceByIndex (IfIndex - 1);

                               Query.Next;
                          end;
                 end;
         end;

     TargetQuery.Free;
     Query.Free;

     SiMain.LeaveMethod ('tInterfaceMapper.LoadFromDatabase');
end;

(* ------------------------------------------- *)

end.
