{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_ipfragment_manager;

interface

uses System.Generics.Collections,
     base_structures,
     definition_ip_addresses,
     class_rawframe;

type tIPFragment = record
//                       IPIdentification : longword;
                         Offset           : integer;
                         Len              : integer;
                         IsSliced         : boolean;
                         Flags            : tIPFlagSet;
                         Fileposition     : tFileFrameLocation;
                         Frame            : tRawFrame;

                         procedure initialize;
                   end;

     tFragmentedPacket = class (tObject)

                               constructor create (const IPAddress : tIPAddress; const IPIdentifier : word);

                               private
                                      priv_IPAddress                : tIPAddress;
                                      priv_IPIdentifier             : word;
                                      priv_lastFragmentSeen         : boolean;
                                      priv_reassembledpayloadlength : integer;
                                      priv_incomplete               : boolean;
                                      priv_hasSlicedFragments       : boolean;
                                      priv_fragmentsOutOfOrder      : boolean;
                                      priv_FragmentMaxPayloadLen    : integer;
                                      priv_FirstFrameTime           : tFrameTimestamp;

                                      priv_fragmentlist             : TDictionary<integer, tIPFragment>;

                                      /// <summary>
                                      ///   Determines if all fragments are
                                      ///   available, meaning there are no
                                      ///   gaps and we have seen the last
                                      ///   fragment
                                      /// </summary>
                                      procedure CheckFragmentCollectionComplete;

                                      /// <summary>
                                      ///   This determines the layer 2 &amp; 3
                                      ///   base for the reassembled frame
                                      /// </summary>
                                      function GetFrameBase (const Frame : tRawFrame) : tFrameChunk;
                                      function GetFragmentPayload (const Frame : tRawFrame) : tFrameChunk;

                               public
                                     property TimestampOfFirstFrame : tFrameTimestamp read priv_FirstFrameTime;
                                     property LastFragmentSeen : boolean read priv_lastFragmentSeen;
                                     property ReassembledPayloadLength : integer read priv_reassembledpayloadlength;
                                     property isIncomplete : boolean read priv_incomplete;
                                     property HasSlicedFragments : boolean read priv_hasSlicedFragments;
                                     property FragmentsAreOutOfOrder : boolean read priv_fragmentsOutOfOrder;
                                     property MaximumIPPayloadLength : integer read priv_FragmentMaxPayloadLen;

                                     /// <summary>
                                     ///   This adds a new frame containing a
                                     ///   fragment piece
                                     /// </summary>
                                     /// <param name="Frame">
                                     ///   Frame containing the fragment
                                     /// </param>
                                     /// <param name="Offset">
                                     ///   Start offset of the fragment
                                     /// </param>
                                     /// <param name="Length">
                                     ///   Length of the fragment
                                     /// </param>
                                     procedure AddPacket (const Frame : tRawFrame; const Offset : integer; const Length : integer; const Flags : tIPFlagSet);
                                     /// <summary>
                                     ///   Creates a new frame build from all
                                     ///   fragments. For this to work all
                                     ///   fragments need to be available and
                                     ///   not sliced.
                                     /// </summary>
                                     /// <returns>
                                     ///   Reassembled frame, or nil if
                                     ///   fragments are sliced or some
                                     ///   fragments are missing
                                     /// </returns>
                                     function GetReassembledFrame : tRawFrame;
                                     function DictionaryKey : string;

                                     destructor Destroy; override;
                         end;


     tIPFragmentManager = class (tObject)

                                constructor create;

                                private
                                       priv_FragmentedPacketList        : TObjectDictionary<string, tFragmentedPacket>;
                                       priv_FirstFrameTime              : tFrameTimestamp;

                                       function CheckIfAllFragmentsAreAvailable : boolean;
                                       function getCount: integer;

                                public
                                      property TimestampOfFirstFrame : tFrameTimestamp read priv_FirstFrameTime;
                                      property count : integer read getCount;
                                      property AllPacketFragmentsAvailable : boolean read CheckIfAllFragmentsAreAvailable;

                                      function FragmentedPacketExists (const IPAddress : tIPAddress; const IPIdentifier : word) : boolean;
                                      /// <summary>
                                      ///   Adds fragment details of a new
                                      ///   packet to the manager so that the
                                      ///   fragment status can be tracked
                                      /// </summary>
                                      /// <param name="IPAddress">
                                      ///   IP address where the fragment comes
                                      ///   from
                                      /// </param>
                                      /// <param name="IPIdentifier">
                                      ///   the IP ID found in the packet
                                      /// </param>
                                      /// <param name="FilePosition">
                                      ///   Position in the trace file
                                      /// </param>
                                      /// <param name="Offset">
                                      ///   Fragment Offset
                                      /// </param>
                                      /// <param name="Length">
                                      ///   Length of the fragment
                                      /// </param>
                                      /// <param name="Flags">
                                      ///   IP flags indicating fragment status
                                      /// </param>
                                      /// <returns>
                                      ///   True if all fragments of a
                                      ///   fragmented packets were found and
                                      ///   added
                                      /// </returns>
                                      function AddPacketFragmentAndCheckComplete (const IPAddress : tIPAddress; const IPIdentifier : word; const Frame : tRawFrame; const Offset : integer; const Length : integer; const Flags : tIPFlagSet) : boolean;
                                      /// <summary>
                                      ///   Returns the fragment entry telling the status of all its parts
                                      /// </summary>
                                      function GetFragmentedPacketEntry (const IPAddress : tIPAddress; const IPIdentifier : word) : tFragmentedPacket;
                                      /// <summary>
                                      ///   Returns the reassembled frame
                                      ///   created from all fragments. This
                                      ///   creates a new tRawFrame containing
                                      ///   the frame as it looked like before
                                      ///   it was fragmented.
                                      /// </summary>
                                      /// <param name="IPAddress">
                                      ///   IP address of the frame
                                      /// </param>
                                      /// <param name="IPIdentifier">
                                      ///   IP Identifier of the frame
                                      /// </param>
                                      /// <returns>
                                      ///   Reassembled frame or NIL if not
                                      ///   possible
                                      /// </returns>
                                      function GetReassembledFrame (const IPAddress : tIPAddress; const IPIdentifier : word) : tRawFrame;
                                      /// <summary>
                                      ///   This function tells how big the IP
                                      ///   payload of a fragment can be at
                                      ///   most, which is determined by
                                      ///   looking at the IP payload size of
                                      ///   the first fragment (with offset 0).
                                      /// </summary>
                                      /// <remarks>
                                      ///   This function allows the last
                                      ///   fragment to realize that it is
                                      ///   smaller than the maximum payload
                                      ///   size, setting the Flags
                                      ///   accordingly.
                                      /// </remarks>
                                      function GetMaximumIPPayloadLength (const IPAddress : tIPAddress; const IPIdentifier : word) : integer;


                                      destructor Destroy; override;

                          end;

     function getFragmentedPacketDictionaryKey (const IPAddress : tIPAddress; const IPIdentifier : longword) : string;
     function getRawFrameFromFrameChunk (Frame: tRawFrame; PayLoad: tFrameChunk; Ticks : int64; Timestamp : tFrameTimestamp): tRawFrame;
     /// <summary>
     ///   Returns a fragment of an IP packet specified by the fragment offset.
     ///   This creates a new tRawFrame.
     /// </summary>
     /// <param name="FullFrame">
     ///   Non-fragmented original frame
     /// </param>
     /// <param name="FragmentOffset">
     ///   Offset of the IP payload, 0 for the first fragment
     /// </param>
     /// <param name="FragmentPayloadSize">
     ///   Size of the payload for the fragment, usually 1480
     /// </param>
     function GetFragmentFromPacket (const FullFrame : tRawFrame; const FragmentOffset, FragmentPayloadSize : integer; const Ticks : int64; const timestamp : tFrameTimestamp) : tRawFrame;

implementation

uses SysUtils,
     Math,
     SiAuto,
     common_parse_helpers,
     class_parsemaster,
     class_parser_baseclass,
     class_parser_ethernet,
     class_parser_ipv4,
     class_parser_ipv6,
     class_parser_ipv6_fragmentationheader,
     class_assembly_ethernet,
     class_assembly_ipv4,
     class_assembly_ipv6,
     class_assembly_ipv6_fragmentationheader;

(* -------------------------------------- *)

function getRawFrameFromFrameChunk (Frame: tRawFrame; PayLoad: tFrameChunk; Ticks : int64; Timestamp : tFrameTimestamp): tRawFrame;

var NewFrame     : tRawFrame;
    TimeDetails  : tTimingDetails;
    WireDetails  : tWireDetails;
    FilePosition : tFileFrameLocation;
    FrameBuffer  : tRawBuffer;
    FinalLen     : integer;
    i            : integer;

begin
     TimeDetails.AbsoluteTime := Timestamp.datetime;
     TimeDetails.TraceTicks := Frame.Ticks;
     TimeDetails.NanoTicks := (Timestamp.nanoseconds mod 1000);            // because right now, Nanosecs also contain microsecs


     WireDetails.LinkType := Frame.LinkType;
     WireDetails.WireSize := PayLoad.Len;
     WireDetails.CaptureSize := Payload.Len;
     WireDetails.InterfaceID := Frame.InterfaceID;
     WireDetails.Dropped := Frame.DropCount;
     WireDetails.FrameFlags := Frame.FrameFlags;
     WireDetails.FrameType := Frame.FrameFileType;

     FilePosition.FileIndex := Frame.TracefileID;
     FilePosition.FileFrameIndex := Frame.FileFrameID;
     FilePosition.FileOffset := 0;

     move (Payload.Data^, FrameBuffer, WireDetails.CaptureSize);

     NewFrame := tRawFrame.create (TimeDetails, WireDetails, FilePosition, FrameBuffer);
     NewFrame.InterfaceLinkID := Frame.InterfaceLinkID;

     for i := 1 to Frame.Commentlist.Count do
         NewFrame.Commentlist.Add(Frame.Commentlist[i-1]);

     result := NewFrame;
end;

(* -------------------------------------- *)

function GetFragmentFromPacket (const FullFrame : tRawFrame; const FragmentOffset, FragmentPayloadSize : integer; const Ticks : int64; const timestamp : tFrameTimestamp) : tRawFrame;

var ParseMaster       : tParseMaster;
    IPv4Parser        : tIPv4Parser;
    IPv6Parser        : tIPv6Parser;
    IPv4Assembler     : tIPv4Assembler;
    IPv6Assembler     : tIPv6Assembler;
    IPv6FragAssembler : tIPv6FragHdrAssembler;
    NewFrame          : tRawFrame;
    HeaderType        : tHeaderType;
    LayerIndex        : integer;
    PayloadLen        : integer;
    L2Chunk           : tFrameChunk;
    IPChunk           : tFrameChunk;
    PayloadChunk      : tFrameChunk;
    BaseChunk         : tFrameChunk;
    MergeChunk        : tFrameChunk;
    found             : boolean;
    i                 : integer;

begin
     SiMain.EnterMethod ('GetFragmentFromPacket');
     ParseMaster := tParseMaster.create (FullFrame, ppmUpToLayer3);

     i := 0;
     found := false;

     repeat
           inc (i);
           HeaderType := ParseMaster.getParserTypeByIndex(i);

           if (HeaderType = NEXT_HEADER_IPv4) or (HeaderType = NEXT_HEADER_IPv4_ENCAPSULATION) then
              begin
                   IPv4Parser := ParseMaster.getParserByIndex(i) as tIPv4Parser;

                   assert (FragmentOffset < IPv4Parser.IPPayloadLength, 'Invalid FragmentOffset (GetFragmentFromPacket)');

                   // Get the L2 stuff (if any)

                   L2Chunk := FullFrame.FrameDataAsChunk (0, IPv4Parser.FrameOffset);

                   // Determine the payload length

                   PayloadLen := min (IPv4Parser.IPPayloadLength - FragmentOffset , FragmentPayloadSize);

                   // Build the new IP header

                   IPv4Assembler := tIPv4Assembler.create (PayloadLen);
                   IPv4Assembler.CopyFromParser (IPv4Parser);

                   // Set the correct fragment offset value

                   IPv4Assembler.Offset := FragmentOffset;

                   // Clear the fragment flags

                   if PayloadLen < FragmentPayloadSize then
                      IPv4Assembler.Flags := []
                   else
                       IPv4Assembler.Flags := [ipfMoreFragments];

                   IPChunk := IPv4Assembler.GetAssembly;
                   IPv4Assembler.Free;

                   // Get the payload for the fragment

                   PayloadChunk := FullFrame.FrameDataAsChunk (IPv4Parser.NextAnchor.Offset + FragmentOffset, PayloadLen);

                   found := True;
              end;

           if (HeaderType = NEXT_HEADER_IPV6) or (HeaderType = NEXT_HEADER_TYPE_IPV6) then
              begin
                   IPv6Parser := ParseMaster.getParserByIndex(i) as tIPv6Parser;

                   // TODO: handle IPv6








                   found := true;
              end;

     until found or (i = ParseMaster.ParserCount);

     ParseMaster.free;

     if found then
        begin
             if L2Chunk.Len > 0 then
                begin
                     BaseChunk := mergeFrameChunks (L2Chunk, IPChunk);

                     // Combine L2 and IP

                     L2Chunk.free;
                     IPChunk.free;

                     MergeChunk := mergeFrameChunks (BaseChunk, PayloadChunk);

                     BaseChunk.free;
                     PayloadChunk.free;
                end
             else
                 begin
                      MergeChunk := mergeFrameChunks (IPChunk, PayloadChunk);

                      IPChunk.free;
                      PayloadChunk.free;
                 end;

             // We don't need the cut markers here

             MergeChunk.clearCutMarkers;
        end;

     // Build the frame

     NewFrame := getRawFrameFromFrameChunk (FullFrame, MergeChunk, Ticks, Timestamp);

     result := NewFrame;

     SiMain.LeaveMethod ('GetFragmentFromPacket');
end;

(* -------------------------------------- *)

function getFragmentedPacketDictionaryKey (const IPAddress : tIPAddress; const IPIdentifier : longword) : string;

begin
     result := IPAddress.ValueAsString + '-' + IntToStr (IPIdentifier);
end;

(* -------------------------------------- *)
{ tIPFragment }
(* -------------------------------------- *)

procedure tIPFragment.initialize;

begin
//   IPIdentification := 0;
     Offset := 0;
     Len := 0;
     IsSliced := False;
     Flags := [];
     Fileposition.initialize;
     Frame := nil;
end;

(* -------------------------------------- *)
{ tFragmentEntry }
(* -------------------------------------- *)

constructor tFragmentedPacket.create (const IPAddress : tIPAddress; const IPIdentifier : word);

begin
     inherited create;

     priv_IPAddress := IPAddress;
     priv_IPIdentifier := IPIdentifier;

     priv_reassembledpayloadlength := 0;
     priv_lastFragmentSeen := false;
     priv_incomplete := True;
     priv_fragmentsOutOfOrder := False;

     priv_fragmentlist := TDictionary<integer, tIPFragment>.create;
end;

(* -------------------------------------- *)

destructor tFragmentedPacket.Destroy;

begin
     priv_fragmentlist.Free;

     inherited;
end;

(* -------------------------------------- *)

procedure tFragmentedPacket.CheckFragmentCollectionComplete;

var TestArray : TArray<integer>;
    i         : integer;
    expect    : integer;
    lastpos   : tFileFrameLocation;
    value     : boolean;

begin
     value := true;
     expect := 0;
     lastpos.initialize;

     priv_hasSlicedFragments := False;

     // Use a TArray<integer> to get sorted access to the dictionary

     TestArray := priv_fragmentlist.Keys.ToArray;
     TArray.Sort<integer>(TestArray);

     // Go through the offsets in sorted order and check if each offset matches
     // the previous plus the previous length

     for i in TestArray do
         begin
              // Check if the fragment is sliced, which means we cannot reassemble

              if priv_fragmentlist.Items[i].IsSliced then
                 priv_hasSlicedFragments := True;

              // Check if the offsets are gapless

              if (priv_fragmentlist.Items[i].Offset <> expect) then
                 value := false
              else
                  expect := priv_fragmentlist.Items[i].Offset + priv_fragmentlist.Items[i].Len;

              // Check if the fragments were out of order

              if lastpos.FileFrameIndex > 0 then
                 begin
                      if priv_fragmentlist.Items[i].FilePosition < lastpos then
                         priv_fragmentsOutOfOrder := True;
                 end;

              lastpos := priv_fragmentlist.Items[i].FilePosition;
         end;

     if value then
        priv_incomplete := False;
end;

(* -------------------------------------- *)

procedure tFragmentedPacket.AddPacket (const Frame : tRawFrame; const Offset, Length: integer; const Flags : tIPFlagSet);

var fragment : tIPFragment;

begin
     if not priv_fragmentlist.ContainsKey (Offset) then
        begin
             // Record the time we got the first fragment

             if priv_fragmentlist.Count = 0 then
                priv_FirstFrameTime := Frame.Timestamp;

             // Got a new fragment (we don't need the IP ID here)

//           fragment.IPIdentification := 0;
             fragment.Offset := Offset;
             fragment.Len := Length;
             fragment.FilePosition := Frame.FilePosition;
             fragment.Flags := Flags;
             fragment.IsSliced := (Frame.FrameSize < Frame.WireSize);
             fragment.Frame := Frame;

             // Remember how big the first fragment was, because we need that for reassembly

             if Offset = 0 then
                priv_FragmentMaxPayloadLen := Length;

             priv_reassembledpayloadlength := priv_reassembledpayloadlength + Length;

             // Check if we have them all. IPv4 doesn't have an explicit flag for this, so we need to check
             // if there's no more and the offset > 0

             if (ipfLastFragment in Flags) or (not (ipfMoreFragments in Flags) and (Offset > 0)) then
                begin
                     fragment.Flags := fragment.Flags + [ipfLastFragment];
                     priv_lastFragmentSeen := True;
                end;

             priv_fragmentlist.Add (Offset, fragment);

             if priv_lastFragmentSeen then
                begin
                     // Check if we have it all

                     CheckFragmentCollectionComplete;
                end;
        end
     else
         begin
              // TODO: handle duplicates (yeah. lazy me, but this is a tough nut to crack)


         end;
end;

(* -------------------------------------- *)

function tFragmentedPacket.GetFrameBase (const Frame: tRawFrame) : tFrameChunk;

var ParseMaster   : tParseMaster;
    IPv4Parser    : tIPv4Parser;
    IPv6Parser    : tIPv6Parser;
    IPv4Assembler : tIPv4Assembler;
    IPv6Assembler : tIPv6Assembler;
    HeaderType    : tHeaderType;
    LayerIndex    : integer;
    Offset        : integer;
    L2Chunk       : tFrameChunk;
    IPChunk       : tFrameChunk;
    BaseChunk     : tFrameChunk;
    PayloadChunk  : tFrameChunk;
    MergeChunk    : tFrameChunk;
    found         : boolean;
    i             : integer;

begin
     ParseMaster := tParseMaster.create (Frame, ppmUpToLayer3);

     // Find the IP layer, then preserve everything before it so we can avoid having to deal
     // with whatever it is

     i := 0;
     found := false;
     IPChunk.initialize;
     MergeChunk.initialize;

     repeat
           inc (i);
           HeaderType := ParseMaster.getParserTypeByIndex(i);

           if (HeaderType = NEXT_HEADER_IPv4) or (HeaderType = NEXT_HEADER_IPv4_ENCAPSULATION) then
              begin
                   IPv4Parser := ParseMaster.getParserByIndex(i) as tIPv4Parser;

                   // Get the L2 stuff (if any)

                   L2Chunk := Frame.FrameDataAsChunk (0, IPv4Parser.FrameOffset);

                   // Build the new IP header

                   IPv4Assembler := tIPv4Assembler.create (priv_reassembledpayloadlength);
                   IPv4Assembler.CopyFromParser (IPv4Parser);

                   // Clear the fragment flags

                   IPv4Assembler.Flags := [];

                   IPChunk := IPv4Assembler.GetAssembly;
                   IPv4Assembler.Free;

                   PayloadChunk := Frame.FrameDataAsChunk (IPv4Parser.NextAnchor.Offset);

                   found := True;
              end;

           if (HeaderType = NEXT_HEADER_IPV6) or (HeaderType = NEXT_HEADER_TYPE_IPV6) then
              begin
                   IPv6Parser := ParseMaster.getParserByIndex(i) as tIPv6Parser;

                   // Get the L2 stuff (if any)

                   L2Chunk := Frame.FrameDataAsChunk (0, IPv6Parser.FrameOffset);

                   // Build the new IP header

                   IPv6Assembler := tIPv6Assembler.create (priv_reassembledpayloadlength);
                   IPv6Assembler.CopyFromParser (IPv6Parser);

                   // Need to set NextHeader to UDP instead of fragmentation header

                   IPv6Assembler.NextHeader := 17;

                   IPChunk := IPv6Assembler.GetAssembly;
                   IPv6Assembler.Free;

                   PayloadChunk := Frame.FrameDataAsChunk (IPv6Parser.NextAnchor.Offset);

                   found := true;
              end;

     until found or (i = ParseMaster.ParserCount);

     ParseMaster.free;

     if found then
        begin
             if L2Chunk.Len > 0 then
                begin
                     BaseChunk := mergeFrameChunks (L2Chunk, IPChunk);

                     // Combine L2 and IP

                     L2Chunk.free;
                     IPChunk.free;
                end
             else
                 BaseChunk := IPChunk;

             // Combine base and IP payload

             MergeChunk := mergeFrameChunks (BaseChunk, PayloadChunk);

             BaseChunk.free;
             PayloadChunk.free;

             // We don't need the cut markers here

             MergeChunk.clearCutMarkers;
        end;

     result := MergeChunk;
end;

(* -------------------------------------- *)

function tFragmentedPacket.GetFragmentPayload (const Frame: tRawFrame): tFrameChunk;

var ParseMaster   : tParseMaster;
    Parser        : tFrameParser;
    HeaderType    : tHeaderType;
    Chunk         : tFrameChunk;
    found         : boolean;
    i             : integer;

begin
     ParseMaster := tParseMaster.create (Frame, ppmUpToLayer3);

     i := 0;
     found := false;

     repeat
           inc (i);
           HeaderType := ParseMaster.getParserTypeByIndex(i);

           if (HeaderType = NEXT_HEADER_IPv4) or (HeaderType = NEXT_HEADER_IPv4_ENCAPSULATION) then
              begin
                   Parser := ParseMaster.getParserByIndex(i);
                   found := True;
              end;

     until found or (i = ParseMaster.ParserCount);

     ParseMaster.free;

     if found then
        Chunk := Frame.FrameDataAsChunk (Parser.NextAnchor.Offset)
     else
         Chunk.initialize;

     result := Chunk;
end;

(* -------------------------------------- *)

function tFragmentedPacket.GetReassembledFrame : tRawFrame;

var TestArray     : TArray<integer>;
    Parser        : tParseMaster;
    FrameChunk    : tFrameChunk;
    FragmentChunk : tFrameChunk;
    MergeChunk    : tFrameChunk;
    FirstFrame    : tRawFrame;
    NewFrame      : tRawFrame;
    i             : integer;

begin
     NewFrame := nil;

     if not (priv_hasSlicedFragments or priv_incomplete) then
        begin
             // Use a TArray<integer> to get sorted access to the dictionary

             TestArray := priv_fragmentlist.Keys.ToArray;
             TArray.Sort<integer>(TestArray);

             // Go through the offsets in sorted order and check if each offset matches
             // the previous plus the previous length

             for i in TestArray do
                 begin
                      if i = 0 then
                         begin
                              FirstFrame := priv_fragmentlist.Items[i].Frame;

                              // first fragment

                              FrameChunk := GetFrameBase (FirstFrame);

//                            FrameChunk.writeToDisk ('d:\temp\Fragment' + IntToStr(i) + '.hex');

                              if FrameChunk.Len = 0 then
                                 exit (NewFrame);
                         end
                      else
                          begin
                               // Get the IP Payload

                               FragmentChunk := GetFragmentPayload (priv_fragmentlist.Items[i].Frame);
//                             FragmentChunk.writeToDisk ('d:\temp\Fragment' + IntToStr(i) + '.hex');

                               if FragmentChunk.Len > 0 then
                                  begin
                                       MergeChunk := mergeFrameChunks (FrameChunk, FragmentChunk);

                                       FrameChunk.free;
                                       FragmentChunk.free;

                                       FrameChunk := MergeChunk;
                                  end;
                          end;
                 end;
        end;

     NewFrame := getRawFrameFromFrameChunk (FirstFrame, FrameChunk, FirstFrame.Ticks, FirstFrame.Timestamp);

     result := NewFrame;
end;

(* -------------------------------------- *)

function tFragmentedPacket.DictionaryKey : string;

begin
     result := getFragmentedPacketDictionaryKey (priv_IPAddress, priv_IPIdentifier);
end;

(* -------------------------------------- *)
{ tIPFragmentManager }
(* -------------------------------------- *)

constructor tIPFragmentManager.create;

begin
     inherited;

     priv_FragmentedPacketList := TObjectDictionary<string, tFragmentedPacket>.create ([doOwnsValues]);
end;

(* -------------------------------------- *)

destructor tIPFragmentManager.Destroy;

begin
     priv_FragmentedPacketList.Free;

     inherited;
end;

(* -------------------------------------- *)

function tIPFragmentManager.CheckIfAllFragmentsAreAvailable: boolean;

var key   : string;
    value : boolean;

begin
     value := True;

     // Check all Packets to see if at least one of them has missing fragments

     for key in priv_FragmentedPacketList.Keys do
         begin
              if priv_FragmentedPacketList.Items[key].isIncomplete then
                 value := False;
         end;

     result := value;
end;

(* -------------------------------------- *)

function tIPFragmentManager.AddPacketFragmentAndCheckComplete (const IPAddress : tIPAddress; const IPIdentifier : word; const Frame : tRawFrame; const Offset : integer; const Length : integer; const Flags : tIPFlagSet) : boolean;

var entry : tFragmentedPacket;
    key   : string;

begin
     assert ((Offset > 0) or (ipfMoreFragments in Flags), 'Frame is not a fragment');

     // Record the time we got the first fragment

     if priv_FragmentedPacketList.Count = 0 then
        priv_FirstFrameTime := Frame.Timestamp;

     key := getFragmentedPacketDictionaryKey (IPAddress, IPIdentifier);

     if priv_FragmentedPacketList.ContainsKey (key) then
        priv_FragmentedPacketList.Items[key].AddPacket (Frame, Offset, Length, Flags)
     else
         begin
              entry := tFragmentedPacket.create (IPAddress, IPIdentifier);
              entry.AddPacket (Frame, Offset, Length, Flags);

              priv_FragmentedPacketList.Add (key, entry);
         end;

     // return if the all fragments were all added or not

     result := priv_FragmentedPacketList.Items[key].isIncomplete;
end;

(* -------------------------------------- *)

function tIPFragmentManager.FragmentedPacketExists (const IPAddress: tIPAddress; const IPIdentifier: word): boolean;

var key : string;

begin
     key := getFragmentedPacketDictionaryKey (IPAddress, IPIdentifier);

     result := priv_FragmentedPacketList.ContainsKey (key);
end;

(* -------------------------------------- *)

function tIPFragmentManager.GetFragmentedPacketEntry (const IPAddress: tIPAddress; const IPIdentifier: word): tFragmentedPacket;

var key : string;

begin
     key := getFragmentedPacketDictionaryKey (IPAddress, IPIdentifier);

     if priv_FragmentedPacketList.ContainsKey (key) then
        result := priv_FragmentedPacketList.Items[key]
     else
         result := nil;
end;

(* -------------------------------------- *)

function tIPFragmentManager.getCount: integer;

begin
     result := priv_FragmentedPacketList.Count;
end;

(* -------------------------------------- *)

function tIPFragmentManager.GetMaximumIPPayloadLength (const IPAddress: tIPAddress; const IPIdentifier: word): integer;

var key : string;

begin
     key := getFragmentedPacketDictionaryKey (IPAddress, IPIdentifier);

     result := priv_FragmentedPacketList.Items[key].MaximumIPPayloadLength;
end;

(* -------------------------------------- *)

function tIPFragmentManager.GetReassembledFrame (const IPAddress: tIPAddress; const IPIdentifier: word): tRawFrame;

var ReFrame : tRawFrame;
    key     : string;

begin
     ReFrame := nil;

     key := getFragmentedPacketDictionaryKey (IPAddress, IPIdentifier);

     if priv_FragmentedPacketList.ContainsKey (key) then
        ReFrame := priv_FragmentedPacketList.Items[key].GetReassembledFrame;

     result := ReFrame;
end;

(* -------------------------------------- *)


end.
