{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_ipv4subnethandler;

interface

uses contnrs,
     definition_ip_addresses;

type tIPv4SubnetDefinition = class (tObject)

                               constructor create (const IPAddress : tIPv4Address); overload;
                               constructor create (const IPAddress : tIPv4Address; const SubnetName : string); overload;
                               constructor create (const SubnetAddress, SubnetMask, SubnetName : string); overload;
                               constructor create (const SubnetAddress, SubnetMask : longword; const SubnetName : string); overload;
                               procedure free;

                               private
                                      priv_Name           : string;
                                      priv_SubnetAddress  : tIPv4Address;
                                      priv_ReplacementNet : tIPv4Address;

{$ifdef debug}
                                      function getReplacementRule : string;
{$endif}
                               public
                                     property Name : string read priv_Name write priv_Name;
                                     property SubnetAddress : tIPv4Address read priv_SubnetAddress write priv_SubnetAddress;
                                     property ReplacementNetAddress : tIPv4Address read priv_ReplacementNet write priv_ReplacementNet;
{$ifdef debug}
                                     property ReplacementRule : string read getReplacementRule;
{$endif}

                                     {$REGION 'Documentation'}
                                     ///	<summary>
                                     ///	  Prepares the IPv6 Subnet masking
                                     ///	  table
                                     ///	</summary>
                                     {$ENDREGION}
                                     procedure copyTo (Target : tIPv4SubnetDefinition);
                                     function isEqualTo (Target : tIPv4SubnetDefinition) : boolean;

                                     {$REGION 'Documentation'}
                                     ///	<summary>
                                     ///	  generates a random IP for the
                                     ///	  replacement subnet range
                                     ///	</summary>
                                     {$ENDREGION}
                                     function getReplacementIPForSubnet (const OriginalAddress : tIPv4Address) : tIPv4Address;
                         end;

     tIPv4SubnetHandler    = class (tObject)

                                   constructor create;
                                   procedure free;

                                   private
                                          priv_subnetlist      : tObjectList;
                                          priv_sorted          : boolean;
                                          priv_LastIPAddress   : tIPv4Address;
                                          priv_LastSubnetIndex : integer;

                                          function getSubnetCount: integer;

                                   public
                                         property Count : integer read getSubnetCount;
                                         property Sorted : boolean read priv_sorted write priv_sorted;

                                         procedure AddSubnet (Subnet : tIPv4SubnetDefinition);
                                         procedure Clear;
                                         function GetSubnetByIndex (const Index : integer) : tIPv4SubnetDefinition;

                                         {$REGION 'Documentation'}
                                         ///	<summary>
                                         ///	  Returns to subnet index that a given
                                         ///	  IP adress belongs to
                                         ///	</summary>
                                         ///	<param name="IPAddress">
                                         ///	  IP address to find a subnet for.
                                         ///	</param>
                                         ///	<returns>
                                         ///	  Index of the subnetdefinition the
                                         ///	  IPAddress belongs to. Returns 0 if
                                         ///	  none was found.
                                         ///	</returns>
                                         {$ENDREGION}
                                         function GetSubnetIndexForIPAddress (const IPAddress : tIPv4Address) : integer;
                                         procedure LoadFromCSVFile (const Filename : string);
                                         procedure SaveToCSVFile (const Filename : string);
                                         procedure Sort;

                             end;

implementation

uses SysUtils,
     toolkit;

(* -------------------------------------- *)

function compareSubnets (item1, item2 : pointer) : integer;

var def1, def2 : tIPv4SubnetDefinition;

begin
     def1 := item1;
     def2 := item2;

     result := CompareValue (def1.priv_SubnetAddress, def2.priv_SubnetAddress);
end;

(* -------------------------------------- *)
{ tSubnetDefinition }
(* -------------------------------------- *)

constructor tIPv4SubnetDefinition.create(const IPAddress: tIPv4Address);

begin
     inherited create;

     priv_Name := '';
     priv_SubnetAddress := IPAddress;
end;

(* -------------------------------------- *)

constructor tIPv4SubnetDefinition.create(const IPAddress: tIPv4Address; const SubnetName: string);

begin
     inherited create;

     priv_Name := SubnetName;
     priv_SubnetAddress := IPAddress;
end;

(* -------------------------------------- *)

constructor tIPv4SubnetDefinition.create(const SubnetAddress, SubnetMask, SubnetName: string);

begin
     inherited create;

     priv_Name := SubnetName;

     priv_SubnetAddress.setValueFromString (SubnetAddress);
     priv_SubnetAddress.setMaskFromString (SubnetMask);
end;

(* -------------------------------------- *)

constructor tIPv4SubnetDefinition.create(const SubnetAddress, SubnetMask: longword; const SubnetName: string);

begin
     inherited create;

     priv_Name := SubnetName;

     priv_SubnetAddress.ValueAsLong := SubnetAddress;
     priv_SubnetAddress.setMaskFromLong(SubnetMask);
end;

(* -------------------------------------- *)

procedure tIPv4SubnetDefinition.free;

begin
     inherited free;
end;

(* -------------------------------------- *)

function tIPv4SubnetDefinition.isEqualTo(Target: tIPv4SubnetDefinition): boolean;

begin
     result := (priv_Name = Target.Name) and (priv_SubnetAddress = Target.SubnetAddress);
end;

(* -------------------------------------- *)

procedure tIPv4SubnetDefinition.copyTo(Target: tIPv4SubnetDefinition);

begin
     Target.Name := priv_Name;
     Target.SubnetAddress := priv_SubnetAddress;
end;

(* -------------------------------------- *)

function tIPv4SubnetDefinition.getReplacementIPForSubnet (const OriginalAddress : tIPv4Address) : tIPv4Address;

var WildCard : tIPv4Address;
    help     : longword;
    randomIP : tIPv4Address;

begin
     result.initialize;

     // Extract Address part from Original

     WildCard := priv_SubnetAddress.NetWildCard;

     if OriginalAddress <> IPv4_NODE_UNSPECIFIED then
        begin
             // take existing host part

             help := OriginalAddress.ValueAsLong;
        end
     else
         begin
              // generate random address part

              randomIP.initialize(true);
              help := randomIP.ValueAsLong;
         end;

     help := help and WildCard.ValueAsLong;

     // Return full network ip

     result.ValueAsLong := (priv_ReplacementNet.ValueAsLong or help);
end;

(* -------------------------------------- *)

{$ifdef debug}
function tIPv4SubnetDefinition.getReplacementRule : string;

begin
     result := priv_SubnetAddress.ValueAsCIDRString + ' -> ' + priv_ReplacementNet.ValueAsCIDRString;
end;
{$endif}

(* -------------------------------------- *)
{ tSubnetHandler }
(* -------------------------------------- *)

constructor tIPv4SubnetHandler.create;

begin
     inherited;

     priv_subnetlist := tObjectList.Create(true);
     priv_sorted := true;

     priv_LastSubnetIndex := -1;
     priv_LastIPAddress.ValueAsLong := 0;
end;

(* -------------------------------------- *)

procedure tIPv4SubnetHandler.free;

begin
     inherited;

     priv_subnetlist.Free;
end;

(* -------------------------------------- *)

function tIPv4SubnetHandler.getSubnetCount: integer;

begin
     result := priv_subnetlist.Count;
end;

(* -------------------------------------- *)

procedure tIPv4SubnetHandler.Clear;

begin
     priv_subnetlist.Clear;
end;

(* -------------------------------------- *)

procedure tIPv4SubnetHandler.AddSubnet(Subnet: tIPv4SubnetDefinition);

begin
     priv_subnetlist.Add(Subnet);

     // If the list needs to be sorted at all times, do it (usually for BinSearch)

     if priv_sorted then
        Sort;
end;

(* -------------------------------------- *)

function tIPv4SubnetHandler.GetSubnetByIndex (const Index: integer): tIPv4SubnetDefinition;

begin
     if (Index > 0) and (Index <= priv_subnetlist.Count) then
        result := priv_subnetlist[Index-1] as tIPv4SubnetDefinition
     else
         result := nil;
end;

(* -------------------------------------- *)

function tIPv4SubnetHandler.GetSubnetIndexForIPAddress (const IPAddress : tIPv4Address) : integer;

var subnet   : tIPv4SubnetDefinition;
    help     : tIPv4Address;
    baseaddr : tIPv4Address;
    current  : integer;
    i        : integer;

begin
     result := 0;
     current := 0;
     help := IPAddress;

     // Check if it is 0.0.0.0

     if (IPAddress.ValueAsLong = 0) then
        Exit (0);

     // Check Cache

     if IPAddress = priv_LastIPAddress then
        Exit (priv_LastSubnetIndex);

     // Search subnet

     for i := 1 to priv_subnetlist.Count do
         begin
              subnet := (priv_subnetlist[i-1] as tIPv4SubnetDefinition);

              // Take current subnet mask and apply it to the IPaddress

              help.MaskBitCount := subnet.priv_SubnetAddress.MaskBitCount;
              baseaddr := help.LowestNetIP;

              // See if the address matches the subnet

              if baseaddr.ValueAsLong = subnet.priv_SubnetAddress.ValueAsLong then
                 begin
                      // if no subnet was found yet, this is the one for now
                      // otherwise, see if the new one has a tighter subnet mask than the last find

                      if current = 0 then
                         current := i
                      else
                          if (priv_subnetlist[current-1] as tIPv4SubnetDefinition).priv_SubnetAddress.MaskBitCount < subnet.priv_SubnetAddress.MaskBitCount then
                             current := i;
                 end;
         end;

     // Update Cache

     if current > 0 then
        begin
             priv_LastSubnetIndex := current;
             priv_LastIPAddress := IPAddress;
        end;

     result := current;
end;

(* -------------------------------------- *)

procedure tIPv4SubnetHandler.LoadFromCSVFile(const Filename: string);

var fHandle        : text;
    sLine          : string;
    boValidFile    : boolean;
    boWasSorted    : boolean;
    Subnet         : tIPv4SubnetDefinition;

begin
     boValidFile := true;

     try
        assignfile (fHandle, Filename);
        reset (fHandle);
     except
           boValidFile := false;
     end;

     if boValidFile then
        begin
             boWasSorted := priv_sorted;
             priv_sorted := false;

             while not eof (fHandle) do
                   begin
                        readln (fHandle, sLine);

                        if pos ('#', sLine) = 0 then
                           begin
                                Subnet := tIPv4SubnetDefinition.Create;

                                Subnet.SubnetAddress.setValueFromString (getEntry (sLine, ',', 1));
                                Subnet.SubnetAddress.setMaskFromString (getEntry (sLine, ',', 2));
                                Subnet.Name := getEntry (sLine, ',', 3);

                                if Subnet.Name = '' then
                                   Subnet.Name := Subnet.SubnetAddress.ValueAsString;

                                priv_Subnetlist.Add(Subnet);
                           end;
                   end;

             Sort;
             priv_sorted := boWasSorted;

             closefile (fHandle);
        end;
end;

(* -------------------------------------- *)

procedure tIPv4SubnetHandler.SaveToCSVFile(const Filename: string);

var fHandle        : text;
    sLine          : string;
    boValidFile    : boolean;
    Subnet         : tIPv4SubnetDefinition;
    i              : integer;

begin
     boValidFile := true;

     try
        assignfile (fHandle, Filename);
        rewrite (fHandle);
     except
           boValidFile := false;
     end;

     if boValidFile then
        begin
             writeln ('# Address,Mask,Name');

             for i := 1 to priv_subnetlist.Count do
                 begin
                      Subnet := priv_subnetlist[i-1] as tIPv4SubnetDefinition;

                      sLine := Subnet.SubnetAddress.ValueAsString + ',' + Subnet.SubnetAddress.MaskAsString + ',' + Subnet.Name;
                      writeln (fHandle, sLine)
                 end;
        end;

     closefile (fHandle);
end;

(* -------------------------------------- *)

procedure tIPv4SubnetHandler.Sort;

begin
     priv_subnetlist.Sort(CompareSubnets);
end;

(* -------------------------------------- *)

end.
