{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_ipv6subnethandler;

interface

uses contnrs,
     definition_ip_addresses;

type tIPv6SubnetDefinition = class (tObject)

                               constructor create (const IPAddress : tIPv6Address); overload;
                               constructor create (const IPAddress : tIPv6Address; const SubnetName : string); overload;
                               procedure free;

                               private
                                      priv_Name           : string;
                                      priv_SubnetAddress  : tIPv6Address;
                                      priv_ReplacementNet : tIPv6Address;

                               public
                                     property Name : string read priv_Name write priv_Name;
                                     property SubnetAddress : tIPv6Address read priv_SubnetAddress write priv_SubnetAddress;
                                     property ReplacementNetAddress : tIPv6Address read priv_ReplacementNet write priv_ReplacementNet;
                                     procedure copyTo (Target : tIPv6SubnetDefinition);
                                     function isEqualTo (Target : tIPv6SubnetDefinition) : boolean;

                                     {$REGION 'Documentation'}
                                     /// <summary>
                                     ///   generates a random IP for the
                                     ///   replacement subnet range, if an
                                     ///   address is specified. Otherwise a
                                     ///   random interface ID is generated and
                                     ///   combined with the replacement subnet
                                     ///   range.
                                     /// </summary>
                                     /// <param name="OriginalAddress">
                                     ///   Address containing the interface ID
                                     ///   to be kept
                                     /// </param>
                                     {$ENDREGION}
                                     function getReplacementIPForSubnet (const OriginalAddress : tIPv6Address) : tIPv6Address;
                         end;

     tIPv6SubnetHandler    = class (tObject)

                                   constructor create;
                                   procedure free;

                                   private
                                          priv_subnetlist      : tObjectList;
                                          priv_sorted          : boolean;
                                          priv_LastIPAddress   : tIPv6Address;
                                          priv_LastSubnetIndex : integer;

                                          function getSubnetCount: integer;

                                          {$REGION 'Documentation'}
                                          ///	<summary>
                                          ///	  Prepare the IPv6 Mask and
                                          ///	  Wildcard table
                                          ///	</summary>
                                          {$ENDREGION}
                                          procedure prepareMaskTable;

                                   public
                                         property Count : integer read getSubnetCount;
                                         property Sorted : boolean read priv_sorted write priv_sorted;

                                         procedure AddSubnet (Subnet : tIPv6SubnetDefinition);
                                         procedure Clear;
                                         function GetSubnetByIndex (const Index : integer) : tIPv6SubnetDefinition;

                                         {$REGION 'Documentation'}
                                         ///	<summary>
                                         ///	  Returns to subnet index that a given
                                         ///	  IP adress belongs to
                                         ///	</summary>
                                         ///	<param name="IPAddress">
                                         ///	  IP address to find a subnet for.
                                         ///	</param>
                                         ///	<returns>
                                         ///	  Index of the subnetdefinition the
                                         ///	  IPAddress belongs to. Returns 0 if
                                         ///	  none was found.
                                         ///	</returns>
                                         {$ENDREGION}
                                         function GetSubnetIndexForIPAddress (const IPAddress : tIPv6Address) : integer;
                                         procedure LoadFromCSVFile (const Filename : string);
                                         procedure SaveToCSVFile (const Filename : string);
                                         procedure Sort;

                             end;

implementation

uses Math,
     SysUtils,
     toolkit;

type tIPv6MaskTableEntry = record
                                 MaskWord     : tIPv6AddressWordStruct;
                                 WildcardWord : tIPv6AddressWordStruct;
                           end;

     tIPv6MaskTable = array [1..128] of tIPv6MaskTableEntry;

var maskTable : tIPv6MaskTable;

(* -------------------------------------- *)

function compareSubnets (item1, item2 : pointer) : integer;

var def1, def2 : tIPv6SubnetDefinition;

begin
     def1 := item1;
     def2 := item2;

     result := CompareValue (def1.priv_SubnetAddress, def2.priv_SubnetAddress);
end;

(* -------------------------------------- *)
{ tIPv6SubnetDefinition }
(* -------------------------------------- *)

constructor tIPv6SubnetDefinition.create (const IPAddress: tIPv6Address);

begin
     inherited create;

     priv_SubnetAddress := IPAddress;
end;

(* -------------------------------------- *)

constructor tIPv6SubnetDefinition.create (const IPAddress: tIPv6Address; const SubnetName: string);

begin
     inherited create;

     priv_SubnetAddress := IPAddress;
     priv_Name := SubnetName;
end;

(* -------------------------------------- *)

procedure tIPv6SubnetDefinition.free;

begin
     inherited free;
end;

(* -------------------------------------- *)

function tIPv6SubnetDefinition.isEqualTo (Target: tIPv6SubnetDefinition): boolean;

begin
     result := (priv_Name = Target.Name) and (priv_SubnetAddress = Target.SubnetAddress);
end;

(* -------------------------------------- *)

procedure tIPv6SubnetDefinition.copyTo (Target: tIPv6SubnetDefinition);

begin
     Target.Name := priv_Name;
     Target.SubnetAddress := priv_SubnetAddress;
end;

(* -------------------------------------- *)

function tIPv6SubnetDefinition.getReplacementIPForSubnet (const OriginalAddress: tIPv6Address): tIPv6Address;

var randomIP : tIPv6Address;
    WildCard : tIPv6AddressWordStruct;
    help     : tIPv6AddressWordStruct;
    i        : integer;

begin
     result.initialize;

     // Extract Address part from Original

     WildCard := MaskTable[priv_SubnetAddress.MaskBitCount].WildcardWord;

     if not OriginalAddress.IsUnspecifiedAddress  then
        begin
             help := OriginalAddress.ValueAsWordStruct;
        end
     else
         begin
              // Generate random address part

              randomIP.initialize (true);
              help := randomIP.ValueAsWordStruct;
         end;

     // Extract host part

     for i := 0 to 7 do
         help[i] := help[i] and WildCard[i];

     // return full network IP

     for i := 0 to 7 do
         result.ValueAsWordStruct[i] := priv_ReplacementNet.ValueAsWordStruct[i] or help[i];
end;

(* -------------------------------------- *)
{ tIPv6SubnetHandler }
(* -------------------------------------- *)

constructor tIPv6SubnetHandler.create;

begin
     inherited;

     priv_subnetlist := tObjectList.Create(true);
     priv_sorted := true;

     priv_LastSubnetIndex := -1;
     priv_LastIPAddress.initialize;

     PrepareMaskTable;
end;

(* -------------------------------------- *)

procedure tIPv6SubnetHandler.free;

begin
     inherited;
end;

(* -------------------------------------- *)

procedure tIPv6SubnetHandler.prepareMaskTable;

var i,j       : integer;
    fullcount : integer;
    bitcount  : integer;
    help      : extended;
    help2     : word;
    value     : word;

begin
     // Prepare Masks

     for i := 1 to 128 do
         begin
              // Determine how many words have all bits on

              fullcount := i div 16;

              if fullcount > 0 then
                 begin
                      // Set them from left to right

//                    for j := 7 downto 8 - fullcount do
                      for j := 0 to fullcount - 1 do
                          MaskTable[i].MaskWord[j] := 65535;
                 end;

              // Set remaining bits

              bitcount := i mod 16;

              if bitcount > 0 then
                 begin
                      value := 0;

//                    for j := 0 to bitcount - 1 do
                      for j := 15 downto 16 - bitcount do
                          begin
                               help := power (2,j);
                               help2 := round (help);

                               value := value or help2;
                          end;

                      MaskTable[i].MaskWord[fullcount] := value;
                 end;

              // Set the wildcards

              for j := 0 to 7 do
                  MaskTable[i].WildcardWord[j] := (65535 - MaskTable[i].MaskWord[j]);
         end;
end;

(* -------------------------------------- *)

procedure tIPv6SubnetHandler.AddSubnet (Subnet: tIPv6SubnetDefinition);

begin
     priv_subnetlist.Add(Subnet);

     // If the list needs to be sorted at all times, do it (usually for BinSearch)

     if priv_sorted then
        Sort;
end;

(* -------------------------------------- *)

function tIPv6SubnetHandler.getSubnetCount: integer;

begin
     result := priv_subnetlist.Count;
end;

(* -------------------------------------- *)

procedure tIPv6SubnetHandler.Clear;

begin
     priv_subnetlist.Clear;
end;

(* -------------------------------------- *)

function tIPv6SubnetHandler.GetSubnetByIndex (const Index: integer): tIPv6SubnetDefinition;

begin
     if (Index > 0) and (Index <= priv_subnetlist.Count) then
        result := priv_subnetlist[Index-1] as tIPv6SubnetDefinition
     else
         result := nil;
end;

(* -------------------------------------- *)

function tIPv6SubnetHandler.GetSubnetIndexForIPAddress (const IPAddress: tIPv6Address): integer;

var subnet   : tIPv6SubnetDefinition;
    help     : tIPv6Address;
    baseaddr : tIPv6Address;
    current  : integer;
    i,j      : integer;

begin
     result := 0;
     current := 0;
     help := IPAddress;

     // Check if it is 0.0.0.0

     if IPAddress.IsUnspecifiedAddress then
        Exit (0);

     // Check Cache

     if IPAddress = priv_LastIPAddress then
        Exit (priv_LastSubnetIndex);

     // Search subnet

     for i := 1 to priv_subnetlist.Count do
         begin
              subnet := (priv_subnetlist[i-1] as tIPv6SubnetDefinition);

              // find out if the IPAddress has exactly the same subnet bits as the current
              // subnet we compare

              baseaddr.MaskBitCount := subnet.priv_SubnetAddress.MaskBitCount;

              for j := 0 to 7 do
                  baseaddr.ValueAsWordStruct[j] := help.ValueAsWordStruct[j] and subnet.priv_SubnetAddress.ValueAsWordStruct[j];

              // See if the address matches the subnet

              if baseaddr = subnet.priv_SubnetAddress then
                 begin
                      // if no subnet was found yet, this is the one for now
                      // otherwise, see if the new one has a tighter subnet mask than the last find

                      if current = 0 then
                         current := i
                      else
                          if (priv_subnetlist[current-1] as tIPv6SubnetDefinition).priv_SubnetAddress.MaskBitCount < subnet.priv_SubnetAddress.MaskBitCount then
                             current := i;
                 end;
         end;

     // Update Cache

     if current > 0 then
        begin
             priv_LastSubnetIndex := current;
             priv_LastIPAddress := IPAddress;
        end;

     result := current;
end;

(* -------------------------------------- *)

procedure tIPv6SubnetHandler.LoadFromCSVFile(const Filename: string);

var fHandle        : text;
    sLine          : string;
    boValidFile    : boolean;
    boWasSorted    : boolean;
    Subnet         : tIPv6SubnetDefinition;
    MaskBitStr     : string;
    HelpAddr       : tIPv6Address;
    Help           : integer;

begin
     boValidFile := true;

     try
        assignfile (fHandle, Filename);
        reset (fHandle);
     except
           boValidFile := false;
     end;

     if boValidFile then
        begin
             boWasSorted := priv_sorted;
             priv_sorted := false;

             while not eof (fHandle) do
                   begin
                        readln (fHandle, sLine);

                        if pos ('#', sLine) = 0 then
                           begin
                                Subnet := tIPv6SubnetDefinition.Create;

                                HelpAddr.setValueFromString(getEntry (sLine, ',', 1));
                                MaskBitStr := (getEntry (sLine, ',', 2));
                                tryStrToInt (MaskBitStr, Help);
                                HelpAddr.MaskBitCount := Help;

                                Subnet.SubnetAddress := HelpAddr;
                                Subnet.Name := getEntry (sLine, ',', 3);

                                if Subnet.Name = '' then
                                   Subnet.Name := Subnet.SubnetAddress.ValueAsString;

                                priv_Subnetlist.Add(Subnet);
                           end;
                   end;

             Sort;
             priv_sorted := boWasSorted;

             closefile (fHandle);
        end;
end;

(* -------------------------------------- *)

procedure tIPv6SubnetHandler.SaveToCSVFile(const Filename: string);

var fHandle        : text;
    sLine          : string;
    boValidFile    : boolean;
    Subnet         : tIPv6SubnetDefinition;
    i              : integer;

begin
     boValidFile := true;

     try
        assignfile (fHandle, Filename);
        rewrite (fHandle);
     except
           boValidFile := false;
     end;

     if boValidFile then
        begin
             writeln ('# Address,MaskBits,Name');

             for i := 1 to priv_subnetlist.Count do
                 begin
                      Subnet := priv_subnetlist[i-1] as tIPv6SubnetDefinition;

                      sLine := Subnet.SubnetAddress.ValueAsString + ',' + IntToStr (Subnet.SubnetAddress.MaskBitCount) + ',' + Subnet.Name;
                      writeln (fHandle, sLine)
                 end;
        end;

     closefile (fHandle);
end;

(* -------------------------------------- *)

procedure tIPv6SubnetHandler.Sort;

begin
     priv_subnetlist.Sort(CompareSubnets);
end;

(* -------------------------------------- *)

end.
