{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_tuplemanager_tcp;

interface

uses System.Generics.Collections,
     base_structures,
     definition_ip_addresses,
     definition_5tuple,
     definition_tcptuple,
     definition_tcp_structures;

type tTCPTupleManager = class (tObject)

                              constructor create;

                              private
                                     priv_tuples           : tDictionary<string, tTCPTupleList>;
                                     priv_currenttuplelist : tTCPTupleList;
                                     priv_currentindex     : integer;

                              public
                                     /// <summary>
                                     ///   Returns a dictionary key without the
                                     ///   ISN, because we need to find them
                                     ///   here without knowing what it is
                                     /// </summary>
                                     function getBaseKey (TCPTuple : tTCPTuple) : string;

                                     procedure AddTCPTuple (Value : tTCPTuple);

                                     function TCPTupleExists (const Tuple : t5Tuple; const Anchor : tConversationAnchor) : boolean;
                                     /// <summary>
                                     ///   Updates the ISN of an existing TCP
                                     ///   tuple in case we found a better one
                                     ///   (e.g. when SYN was out-of-order and
                                     ///   we see a later sequence first)
                                     /// </summary>
                                     /// <param name="Tuple">
                                     ///   5-Tuple of the connection
                                     /// </param>
                                     /// <param name="Anchor">
                                     ///   TCP Connection Anchor
                                     /// </param>
                                     /// <param name="NewISN">
                                     ///   SequenceNumber seen in the SYN packet
                                     /// </param>
                                     /// <param name="NewISNTimestamp">
                                     ///   Timestamp of the SYN packet
                                     /// </param>
                                     procedure UpdateISN (const Tuple : t5Tuple; const Anchor : tConversationAnchor; const NewISN : tTCPSequence; const NewISNTimestamp : tFrametimestamp);
                                     /// <summary>
                                     ///   Returns the current TCP Tuple for the
                                     ///   given 5-Tuple/Anchor. This is
                                     ///   basically the same as the two values,
                                     ///   but with the additional ISN to be
                                     ///   able to distinguish between multiple
                                     ///   TCP connections reusing socketpairs
                                     ///   over time.
                                     /// </summary>
                                     /// <param name="Tuple">
                                     ///   5-Tuple of the TCP connection
                                     /// </param>
                                     /// <param name="Anchor">
                                     ///   Conversation Anchor of the TCP
                                     ///   connection
                                     /// </param>
                                     /// <returns>
                                     ///   TCP Tuple currently in use for the
                                     ///   given 5-Tuple and Conversation Anchor
                                     /// </returns>
                                     function getCurrentTCPTuple (const Tuple : t5Tuple; const Anchor : tConversationAnchor) : tTCPTuple;
                                     /// <summary>
                                     ///   Returns the number of TCP tuples for
                                     ///   a given 5-Tuple and conversation
                                     ///   anchor. This can only be &gt;1 if
                                     ///   there are socket pairs being reused.
                                     /// </summary>
                                     /// <param name="Tuple">
                                     ///   5-Tuple of the TCP connection
                                     /// </param>
                                     /// <param name="Anchor">
                                     ///   Conversation anchor of the TCP
                                     ///   connection
                                     /// </param>
                                     /// <returns>
                                     ///   number of connections on the same
                                     ///   socket pair and anchor
                                     /// </returns>
                                     function getTCPTupleCount (const Tuple : t5Tuple; const Anchor : tConversationAnchor) : integer;
                                     function getFirstTCPTUple (const Tuple : t5Tuple; const Anchor : tConversationAnchor) : tTCPTuple;
                                     function getNextTCPTUple (const Tuple : t5Tuple; const Anchor : tConversationAnchor) : tTCPTuple;

                                     destructor Destroy; override;
                        end;

implementation

(* ------------------------------------------- *)
{ tTCPTupleManager }
(* ------------------------------------------- *)

constructor tTCPTupleManager.create;

begin
     inherited create;

     priv_tuples := tDictionary<string, tTCPTupleList>.create;
end;

(* ------------------------------------------- *)

destructor tTCPTupleManager.Destroy;

begin
     priv_tuples.Free;

     inherited;
end;

(* ------------------------------------------- *)

function tTCPTupleManager.getBaseKey (TCPTuple : tTCPTuple): string;

begin
//   result := TCPTuple.Anchor.DictionaryKey + '-' + TCPTuple.Tuple.DictionaryKey;
     result := TCPTuple.Base.DictionaryKey;
end;

(* ------------------------------------------- *)

function tTCPTupleManager.TCPTupleExists (const Tuple: t5Tuple; const Anchor: tConversationAnchor): boolean;

var value : tTCPTuple;
    key   : string;

begin
     value.initialize (Tuple);
//   value.Anchor := Anchor;
     value.Base.Anchor := Anchor;

     key := getBaseKey (value);

     result := priv_tuples.ContainsKey (Key);
end;

(* ------------------------------------------- *)

procedure tTCPTupleManager.AddTCPTuple (Value: tTCPTuple);

var key  : string;
    list : tTCPTupleList;
    len  : integer;
    i    : integer;

begin
     // We use the base key, as the dictionary key also includes ISN
     // and we need to be able to find the list of tuples without it

     key := getBaseKey (Value);

     if not priv_tuples.ContainsKey (Key) then
        begin
             setLength (list, 1);
             list[0] := Value;

             priv_tuples.Add (key, list);
        end
     else
         begin
              list := priv_tuples.Items[key];
              len := length (list);

              // Make sure this is really a new entry

              for i := 1 to len do
                  assert (list[i].ISN <> Value.ISN);

              // add the new value to the list

              inc (len);
              setLength (list, len);

              list[len-1] := Value;

              priv_tuples.Items[key] := list;
         end;
end;

(* ------------------------------------------- *)

function tTCPTupleManager.getCurrentTCPTuple (const Tuple: t5Tuple; const Anchor: tConversationAnchor): tTCPTuple;

var value : tTCPTuple;
    key   : string;
    list  : tTCPTupleList;
    len   : integer;

begin
//   value.initialize;
//   value.Tuple := Tuple;
//   value.Anchor := Anchor;
     value.initialize (tuple);
     value.Base.Anchor := Anchor;

     key := getBaseKey (value);

     assert (priv_tuples.ContainsKey (Key), 'TCPTuple not found (tTCPTupleManager.getCurrentTCPTuple)');

     list := priv_tuples.Items[key];
     len := length (list);

     assert (len > 0, 'TCPTuple List empty (tTCPTupleManager.getCurrentTCPTuple)');

     // Return the last entry, as it's the current one

     value := list[len-1];

     result := value;
end;

(* ------------------------------------------- *)

function tTCPTupleManager.getTCPTupleCount(const Tuple: t5Tuple; const Anchor: tConversationAnchor): integer;

var value : tTCPTuple;
    key   : string;
    list  : tTCPTupleList;
    len   : integer;

begin
//   value.initialize;
//   value.Tuple := Tuple;
//   value.Anchor := Anchor;
     value.initialize (tuple);
     value.Base.Anchor := Anchor;

     key := getBaseKey (value);

     if priv_tuples.ContainsKey (Key) then
        begin
              list := priv_tuples.Items[key];
              len := length (list);
        end;

     result := len;
end;

(* ------------------------------------------- *)

function tTCPTupleManager.getFirstTCPTUple (const Tuple: t5Tuple; const Anchor: tConversationAnchor): tTCPTuple;

var value : tTCPTuple;
    key   : string;
    list  : tTCPTupleList;
    len   : integer;

begin
//   value.initialize;
//   value.Tuple := Tuple;
//   value.Anchor := Anchor;
     value.initialize (tuple);
     value.Base.Anchor := Anchor;

     if TCPTupleExists (Tuple, Anchor) then
        begin
             key := getBaseKey (value);
             priv_currenttuplelist := priv_tuples.Items[key];

             value := priv_currenttuplelist[0];
             priv_currentindex := 1;
        end
     else
         begin
              setLength (priv_currenttuplelist, 0);
              value.initialize;
              priv_currentindex := 0;
         end;

     result := value;
end;

(* ------------------------------------------- *)

function tTCPTupleManager.getNextTCPTUple(const Tuple: t5Tuple; const Anchor: tConversationAnchor): tTCPTuple;

var value : tTCPTuple;
    key   : string;
    len   : integer;

begin
     len := length (priv_currenttuplelist);

     if (len > 0) and (priv_currentindex < len) then
        begin
             inc (priv_currentindex);
             value := priv_currenttuplelist[priv_currentindex - 1];
        end
     else
         value.initialize;
end;

(* ------------------------------------------- *)

procedure tTCPTupleManager.UpdateISN (const Tuple: t5Tuple; const Anchor: tConversationAnchor; const NewISN: tTCPSequence; const NewISNTimestamp : tFrametimestamp);

var value : tTCPTuple;
    key   : string;
    list  : tTCPTupleList;
    len   : integer;

begin
//   value.initialize;
//   value.Tuple := Tuple;
//   value.Anchor := Anchor;
     value.initialize (tuple);
     value.Base.Anchor := Anchor;

     key := getBaseKey (value);

     if priv_tuples.ContainsKey (Key) then
        begin
              len := length (priv_tuples.Items[key]);

              assert (len > 0, 'TCPTuple List empty (tTCPTupleManager.UpdateISN)');

              // Return the last entry, as it's the current one

              priv_tuples.Items[key][len-1].ISN := NewISN;
              priv_tuples.Items[key][len-1].ISNTimestamp := NewISNTimestamp;
        end;
end;

(* ------------------------------------------- *)

end.
