{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit inteldb_tools;

interface

uses

     // FireDAC

     FireDAC.Stan.Intf, FireDAC.Stan.Option, FireDAC.Stan.Error, FireDAC.UI.Intf,
     FireDAC.Phys.Intf, FireDAC.Stan.Def, FireDAC.Stan.Pool, FireDAC.dAPT,
     FireDAC.Stan.Async, FireDAC.Phys, FireDAC.Stan.ExprFuncs,
     FireDAC.VCLUI.Wait, FireDAC.Comp.UI, FireDAC.Phys.SQLite,
     Data.DB, FireDAC.Comp.Client,

     base_structures;


type tFileIntelStatusAttr = (fisNotFound, fisFilenameMatch, fisWrongPath, fisWrongFileDate, fisWrongSize, fisWrongFirstFrameDate, fisWrongLastFrameDate, fisUnknownLastDate);
     tFileIntelStatusSet = set of tFileIntelStatusAttr;
     tFileIntelStatus = record
                              DBIndex    : integer;
                              Attributes : tFileIntelStatusSet;

                              function toString : string;
                        end;

     function getIntelDBFileStatus (IntelDB : tFDConnection; const Filename: string; const FileSize : int64; const FileDate : tDateTime; const FirstFrameTime, LastFrameTime : tFrameTimeStamp) : tFileIntelStatus;

     {$REGION 'Documentation'}
     /// <summary>
     ///   Determines if an entry in the file table is considered a match to
     ///   the file in the filelist by examining its file status. The result is
     ///   determined by first frame time, file size, file name, file path, and
     ///   last frame time in case of PCAPng.
     /// </summary>
     /// <param name="FileStatus">
     ///   true if the filestatus indicates a match
     /// </param>
     /// <remarks>
     ///   PCAPng is special in two ways: it can change file size even if the
     ///   packets stay exactly the same, because the file format allows meta
     ///   information to be added, e.g. comments. That makes relying on file
     ///   size as a marker impossible; fortunately, PCAPng also allows reading
     ///   the last frame without scanning the whole file, so in that case the
     ///   timestamp of the last frame is used to see if the file is unchanged.
     /// </remarks>
     {$ENDREGION}
     function DBFileEntryMatchFound (const FileStatus : tFileIntelStatus) : boolean;
     function readGeneralStatisticsFromDB (Database : tFDConnection; FileIndex : integer) : tGeneralTraceStats;

     /// <summary>
     ///   Compares two frame time stamps while ignoring micro and nano seconds
     /// </summary>
     function FrameDateCorrect (date1, date2 : tFrametimestamp) : boolean;


implementation

uses SysUtils,
     SiAuto;

(* ------------------------------------------- *)

function FrameDateCorrect (date1, date2 : tFrametimestamp) : boolean;

var help1 : integer;
    help2 : integer;

begin
     help1 := (date1.nanoseconds DIV 1000000);
     help2 := (date2.nanoseconds DIV 1000000);

     result := (date1.datetime = date2.datetime) and (help1 = help2);
end;

(* ------------------------------------------- *)
{ tFileIntelStatus }
(* ------------------------------------------- *)

function tFileIntelStatus.toString: string;

var value : string;
    help  : string;
    attr  : tFileIntelStatusAttr;

begin
     value := 'DBIndex ' + IntToStr (DBIndex) + ', Attributes (';
     help := '';

     for attr in Attributes do
         begin
              case attr of
               fisNotFound            : help := help + 'not found, ';
               fisFilenameMatch       : help := help + 'filename match, ';
               fisWrongPath           : help := help + 'wrong path, ';
               fisWrongFileDate       : help := help + 'wrong file date, ';
               fisWrongSize           : help := help + 'wrong size, ';
               fisWrongFirstFrameDate : help := help + 'wrong first frame date, ';
               fisWrongLastFrameDate  : help := help + 'wrong first frame date, ';
               fisUnknownLastDate     : help := help + 'unknown last frame date, ';
              end;

         end;

     // Strip the last comma

     if help <> '' then
        help := copy (help, 1, length (help) - 2);

     value := value + help + ')';

     result := value;
end;

(* ------------------------------------------- *)

function getIntelDBFileStatus (IntelDB : tFDConnection; const Filename: string; const FileSize : int64; const FileDate : tDateTime; const FirstFrameTime, LastFrameTime : tFrameTimeStamp) : tFileIntelStatus;

var Query        : TFDQUery;
    Status       : tFileIntelStatus;
    DBPath       : string;
    DBTime       : tDateTime;
    DBFirstFrame : tFrameTimeStamp;
    DBLastFrame  : tFrameTimeStamp;
    DBSize       : int64;
    help         : string;

begin
     SiMain.EnterMethod('getIntelDBFileStatus');

     Status.DBIndex := 0;
     Status.Attributes := [];

     help := lowercase (extractFilename (Filename));

     Query := TFDQUery.Create(nil);
     Query.Connection := IntelDB;

     Query.SQL.Text := 'SELECT * FROM filelist WHERE filename=:filename';
     Query.ParamByName ('filename').AsString := help;
     Query.Open;

     if Query.RecordCount > 0 then
        begin
             SiMain.LogVerbose ('Looking for file "' + help + '", ' + IntToStr (Query.RecordCount) + ' records found, verifying match');

             DBPath := lowercase (Query.FieldByName('filepath').AsString);
             DBTime := Query.FieldByName('filedate').AsDateTime;
             DBSize := Query.FieldByName('filesize').AsLargeInt;
             DBFirstFrame.datetime := Query.FieldByName('firstframetimestamp').AsDateTime;
             DBFirstFrame.nanoseconds := Query.FieldByName('firstframenanoseconds').AsInteger;
             DBLastFrame.datetime := Query.FieldByName('lastframetimestamp').AsDateTime;
             DBLastFrame.nanoseconds := Query.FieldByName('lastframenanoseconds').AsInteger;

             Status.DBIndex := Query.FieldByName ('id').AsInteger;

             Status.Attributes := [fisFilenameMatch];

             // Check if the file matches all of these, and if not, mark it in the attribute set

             if not (DBPath = LowerCase (ExtractFilePath (Filename))) then
                Status.Attributes := [fisWrongPath];

             if not (DBTime = FileDate) then
                Status.Attributes := Status.Attributes + [fisWrongFileDate];

             if not FrameDateCorrect (DBFirstFrame, FirstFrameTime) then
                 Status.Attributes := Status.Attributes + [fisWrongFirstFrameDate];

             if DBLastFrame.datetime = 0 then
                Status.Attributes := Status.Attributes + [fisUnknownLastDate];

             if (LastFrameTime.datetime > 0) and not FrameDateCorrect (DBLastFrame, LastFrameTime) then
                Status.Attributes := Status.Attributes + [fisWrongLastFrameDate];

             if not (DBSize = FileSize) then
                Status.Attributes := Status.Attributes + [fisWrongSize];

             SiMain.LogVerbose ('Verification results: ' + Status.toString);
        end
     else
         begin
              SiMain.LogVerbose ('Looking for file "' + help + '", 0 records found, setting DBIndex to zero');

              Status.DBIndex := 0;
              Status.Attributes := [fisNotFound];
         end;

     Query.Free;

     result := Status;

     SiMain.LeaveMethod('getIntelDBFileStatus');
end;

(* ------------------------------------------- *)

function DBFileEntryMatchFound (const FileStatus : tFileIntelStatus) : boolean;

var boMatch : boolean;

begin
     // Check if we found an entry that has the correct timestamp for the first frame in the file

     boMatch := not ((FileStatus.DBIndex = 0) or (fisWrongFirstFrameDate in FileStatus.Attributes));

     if boMatch then
        begin
             // There IS an entry in the database, but we need to check if it is outdated/not for this file

             if not (fisUnknownLastDate in FileStatus.Attributes) then
                begin
                     // We know the time stamp of the last frame, so this must be a PCAPng file

                     boMatch := not (fisWrongLastFrameDate in FileStatus.Attributes);
                end
             else
                 begin
                      // it's not PCAPng, so we need to check file size

                      boMatch := not (fisWrongSize in FileStatus.Attributes);
                 end;
        end;

     result := boMatch;
end;

(* ------------------------------------------- *)

function readGeneralStatisticsFromDB (Database : tFDConnection; FileIndex : integer) : tGeneralTraceStats;

var Query : tFDQuery;
    stats : tGeneralTraceStats;

begin
     stats.initialize;

     Query := TFDQUery.Create (nil);
     Query.Connection := Database;

     try
        Query.SQL.Text := 'SELECT * FROM filestatistics WHERE fileid=:fileid;';
        Query.ParamByName ('fileid').AsInteger := FileIndex;
        Query.Open;

        stats.FirstFrameTime.datetime := Query.FieldByName ('firstframetime').AsDateTime;
        stats.FirstFrameTime.nanoseconds := Query.FieldByName ('firstframetimenanosecs').AsInteger;
        stats.FirstFrameOffset := Query.FieldByName ('firstframeoffset').AsLargeInt;
        stats.LastFrameTime.datetime := Query.FieldByName ('lastframetime').AsDateTime;
        stats.LastFrameTime.nanoseconds := Query.FieldByName ('lastframetimenanosecs').AsInteger;
        stats.LastFrameOffset := Query.FieldByName ('lastframeoffset').AsLargeInt;
        stats.MinFrameSize := Query.FieldByName ('minframesize').AsInteger;
        stats.MaxFrameSize := Query.FieldByName ('maxframesize').AsInteger;
        stats.AverageFrameSize := Query.FieldByName ('averageframesize').AsInteger;
        stats.FrameCount := Query.FieldByName ('framecount').AsInteger;
        stats.TotalFrameBytes := Query.FieldByName ('totalframebytes').AsLargeInt;
        stats.TotalWireBytes := Query.FieldByName ('totalwirebytes').AsLargeInt;
        stats.isSliced := Query.FieldByName ('isFrameSliced').AsBoolean;
        stats.SliceSizeConstant := Query.FieldByName ('SliceIsConstant').AsBoolean;
        stats.SliceSize := Query.FieldByName ('slicesize').AsInteger;
        stats.TimeOrderCorrect := Query.FieldByName ('TimeOrderIsCorrect').AsBoolean;
        stats.OutOfTimeFrames := Query.FieldByName ('OutOfTimeFrames').AsInteger;

     finally
            Query.Free;
     end;

     result := stats;
end;

(* ------------------------------------------- *)


end.
