{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_taskprocessor_baseclass;

interface

uses system.Diagnostics,
     system.TimeSpan,
     base_structures,
     class_tracefilelist;

type tTaskErrorCode = (ecSuccess, ecWarning, ecFailure, ecAbort);

//   tOnProcessingError = procedure (Sender : tObject; const ErrorMessage : string) of Object;

     tOnStartProcessingFile = procedure (FileEntry : tTracefileEntry) of object;
     tOnSkippingFile = procedure (FileEntry : tTracefileEntry) of object;
     tOnFinishProcessingFile = procedure (FileEntry : tTracefileEntry; const ErrorCount : integer) of object;
//   tOnAdvancePercentProcessed = procedure (const FileID : integer; FileEntry : tTracefileEntry; const PercentageParsed : integer) of object;
     tOnAdvancePercentProcessed = procedure (Sender : tObject; FileEntry : tTracefileEntry; const PercentageParsed : integer) of object;
     tOnErrorProcessingTask = procedure (const FileEntry : tTracefileEntry; const FrameIndex : uint64; const ErrorLocation : tATErrorLocation; var AbortTask : boolean) of object;
     tOnCompleteScanOfFile = procedure (const FileID : integer; FileEntry : tTracefileEntry; const ErrorCount : integer) of object;

     tTaskProcessor = class (tObject)

                            constructor create; overload;
                            constructor create (InputFile : tTracefileEntry; Output : tOutputlocation); overload;
                            procedure free;

                            private
//                                 priv_OnProcessingError : tOnProcessingError;

                            protected
                                     EventOnStartProcessingOfFile      : tOnStartProcessingFile;
                                     EventOnSkippingFile               : tOnSkippingFile;
                                     EventOnFinishProcessingFile       : tOnFinishProcessingFile;
                                     EventOnAdvancePercentProcessed    : tOnAdvancePercentProcessed;
                                     EventOnErrorProcessingFrame       : tOnErrorProcessingTask;
                                     EventOnCompleteScanOfFile         : tOnCompleteScanOfFile;

                                     prot_Inputfile                    : tTracefileEntry;
                                     prot_OutputLocation               : tOutputlocation;
                                     prot_ParseErrorCount              : integer;
                                     prot_percentagescanned            : integer;
                                     prot_ProgressSampleRate           : integer;

                                     prot_fps                          : integer;

                                     prot_intervalstart                : tDateTime;
                                     prot_intervalframesprocessed      : int64;

                                     prot_StopWatch                    : tStopWatch;
                                     prot_Elapsed                      : tTimespan;

                                     procedure ResetStatistics;

                                     procedure ReportProgress (const FrameOffset : uint64); overload;
                                     procedure ReportProgress (const FramesProcessed, TotalFrames : uint64); overload;

                            public
                                  property InputFile : tTracefileEntry read prot_Inputfile;
                                  property OutputLocation : tOutputlocation read prot_OutputLocation;
                                  property ErrorCount : integer read prot_ParseErrorCount;
                                  property onStartProcessingFile : tOnStartProcessingFile read EventOnStartProcessingOfFile write EventOnStartProcessingOfFile;
                                  property onSkippingFile : tOnSkippingFile read EventOnSkippingFile write EventOnSkippingFile;
                                  property onFinishProcessingFile : tOnFinishProcessingFile read EventOnFinishProcessingFile write EventOnFinishProcessingFile;
                                  property OnAdvancePercentageProcessed : tOnAdvancePercentProcessed read EventOnAdvancePercentProcessed write EventOnAdvancePercentProcessed;
                                  property onErrorProcessingFrame : tOnErrorProcessingTask read EventOnErrorProcessingFrame write EventOnErrorProcessingFrame;
                                  property onCompleteScanOfFile : tOnCompleteScanOfFile read EventOnCompleteScanOfFile write EventOnCompleteScanOfFile;
                                  property FramesPerSecond : integer read prot_fps;
                                  property ProgressSampleRate : integer read prot_ProgressSampleRate;

                                  function Execute : tTaskErrorCode; virtual; abstract;
                      end;

implementation

uses SysUtils,
     DateUtils;

(* ------------------------------------------- *)
{ tTaskProcessor }
(* ------------------------------------------- *)

constructor tTaskProcessor.create;

begin
     inherited;

     prot_inputfile := nil;

     prot_ParseErrorCount := 0;
     prot_ProgressSampleRate := 1000;

     ResetStatistics;
end;

(* ------------------------------------------- *)

constructor tTaskProcessor.create (InputFile : tTracefileEntry; Output : tOutputlocation);

begin
     inherited create;

     prot_Inputfile := InputFile;
     prot_OutputLocation := Output;

     prot_ParseErrorCount := 0;
     prot_percentagescanned := 0;
     prot_ProgressSampleRate := 1000;

     ResetStatistics;
end;

(* ------------------------------------------- *)

procedure tTaskProcessor.free;

begin
     inherited;
end;

(* ------------------------------------------- *)

procedure tTaskProcessor.ResetStatistics;

begin
     prot_percentagescanned := 0;

     prot_intervalstart := 0;
     prot_intervalframesprocessed := 0;
end;

(* ------------------------------------------- *)

procedure tTaskProcessor.ReportProgress (const FrameOffset: uint64);

var percentagescanned : integer;
    iHelp             : integer;
    dHelp             : double;

begin
     if (prot_Inputfile.FileSize > 0) and assigned (EventOnAdvancePercentProcessed) then
        begin
             dHelp := 100 * frameoffset / prot_Inputfile.FileSize;
             iHelp := trunc (dHelp);

             if iHelp > prot_percentagescanned then
                begin
                     prot_percentagescanned := iHelp;

                     if assigned (EventOnAdvancePercentProcessed) then
                        EventOnAdvancePercentProcessed (self, prot_Inputfile, prot_percentagescanned);
                end;
        end;
end;

(* ------------------------------------------- *)

procedure tTaskProcessor.ReportProgress (const FramesProcessed, TotalFrames: uint64);

var percentagescanned : integer;
    iHelp             : integer;
    dHelp             : double;
    deltasecs         : integer;
    deltaframes       : int64;

begin
     if (TotalFrames > 0) and assigned (EventOnAdvancePercentProcessed) then
        begin
             dHelp := 100 * FramesProcessed / TotalFrames;
             iHelp := trunc (dHelp);

             if iHelp > prot_percentagescanned then
                begin
                     prot_percentagescanned := iHelp;

                     if (prot_intervalstart > 0) then
                        begin
                             deltasecs := secondsbetween (Now, prot_intervalstart);
                             deltaframes := FramesProcessed - prot_intervalframesprocessed;

                             if (deltasecs > 0) then
                                prot_fps := deltaframes div deltasecs
                             else
                                 prot_fps := 0;
                        end
                     else
                         prot_fps := 0;

                     if assigned (EventOnAdvancePercentProcessed) then
                        EventOnAdvancePercentProcessed (self, prot_Inputfile, prot_percentagescanned);

                     prot_intervalframesprocessed := FramesProcessed;
                     prot_intervalstart := Now;
                end;
        end;
end;

(* ------------------------------------------- *)


end.
