{ TraceWrangler - Network Capture File Toolkit
  By Jasper Bongertz <jasper@packet-foo.com>
  Copyright 2016 Jasper Bongertz
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
}

unit class_taskprocessor_extract;

interface

uses Classes,
     base_structures,
     definition_interface_mapping_types,
     definition_filter_structures,
     definition_conversation_types,
     class_tracefilelist,
     class_tracefilesetloader,
     class_taskprocessor_baseclass,
     pcapng_structures,
     class_interfacelist,
     class_interfacemapper,
     class_processchain,
     class_frameprocessor_multiwriter,
     class_parse_snortalertlog,
     class_conversationstorage;

type tExtractionMode = (xmNone, xmSingleFile, xmFilePerConversation);

     tExtractionParameters = record
                                   TraceFileList           : tTracefileList;
                                   FilterList              : tFilterSet;
                                   ExtractionMode          : tExtractionMode;
                                   InterfaceTransferScope  : tInterfaceTransferScope;      // not used yet; will be used to determine if to keep unused interfaces on write
                                   OutputSchema            : string;
                                   InterfaceMappingMode    : tInterfaceMapMode;
                                   ActiveInterfaceList     : tInterfaceFileIDList;         // Interfaces in use by the conversations/endpoints
                                   TotalExpectedFrameCount : int64;
                                   WriteMode               : tMultiWriteMode;
                                   MatchAllOnEmptyList     : boolean;
                                   AdjustFileTimestamp     : boolean;
                                   AdjustTimestampMode     : tFileTimestampMode;
                                   ManualTimestamp         : tDateTime;
                                   MetaDataSet             : tMetaDataSet;
                                   TruncateAfterLayer      : byte;                         // 0 if none

                                   procedure initialize;
                             end;

     tOnMetaInfoUpdate = procedure (Sender : tObject; const Statistics : tMultiWriteStatistics) of Object;

     tTaskProcessorExtract = class (tTaskProcessor)

                                   /// <summary>
                                   ///   Creates the Extractionprocessor. Keep
                                   ///   in mind that the filterlist subitem of
                                   ///   the Parameters param will be freed
                                   ///   after the task is complete.
                                   /// </summary>
                                   /// <param name="ConversationStore">
                                   ///   Conversation store containing metadata
                                   ///   of all files involved
                                   /// </param>
                                   /// <param name="Parameters">
                                   ///   Extraction parameters, with filterlist
                                   ///   subitem that is freed after the task
                                   ///   is complete
                                   /// </param>
                                   constructor create (const ConversationStore : tConversationStorage; const Parameters: tExtractionParameters);
                                   procedure free;

                                   private
                                          EventOnNextFile            : tOnNextFile;
                                          EventOnMetaInfoUpdate      : tOnMetaInfoUpdate;

                                          priv_conversationstore     : tConversationStorage;

                                          priv_firstframelocation    : tFileFrameLocation;
                                          priv_lastframelocation     : tFileFrameLocation;
                                          priv_totalframestoscan     : int64;

                                          priv_params                : tExtractionParameters;
                                          priv_processchain          : tProcessChain;
                                          priv_interfaceMapper       : tInterfaceMapper;
//                                        priv_CurrentFileEntry      : tTracefileEntry;      there is prot_inputfile for this in the base class already

                                          // persistent processor for Multiwriting so we can update the file name property for file name generation

                                          priv_multiloader           : tTracefileSetLoader;
                                          priv_multiwriteprocessor   : tFrameProcessorMultiWriter;

                                          priv_filesToProcess        : integer;
                                          priv_filesProcessed        : integer;

                                          // Statistics read from tracesetloader

                                          priv_currentfilereadframecount  : int64;
                                          priv_currentfilereadframebytes  : int64;
                                          priv_totalreadframecount        : int64;
                                          priv_totalreadframebytes        : int64;
                                          priv_currentfileframesprocessed : int64;

                                          procedure OnMultiLoaderProcessNextFile (Sender : tObject; const FileIndex : integer; const FileEntry : tTracefileEntry; var SkipFile : boolean);
                                          procedure OnMultiLoaderCloseFile (Sender : tObject; const FileIndex : integer; const FileEntry : tTracefileEntry);

                                          procedure OnMultiWriterChange (Sender : tObject; const Statistics : tMultiWriteStatistics);
                                          procedure OnMetaInfoMatch (Sender : tObject; const Statistics : tMultiWriteStatistics);

                                          function getMultiWriteStatistics : tMultiWriteStatistics;
                                          function getExtractionMode : tExtractionMode;

                                          {$REGION 'Documentation'}
                                          /// <summary>
                                          ///   Determines the overall first
                                          ///   and last frame location to
                                          ///   allow processing only relevant
                                          ///   files
                                          /// </summary>
                                          {$ENDREGION}
                                          procedure DetermineFrameLocationMarkers;
                                          function FrameInExtractScope (FileIndex : integer; FileFrameIndex, FileFrameOffset : uint64) : boolean;

                                          procedure ExecuteSingleFileProcessChain;
                                          procedure ExecuteMultiFileProcessChain;
                                          procedure ExtractToSingleFile;
                                          procedure ExtractToMultipleFiles;

                                   protected

                                   public
                                         property FilesToProcess : integer read priv_filesToProcess;
                                         property FilesProcessed : integer read priv_filesProcessed;
                                         property MultiWriteStatistics : tMultiWriteStatistics read getMultiWriteStatistics;
                                         property TotalFramesToScan : int64 read priv_totalframestoscan;
                                         property CurrentFileFrameReadCount : int64 read priv_currentfilereadframecount;
                                         property CurrentFileFrameByteCount : int64 read priv_currentfilereadframebytes;
                                         property CurrentFileFramesProcessed : int64 read priv_currentfileframesprocessed;
                                         property TotalFrameReadCount : int64 read priv_totalreadframecount;
                                         property TotalFrameByteCount : int64 read priv_totalreadframebytes;

                                         property ExtractionMode : tExtractionMode read getExtractionMode;
                                         property onNextFile : tOnNextFile read EventOnNextFile write EventOnNextFile;
//                                       property CurrentFile : tTracefileEntry read priv_currentfileentry;

                                         function execute : tTaskErrorCode;
                              end;

implementation

uses SysUtils,
     SiAuto,
     class_rawframe,
     class_traceloader,
     class_tracewriter,
     class_writer_pcapng,
     class_frameprocessor_baseclass,
     class_frameprocessor_filter,
     class_frameprocessor_metaenhancer;

(* ------------------------------------------ *)
{ tExtractionParameters }
(* ------------------------------------------ *)

procedure tExtractionParameters.initialize;

begin
     TraceFileList := nil;
     Filterlist := nil;
     ExtractionMode := xmNone;
     InterfaceTransferScope := itsKeepAll;
     OutputSchema := '';
     InterfaceMappingMode := immMerge;
     setlength (ActiveInterfaceList, 0);
     TotalExpectedFrameCount := 0;
     WriteMode := mwmNone;
     MatchAllOnEmptyList := false;

     AdjustFileTimestamp := false;
     AdjustTimestampMode := tsmNone;
     ManualTimestamp := 0;

     TruncateAfterLayer := 0;

     setlength (MetaDataSet, 0);
end;

(* ------------------------------------------ *)
{ tTaskProcessorExtract }
(* ------------------------------------------ *)

constructor tTaskProcessorExtract.create (const ConversationStore : tConversationStorage; const Parameters: tExtractionParameters);

begin
     SiMain.EnterMethod ('tTaskProcessorExtract.create');

     inherited create;

     priv_multiwriteprocessor := nil;
     priv_multiloader := nil;

     Parameters.TraceFileList.updateFirstAndLastFrameLocators;
     priv_firstframelocation := Parameters.TraceFileList.FirstFrameLocation;
     priv_lastframelocation := Parameters.TraceFileList.LastFrameLocation;

     priv_conversationstore := ConversationStore;
     priv_params := Parameters;

     SiMain.LogDebug ('ActiveInterface count: ' + length (Parameters.ActiveInterfaceList).ToString);

     priv_filesToProcess := 0;
     priv_filesProcessed := 0;

     SiMain.LeaveMethod ('tTaskProcessorExtract.create');
end;

(* ------------------------------------------ *)

procedure tTaskProcessorExtract.free;

begin
//   priv_params.FilterList.Free;    This ain't ours! Caller needs to handle this.

     inherited free;
end;

(* ------------------------------------------ *)

function tTaskProcessorExtract.getExtractionMode: tExtractionMode;

begin
     result := priv_params.ExtractionMode;
end;

(* ------------------------------------------ *)

function tTaskProcessorExtract.getMultiWriteStatistics: tMultiWriteStatistics;

begin
     if assigned (priv_multiwriteprocessor) then
        result := priv_multiwriteprocessor.Statistics
     else
         result.initialize;
end;

(* ------------------------------------------ *)

procedure tTaskProcessorExtract.OnMultiLoaderProcessNextFile (Sender: tObject; const FileIndex : integer; const FileEntry: tTracefileEntry; var SkipFile : boolean);

var isRelevant : boolean;

begin
     // Check if relevant for extraction - if we can determine it. Otherwise, it is.

     isRelevant := FileIndexExistsInList (priv_params.ActiveInterfaceList, FileIndex);

     if isRelevant then
        begin
             prot_Inputfile := FileEntry;

             // Reset the per-file counters and stats

             priv_currentfileframesprocessed := 0;
             ResetStatistics;

             // Update MultiWriter if active

             if assigned (priv_multiwriteprocessor) then
                priv_multiwriteprocessor.CurrentInputFilename := FileEntry.FileName;

             // Pass the file entry up

             if assigned (EventOnNextFile) then
                EventOnNextFile (Sender, FileIndex, FileEntry, SkipFile);

             // Trigger main window file list update, if relevant

             if assigned (EventOnStartProcessingOfFile) then
                EventOnStartProcessingOfFile (FileEntry);
        end
     else
         begin
              SkipFile := true;

              // Pass the file entry up

             if assigned (EventOnNextFile) then
                EventOnNextFile (Sender, FileIndex, FileEntry, SkipFile);

             if assigned (EventOnSkippingFile) then
                EventOnSkippingFile (FileEntry);

              SiMain.LogVerbose ('Skipping File "' + FileEntry.FullName + '" - not in scope of extraction');
         end;
end;

(* ------------------------------------------ *)

procedure tTaskProcessorExtract.OnMultiLoaderCloseFile (Sender: tObject; const FileIndex: integer; const FileEntry: tTracefileEntry);

begin
     // Pass the file entry up by using the close event provided by the base class

     if assigned (EventOnFinishProcessingFile) then
        EventOnFinishProcessingFile (FileEntry, prot_ParseErrorCount);
end;

(* ------------------------------------------ *)

procedure tTaskProcessorExtract.OnMetaInfoMatch (Sender: tObject; const Statistics: tMultiWriteStatistics);

begin
     if assigned (EventOnMetaInfoUpdate) then
        EventOnMetaInfoUpdate (Sender, Statistics);
end;

(* ------------------------------------------ *)

procedure tTaskProcessorExtract.OnMultiWriterChange (Sender: tObject; const Statistics: tMultiWriteStatistics);

begin
     if assigned (EventOnMetaInfoUpdate) then
        EventOnMetaInfoUpdate (Sender, Statistics);
end;

(* ------------------------------------------ *)

procedure tTaskProcessorExtract.DetermineFrameLocationMarkers;

var filter : tFrameFilter;
    entry  : tTracefileEntry;
    i,j    : integer;

begin
     SiMain.EnterMethod ('tTaskProcessorExtract.DetermineFrameLocationMarkers');

     // If there is a filter we determine the location markers, otherwise they cover the whole file (set on .create)

     if priv_params.FilterList.count > 0 then
        begin
             // set initial values from first filter

             filter := priv_params.FilterList.getFirst;

             priv_firstframelocation := filter.HelpFirstSeen;
             priv_lastframelocation := filter.HelpLastSeen;

             repeat
                   filter := priv_params.FilterList.getNext;

                   if not filter.isUnspecified then
                      begin
                           priv_firstframelocation.UpdateIfLower (filter.HelpFirstSeen);
                           priv_lastframelocation.UpdateIfHigher (filter.HelpLastSeen);
                      end;

             until filter.isUnspecified;

             // If we have no clue what we're extracting, assume all and signal it with a zero

             if priv_firstframelocation.isZero and priv_lastframelocation.isZero then
                priv_totalframestoscan := 0
             else
                 begin
                      // Calculate total number of frames we will scan. We have to read the traces from the
                      // beginning, but we can abort reading when we know we're done. So reading frame 58 means
                      // we have to read 1-58.

                      if priv_firstframelocation = priv_lastframelocation then
                         priv_totalframestoscan := priv_firstframelocation.FileFrameIndex
                      else
                          begin
                               priv_totalframestoscan := 0;

                               for i := priv_firstframelocation.FileIndex to priv_lastframelocation.FileIndex - 1 do
                                   begin
                                        for j := 1 to length (priv_params.ActiveInterfaceList) do
                                            begin
                                                 if priv_params.ActiveInterfaceList[j-1].FileIndex = i then
                                                    begin
                                                         entry := priv_params.TraceFileList.getFile(i);
                                                         priv_totalframestoscan := priv_totalframestoscan + entry.FrameCount;
                                                    end;
                                            end;
                                   end;

                               // Add the remaining frames

                               priv_totalframestoscan := priv_totalframestoscan + priv_lastframelocation.FileFrameIndex;
                          end;
                 end;
        end;

     // TODO: when we can position the loader on specific offsets we need to substract the leading and trailing number of frames from the total count

     SiMain.LeaveMethod ('tTaskProcessorExtract.DetermineFrameLocationMarkers');
end;

(* ------------------------------------------ *)

function tTaskProcessorExtract.FrameInExtractScope (FileIndex : integer; FileFrameIndex, FileFrameOffset : uint64): boolean;

var help : tFileFrameLocation;

begin
     help.FileIndex := FileIndex;
     help.FileFrameIndex := FileFrameIndex;
     help.FileOffset := FileFrameOffset;

     // If we have no first/last helpers, we need to assume everything is in scope

     if (priv_firstframelocation.FileFrameIndex = 0) and (priv_lastframelocation.FileFrameIndex = 0) then
        result := true
     else
         result := (help >= priv_firstframelocation) and (help <= priv_lastframelocation);
end;

(* ------------------------------------------ *)

procedure tTaskProcessorExtract.ExecuteSingleFileProcessChain;

var frame       : tRawFrame;
    IFaceEntry  : tCaptureInterface;
    iFramesRead : uint64;
    i           : integer;

begin
     SiMain.EnterMethod ('tTaskProcessorExtract.ExecuteSingleFileProcessChain');

     // Create a trace set loader and adjust the read scope

     priv_multiloader := tTracefileSetLoader.create (priv_params.TraceFileList);
     priv_multiloader.onNextFile := OnMultiLoaderProcessNextFile;
     priv_multiloader.onCloseFile := OnMultiLoaderCloseFile;
     priv_multiloader.reset (priv_firstframelocation.FileIndex, priv_lastframelocation.FileIndex);

     if priv_totalframestoscan = 0 then
        priv_totalframestoscan := priv_multiloader.TotalFrameCount;

     priv_filesToProcess := priv_multiloader.FilesToProcess;
     priv_filesProcessed := priv_multiloader.FilesProcessed;

     iFramesRead := 0;

     repeat
           priv_multiloader.readBatch;

            for i := 1 to priv_multiloader.FramesAvailable do
               begin
                    frame := priv_multiloader.getFrameByIndex(i);
                    inc (iFramesRead);

                    // If the frame is within the range of where we know to find matching frames, work with it

                    if FrameInExtractScope (frame.TracefileID, frame.FileFrameID, frame.FileOffset) then
                       begin
                            // Assign the interface to the frame. InterfaceLinkID is used as a failsafe to make sure the frame
                            // is written with the correct interface index. The current file entry is updated by an event that
                            // is triggered whenever the multiloader opens the next file

                            IFaceEntry := prot_Inputfile.CaptureInterfaces.getInterfaceByIndex (Frame.InterfaceID);

                            // Some exceptions if we messed up - those should not happen anymore once the code is stable, so the next two if statements may be removed then

                            if IFaceEntry = nil then
                               begin
                                    SiMain.LogFatal ('Could not get interface #' + IntToStr (Frame.InterfaceID) + ' for frame #' + IntToStr (frame.FileFrameID) + ' (tTaskProcessorExtract.ExecuteSingleFileProcessChain)');
                                    raise Exception.Create('Could not get interface for frame #' + IntToStr (frame.FileFrameID) + ' (tTaskProcessorExtract.ExecuteSingleFileProcessChain)');
                               end;

                            if IFaceEntry.MappedToInterface = nil then
                               begin
                                    // if the interface is mapped to nothing, this interface is unused by the extracted frames (unless we have a bug in the mapping process)

                                    frame.free;
                                    frame := nil;

//                                  SiMain.LogFatal ('Could not get mapped interface #' + IntToStr (Frame.InterfaceID) + ' for frame #' + IntToStr (frame.FileFrameID) + ' (tTaskProcessorExtract.ExecuteSingleFileProcessChain)');
//                                  raise Exception.Create ('Could not get mapped interface for frame #' + IntToStr (frame.FileFrameID) + ' (tTaskProcessorExtract.ExecuteSingleFileProcessChain)');
                               end
                            else
                                begin
                                     Frame.InterfaceID := IFaceEntry.MappedToInterface.InterfaceID;
                                     Frame.InterfaceLinkID := IFaceEntry.MappedToInterface.PacketLinkID;

                                     // Toss frame into process chain

                                     priv_processchain.ProcessFrame (frame);
                                end;
                       end
                    else
                        begin
                             frame.free;
                             frame := nil;
                        end;

                    // Report Progress for the current file

                    ReportProgress (iFramesRead, priv_totalframestoscan);
               end;

     until priv_multiloader.EOF;

     priv_multiloader.free;

     SiMain.LeaveMethod ('tTaskProcessorExtract.ExecuteSingleFileProcessChain');
end;

(* ------------------------------------------ *)

procedure tTaskProcessorExtract.ExecuteMultiFileProcessChain;

var frame       : tRawFrame;
    IFaceEntry  : tCaptureInterface;
    i           : integer;

begin
     SiMain.EnterMethod ('tTaskProcessorExtract.ExecuteMultiFileProcessChain');

     priv_totalreadframecount := 0;
     priv_totalreadframebytes := 0;

     // Create a trace set loader and adjust the read scope

     priv_multiloader := tTracefileSetLoader.create (priv_params.TraceFileList);
     priv_multiloader.onNextFile := OnMultiLoaderProcessNextFile;
     priv_multiloader.onCloseFile := OnMultiLoaderCloseFile;

     if ((priv_firstframelocation.FileIndex = 0) and (priv_lastframelocation.FileIndex = 0)) then
        priv_multiloader.reset
     else
         priv_multiloader.reset (priv_firstframelocation.FileIndex, priv_lastframelocation.FileIndex);

     if priv_totalframestoscan = 0 then
        priv_totalframestoscan := priv_multiloader.TotalFrameCount;

     repeat
           priv_multiloader.readBatch;

           prot_stopwatch.StartNew;

           for i := 1 to priv_multiloader.FramesAvailable do
               begin
                    frame := priv_multiloader.getFrameByIndex(i);
                    inc (priv_currentfileframesprocessed);

                    priv_totalreadframecount := priv_totalreadframecount + 1;
                    priv_totalreadframebytes := priv_totalreadframebytes + frame.FrameSize;

                    // If the frame is within the range of where we know to find matching frames, work with it

                    if FrameInExtractScope (frame.TracefileID, frame.FileFrameID, frame.FileOffset) then
                       begin
                            // Assign the interface to the frame. InterfaceLinkID is used as a failsafe to make sure the frame
                            // is written with the correct interface index. The current file entry is updated by an event that
                            // is triggered whenever the multiloader opens the next file

                            IFaceEntry := prot_Inputfile.CaptureInterfaces.getInterfaceByIndex (Frame.InterfaceID);

                            // Some exceptions if we messed up - those should not happen anymore once the code is stable, so the next two if statements may be removed then

                            if IFaceEntry = nil then
                               begin
                                    SiMain.LogFatal ('Could not get interface #' + IntToStr (Frame.InterfaceID) + ' for frame #' + IntToStr (frame.FileFrameID) + ' (tTaskProcessorExtract.ExecuteMultiFileProcessChain)');
                                    raise Exception.Create('Could not get interface for frame #' + IntToStr (frame.FileFrameID) + ' (tTaskProcessorExtract.ExecuteMultiFileProcessChain)');
                               end;

                            if IFaceEntry.MappedToInterface = nil then
                               begin
                                    // if the interface is mapped to nothing, this interface is unused by the extracted frames (unless we have a bug in the mapping process)

                                    frame.free;
                                    frame := nil;

//                                  SiMain.LogFatal ('Could not get mapped interface #' + IntToStr (Frame.InterfaceID) + ' for frame #' + IntToStr (frame.FileFrameID) + ' (tTaskProcessorExtract.ExecuteMultiFileProcessChain)');
//                                  raise Exception.Create ('Could not get mapped interface for frame #' + IntToStr (frame.FileFrameID) + ' (tTaskProcessorExtract.ExecuteMultiFileProcessChain)');
                               end
                            else
                                begin
                                     Frame.InterfaceID := IFaceEntry.MappedToInterface.InterfaceID;
                                     Frame.InterfaceLinkID := IFaceEntry.MappedToInterface.PacketLinkID;

                                     // Toss frame into process chain

                                     priv_processchain.ProcessFrame (frame);
                                end;
                       end
                    else
                        begin
                             frame.free;
                             frame := nil;
                        end;

                    // Report Progress and transfer details from loader

                    priv_currentfilereadframecount := priv_multiloader.CurrentFileFrameReadCount;
                    priv_currentfilereadframebytes := priv_multiloader.CurrentFileFrameByteCount;

                    priv_filesProcessed := priv_multiloader.FilesProcessed;

                    if (priv_currentfileframesprocessed mod prot_ProgressSampleRate) = 0 then
                       ReportProgress (priv_currentfileframesprocessed, prot_inputfile.FrameCount);
               end;

           prot_elapsed := prot_stopwatch.Elapsed;

           SiMain.LogDebug ('Multiloader batch processing of ' + priv_multiloader.FramesAvailable.ToString + ' took ' + prot_elapsed.TotalMilliseconds.ToString + 'ms');

     until priv_multiloader.EOF;

     priv_multiloader.free;

     SiMain.LeaveMethod ('tTaskProcessorExtract.ExecuteMultiFileProcessChain');
end;

(* ------------------------------------------ *)

procedure tTaskProcessorExtract.ExtractToSingleFile;

var processor : tFrameProcessor;
    writer    : tPCAPngWriter;
    list      : tCaptureInterfaceList;

begin
     SiMain.EnterMethod ('tTaskProcessorExtract.ExtractToSingleFile');

     // Map interfaces, but only those in use by the conversations/endpoints (ActiveInterfaceList)

     priv_interfaceMapper := tInterfaceMapper.create (priv_params.TraceFileList, priv_params.InterfaceMappingMode);
     priv_interfaceMapper.RelevantInterfaceList := priv_params.ActiveInterfaceList;
     priv_interfaceMapper.doAutoMapping;

     list := tCaptureInterfaceList.create (true);
     priv_interfaceMapper.getTargetInterfaceList (list);

     // Prepare writer

     writer := tPCAPngWriter.create (priv_params.OutputSchema, true);
     processor := tFrameProcessor.create ('Output', writer);

     writer.Open;

     // Prepare comment, mimicking Wireshark's mergecap except for the mentioning of TraceWrangler

     writer.HeaderCommentList.Add ('File extracted from ' + IntToStr (priv_params.TraceFileList.Count) + ' source files by TraceWrangler');

     // Write header

     writer.WriteGenericHeader (list);

     // Create process chain

     priv_processchain := tProcessChain.create;

     // Add filter processor to keep only valid frames

     priv_processchain.AddProcessor(processor);
     processor := tFrameProcessorFilter.create ('FilterSet', priv_params.FilterList, priv_params.MatchAllOnEmptyList);

     priv_processchain.AddProcessor (processor);

     // Process

     ExecuteSingleFileProcessChain;

     // Release process chain, which will transfer the remaining frames
     // from the frame storage to the writer

     priv_processchain.free;

     // Finalize Writer

     writer.Flush;
     writer.Close;

     // clean up

     writer.free;
     list.free;
     priv_interfaceMapper.free;

     SiMain.LeaveMethod ('tTaskProcessorExtract.ExtractToSingleFile');
end;

(* ------------------------------------------ *)

procedure tTaskProcessorExtract.ExtractToMultipleFiles;

var processor   : tFrameProcessor;
    multiparams : tMultiWriteParams;

begin
     SiMain.EnterMethod ('tTaskProcessorExtract.ExtractToMultipleFiles');

     multiparams.initialize;

     // Create process chain, and prevent autocreating an output stub (since we're not using a single standard tracewriter object this time)

     priv_processchain := tProcessChain.create;
     priv_processchain.DoNotAddStub := true;

     // Map interfaces, but only those in use by the conversations/endpoints (ActiveInterfaceList)

     priv_interfaceMapper := tInterfaceMapper.create (priv_params.TraceFileList, priv_params.InterfaceMappingMode);
     priv_interfaceMapper.RelevantInterfaceList := priv_params.ActiveInterfaceList;
     priv_interfaceMapper.doAutoMapping;

     multiparams.InterfaceList := tCaptureInterfaceList.create (true);
     priv_interfaceMapper.getTargetInterfaceList (multiparams.InterfaceList);

     // Prepare remaining params for multiwriter

     multiparams.FileSchema := priv_params.OutputSchema;
     multiparams.ConversationStore := priv_conversationstore;
     multiparams.MaxFrameCountBeforeWrite := 10000;
     multiparams.TruncateAfterLayer := priv_params.TruncateAfterLayer;
//   multiparams.WriteMode := priv_params.WriteMode;

     if priv_params.AdjustFileTimestamp then
        begin
             multiparams.AbsoluteTimestamp := priv_params.ManualTimestamp;
             multiparams.FileTimestampMode := priv_params.AdjustTimestampMode;
        end
     else
         begin
              multiparams.AbsoluteTimestamp := 0;
              multiparams.FileTimestampMode := tsmNone;
         end;

     // create multiwriter processor

     priv_multiwriteprocessor := tFrameProcessorMultiWriter.create ('MultiOutput', multiparams);
//   priv_multiwriteprocessor.AddMetaDataSet (priv_params.MetaDataSet);

     priv_processchain.AddProcessor (priv_multiwriteprocessor);

     // Add meta processor if information is available

     if length (priv_params.MetaDataSet) > 0 then
        begin
             processor := tFrameProcessorMetaEnhancer.create ('MetaEnhancer', priv_params.MetaDataSet);
             priv_processchain.AddProcessor (processor);
        end;

     // Add filter processor to keep only valid frames

     if priv_params.FilterList.count > 0 then
        begin
             processor := tFrameProcessorFilter.create ('FilterSet', priv_params.FilterList, priv_params.MatchAllOnEmptyList);
             priv_processchain.AddProcessor (processor);
        end;

     // Process

     ExecuteMultiFileProcessChain;

     // Release process chain, which will transfer the remaining frames
     // from the frame storage to the writer

     priv_processchain.free;

     // Cleanup

     multiparams.InterfaceList.free;
     priv_interfaceMapper.free;

     SiMain.LeaveMethod ('tTaskProcessorExtract.ExtractToMultipleFiles');
end;

(* ------------------------------------------ *)

function tTaskProcessorExtract.execute: tTaskErrorCode;

begin
     SiMain.EnterMethod ('tTaskProcessorExtract.execute');

     DetermineFrameLocationMarkers;
     priv_conversationstore.UpdateFirstAndLastDictionaries (priv_params.WriteMode, true);

     case priv_params.ExtractionMode of
      xmSingleFile          : ExtractToSingleFile;
      xmFilePerConversation : ExtractToMultipleFiles;
     end;

     SiMain.LeaveMethod ('tTaskProcessorExtract.execute');
end;

(* ------------------------------------------ *)

end.
